/*
 * vcp2_api_int.h
 *
 * This file declares internal functions, types, and macros for the VCP2
 * library.
 */

#ifndef _VCP2_API_INT_H
#define _VCP2_API_INT_H

#include "vcp2_hbi.h"
#include "vp_api_cfg.h"
#include "vp_api_common.h"
#include "vcp2_ptable.h"


/* ====================================
    Macros Used Internally by VCP2_API
   ==================================== */

/* Worst-case VCP2 core frequency in cycles per millisecond (kHz) */
#define CORE_CLK_KHZ 98304

/* Worst-case PCLK frequency in cycles per millisecond (kHz) */
#define PCLK_KHZ 1024

/* Worst-case System Reset time (in PCLK frames) according to section 5.1.1
   in VCP2 SDS */
#if (VP_VCP2_BOARD_CLKSRC == 1)
#define SYSTEM_RESET_TIME_US 100 + 219
#else
#define SYSTEM_RESET_TIME_US 100 + ((1200000UL + (PCLK_KHZ) - 1) / (PCLK_KHZ))
#endif
#define SYSTEM_RESET_TIME_FRAMES ((SYSTEM_RESET_TIME_US + 124) / 125)

/*
 * Calculate the worst-case execution time (in 30-ms increments) of the memory
 * self-test.
 * test time = ceiling(32768000 / VCP2 core clock frequency in kHz)
 * polling iterations = 1 + ceiling(test time / 30 ms)
 */
#define MBIST_TIME_MS ((32768000UL + (CORE_CLK_KHZ) - 1) / (CORE_CLK_KHZ))
#define MBIST_POLL_ITERATIONS (1 + (((MBIST_TIME_MS) + 29) / 30))

/* If using the emulator, the wait times need to be increased. */
#ifdef VP_VCP2_EMU_SUPPORT
#undef  SYSTEM_RESET_TIME_FRAMES
#define SYSTEM_RESET_TIME_FRAMES 255
#undef  MBIST_POLL_ITERATIONS
#define MBIST_POLL_ITERATIONS    30
#endif

/* Using the preprocessor, calculate the length (in words) of the longest VCP2
 * profile type: */
#define MAX_PROFILE_LEN VCP2_DEV_PROFILE_LEN
#if (VCP2_DC_PROFILE_LEN > MAX_PROFILE_LEN)
    #undef MAX_PROFILE_LEN
    #define MAX_PROFILE_LEN VCP2_DC_PROFILE_LEN
#endif
#if (VCP2_AC_PROFILE_LEN > MAX_PROFILE_LEN)
    #undef MAX_PROFILE_LEN
    #define MAX_PROFILE_LEN VCP2_AC_PROFILE_LEN
#endif
#if (VCP2_CID_PROFILE_LEN > MAX_PROFILE_LEN)
    #undef MAX_PROFILE_LEN
    #define MAX_PROFILE_LEN VCP2_CID_PROFILE_LEN
#endif
#if (VCP2_MTR_PROFILE_LEN > MAX_PROFILE_LEN)
    #undef MAX_PROFILE_LEN
    #define MAX_PROFILE_LEN VCP2_MTR_PROFILE_LEN
#endif
#if (VCP2_TONE_PROFILE_LEN > MAX_PROFILE_LEN)
    #undef MAX_PROFILE_LEN
    #define MAX_PROFILE_LEN VCP2_TONE_PROFILE_LEN
#endif
#if (VCP2_CAD_PROFILE_LEN > MAX_PROFILE_LEN)
    #undef MAX_PROFILE_LEN
    #define MAX_PROFILE_LEN VCP2_CAD_PROFILE_LEN
#endif
#if (VCP2_RING_PROFILE_LEN > MAX_PROFILE_LEN)
    #undef MAX_PROFILE_LEN
    #define MAX_PROFILE_LEN VCP2_RING_PROFILE_LEN
#endif

/* Number of input types supported by VCP2: */
#define NUM_INPUT_TYPES         (VP_INPUT_BAT3_FLT + 1)

/* Maximum CID data length (in bytes) for VpSendCid() and VpContinueCid(): */
#define MAX_CID_DATA_LENGTH             32
#define MAX_CID_CONTINUE_DATA_LENGTH    16

/* Maximum length of MPI command+data for VpLowLevelCmd(): */
#define MAX_MPI_DATA_LENGTH             16

/* VCP2 doesn't use the standard test ID for AC_RLOOP_PHASE. */
#define VCP2_TEST_ID_AC_RLOOP_PHASE 23

/* Number of bytes required to store a flag for each VCP2 line: */
#define VCP2_LINE_FLAG_BYTES ((VP_VCP2_MAX_NUM_LINES + 7) / 8)

/* Utility macros: */
#define MAX(a, b) ((a) > (b) ? (a) : (b))
#define MIN(a, b) ((a) < (b) ? (a) : (b))

/* Size (in words) of the various mailbox fields: */
#define HEADER_FLD          2
#define PROFILE_FLD(type)   + (1 + VCP2_ ## type ## _PROFILE_LEN)
#define ARG_FLD             + 1
#define HANDLE_FLD          + 1
#define CID_DATA_FLD        + ((MAX_CID_DATA_LENGTH + 1) / 2)
#define MPI_DATA_FLD        + (MAX_MPI_DATA_LENGTH / 2)
#define LLC16_CMD_DATA_FLD  + 123
#define LLC16_RSP_DATA_FLD  + 125
#define CID_CONT_FLD        + ((MAX_CID_CONTINUE_DATA_LENGTH + 1) / 2)

/* Calculate maximum mailbox space requirements (in words) for each command: */
#define BUFSIZE_INITDEVICE      (HEADER_FLD ARG_FLD PROFILE_FLD(DEV) PROFILE_FLD(AC) PROFILE_FLD(DC) PROFILE_FLD(RING))
#define BUFSIZE_CONFIGLINE      (HEADER_FLD PROFILE_FLD(AC) PROFILE_FLD(DC) PROFILE_FLD(RING))
#define BUFSIZE_CALCODEC        (HEADER_FLD ARG_FLD)
#define BUFSIZE_CALLINE         (HEADER_FLD)
#define BUFSIZE_INITRING        (HEADER_FLD PROFILE_FLD(RINGCAD) PROFILE_FLD(CID))
#define BUFSIZE_INITMETER       (HEADER_FLD PROFILE_FLD(MTR) PROFILE_FLD(CID))
#define BUFSIZE_SETBATTERIES    (HEADER_FLD ARG_FLD ARG_FLD*3)
#define BUFSIZE_FLUSHEVENTS     (HEADER_FLD)
#define BUFSIZE_GETLOOPCOND     (HEADER_FLD HANDLE_FLD)
#define BUFSIZE_SETLINESTATE    (HEADER_FLD ARG_FLD)
#define BUFSIZE_SETLINETONE     (HEADER_FLD PROFILE_FLD(TONE) PROFILE_FLD(TONECAD) ARG_FLD ARG_FLD)
#define BUFSIZE_SETRELAYSTATE   (HEADER_FLD ARG_FLD)
#define BUFSIZE_SETRELGAIN      (HEADER_FLD HANDLE_FLD ARG_FLD ARG_FLD)
#define BUFSIZE_SENDCID         (HEADER_FLD PROFILE_FLD(CID) ARG_FLD CID_DATA_FLD)
#define BUFSIZE_STARTMETER      (HEADER_FLD ARG_FLD ARG_FLD ARG_FLD)
#define BUFSIZE_SETOPTION       (HEADER_FLD ARG_FLD*MAX(8,(VP_VCP2_MAX_NUM_LINES+1)/2))
#define BUFSIZE_GETOPTION       (HEADER_FLD HANDLE_FLD ARG_FLD)
#define BUFSIZE_SELFTEST        (HEADER_FLD)
#define BUFSIZE_TESTLINE        (HEADER_FLD HANDLE_FLD ARG_FLD*16)
#define BUFSIZE_CODECHECKSUM    (HEADER_FLD HANDLE_FLD)
#define BUFSIZE_LOWLEVELCMD     (HEADER_FLD HANDLE_FLD ARG_FLD ARG_FLD MPI_DATA_FLD)
#define BUFSIZE_LOWLEVELCMD16   (HEADER_FLD HANDLE_FLD ARG_FLD ARG_FLD LLC16_CMD_DATA_FLD)
#define BUFSIZE_INITCID         (HEADER_FLD ARG_FLD CID_DATA_FLD)
#define BUFSIZE_CONTINUECID     (HEADER_FLD ARG_FLD CID_CONT_FLD)
#define BUFSIZE_INITLINE        (BUFSIZE_CONFIGLINE)
#define BUFSIZE_SENDSIGNAL      (HEADER_FLD ARG_FLD*5)
#define BUFSIZE_DEVICEIOACCESS  (HEADER_FLD ARG_FLD + ARG_FLD*(VP_VCP2_MAX_NUM_LINES+1)/2)
#define BUFSIZE_LINEIOACCESS    (HEADER_FLD ARG_FLD*3)
#define BUFSIZE_SLACBOOTLOAD    (HEADER_FLD ARG_FLD)

/* Calculate maximum mailbox space requirements (in words) for each response,
   not including the first two arguments (event type and response length): */
#define BUFSIZE_CHECKSUM_RESP       (HEADER_FLD ARG_FLD*6)
#define BUFSIZE_LOWLEVCMD_RESP      (HEADER_FLD ARG_FLD ARG_FLD*MAX(MPI_DATA_FLD, LLC16_RSP_DATA_FLD))
#define BUFSIZE_RDLOOPCOND_RESP     (HEADER_FLD ARG_FLD*10)
#define BUFSIZE_GAINCMP_RESP        (HEADER_FLD ARG_FLD*3)
#define BUFSIZE_RDOPTION_RESP       BUFSIZE_SETOPTION
#define BUFSIZE_DEVIOACCESS_RESP    (HEADER_FLD ARG_FLD ARG_FLD*(VP_VCP2_MAX_NUM_LINES+1)/2)
#define BUFSIZE_LINEIOACCESS_RESP   (HEADER_FLD ARG_FLD*2)
#define BUFSIZE_TESTCMP_RESP        (HEADER_FLD ARG_FLD*16)

/* Create shortcut macros for conveniently accessing HBI registers using
 * VpHbiWr() and VpHbiRd(): */
#define INTIND              VCP2_HBI_INTIND_PAG,VCP2_HBI_INTIND_OFF,VCP2_HBI_INTIND_LEN-1
#define INTPARAM            VCP2_HBI_INTPARAM_PAG,VCP2_HBI_INTPARAM_OFF,VCP2_HBI_INTPARAM_LEN-1
#define POFFSET             VCP2_HBI_POFFSET_PAG,VCP2_HBI_POFFSET_OFF,VCP2_HBI_POFFSET_LEN-1
#define SW_FLAGS            VCP2_HBI_SW_FLAGS_PAG,VCP2_HBI_SW_FLAGS_OFF,VCP2_HBI_SW_FLAGS_LEN-1
#define P255_CHECK          VCP2_HBI_P255_CHECK_PAG,VCP2_HBI_P255_CHECK_OFF,VCP2_HBI_P255_CHECK_LEN-1
#define P255_BASE           VCP2_HBI_P255_BASE_PAG,VCP2_HBI_P255_BASE_OFF,VCP2_HBI_P255_BASE_LEN-1
#define P255_BASE_HI        VCP2_HBI_P255_BASE_HI_PAG,VCP2_HBI_P255_BASE_HI_OFF,VCP2_HBI_P255_BASE_HI_LEN-1
#define PCLKSEL             VCP2_HBI_PCLKSEL_PAG,VCP2_HBI_PCLKSEL_OFF,VCP2_HBI_PCLKSEL_LEN-1
#define CLKGEN_STATUS       VCP2_HBI_CLKGEN_STATUS_PAG,VCP2_HBI_CLKGEN_STATUS_OFF,VCP2_HBI_CLKGEN_STATUS_LEN-1
#define CLKGEN_STATUS_HI    VCP2_HBI_CLKGEN_STATUS_HI_PAG,VCP2_HBI_CLKGEN_STATUS_HI_OFF,VCP2_HBI_CLKGEN_STATUS_HI_LEN-1
#define SYSINTSTAT          VCP2_HBI_SYSINTSTAT_PAG,VCP2_HBI_SYSINTSTAT_OFF,VCP2_HBI_SYSINTSTAT_LEN-1
#define SYSINTMASK          VCP2_HBI_SYSINTMASK_PAG,VCP2_HBI_SYSINTMASK_OFF,VCP2_HBI_SYSINTMASK_LEN-1
#define BOOT_SENSE          VCP2_HBI_BOOT_SENSE_PAG,VCP2_HBI_BOOT_SENSE_OFF,VCP2_HBI_BOOT_SENSE_LEN-1
#define RESET_TYPE          VCP2_HBI_RESET_TYPE_PAG,VCP2_HBI_RESET_TYPE_OFF,VCP2_HBI_RESET_TYPE_LEN-1
#define ENTRY_ADDR          VCP2_HBI_ENTRY_ADDR_PAG,VCP2_HBI_ENTRY_ADDR_OFF,VCP2_HBI_ENTRY_ADDR_LEN-1
#define LINE_STATUS(n)      VCP2_HBI_LINE_STATUS_PAG(n),VCP2_HBI_LINE_STATUS_OFF(n),VCP2_HBI_LINE_STATUS_LEN-1
#define DTMF_STATUS         VCP2_HBI_DTMF_STATUS_PAG,VCP2_HBI_DTMF_STATUS_OFF,VCP2_HBI_DTMF_STATUS_LEN-1
#define DTMF_FREE           VCP2_HBI_DTMF_FREE_PAG,VCP2_HBI_DTMF_FREE_OFF,VCP2_HBI_DTMF_FREE_LEN-1
#define RAMCODE_REV         VCP2_HBI_RAMCODE_REV_PAG,VCP2_HBI_RAMCODE_REV_OFF,VCP2_HBI_RAMCODE_REV_LEN-1
#define PRODUCT_ID          VCP2_HBI_PRODUCT_ID_PAG,VCP2_HBI_PRODUCT_ID_OFF,VCP2_HBI_PRODUCT_ID_LEN-1
#define TESTCHAN(n)         VCP2_HBI_TESTCHAN_PAG(n),VCP2_HBI_TESTCHAN_OFF(n),VCP2_HBI_TESTCHAN_LEN-1
#define TESTCURSOR(n)       VCP2_HBI_TESTCURSOR_PAG(n),VCP2_HBI_TESTCURSOR_OFF(n),VCP2_HBI_TESTCURSOR_LEN-1
#define TIMESTAMP           VCP2_HBI_TIMESTAMP_PAG,VCP2_HBI_TIMESTAMP_OFF,VCP2_HBI_TIMESTAMP_LEN-1
#define LINESTATE(n)        VCP2_HBI_LINESTATE_PAG(n),VCP2_HBI_LINESTATE_OFF(n),VCP2_HBI_LINESTATE_LEN-1
#define DEV_PROFILE(n)      DEV_PROFILE_PAG,DEV_PROFILE_OFF,DEV_PROFILE_LEN-1
#define DC_PROFILE(n)       DC_PROFILE_PAG(n),DC_PROFILE_OFF(n),DC_PROFILE_LEN-1
#define AC_PROFILE(n)       AC_PROFILE_PAG(n),AC_PROFILE_OFF(n),AC_PROFILE_LEN-1
#define CID_PROFILE(n)      CID_PROFILE_PAG(n),CID_PROFILE_OFF(n),CID_PROFILE_LEN-1
#define MTR_PROFILE(n)      MTR_PROFILE_PAG(n),MTR_PROFILE_OFF(n),MTR_PROFILE_LEN-1
#define TONE_PROFILE(n)     TONE_PROFILE_PAG(n),TONE_PROFILE_OFF(n),TONE_PROFILE_LEN-1
#define CAD_PROFILE(n)      CAD_PROFILE_PAG(n),CAD_PROFILE_OFF(n),CAD_PROFILE_LEN-1
#define RING_PROFILE(n)     RING_PROFILE_PAG(n),RING_PROFILE_OFF(n),RING_PROFILE_LEN-1
#define TERMTYPE(n)         VCP2_HBI_TERMTYPE_PAG(n),VCP2_HBI_TERMTYPE_OFF(n),VCP2_HBI_TERMTYPE_LEN-1

/* Compose a set of #defines indicating whether the user-selected default
   options in vp_api_cfg.h equal the device's defaults: */
#define CUSTOM_PULSE_DEFAULT ( \
    (VCP2_DFLT_PULSE_BREAK_MIN != VP_OPTION_DEFAULT_DP_BREAK_MIN) \
    || (VCP2_DFLT_PULSE_BREAK_MAX != VP_OPTION_DEFAULT_DP_BREAK_MAX) \
    || (VCP2_DFLT_PULSE_MAKE_MIN != VP_OPTION_DEFAULT_DP_MAKE_MIN) \
    || (VCP2_DFLT_PULSE_MAKE_MAX != VP_OPTION_DEFAULT_DP_MAKE_MAX) \
    || (VCP2_DFLT_PULSE_INTER_DIG_MIN != VP_OPTION_DEFAULT_DP_INTER_DIG_MIN) \
    || (VCP2_DFLT_PULSE_FLASH_MIN != VP_OPTION_DEFAULT_DP_FLASH_MIN) \
    || (VCP2_DFLT_PULSE_FLASH_MAX != VP_OPTION_DEFAULT_DP_FLASH_MAX))
#define CUSTOM_CRITICAL_FLT_DEFAULT ( \
    (VCP2_DFLT_CRITICAL_FLT_AC_DIS_EN != VP_OPTION_DEFAULT_CF_AC_DIS_EN) \
    || (VCP2_DFLT_CRITICAL_FLT_DC_DIS_EN != VP_OPTION_DEFAULT_CF_DC_DIS_EN) \
    || (VCP2_DFLT_CRITICAL_FLT_THERM_DIS_EN != VP_OPTION_DEFAULT_CF_THERMAL_DIS_EN))
#define CUSTOM_ZERO_CROSS_DEFAULT \
    (VCP2_DFLT_ZERO_CROSS != VP_OPTION_DEFAULT_ZERO_CROSS)
#define CUSTOM_RAMP2STBY_DEFAULT \
    (VP_OPTION_DEFAULT_RAMP2STBY != VCP2_DFLT_RAMP2STBY)
#define CUSTOM_PULSE_MODE_DEFAULT \
    (VP_OPTION_DEFAULT_PULSE_MODE != VCP2_DFLT_PULSE_MODE)
#define CUSTOM_CODEC_DEFAULT \
    (VP_OPTION_DEFAULT_CODEC_MODE != VCP2_DFLT_CODEC)
#define CUSTOM_PCM_HWY_DEFAULT \
    (VP_OPTION_DEFAULT_PCM_HWY != VCP2_DFLT_PCM_HWY)
#define CUSTOM_LOOPBACK_DEFAULT \
    (VP_OPTION_DEFAULT_LOOP_BACK != VCP2_DFLT_LOOPBACK)
#define CUSTOM_LINE_STATE_DEFAULT ( \
    (VP_OPTION_DEFAULT_LS_BAT_REV != VCP2_DFLT_LINE_STATE_BATTREV) \
    || (VP_OPTION_DEFAULT_LS_BAT != VCP2_DFLT_LINE_STATE_BAT))
#define CUSTOM_EVENT_MASK_DEFAULT ( \
    (VP_OPTION_DEFAULT_FAULT_EVENT_MASK != VCP2_DFLT_EVENT_MASK_FAULT) \
    || (VP_OPTION_DEFAULT_SIGNALING_EVENT_MASK != VCP2_DFLT_EVENT_MASK_SIGNALING) \
    || (VP_OPTION_DEFAULT_RESPONSE_EVENT_MASK != VCP2_DFLT_EVENT_MASK_RESPONSE) \
    || (VP_OPTION_DEFAULT_TEST_EVENT_MASK != VCP2_DFLT_EVENT_MASK_TEST) \
    || (VP_OPTION_DEFAULT_PROCESS_EVENT_MASK != VCP2_DFLT_EVENT_MASK_PROCESS) \
    || (VP_OPTION_DEFAULT_FXO_EVENT_MASK != VCP2_DFLT_EVENT_MASK_FXO) \
    || (VP_OPTION_DEFAULT_PACKET_EVENT_MASK != VCP2_DFLT_EVENT_MASK_PACKET))
#define CUSTOM_RING_CNTRL_DEFAULT ( \
    (VP_OPTION_DEFAULT_RC_ZERO_CROSS != VCP2_DFLT_RING_CNTRL_ZEROCROSS) \
    || (VP_OPTION_DEFAULT_RC_RING_EXIT_DBNC_VAL != VCP2_DFLT_RING_CNTRL_RINGEXIT_DBNC) \
    || (VP_OPTION_DEFAULT_RC_RING_EXIT_STATE != VCP2_DFLT_RING_CNTRL_RINGEXIT_STATE))
#define CUSTOM_DTMF_MODE_DEFAULT ( \
    (VP_OPTION_DEFAULT_DM_DTMF_CONTROL_MODE != VCP2_DFLT_DTMF_MODE_CONTROL) \
    || (VP_OPTION_DEFAULT_DM_DIRECTION != VCP2_DFLT_DTMF_MODE_DIRECTION))
#define CUSTOM_PCM_TXRX_CNTRL_DEFAULT \
    (VP_OPTION_DEFAULT_PCM_TXRX_CNTRL != VCP2_DFLT_PCM_TXRX_CNTRL)

/* Byte offsets for header fields in a VCP2-VP792 firmware image: */
#define VCP2_792_IMAGE_ID            0
#define VCP2_792_IMAGE_FILE_VER      8
#define VCP2_792_IMAGE_MIN_API_VER   10
#define VCP2_792_IMAGE_LENGTH        12
#define VCP2_792_IMAGE_VCP2_START    18
#define VCP2_792_IMAGE_VCP2_LENGTH   20
#define VCP2_792_IMAGE_VCP2_MIN_VER  24
#define VCP2_792_IMAGE_VCP2_MAX_VER  26
#define VCP2_792_IMAGE_792_START     34
#define VCP2_792_IMAGE_792_LENGTH    36
#define VCP2_792_IMAGE_792_MIN_VER   40
#define VCP2_792_IMAGE_792_MAX_VER   42

/* Number of bits a VCP2 timestamp has to be right-shifted by, in order to
   match the timestamp resolution that the API exposes.  Currently VCP2
   timestamp granularity is 0.125 ms, API timestamp granularity is .5 ms, so
   the shift is 2 bits.  This is also the number of high-order bits that the
   API has to maintain in the device object. */
#define VCP2_TIMESTAMP_SHIFT    (2)


/* =============================================
    Prototypes for Top-Level VCP2_API Functions
   ============================================= */

/* System Configuration functions (vcp2_config.c): */

VpStatusType
Vcp2MakeLineObject(
    VpTermType termType,
    uint8 channelId,
    VpLineCtxType *pLineCtx,
    void *pLineObj,
    VpDevCtxType *pDevCtx);

/* Control functions (vcp2_control.c): */

EXTERN VpStatusType
Vcp2ContinueCid(
    VpLineCtxType *pLineCtx,
    uint8 length,
    uint8p pCidData);

EXTERN VpStatusType
Vcp2DeviceIoAccess(
    VpDevCtxType *pDevCtx,
    VpDeviceIoAccessDataType *pDeviceIoData);

EXTERN VpStatusType
Vcp2DeviceIoAccessExt(
    VpDevCtxType *pDevCtx,
    VpDeviceIoAccessExtType *pDeviceIoAccess);

EXTERN VpStatusType
Vcp2FillTestBuf(
    VpLineCtxType *pLineCtx,
    uint16 length,
    VpVectorPtrType pData);

EXTERN VpStatusType
Vcp2LineIoAccess(
    VpLineCtxType *pLineCtx,
    VpLineIoAccessType *pLineIoAccess,
    uint16 handle);

EXTERN VpStatusType
Vcp2LowLevelCmd(
    VpLineCtxType *pLineCtx,
    uint8p pMpiCmdData,
    uint8 length,
    uint16 handle);

EXTERN VpStatusType
Vcp2LowLevelCmd16(
    VpLineCtxType *pLineCtx,
    VpLowLevelCmdType cmdType,
    uint16 *writeWords,
    uint8 numWriteWords,
    uint8 numReadWords,
    uint16 handle);

EXTERN VpStatusType
Vcp2SendCid(
    VpLineCtxType *pLineCtx,
    uint8 length,
    VpProfilePtrType pCidProfile,
    uint8p pCidData);

EXTERN VpStatusType
Vcp2SendSignal(
    VpLineCtxType *pLineCtx,
    VpSendSignalType sigType,
    void *pStruct);

EXTERN VpStatusType
Vcp2SetLineState(
    VpLineCtxType *pLineCtx,
    VpLineStateType state);

EXTERN VpStatusType
Vcp2SetLineTone(
    VpLineCtxType *pLineCtx,
    VpProfilePtrType pToneProfile,
    VpProfilePtrType pToneCadProfile,
    VpDtmfToneGenType *pDtmfControl);

EXTERN VpStatusType
Vcp2SetOption(
    VpLineCtxType *pLineCtx,
    VpDevCtxType *pDevCtx,
    VpOptionIdType optionId,
    void *value);

EXTERN VpStatusType
Vcp2SetRelayState(
    VpLineCtxType *pLineCtx,
    VpRelayControlType rState);

EXTERN VpStatusType
Vcp2SetRelGain(
    VpLineCtxType *pLineCtx,
    uint16 txLevel,
    uint16 rxLevel,
    uint16 handle);

EXTERN VpStatusType
Vcp2SelfTest(
    VpLineCtxType *pLineCtx);

EXTERN VpStatusType
Vcp2StartMeter(
    VpLineCtxType *pLineCtx,
    uint16 onTime,
    uint16 offTime,
    uint16 numMeters);

/* Initialization functions (vcp2_init.c): */

EXTERN VpStatusType
Vcp2BootLoad(
    VpDevCtxType *pDevCtx,
    VpBootStateType state,
    VpImagePtrType pImageBuffer,
    uint32 bufferSize,
    VpScratchMemType *pScratchMem,
    VpBootModeType validation);

EXTERN VpStatusType
Vcp2CalCodec(
    VpLineCtxType *pLineCtx,
    VpDeviceCalType mode);

EXTERN VpStatusType
Vcp2CalLine(
    VpLineCtxType *pLineCtx);

EXTERN VpStatusType
Vcp2ConfigLine(
    VpLineCtxType *pLineCtx,
    VpProfilePtrType pAcProfile,
    VpProfilePtrType pDcProfile,
    VpProfilePtrType pRingProfile);

EXTERN VpStatusType
Vcp2InitCid(
    VpLineCtxType *pLineCtx,
    uint8 length,
    uint8p pCidData);

EXTERN VpStatusType
Vcp2InitDevice(
    VpDevCtxType *pDevCtx,
    VpProfilePtrType pDevProfile,
    VpProfilePtrType pAcProfile,
    VpProfilePtrType pDcProfile,
    VpProfilePtrType pRingProfile,
    VpProfilePtrType pFxoAcProfile,
    VpProfilePtrType pFxoCfgProfile);

EXTERN VpStatusType
Vcp2InitLine(
    VpLineCtxType *pLineCtx,
    VpProfilePtrType pAcProfile,
    VpProfilePtrType pDcProfile,
    VpProfilePtrType pRingProfile);

EXTERN VpStatusType
Vcp2InitMeter(
    VpLineCtxType *pLineCtx,
    VpProfilePtrType pMeterProfile);

EXTERN VpStatusType
Vcp2InitProfile(
    VpDevCtxType *pDevCtx,
    VpProfileType type,
    VpProfilePtrType pProfileIndex,
    VpProfilePtrType pProfile);

EXTERN VpStatusType
Vcp2InitRing(
    VpLineCtxType *pLineCtx,
    VpProfilePtrType pRingCadProfile,
    VpProfilePtrType pCidProfile);

EXTERN VpStatusType
Vcp2SetBatteries(
    VpLineCtxType *pLineCtx,
    VpBatteryModeType battMode,
    VpBatteryValuesType *pBatt);

EXTERN VpStatusType
Vcp2SoftReset(
    VpDevCtxType *pDevCtx);

/* Status and Query functions (vcp2_query.c): */

EXTERN VpStatusType
Vcp2ClearResults(
    VpDevCtxType *pDevCtx);

EXTERN VpStatusType
Vcp2CodeCheckSum(
    VpDevCtxType *pDevCtx,
    uint16 handle);

EXTERN VpStatusType
Vcp2FlushEvents(
    VpDevCtxType *pDevCtx);

EXTERN bool
Vcp2GetEvent(
    VpDevCtxType *pDevCtx,
    VpEventType *pEvent);

EXTERN VpStatusType
Vcp2GetDeviceStatus(
    VpDevCtxType *pDevCtx,
    VpInputType input,
    uint32 *pDevStatus);

EXTERN VpStatusType
Vcp2GetDeviceStatusExt(
    VpDevCtxType *pDevCtx,
    VpDeviceStatusType *pDeviceStatus);

EXTERN VpStatusType
Vcp2GetLineState(
    VpLineCtxType *pLineCtx,
    VpLineStateType *pCurrentState);

EXTERN VpStatusType
Vcp2GetLineStatus(
    VpLineCtxType *pLineCtx,
    VpInputType input,
    bool *pStatus);

EXTERN VpStatusType
Vcp2GetLoopCond(
    VpLineCtxType *pLineCtx,
    uint16 handle);

EXTERN VpStatusType
Vcp2GetOption(
    VpLineCtxType *pLineCtx,
    VpDevCtxType *pDevCtx,
    VpOptionIdType option,
    uint16 handle);

EXTERN VpStatusType
Vcp2GetResults(
    VpEventType *pEvent,
    void *pResults);

/* Line testing functions (vcp2_test.c): */

EXTERN VpStatusType
Vcp2TestLine(
    VpLineCtxType *pLineCtx,
    VpTestIdType testId,
    const void *pArgs,
    uint16 handle);


/* =============================================
    Prototypes for Internal VCP2_API Functions
   ============================================= */

/* Support functions (vcp2_common.c): */

EXTERN bool
Vcp2BadProfileArg(
    VpVcp2DeviceObjectType *pDevObj,
    VpProfileType profileType,
    VpProfilePtrType pProfile);

EXTERN bool
Vcp2DeviceOptionSupported(
    VpVcp2DeviceObjectType *pDevObj,
    VpOptionIdType optionId);

EXTERN bool
Vcp2LineOptionSupported(
    VpVcp2DeviceObjectType *pDevObj,
    VpOptionIdType optionId);

EXTERN void
Vcp2MailboxAddBytes(
    uint16 **ppBuf,
    const uint8 *pData,
    int length);

EXTERN void
Vcp2MailboxAddProfile(
    uint16 **ppBuf,
    VpProfileType profileType,
    VpProfilePtrType pProfile);

EXTERN void
Vcp2MailboxInit(
    uint16 **ppBuf,
    MboxCmdIdType mailBoxCmd,
    uint8 channelId);

EXTERN VpStatusType
Vcp2MailboxSend(
    VpDeviceIdType deviceId,
    uint16 *buf,
    uint16 *pBuf);

EXTERN void
Vcp2LocateProfileTableEntry(
    int index,
    VpProfileType profileType,
    int *pPage,
    int *pOffset);

EXTERN uint8
Vcp2ProfileTableSize(
    VpProfileType profileType);

EXTERN VpStatusType
Vcp2ReadDtmfStatusRegs(
    VpDeviceIdType deviceId,
    VpOptionDtmfModeType *pDtmfControl,
    bool waitForMbox);

EXTERN void
Vcp2ResetDeviceContext(
    VpDevCtxType *pDevCtx);

EXTERN void
Vcp2ResetDeviceObject(
    VpVcp2DeviceObjectType *pDevObj);

#endif /* _VCP2_API_INT_H */
