/** \file vcp2_common.c
 * vcp2_common.c
 *
 * This file contains support functions used by other functions in the
 * VCP2_API library.  If the function appears in this file, it is called from
 * at least two other modules in the VCP2_API library.
 *
 * The functions herein are used internally by the API and should not be
 * called directly by the application.
 *
 * Copyright (c) 2008, Zarlink Semiconductor, Inc.
 */

/* Only compile this file if specified in vp_api_cfg.h: */
#include "vp_api_cfg.h"
#ifdef VP_CC_VCP2_SERIES

#include "vcp2_api.h"
#include "vp_api_int.h"
#include "vcp2_api_int.h"
#include "boot_common.h"
#include "vp_debug.h"


/* =================================
    Prototypes for Static Functions
   ================================= */

static VpStatusType
CmdMailboxAcquire(
    VpDeviceIdType deviceId);

static bool
ProfileTableEntryEmpty(
    VpVcp2DeviceObjectType *pDevObj,
    VpProfileType profileType,
    VpProfilePtrType pProfileIndex);

static bool
VerifyProfileLength(
    VpProfileType profileType,
    VpProfilePtrType pProfile);


/* ===================
    Support Functions
   =================== */

/**
 * This internal function determines whether the supplied profile argument
 * is valid.  If the pProfile argument is NULL, this function returns TRUE.
 * If the pProfile argument contains a valid Profile Table index, this
 * function returns TRUE.  Otherwise this function returns TRUE only if the
 * type argument matches the type of the profile struct itself.
 *
 * \param[in] pDevObj       Pointer to the Device Object
 * \param[in] profileType   Type of profile expected
 * \param[in] pProfile      Profile argument to be verified
 *
 * \retval  ::TRUE          Profile argument is invalid
 * \retval  ::FALSE         Profile argument is valid
 */
bool
Vcp2BadProfileArg(
    VpVcp2DeviceObjectType *pDevObj,
    VpProfileType profileType,
    VpProfilePtrType pProfile)
{

    if (pProfile == VP_PTABLE_NULL) {
        /* No profile argument specified. */
        return FALSE;
    } else if ((pProfile >= VP_PTABLE_INDEX1) && (pProfile <= VP_PTABLE_INDEX15)) {
        /* Profile Table index given. */
        if (ProfileTableEntryEmpty(pDevObj, profileType, pProfile)) {
            VP_DOUT(VP_DBG_ERROR, ("Vcp2BadProfileArg(): Profile table entry empty\n"));
            return TRUE;
        } else {
            return FALSE;
        }
    } else {
        /* Pointer to profile given. */
        if (VpVerifyProfileType(profileType, pProfile) && VerifyProfileLength(profileType, pProfile)) {
            /* Profile is correct type and length. */
            return FALSE;
        } else {
            VP_DOUT(VP_DBG_ERROR, ("Vcp2BadProfileArg(): Bad profile %s\n", VpVerifyProfileType(profileType, pProfile) ? "length" : "type"));
            return TRUE;
        }
    }
}

/**
 * This internal function verifies that the selected option ID is supported
 * by the specified VCP2 device, based on the firmware product type, and that
 * the option is line-specific (as opposed to device-specific).
 *
 * \param[in] pDevObj   Device Object of the VCP2 device
 * \param[in] optionId  Option ID to be verified
 *
 * \retval ::TRUE   Option is line-specific and supported.
 * \retval ::FALSE  Option is not supported by firmware product type, or is
 *                  device-specific.
 */
bool
Vcp2LineOptionSupported(
    VpVcp2DeviceObjectType *pDevObj,
    VpOptionIdType optionId)
{

    /* Currently all VCP2 firmware product types support the same options. */
    if (
        (optionId > VP_OPTION_ID_DTMF_SPEC) ||
        (optionId == VP_OPTION_ID_RESERVED_1) ||
        (optionId == VP_OPTION_ID_RESERVED_2) ||
        (optionId == VP_DEVICE_OPTION_ID_PULSE2)
    ) {
        /* These options are not understood by the VCP2. */
        return FALSE;
    }

    return !Vcp2DeviceOptionSupported(pDevObj, optionId);
}

/**
 * This internal function calculates the HBI page and offset for the requested
 * Profile Table entry, based on the profile type and index.
 *
 * \note There is no error checking in this function.  Arguments are presumed
 *       valid.
 *
 * \param[in]   index       Profile Table index
 * \param[in]   profileType Type of profile
 * \param[out]  pPage       Page of the requested Profile Table entry
 * \param[out]  pOffset     Offset of the requested Profile Table entry
 */
void
Vcp2LocateProfileTableEntry(
    int index,
    VpProfileType profileType,
    int *pPage,
    int *pOffset)
{
    int page = VCP2_DEV_PROFILE_PAG(index);
    int offset = VCP2_DEV_PROFILE_OFF(index);

    switch (profileType) {
        case VP_PROFILE_AC:
            page = VCP2_AC_PROFILE_PAG(index);
            offset = VCP2_AC_PROFILE_OFF(index);
            break;
        case VP_PROFILE_DC:
            page = VCP2_DC_PROFILE_PAG(index);
            offset = VCP2_DC_PROFILE_OFF(index);
            break;
        case VP_PROFILE_RING:
            page = VCP2_RING_PROFILE_PAG(index);
            offset = VCP2_RING_PROFILE_OFF(index);
            break;
        case VP_PROFILE_RINGCAD:
            page = VCP2_RINGCAD_PROFILE_PAG(index);
            offset = VCP2_RINGCAD_PROFILE_OFF(index);
            break;
        case VP_PROFILE_TONE:
            page = VCP2_TONE_PROFILE_PAG(index);
            offset = VCP2_TONE_PROFILE_OFF(index);
            break;
        case VP_PROFILE_METER:
            page = VCP2_MTR_PROFILE_PAG(index);
            offset = VCP2_MTR_PROFILE_OFF(index);
            break;
        case VP_PROFILE_CID:
            page = VCP2_CID_PROFILE_PAG(index);
            offset = VCP2_CID_PROFILE_OFF(index);
            break;
        case VP_PROFILE_TONECAD:
            page = VCP2_TONECAD_PROFILE_PAG(index);
            offset = VCP2_TONECAD_PROFILE_OFF(index);
        default:
            /* No error checking here. */
            break;
    }

    *pPage = page;
    *pOffset = offset;
}

/**
 * This internal function verifies that the selected option ID is supported
 * by the specified VCP2 device, based on the firmware product type, and that
 * the option is device-specific (as opposed to line-specific).
 *
 * \param[in] pDevObj   Device Object of the VCP2 device
 * \param[in] optionId  Option ID to be verified
 *
 * \retval ::TRUE   Option is device-specific and supported.
 * \retval ::FALSE  Option is not supported by firmware product type, or is
 *                  line-specific.
 */
bool
Vcp2DeviceOptionSupported(
    VpVcp2DeviceObjectType *pDevObj,
    VpOptionIdType optionId)
{

    /* Currently all VCP2 firmware product types support the same options. */

    if (
        (optionId == VP_DEVICE_OPTION_ID_PULSE) ||
        (optionId == VP_DEVICE_OPTION_ID_CRITICAL_FLT) ||
        (optionId == VP_DEVICE_OPTION_ID_RAMP2STBY) ||
        (optionId == VP_DEVICE_OPTION_ID_DEVICE_IO) ||
        (optionId == VP_DEVICE_OPTION_ID_DEV_IO_CFG)
     ) {

        /* These are device-specific options in the VCP2 device. */
        return TRUE;
    }

    /* The requested option ID is line-specific or unsupported. */
    return FALSE;
}

/**
 * This function adds a string of bytes to the command mailbox buffer.
 * If the length is odd, it pads the string with an extra 0 byte.
 *
 * \param[in,out] ppBuf  Pointer to the calling function's pBuf variable
 * \param[in]     pData  Bytestring to be added to the Device Object's mailbox
 *                       buffer.
 * \param[in]     length Length of the bytestring (in bytes)
 */
void
Vcp2MailboxAddBytes(
    uint16 **ppBuf,
    const uint8 *pData,
    int length)
{
    uint16 byte0, byte1, word;
    uint16 *pBuf = *ppBuf;

    while (length > 0) {
        byte0 = *(pData++);

        /* If length is odd, add zero byte padding. */
        if (length == 1) {
            byte1 = 0;
        } else {
            byte1 = *(pData++);
        }

        /* VCP2 is big-endian.  Pack bytes accordingly. */
        word = (byte0 << 8) | byte1;

        /* Add word to command mailbox buffer. */
        *pBuf++ = word;

        length -= 2;
    }

    *ppBuf = pBuf;
}

/**
 * This function adds a profile argument to the specified command mailbox
 * buffer.
 *
 * \param[in] ppBuf         Pointer to the calling function's pBuf variable
 * \param[in] profileType   Type of the profile argument passed in pProfile.
 * \param[in] pProfile      Profile argument to be added to the buffer.
 */
void
Vcp2MailboxAddProfile(
    uint16 **ppBuf,
    VpProfileType profileType,
    VpProfilePtrType pProfile)
{
    uint16 *pBuf = *ppBuf;

    if (pProfile == VP_PTABLE_NULL) {

        /* No profile argument specified. */
        *pBuf++ = VCP2_HBI_PROFILE_OMITTED;
    } else if (
        (pProfile >= VP_PTABLE_INDEX1) &&
        (pProfile <= VP_PTABLE_INDEX15)
    ) {

        /* Argument is a Profile Table index.  Locate the entry. */
        int page, offset, idx = VpGetProfileIndex(pProfile);
        Vcp2LocateProfileTableEntry(idx, profileType, &page, &offset);

        /* Notify VCP2 of the page and offset where it can find the profile. */
        *pBuf++ = VCP2_HBI_PROFILE_IN_TABLE;
        *pBuf++ = VCP2_HBI_PROFILE_LOCATION(page, offset);
    } else {

        /* Pointer to profile given. */
        int lenBytes = pProfile[VP_PROFILE_LENGTH] + 4;

        /* Copy the complete profile into the mailbox. */
        *pBuf++ = VCP2_HBI_PROFILE_FOLLOWS;
        Vcp2MailboxAddBytes(&pBuf, (uint8p)pProfile, lenBytes);
    }

    /* Update the calling function's mailbox buffer pointer. */
    *ppBuf = pBuf;
}

/**
 * Inserts a standard command header, composed of the specified arguments,
 * into the specified mailbox buffer.
 *
 * \param[in] ppBuf         Pointer to the calling function's pBuf variable
 * \param[in] mailBoxCmd    Mailbox command ID
 * \param[in] channelId     Channel on which the command is to be performed
 */
void
Vcp2MailboxInit(
    uint16 **ppBuf,
    MboxCmdIdType mailBoxCmd,
    uint8 channelId)
{
    uint16 *pBuf = *ppBuf;

    /* Add a placeholder for the command length.  This will be filled in by
       Vcp2MailboxSend(). */
    *pBuf++ = 0;

    /* Add the command ID and channel ID to the mailbox buffer. */
    *pBuf++ = ((uint16)mailBoxCmd << 8) | channelId;

    /* Update the calling function's mailbox buffer pointer. */
    *ppBuf = pBuf;
}

/**
 * This function sends the contents of the supplied mailbox buffer
 * to the VCP2 and relinquishes control of the mailbox to the VCP2.
 *
 * \note This function must be called from within an HBI critical section.
 *
 * \param[in] deviceId  Device ID of the VCP2 device
 * \param[in] buf       Mailbox buffer
 * \param[in] pBuf      Pointer to the end of the mailbox buffer
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2MailboxSend(
    VpDeviceIdType deviceId,
    uint16 *buf,
    uint16 *pBuf)
{
    uint16 length = (uint16)(pBuf - buf);
    uint16 flag = VCP2_HBI_SWFLAG_CMD_MBOX;
    VpStatusType retval;

    /* Fill in the payload length field in the mailbox header. */
    buf[VCP2_HBI_MBOX_LENGTH] = length - HEADER_FLD;

    /* Acquire exclusive access to the command mailbox. */
    retval = CmdMailboxAcquire(deviceId);
    if (retval != VP_STATUS_SUCCESS) {
        return retval;
    }

    /* Dump the mailbox buffer to the VCP2's command mailbox. */
    if (VpHbiWr(deviceId, VCP2_HBI_CMD_MBOX_PAG, VCP2_HBI_CMD_MBOX_OFF(0),
        length - 1, buf)
        == FALSE) {
        return VP_STATUS_ERR_HBI;
    }

    /* Release control of the mailbox to the VCP2. */
    if (VpHbiWr(deviceId, SW_FLAGS, &flag) == FALSE) {
        return VP_STATUS_ERR_HBI;
    }

    return VP_STATUS_SUCCESS;
}

/**
 * This internal function returns the number of slots in the Profile Table
 * for a given type of profile.
 *
 * \param[in] type      Type of profile expected
 *
 * \returns
 * Maximum Profile Table index, plus 1.
 *
 */
uint8
Vcp2ProfileTableSize(
    VpProfileType profileType)
{
    switch (profileType) {
        case VP_PROFILE_DEVICE:
            return VCP2_DEV_PROFILE_COUNT;
        case VP_PROFILE_DC:
            return VCP2_DC_PROFILE_COUNT;
        case VP_PROFILE_AC:
            return VCP2_AC_PROFILE_COUNT;
        case VP_PROFILE_CID:
            return VCP2_CID_PROFILE_COUNT;
        case VP_PROFILE_METER:
            return VCP2_MTR_PROFILE_COUNT;
        case VP_PROFILE_TONE:
            return VCP2_TONE_PROFILE_COUNT;
        case VP_PROFILE_RINGCAD:
            return VCP2_RINGCAD_PROFILE_COUNT;
        case VP_PROFILE_TONECAD:
            return VCP2_TONECAD_PROFILE_COUNT;
        case VP_PROFILE_RING:
            return VCP2_RING_PROFILE_COUNT;
        /* Other profile types are not supported by the VCP2. */
        default:
            return 0;
    }
}

/**
 * This internal function reads the VCP2's DTMF status registers into the
 * supplied VpOptionDtmfModeType struct.  If the waitForMbox argument is TRUE,
 * this function waits for ownership of the command mailbox before reading
 * the registers, to be sure all previous VpSetOption(DTMF_MODE) commands have
 * been processed.
 *
 * \param[in]   deviceId      Device ID for the VCP2 device
 * \param[in]   pDtmfControl  VP_OPTION_ID_DTMF_MODE arguments
 * \param[in]   waitForMbox   Whether to wait for the command mailbox
 */
VpStatusType
Vcp2ReadDtmfStatusRegs(
    VpDeviceIdType deviceId,
    VpOptionDtmfModeType *pDtmfControl,
    bool waitForMbox)
{
    uint16 dtmfStatus[(VCP2_LINE_FLAG_BYTES + 1) / 2];
    int byte;

    /* Must not read the direct page DTMF status registers until the mailbox
       is available.  (Bug fix #372.)  */
    if (waitForMbox) {
        VpStatusType retval = CmdMailboxAcquire(deviceId);
        if (retval != VP_STATUS_SUCCESS) {
            return retval;
        }
    }

    /* Read DTMF resources register. */
    if (VpHbiRd(deviceId, DTMF_FREE, dtmfStatus) == FALSE) {
        return VP_STATUS_ERR_HBI;
    }

    /* Copy resources register into user's struct. */
    pDtmfControl->dtmfResourcesRemaining = (uint8)(dtmfStatus[0] & 0xFF);

    /* Read DTMF status registers. */
    if (VpHbiRd(deviceId, DTMF_STATUS, dtmfStatus) == FALSE) {
        return VP_STATUS_ERR_HBI;
    }

    /* Copy status registers into user's struct. */
    for (byte = 0; byte < VCP2_LINE_FLAG_BYTES; byte++) {
        if (byte % 2) {
            pDtmfControl->dtmfDetectionEnabled[byte] = (uint8)(dtmfStatus[byte / 2] >> 8);
        } else {
            pDtmfControl->dtmfDetectionEnabled[byte] = (uint8)(dtmfStatus[byte / 2] & 0xFF);
        }
    }

    return VP_STATUS_SUCCESS;
}

/**
 * This internal function sets the function pointers in the supplied VCP2
 * Device Context, based on the product type specified in the supplied Device
 * Object.
 *
 * \param[in] pDevCtx   Pointer to the Device Context struct to be updated
 */
void
Vcp2ResetDeviceContext(
    VpDevCtxType *pDevCtx)
{
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpVcp2ProductType product = pDevObj->product;
    ApiFunctions *pFunc = &pDevCtx->funPtrsToApiFuncs;

    /* Clear the function pointers array, so that any pointers that are not
       explicitly set below will return VP_STATUS_FUNC_NOT_SUPPORTED. */
    VpMemSet(pFunc, 0x00, sizeof(ApiFunctions));

    /* Functions which are supported by all VCP2 product types: */
    pFunc->MakeLineObject = Vcp2MakeLineObject;
    pFunc->BootLoad = Vcp2BootLoad;
    pFunc->InitDevice = Vcp2InitDevice;
    pFunc->InitLine = Vcp2InitLine;
    pFunc->ConfigLine = Vcp2ConfigLine;
    pFunc->CalCodec = Vcp2CalCodec;
    pFunc->CalLine = Vcp2CalLine;
    pFunc->InitRing = Vcp2InitRing;
    pFunc->InitCid = Vcp2InitCid;
    pFunc->InitMeter = Vcp2InitMeter;
    pFunc->InitProfile = Vcp2InitProfile;
    pFunc->SoftReset = Vcp2SoftReset;
    pFunc->SetLineState = Vcp2SetLineState;
    pFunc->SetLineTone = Vcp2SetLineTone;
    pFunc->SetRelayState = Vcp2SetRelayState;
    pFunc->SetRelGain = Vcp2SetRelGain;
    pFunc->SendSignal = Vcp2SendSignal;
    pFunc->SendCid = Vcp2SendCid;
    pFunc->ContinueCid = Vcp2ContinueCid;
    pFunc->StartMeter = Vcp2StartMeter;
    pFunc->SetOption = Vcp2SetOption;
    pFunc->DeviceIoAccess = Vcp2DeviceIoAccess;
    pFunc->DeviceIoAccessExt = Vcp2DeviceIoAccessExt;
    pFunc->LineIoAccess = Vcp2LineIoAccess;
    pFunc->GetLineState = Vcp2GetLineState;
    pFunc->LowLevelCmd = Vcp2LowLevelCmd;
    pFunc->LowLevelCmd16 = Vcp2LowLevelCmd16;
    pFunc->GetEvent = Vcp2GetEvent;
    pFunc->GetLineStatus = Vcp2GetLineStatus;
    pFunc->GetDeviceStatus = Vcp2GetDeviceStatus;
    pFunc->GetDeviceStatusExt = Vcp2GetDeviceStatusExt;
    pFunc->GetLoopCond = Vcp2GetLoopCond;
    pFunc->GetOption = Vcp2GetOption;
    pFunc->FlushEvents = Vcp2FlushEvents;
    pFunc->GetResults = Vcp2GetResults;
    pFunc->ClearResults = Vcp2ClearResults;
    pFunc->CodeCheckSum = Vcp2CodeCheckSum;
    pFunc->SelfTest = Vcp2SelfTest;

    if ((product & VP_VCP_PROD_FEAT_MASK) != VP_VCP_PROD_FEAT_NT) {

        /* Line-testing functions: */
        pFunc->TestLine = Vcp2TestLine;
        pFunc->FillTestBuf = Vcp2FillTestBuf;
    }

    if ((product & VP_VCP_PROD_SLAC_MASK) == VP_VCP_PROD_SLAC_VE792) {

        /* Functions which are supported only by VCP2-792: */
        pFunc->SetBatteries = Vcp2SetBatteries;
    }
}

/**
 * This internal function initializes the supplied VCP2 Device Object struct
 * by setting all members to default values.
 *
 * \param[in] pDevObj   Pointer to the Device Object struct to be initialized
 */
void
Vcp2ResetDeviceObject(
    VpVcp2DeviceObjectType *pDevObj)
{

    /* Initialize variables needed for argument-checking future API calls. */
    int i = VP_VCP2_NUM_PROFILE_TYPES;
    while (i) {
        pDevObj->profTabInit[--i] = 0;
    }
    pDevObj->slacsFound = 0;

    /* Initialize other Device Object members. */
    pDevObj->devInit = FALSE;
    pDevObj->product = VP_VCP_PROD_UNKNOWN;
    pDevObj->numChannels = VP_VCP2_MAX_NUM_LINES;
    pDevObj->recCnt = 0;

    /* Note: pDevObj->deviceId is initialized in VpMakeDeviceObject(). */
}


/* ==================
    Static Functions
   ================== */

/**
 *  This function ensures that the host has the right to write to the VCP2's
 *  command mailbox.  If the VCP2 owns the mailbox (i.e. is busy), this
 *  function returns VP_STATUS_MAILBOX_BUSY.
 *
 * \note If VP_VCP2_MAILBOX_SPINWAIT is defined, this function continuously
 *       polls the VCP2 until the mailbox is available.  This function polls
 *       VP_VCP2_MAILBOX_SPINWAIT times before giving up.
 *
 * \param[in] deviceId  Device ID identifying the VCP2 device.
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
static VpStatusType
CmdMailboxAcquire(
    VpDeviceIdType deviceId)
{
    uint16 flags;

#if (VP_VCP2_MAILBOX_SPINWAIT > 0)
    uint32 i;
    for (i = VP_VCP2_MAILBOX_SPINWAIT; i; i--) {
#endif

        /* Read mailbox flags register. */
        if (VpHbiRd(deviceId, SW_FLAGS, &flags) == FALSE) {
            return VP_STATUS_ERR_HBI;
        }

        /* Check command mailbox flag: 0 => host owns; 1=> VCP2 owns. */
        if ((flags & VCP2_HBI_SWFLAG_CMD_MBOX) == 0) {
            return VP_STATUS_SUCCESS;
        }

#if (VP_VCP2_MAILBOX_SPINWAIT > 0)
        if (i % (VP_VCP2_MAILBOX_SPINWAIT / 10) == 0) {
            VP_DOUT(DBG_WARNING, ("Waiting for VCP2 command mailbox"));
        }
    }
#endif

    return VP_STATUS_MAILBOX_BUSY;
}

/**
 * This internal function determines whether the specified entry in the VCP2
 * Profile Table has been initialized by a previous call to VpInitProfile().
 *
 * \param[in] pDevObj       Pointer to the Device Object
 * \param[in] profileType   Type of Profile Table entry to check
 * \param[in] pProfile      Profile index to check
 *
 * \retval  ::TRUE          Profile Table entry is empty or invalid
 * \retval  ::FALSE         Profile Table entry is nonempty
 */
static bool
ProfileTableEntryEmpty(
    VpVcp2DeviceObjectType *pDevObj,
    VpProfileType profileType,
    VpProfilePtrType pProfileIndex)
{
    uint16 profTabInit = pDevObj->profTabInit[profileType];
    uint16 mask;

    /* Obtain the profile index and verify that it is within limits. */
    int idx = VpGetProfileIndex(pProfileIndex);
    int maxIdx = Vcp2ProfileTableSize(profileType) - 1;
    if ((idx < 0) || (idx > maxIdx)) {
        VP_DOUT(VP_DBG_ERROR, ("ProfileTableEntryEmpty(): invalid index %d (max %d)\n", idx, maxIdx));
        return TRUE;
    }

    /* Determine if the Profile Table entry was previously initialized. */
    mask = 1 << idx;
    if (profTabInit & mask) {
        return FALSE;
    }

    return TRUE;
}

/**
 * This function verifies that the length of the supplied profile is less than
 * or equal to the maximum length for the given type of profile.
 *
 * \param[in] profileType   Type of Profile supplied in pProfile
 * \param[in] pProfile      Profile to be checked
 *
 * \retval  ::TRUE          Profile is within length limit
 * \retval  ::FALSE         Profile is too large
 */
static bool
VerifyProfileLength(
    VpProfileType profileType,
    VpProfilePtrType pProfile)
{
    int expectedLength = 0;                                 /* words */
    int actualLength = pProfile[VP_PROFILE_LENGTH] + 4;     /* bytes */

    switch (profileType) {
        case VP_PROFILE_DEVICE:
            expectedLength = VCP2_DEV_PROFILE_LEN;
            break;
        case VP_PROFILE_AC:
            expectedLength = VCP2_AC_PROFILE_LEN;
            break;
        case VP_PROFILE_DC:
            expectedLength = VCP2_DC_PROFILE_LEN;
            break;
        case VP_PROFILE_RING:
            expectedLength = VCP2_RING_PROFILE_LEN;
            break;
        case VP_PROFILE_RINGCAD:
            expectedLength = VCP2_RINGCAD_PROFILE_LEN;
            break;
        case VP_PROFILE_TONE:
            expectedLength = VCP2_TONE_PROFILE_LEN;
            break;
        case VP_PROFILE_METER:
            expectedLength = VCP2_MTR_PROFILE_LEN;
            break;
        case VP_PROFILE_CID:
            expectedLength = VCP2_CID_PROFILE_LEN;
            break;
        case VP_PROFILE_TONECAD:
            expectedLength = VCP2_TONECAD_PROFILE_LEN;
        default:
            /* No error checking here. */
            ;
    }


    /* Compare profile length field with maximum length allowed. */
    return (actualLength <= expectedLength * 2);
}

#endif /* VP_CC_VCP2_SERIES */
