/** \file vcp2_config.c
 * vcp2_config.c
 *
 * This file contains the implementation for the system configuration functions
 * in the VCP2 library.
 *
 * The functions herein are used internally by the API and should not be
 * called directly by the application.
 *
 * Copyright (c) 2008, Zarlink Semiconductor, Inc.
 */

/* Only compile this file if specified in vp_api_cfg.h: */
#include "vp_api_cfg.h"
#if defined (VP_CC_VCP2_SERIES)

#include "vcp2_api.h"
#include "vp_api_int.h"
#include "vcp2_api_int.h"
#include "vp_debug.h"


/* ================================
    System Configuration Functions
   ================================ */

/**
 * This function associates the supplied Device Context struct with the
 * supplied VCP2 Device Object struct.  This function should be called after
 * initializing the Device Object with VpMakeDeviceObject(), and can be
 * called multiple times to create multiple Device Contexts for use in
 * different memory address spaces.  The contents of the Device Object are not
 * modified by this function.
 *
 * \param[in] pDevCtx   Pointer to the Device Context struct to be initialized
 * \param[in] pDevObj   Pointer to the Device Object struct that the Device
 *                      Context will be associated with
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG  One or more arguments were NULL.
 */
VpStatusType
Vcp2MakeDeviceCtx(
    VpDevCtxType *pDevCtx,
    VpVcp2DeviceObjectType *pDevObj)
{

    /* Initialize members in the Device Context. */
    pDevCtx->pDevObj = pDevObj;
    pDevCtx->deviceType = VP_DEV_VCP2_SERIES;

    /* Set up function pointers in the Device Context. */
    Vcp2ResetDeviceContext(pDevCtx);

    return VP_STATUS_SUCCESS;
}

/**
 * This function initializes the supplied VCP2 Device Context and VCP2 Device
 * Object structs.  This function must be called before calling any other
 * VCP2_API functions.  The initialized Device Context struct is used to
 * uniquely identify the device in subsequent VP-API function calls.  If
 * pDevCtx is NULL, only the Device Object is initialized.
 *
 * \note
 * Memory for the VpDevCtxType and VpVcp2DeviceObjectType structs must be
 * allocated (statically or dynamically) by the application before calling
 * this function.  The application does not need to initialize the memory.
 *
 * \param[in] pDevCtx   Pointer to the Device Context struct to be initialized
 * \param[in] pDevObj   Pointer to the Device Object struct to be initialized
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG  ::Vcp2MakeDeviceCtx() failed.
 */
VpStatusType
Vcp2MakeDeviceObject(
    VpDevCtxType *pDevCtx,
    VpVcp2DeviceObjectType *pDevObj)
{
    /* Reset all members of the Device Object to their default values. */
    Vcp2ResetDeviceObject(pDevObj);

    if (pDevCtx == VP_NULL) {
        return VP_STATUS_SUCCESS;
    }

    return Vcp2MakeDeviceCtx(pDevCtx, pDevObj);
}

/**
 * This function initializes a Line Context and a VCP2 Line Object using the
 * supplied.  It also establishes a link between the Device Context and the
 * Line Context. This function is internal to the API and should never be
 * called by the application.
 *
 * \note
 * Memory for the VpLineCtxType and VpVcp2LineObjectType structs must be
 * allocated (statically or dynamically) by the application before calling
 * this function.  The application does not need to initialize the memory.
 * VpMakeDeviceObject() must be called prior to this function, to initialize
 * the Device Context struct.  This function should only be called after the
 * VCP2 has been booted with VpBootLoad().
 *
 * \param[in] termType  Line termination type for the indicated line
 * \param[in] channelId Channel ID of the VCP2 channel to be controlled
 * \param[in] pLineCtx  Pointer to the Line Context struct to be initialized
 * \param[in] pLineObj  Pointer to the Line Object struct to be initialized
 * \param[in] pDevCtx   Pointer to the Device Context struct that the Line
 *                      Context will be associated with
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_ERR_VTD_CODE Termination type (termType) unsupported.
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2MakeLineObject(
    VpTermType termType,
    uint8 channelId,
    VpLineCtxType *pLineCtx,
    void *pVoidLineObj,
    VpDevCtxType *pDevCtx)
{
    VpVcp2LineObjectType *pLineObj = pVoidLineObj;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 termTypeReg;
    VpStatusType retval = VP_STATUS_SUCCESS;

    /* Check arguments. */
    if (channelId >= pDevObj->numChannels) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2MakeLineObject(): Channel ID out of range.\n"));
        return VP_STATUS_INVALID_ARG;
    }
    switch (termType) {
        case VP_TERM_FXS_GENERIC:
        case VP_TERM_FXS_75181:
        case VP_TERM_FXS_TO_TL:
        case VP_TERM_FXS_75282:
        case VP_TERM_FXS_TITO_TL_R:
        case VP_TERM_FXS_CO_TL:
        case VP_TERM_FXS_RR:
        case VP_TERM_FXS_TI:
            break;
        default:
            return VP_STATUS_ERR_VTD_CODE;
    }

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Initialize Line Context. */
    pLineCtx->pLineObj = pLineObj;

    /* Establish the link between Device Context and Line Context. */
    pLineCtx->pDevCtx = pDevCtx;
    pDevCtx->pLineCtx[channelId] = pLineCtx;

    /* Initialize the Line Object. */
    pLineObj->termType = termType;
    pLineObj->channelId = channelId;

    /*
     * Note that we are not initializing the lineId member, because we do not
     * know what it should be initialized to, because VpLineIdType is defined
     * by the customer.  The lineId member is initialized by VpMapLineId().
     */

    /* Read line termination type register. */
    if (VpHbiRd(deviceId, TERMTYPE(channelId), &termTypeReg) == FALSE) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2MakeLineObject(): %s.\n", "HBI error"));
        retval = VP_STATUS_ERR_HBI;
    }

    /* Modify line termination type register. */
    if (channelId & 1) {
        termTypeReg &= 0xFF00;
        termTypeReg |= termType;
    } else {
        termTypeReg &= 0x00FF;
        termTypeReg |= ((uint16)termType << 8);
    }

    /* Write line termination type register. */
    if (VpHbiWr(deviceId, TERMTYPE(channelId), &termTypeReg) == FALSE) {
        retval = VP_STATUS_ERR_HBI;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return VP_STATUS_SUCCESS;
}

#endif /* VP_CC_VCP2_SERIES */
