/** \file vcp2_control.c
 * vcp2_control.c
 *
 * This file contains the implementation for the Control functions
 * in the VCP2 library.
 *
 * The functions herein are used internally by the API and should not be
 * called directly by the application.
 *
 * Copyright (c) 2008, Zarlink Semiconductor, Inc.
 */

/* Only compile this file if specified in vp_api_cfg.h: */
#include "vp_api_cfg.h"
#if defined (VP_CC_VCP2_SERIES)

#include "vcp2_api.h"
#include "vp_api_int.h"
#include "vcp2_api_int.h"


/* =================================
    Prototypes for Static Functions
   ================================= */

static VpStatusType
DeviceIoAccessInt(
    VpDevCtxType *pDevCtx,
    VpDeviceIoAccessExtType *pDeviceIoAccess,
    uint8 version);

static uint8
GetTotalDtmfResources(
    VpVcp2DeviceObjectType *pDevObj);

static bool
LineStateNotSupported(
    VpVcp2DeviceObjectType *pDevObj,
    VpLineStateType state);

static void
MailboxAddOptionValue(
    VpDevCtxType *pDevCtx,
    uint16 **ppBuf,
    VpOptionIdType optionId,
    void *value);

static void
MailboxAddSignalArgs(
    uint16 **ppBuf,
    VpSendSignalType sigType,
    void *pArgs);

static void
MakeNewDevIoAccessStruct(
    VpDevCtxType *pDevCtx,
    VpDeviceIoAccessDataType *pDeviceIoData,
    VpDeviceIoAccessExtType *pDeviceIoAccess);

static void
MakeNewDevIoConfigStruct(
    VpDevCtxType *pDevCtx,
    VpOptionDeviceIoType *pDeviceIo,
    VpOptionDeviceIoConfigType *pDeviceIoConfig);

static VpStatusType
MaskDedicatedPins(
    VpTermType termType,
    uint8 *pMask);

static bool
OptionValueNotSupported(
    VpLineCtxType *pLineCtx,
    VpVcp2DeviceObjectType *pDevObj,
    VpOptionIdType optionId,
    void *value);

static VpStatusType
ProcessOptionValue(
    VpLineCtxType *pLineCtx,
    VpVcp2DeviceObjectType *pDevObj,
    VpOptionIdType optionId,
    void *value,
    bool *pSendCommand);

static bool
RelayStateNotSupported(
    VpRelayControlType rState,
    VpTermType termType);

static bool
SetCodeLoadBase(
    VpDeviceIdType deviceId,
    uint32 ambaAddr);

static bool
SignalNotSupported(
    VpVcp2DeviceObjectType *pDevObj,
    VpSendSignalType sigType,
    void *pArgs);

static VpStatusType
UpdateDtmfStatusMembers(
    VpLineCtxType *pLineCtx,
    VpVcp2DeviceObjectType *pDevObj,
    VpOptionDtmfModeType *pDtmfControl,
    bool *pSendCommand);


/* ===================
    Control Functions
   =================== */

/**
 * This function continues a Caller ID sequence by updating the FSK data
 * buffer.  Refer to the VoicePath API User's Guide for details about
 * ::VpContinueCid().
 *
 * \param[in] pLineCtx      Line Context struct for the line
 * \param[in] length        Length of FSK data to be buffered
 * \param[in] pCidData      FSK data to be buffered
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_INVALID_ARG      Invalid length, or pCidData was NULL
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2ContinueCid(
    VpLineCtxType *pLineCtx,
    uint8 length,
    uint8p pCidData)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_CONTINUECID];
    uint16 *pBuf = buf;
    VpStatusType retval;

    if (pDevObj->devInit == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Check arguments. */
    if ((length > MAX_CID_CONTINUE_DATA_LENGTH) || (length == 0)) {
        return VP_STATUS_INVALID_ARG;
    }
    if (pCidData == VP_NULL) {
        return VP_STATUS_INVALID_ARG;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_CONTINUECID, channelId);

    /* Add FSK data length to the mailbox buffer. */
    *pBuf++ = (uint16)length;

    /* Add FSK data to the mailbox buffer. */
    Vcp2MailboxAddBytes(&pBuf, pCidData, length);

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function is used to set GPIO pins provided in the VTD.  Please see
 * VP-API Reference Guide for details about ::VpDeviceIoAccess().  This
 * function is now deprecated in favor of ::VpDeviceIoAccessExt().
 *
 * \note This function can be used before ::VpInitDevice() has been called.
 *
 * \param[in]   pDevCtx         Device Context for the VCP2 whose I/O
 *                              lines are to be toggled
 * \param[in]   pDeviceIoData   Struct containing access type and values
 *                              to be written
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG    invalid access type
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2DeviceIoAccess(
    VpDevCtxType *pDevCtx,
    VpDeviceIoAccessDataType *pDeviceIoData)
{
    VpDeviceIoAccessExtType deviceIoAccess;

    /* Convert old-format struct into new format. */
    MakeNewDevIoAccessStruct(pDevCtx, pDeviceIoData, &deviceIoAccess);

    /* This function is just a wrapper. */
    return DeviceIoAccessInt(pDevCtx, &deviceIoAccess, 1);
}

/**
 * This function is used to set GPIO pins provided in the VTD.  Please see
 * VP-API Reference Guide for details about ::VpDeviceIoAccessExt().
 *
 * \note This function can be used before ::VpInitDevice() has been called.
 *
 * \param[in]   pDevCtx         Device Context for the VCP2 whose I/O
 *                              lines are to be toggled
 * \param[in]   pDeviceIoAccess Struct containing access type and values
 *                              to be written
 * \param[in]   handle          Handle to be returned with the event
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG    invalid access type
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 * \retval ::VP_STATUS_DEDICATED_PINS attempted to toggle pins that belong to
 *                                    the line termination type
 */
VpStatusType
Vcp2DeviceIoAccessExt(
    VpDevCtxType *pDevCtx,
    VpDeviceIoAccessExtType *pDeviceIoAccess)
{
    /* This function is just a wrapper. */
    return DeviceIoAccessInt(pDevCtx, pDeviceIoAccess, 0);
}

/**
 * This function loads test vector data into the VCP2's internal buffers for
 * algorithm testing.  Refer to the VoicePath API User's Guide for details
 * about ::VpFillTestBuf().
 *
 * \note This function can be used before ::VpInitDevice() has been called.
 *
 * \param[in]   pLineCtx    Line Context for a line to which a test buffer
 *                          has been assigned
 * \param[in]   length      number of samples (16-bit words) of data to be
 *                          copied into the test buffer
 * \param[in]   pData       buffer containing samples to be copied
 *
 * \retval ::VP_STATUS_INVALID_ARG  length exceeds test buffer size
 * \retval ::VP_STATUS_DEVICE_BUSY  VpTestLine(PREPARE) has not previously
 *                                  been called on the specified line.
 * \retval ::VP_STATUS_ERR_HBI
 * \retval ::VP_STATUS_SUCCESS
 */
VpStatusType
Vcp2FillTestBuf(
    VpLineCtxType *pLineCtx,
    uint16 length,
    VpVectorPtrType pData)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    int testBuf;
    VpStatusType retval = VP_STATUS_DEVICE_BUSY;

    /* Check arguments. */
    if (length > VCP2_HBI_TESTBUF_LEN) {
        return VP_STATUS_INVALID_ARG;
    }

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* See which test buffer (if any) is allocated to the line. */
    for (testBuf = 0; testBuf < VCP2_HBI_NUM_TEST_BUFS; testBuf++) {
        uint16 testChan;
        if (VpHbiRd(deviceId, TESTCHAN(testBuf), &testChan) == FALSE) {
            retval = VP_STATUS_ERR_HBI;
            break;
        }

        /* Even/odd line stored in upper/lower byte, respectively. */
        if ((testBuf % 2) == 0) {
            testChan >>= 8;
        }
        testChan &= 0xFF;

        /* See if this buffer is allocated to the specified line. */
        if (testChan == channelId) {
            retval = VP_STATUS_SUCCESS;
            break;
        }
    }

    /* No point in continuing further if no test buffer has been allocated. */
    if (retval == VP_STATUS_DEVICE_BUSY) {
        VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
        return retval;
    }

    /* Reset the test buffer cursor to 0. */
    if (VpHbiWr(deviceId, TESTCURSOR(testBuf), 0) == FALSE) {
        retval = VP_STATUS_ERR_HBI;
    }

    /* Write the supplied data starting at the beginning of the test
       buffer. */
    if (VpHbiXfer(TRUE, deviceId, VCP2_HBI_TESTBUF_AMBA(testBuf),
        length, pData, TRUE, SetCodeLoadBase) == FALSE) {
        retval = VP_STATUS_ERR_HBI;
    }

    /* Fill the remainder of the test buffer with zeroes. */
    if (VpHbiXfer(TRUE, deviceId, VCP2_HBI_TESTBUF_AMBA(testBuf) + length * 2,
        VCP2_HBI_TESTBUF_LEN - length, VP_NULL, TRUE, SetCodeLoadBase) == FALSE) {
        retval = VP_STATUS_ERR_HBI;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function is used to set GPIO pins associated with a particular line.
 * Please see VP-API Reference Guide for details about ::VpLineIoAccess().
 *
 * \note This function can be used before ::VpInitDevice() has been called.
 *
 * \param[in]   pLineCtx        Linee Context for the VCP2 line whose I/O
 *                              lines are to be toggled
 * \param[in]   pLineIoAccess   Struct containing access type and values
 *                              to be written
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG    pDeviceIoData was NULL, or invalid access type
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 * \retval ::VP_STATUS_DEDICATED_PINS attempted to toggle pins that belong to
 *                                    the line termination type
 */
VpStatusType
Vcp2LineIoAccess(
    VpLineCtxType *pLineCtx,
    VpLineIoAccessType *pLineIoAccess,
    uint16 handle)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    VpVcp2DeviceObjectType *pDevObj = pLineCtx->pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint8 channelId = pLineObj->channelId;
    VpIoDirectionType direction = pLineIoAccess->direction;
    uint8 *pMask = &pLineIoAccess->ioBits.mask;
    uint8 data = pLineIoAccess->ioBits.data;
    uint16 buf[BUFSIZE_LINEIOACCESS];
    uint16 *pBuf = buf;
    VpStatusType retval, err;

    /* Check arguments. */
    if ((direction != VP_IO_WRITE) && (direction != VP_IO_READ)) {
        return VP_STATUS_INVALID_ARG;
    }

    /* Modify the access mask to exclude GPIO pins dedicated to the line
       termination type. */
    retval = MaskDedicatedPins(pLineObj->termType, pMask);

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_LINEIOACCESS, channelId);

    /* Add the handle argument to the mailbox buffer. */
    *pBuf++ = handle;

    /* Add the direction argument to the mailbox buffer. */
    *pBuf++ = (uint16)direction;

    /* Add the access mask and data to the mailbox buffer. */
    *pBuf++ = ((uint16)*pMask << 8) | data;

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    err = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    /* Errors override the VP_STATUS_DEDICATED_PINS return code. */
    if (err != VP_STATUS_SUCCESS) {
        retval = err;
    }
    return retval;
}

/**
 * This function executes a direct SLAC register access through the VCP2.
 * Refer to the VoicePath API User's Guide for details about
 * ::VpLowLevelCmd().
 *
 * \note This function can be used before ::VpInitDevice() has been called.
 *
 * \param[in]   pLineCtx    Line Context for the desired SLAC channel
 * \param[in]   pMpiCmdData Pointer to MPI command byte and zero or more data
 *                          bytes to be sent to the MPI bus
 * \param[in]   length      Number of data bytes to be sent or received
 * \param[in]   handle      Handle to be returned with the completion event
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG  cmd was NULL, or length was too large
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2LowLevelCmd(
    VpLineCtxType *pLineCtx,
    uint8p pMpiCmdData,
    uint8 length,
    uint16 handle)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_LOWLEVELCMD];
    uint16 *pBuf = buf;
    VpStatusType retval;

    /* Check arguments. */
    if (
        (pMpiCmdData == VP_NULL) ||
        (length > (MAX_MPI_DATA_LENGTH - 1))
    ) {
        return VP_STATUS_INVALID_ARG;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_LOWLEVELCMD, channelId);

    /* Add the handle field to the mailbox buffer. */
    *pBuf++ = (uint16)handle;

    /* Add the MPI command to the mailbox buffer. */
    *pBuf++ = (uint16)*pMpiCmdData;

    /* Add the MPI data length to the mailbox buffer. */
    *pBuf++ = (uint16)length;

    /* For write commands, add the MPI data to the mailbox buffer. */
    if ((*pMpiCmdData & 1) == 0) {
        Vcp2MailboxAddBytes(&pBuf, &pMpiCmdData[1], length);
    }

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function executes a direct SLAC access through the VCP2.
 * Refer to the VoicePath API User's Guide for details about
 * ::VpLowLevelCmd16().
 *
 * \note This function can be used before ::VpInitDevice() has been called.
 *
 * \param[in]   pLineCtx        Line Context for the desired SLAC channel
 * \param[in]   cmdType         Value of type VpLowLevelCmdType to determine which
 *                              type of low level access is to be performed - Some
 *                              combination of read or write and page or mailbox.
 * \param[in]   *writeWords     Pointer to command and data words to be written
 *                              to the SLAC interface.
 * \param[in]   numWriteWords   Total number of command and data words to write
 *                              to the SLAC interface.
 * \param[in]   numReadWords    Number of data words to be read from the SLAC
 *                              interface.  This parameter is only used for read
 *                              operations.
 * \param[in]   handle          Handle to be returned with the completion event.
 *                              This parameter is only used for read operations.
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG  invalid cmdType or out of range r/w words
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2LowLevelCmd16(
    VpLineCtxType *pLineCtx,
    VpLowLevelCmdType cmdType,
    uint16 *writeWords,
    uint8 numWriteWords,
    uint8 numReadWords,
    uint16 handle)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_LOWLEVELCMD16];
    uint16 *pBuf = buf;
    VpStatusType retval;

    /* Check arguments. */
    if ((numWriteWords < 1) || (numWriteWords > LLC16_CMD_DATA_FLD)) {
        VP_DOUT(VP_DBG_ERROR,("Vcp2LowLevelCmd16(): numWriteWords out of range (%d)", numWriteWords));
        return VP_STATUS_INVALID_ARG;
    }
    switch (cmdType) {
        case VP_LOWLEV_PAGE_WR:
            break;
        case VP_LOWLEV_PAGE_RD:
            /* Must be 1 or 2 write words.  2 allows for page select followed by
             * read command for reading mailbox or codeload pages. */
            if ((numWriteWords < 1) || (numWriteWords > 2)) {
                VP_DOUT(VP_DBG_ERROR,("Vcp2LowLevelCmd16(): numWriteWords out of range (%d)", numWriteWords));
                return VP_STATUS_INVALID_ARG;
            }
            /* Read words must be at least 1 and no more than 128 */
            if ((numReadWords < 1) || (numReadWords > 128)) {
                VP_DOUT(VP_DBG_ERROR,("Vcp2LowLevelCmd16(): numReadWords out of range (%d)", numReadWords));
                return VP_STATUS_INVALID_ARG;
            }
            break;
        case VP_LOWLEV_MBOX_WR:
            /* Need at least 2 words to create a valid SLAC mailbox command */
            if (numWriteWords < 2) {
                VP_DOUT(VP_DBG_ERROR,("Vcp2LowLevelCmd16(): numWriteWords out of range (%d)", numWriteWords));
                return VP_STATUS_INVALID_ARG;
            }
            break;
        case VP_LOWLEV_MBOX_RD:
            /* Need at least 2 words to create a valid SLAC mailbox command */
            if (numWriteWords < 2) {
                VP_DOUT(VP_DBG_ERROR,("Vcp2LowLevelCmd16(): numWriteWords out of range (%d)", numWriteWords));
                return VP_STATUS_INVALID_ARG;
            }
            /* Read words must be at least 1 and no more than 128 */
            if ((numReadWords < 1) || (numReadWords > 128)) {
                VP_DOUT(VP_DBG_ERROR,("Vcp2LowLevelCmd16(): numReadWords out of range (%d)", numReadWords));
                return VP_STATUS_INVALID_ARG;
            }
            break;
        default:
            VP_DOUT(VP_DBG_ERROR,("Vcp2LowLevelCmd16(): cmdType invalid (%d)", cmdType));
            return VP_STATUS_INVALID_ARG;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_LOWLEVELCMD16, channelId);

    /* Add the handle field to the mailbox buffer. */
    *pBuf++ = (uint16)handle;

    /* Add the command type to the mailbox buffer. */
    *pBuf++ = (uint16)cmdType;

    /* Add the write and read lengths to the mailbox buffer. */
    *pBuf++ = (uint16)(((uint16)numWriteWords << 8) | numReadWords);

    /* Add the write words to the mailbox buffer. */
    Vcp2MailboxAddBytes(&pBuf, (uint8p)writeWords, numWriteWords * 2);

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
} /* Vcp2LowLevelCmd16() */


/**
 * This function initiates a self test on the specified VCP2-controlled line.
 * Refer to the VoicePath API User's Guide for details about ::VpSelfTest().
 *
 * \note This function can be used before ::VpInitDevice() has been called.
 *
 * \param[in]   pLineCtx    Line Context for the line to be tested
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2SelfTest(
    VpLineCtxType *pLineCtx)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_SELFTEST];
    uint16 *pBuf = buf;
    VpStatusType retval;

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_SELFTEST, channelId);

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function manually initiates a Caller ID sequence.  Refer to the
 * VoicePath API User's Guide for details about ::VpSendCid().
 *
 * \param[in] pLineCtx      Line Context struct for the line
 * \param[in] length        Length of FSK data to be buffered
 * \param[in] pCidProfile   Caller ID Profile to apply to the line
 * \param[in] pCidData      FSK data to be buffered
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_INVALID_ARG      Invalid length, or pCidData was NULL,
 *                                      or pCidProfile was not a Profile Table
 *                                      index.
 * \retval ::VP_STATUS_ERR_PROFILE
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2SendCid(
    VpLineCtxType *pLineCtx,
    uint8 length,
    VpProfilePtrType pCidProfile,
    uint8p pCidData)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_SENDCID];
    uint16 *pBuf = buf;
    VpStatusType retval;

    /* Check arguments. */
    int idx = VpGetProfileIndex(pCidProfile);
    if ((length > MAX_CID_DATA_LENGTH) || (length == 0)) {
        return VP_STATUS_INVALID_ARG;
    }
    if (pCidData == VP_NULL) {
        return VP_STATUS_INVALID_ARG;
    }

    /* Verify that VpInitDevice() has been called. */
    if (pDevObj->devInit == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Currently, the VCP2 only supports passing a Profile Table index (not a
       pointer to a profile) for Caller ID profiles. */
    if (idx < 0) {
        return VP_STATUS_INVALID_ARG;
    }

    if (Vcp2BadProfileArg(pDevObj, VP_PROFILE_CID, pCidProfile)) {
        return VP_STATUS_ERR_PROFILE;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_SENDCID, channelId);

    /* Add Caller ID Profile to the mailbox buffer. */
    Vcp2MailboxAddProfile(&pBuf, VP_PROFILE_CID, pCidProfile);

    /* Add FSK data length to the mailbox buffer. */
    *pBuf++ = (uint16)length;

    /* Add FSK data to the mailbox buffer. */
    Vcp2MailboxAddBytes(&pBuf, pCidData, length);

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function is used to send a specified signal on the line.  Refer to
 * the VoicePath API User's Guide for details about ::VpSendSignal().
 *
 * \param[in] pLineCtx  Line Context struct for the line
 * \param[in] sigType   Type of signal to be applied to the line
 * \param[in] pStruct   "Message waiting" signal parameters (optional)
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_INVALID_ARG      Signal type invalid or unsupported.
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2SendSignal(
    VpLineCtxType *pLineCtx,
    VpSendSignalType sigType,
    void *pStruct)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_SENDSIGNAL];
    uint16 *pBuf = buf;
    VpStatusType retval;

    if (pDevObj->devInit == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Signal-specific argument checking: */
    if (SignalNotSupported(pDevObj, sigType, pStruct)) {
        return VP_STATUS_INVALID_ARG;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_SENDSIGNAL, channelId);

    /* Add the signal type field to the mailbox buffer. */
    *pBuf++ = (uint16)sigType;

    /* Add signal-specific argument(s) to mailbox buffer. */
    MailboxAddSignalArgs(&pBuf, sigType, pStruct);

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function sets a line controlled by the VCP2 into the given state.
 * Refer to the VP API User's Guide for details about ::VpSetLineState().
 *
 * \param[in] pLineCtx      Line Context struct for the line
 * \param[in] state         Line state to be entered
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG          Invalid or unsupported line state
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2SetLineState(
    VpLineCtxType *pLineCtx,
    VpLineStateType state)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_SETLINESTATE];
    uint16 *pBuf = buf;
    VpStatusType retval;

    if (pDevObj->devInit == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Check arguments. */
    if (LineStateNotSupported(pDevObj, state)) {
        return VP_STATUS_INVALID_ARG;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_SETLINESTATE, channelId);

    /* Add the line state argument. */
    *pBuf++ = (uint16)state;

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    /* Work-around for a VCP2 firmware erratum:  Keep track of whether we are
       in the tip-open line state, so we can mask hook events in that state. */
    pLineObj->tipOpenState = (state == VP_LINE_TIP_OPEN) ? TRUE : FALSE;

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function requests the VCP2 to emit a cadenced tone on the indicated
 * line.  Refer to the VP-API User's Guide for details about ::VpSetLineTone().
 *
 * \param[in] pLineCtx          Line Context struct for the line
 * \param[in] pToneProfile      Tone Profile to be applied to the line
 * \param[in] pToneCadProfile   Tone Cadence Profile to be applied to the line
 * \param[in] pDtmfControl      (unused)
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG  DTMF generation not supported by VCP2-790
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_ERR_PROFILE
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2SetLineTone(
    VpLineCtxType *pLineCtx,
    VpProfilePtrType pToneProfile,
    VpProfilePtrType pToneCadProfile,
    VpDtmfToneGenType *pDtmfControl)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_SETLINETONE];
    uint16 *pBuf = buf;
    VpStatusType retval;

    if (pDevObj->devInit == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Check arguments. */
    if (
        Vcp2BadProfileArg(pDevObj, VP_PROFILE_TONE, pToneProfile) ||
        Vcp2BadProfileArg(pDevObj, VP_PROFILE_TONECAD, pToneCadProfile)
    ) {
        return VP_STATUS_ERR_PROFILE;
    }
    if (
        ((pDevObj->product & VP_VCP_PROD_SLAC_MASK) == VP_VCP_PROD_SLAC_VE790) &&
        (pDtmfControl != VP_NULL)
    ) {
        /* VCP2-790 does not support DTMF tone generation yet. */
        return VP_STATUS_INVALID_ARG;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_SETLINETONE, channelId);

    /* Add the Tone Profile. */
    Vcp2MailboxAddProfile(&pBuf, VP_PROFILE_TONE, pToneProfile);

    /* Add the Tone Cadence Profile. */
    Vcp2MailboxAddProfile(&pBuf, VP_PROFILE_TONECAD, pToneCadProfile);

    /* Add DTMF tone generation info to the mailbox buffer. */
    if (pDtmfControl == VP_NULL) {

        /* DTMF generation not requested. */
        *pBuf++ = (uint16)VP_DIG_NONE;
        *pBuf++ = 0;
    } else {

        /* DTMF generation requested. */
        *pBuf++ = (uint16)pDtmfControl->toneId;
        *pBuf++ = (uint16)pDtmfControl->dir;
    }

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function is used to set various options that are supported by the
 * VCP2.  For a detailed description of ::VpSetOption() and which options are
 * available for the VCP2, please see the VP-API Reference Guide.
 *
 * \param[in] pLineCtx      Line Context struct for line-specific options
 * \param[in] pDevCtx       Device Context struct for device-specific options
 * \param[in] optionId      Option ID to identify which option is to be set
 * \param[in] value         Struct containing option value argument(s)
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG      Option ID was invalid or unsupported,
 *                                      or a Line Context was passed with a
 *                                      device-specific option, or vice versa.
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2SetOption(
    VpLineCtxType *pLineCtx,
    VpDevCtxType *pDevCtx,
    VpOptionIdType optionId,
    void *value)
{
    uint8 channelId;
    VpVcp2DeviceObjectType *pDevObj;
    VpDeviceIdType deviceId;
    uint16 buf[BUFSIZE_SETOPTION];
    uint16 *pBuf = buf;
    bool sendCommand = TRUE;
    VpStatusType retval = VP_STATUS_SUCCESS;

    VP_DOUT(DBG_API_FUNC, ("Entering Vcp2SetOption()"));

    /* Get Device Object, device ID, and channel ID. */
    if (pDevCtx != VP_NULL) {
        pDevObj = pDevCtx->pDevObj;
        channelId = VP_ALL_LINES;
    } else {
        VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
        pDevObj = pLineCtx->pDevCtx->pDevObj;
        channelId = pLineObj->channelId;
    }
    deviceId = pDevObj->deviceId;

    if (pDevObj->devInit == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Is the requested option supported by VCP2? */
    if (Vcp2DeviceOptionSupported(pDevObj, optionId)) {
        /* For device options, must specify a Device Context. */
        if (pDevCtx == VP_NULL) {
            return VP_STATUS_INVALID_ARG;
        }
    } else if (!Vcp2LineOptionSupported(pDevObj, optionId)) {
        /* Option is not supported by this device. */
        return VP_STATUS_INVALID_ARG;
    }

    /* Option-specific argument checking: */
    if (OptionValueNotSupported(pLineCtx, pDevObj, optionId, value)) {
        return VP_STATUS_INVALID_ARG;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_SETOPTION, channelId);

    /* Add option ID to the mailbox buffer. */
    *pBuf++ = (uint16)optionId;

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Some options require special processing (e.g. updating the device
       object or reading VCP direct page registers).  If we don't need to send
       a mailbox command, the following function will set sendCommand =
       FALSE. */
    retval = ProcessOptionValue(pLineCtx, pDevObj, optionId, value, &sendCommand);

    /* Add the requested option value to the mailbox buffer. */
    MailboxAddOptionValue(pDevCtx, &pBuf, optionId, value);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    if (sendCommand && (retval == VP_STATUS_SUCCESS)) {
        retval = Vcp2MailboxSend(deviceId, buf, pBuf);
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    VP_DOUT(DBG_API_FUNC, ("Vcp2SetOption() = %d", retval));
    return retval;
}

/**
 * This function configures VCP2-controlled board relays associated with the
 * specified line.  Refer to the VP API User's Guide for details about
 * ::VpSetRelayState().
 *
 * \param[in] pLineCtx  Line Context struct for the line
 * \param[in] rstate    Relay state to be applied to the line
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG          Relay state invalid or unsupported
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2SetRelayState(
    VpLineCtxType *pLineCtx,
    VpRelayControlType rState)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_SETRELAYSTATE];
    uint16 *pBuf = buf;
    VpStatusType retval;

    if (pDevObj->devInit == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Check arguments. */
    if (RelayStateNotSupported(rState, pLineObj->termType)) {
        return VP_STATUS_INVALID_ARG;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_SETRELAYSTATE, channelId);

    /* Add the relay state to the mailbox buffer. */
    *pBuf++ = (uint16)rState;

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function adjusts the relative gain for the transmit and receive paths
 * for the specified line.  Refer to the VoicePath API User's Guide for
 * details about ::VpSetRelGain().
 *
 * \param[in] pLineCtx  Line Context struct for the line to be adjusted
 * \param[in] rxLevel   Relative receive gain level
 * \param[in] txLevel   Relative transmit gain level
 * \param[in] handle    Handle to be returned with the gain-complete event
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2SetRelGain(
    VpLineCtxType *pLineCtx,
    uint16 txLevel,
    uint16 rxLevel,
    uint16 handle)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_SETRELGAIN];
    uint16 *pBuf = buf;
    VpStatusType retval;

    if (pDevObj->devInit == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_SETRELGAIN, channelId);

    /* Add the handle to the mailbox buffer. */
    *pBuf++ = (uint16)handle;

    /* Add the level arguments to the mailbox buffer. */
    *pBuf++ = (uint16)txLevel;
    *pBuf++ = (uint16)rxLevel;

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function places a cadenced metering tone onto VCP2-controlled line.
 * Refer to the VoicePath API User's Guide for details about ::VpStartMeter().
 *
 * \param[in] pLineCtx      Line Context struct for the line
 * \param[in] onTime        Metering pulse make time, in units of 10 ms
 * \param[in] offTime       Metering pulse break time, in units of 10 ms
 * \param[in] numMeters     Number of metering pulses to automatically perform
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2StartMeter(
    VpLineCtxType *pLineCtx,
    uint16 onTime,
    uint16 offTime,
    uint16 numMeters)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_STARTMETER];
    uint16 *pBuf = buf;
    VpStatusType retval;

    if (pDevObj->devInit == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_STARTMETER, channelId);

    /* Add cadence arguments to the mailbox buffer. */
    *pBuf++ = (uint16)onTime;
    *pBuf++ = (uint16)offTime;
    *pBuf++ = (uint16)numMeters;

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}


/* ==================
    Static Functions
   ================== */

/*
 * This internal function implements the internals of Vcp2DeviceIoAccess() and
 * Vcp2DeviceIoAccessExt().  It accepts a "version" argument to distinguish
 * between the two.
 *
 * \param[in]   pDevCtx         Device Context for the VCP2 whose I/O
 *                              lines are to be toggled
 * \param[in]   pDeviceIoAccess Struct containing access type and values
 *                              to be written
 * \param[in]   version         0 = Vcp2DeviceIoAccessExt(); 1 = Vcp2DeviceIoAccess()
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG    invalid access type
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 * \retval ::VP_STATUS_DEDICATED_PINS attempted to toggle pins that belong to
 *                                    the line termination type
 */
static VpStatusType
DeviceIoAccessInt(
    VpDevCtxType *pDevCtx,
    VpDeviceIoAccessExtType *pDeviceIoAccess,
    uint8 version)
{
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceType deviceId = pDevObj->deviceId;
    VpIoDirectionType direction = pDeviceIoAccess->direction;
    uint16 buf[BUFSIZE_DEVICEIOACCESS];
    uint16 *pBuf = buf;
    uint16 temp = 0; /* silence compiler warning */
    uint8 channelId, lineByte;
    VpStatusType retval = 0, err;

    /* Check arguments. */
    if ((direction != VP_IO_WRITE) && (direction != VP_IO_READ)) {
        return VP_STATUS_INVALID_ARG;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_DEVICEIOACCESS, VP_ALL_LINES);

    /* Add the version and direction arguments to the mailbox buffer. */
    *pBuf++ = ((uint16)version << 8) | direction;

    for (channelId = 0; channelId < VP_VCP2_MAX_NUM_LINES; channelId++) {
        VpLineCtxType *pLineCtx = pDevCtx->pLineCtx[channelId];
        VpLineIoBitsType *pLineIoBits = &pDeviceIoAccess->lineIoBits[channelId];

        if (pLineCtx != VP_NULL) {
            VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;

            /* Modify the access mask to exclude GPIO pins dedicated to the line
               termination type.  Note that it is OK to use "|=" here because
               MaskDedicatedPins() only has two return values: 0 and
               VP_STATUS_DEDICATED_PINS. */
            retval |= MaskDedicatedPins(pLineObj->termType, &pLineIoBits->mask);
        }

        /* Add the access mask and data to the mailbox buffer. */
        lineByte = (pLineIoBits->mask << 4) | (pLineIoBits->data & 0x0F);
        if (channelId % 2) {
            *pBuf++ = (temp | lineByte);
        } else {
            temp = (uint16)lineByte << 8;
        }
    }

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    err = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    /* Errors override the VP_STATUS_DEDICATED_PINS return code. */
    if (err != VP_STATUS_SUCCESS) {
        retval = err;
    }
    return retval;
}

/**
 * This internal function returns the number of lines on which the specified
 * VCP2 device can detect DTMF at the same time, based on the firmware type.
 *
 * \param[in] pDevObj   Device object for the VCP2 device
 */
static uint8
GetTotalDtmfResources(
    VpVcp2DeviceObjectType *pDevObj)
{
    /* Could switch on firmware type here.  For now, all VCP2 devices support
       32 lines of DTMF detection. */
    return (uint8)32;
}

/**
 * This internal function verifies that the selected line state is supported
 * by the specified VCP2 device, based on the firmware product type.
 *
 * \param[in] pDevObj   Device Object of the VCP2 device
 * \param[in] state     Line state to be verified
 *
 * \retval ::TRUE   Line state is invalid or unsupported.
 * \retval ::FALSE  Line state is OK.
 */
static bool
LineStateNotSupported(
    VpVcp2DeviceObjectType *pDevObj,
    VpLineStateType state)
{
    /* Currently all VCP2 firmware product types support the same options. */
    pDevObj = pDevObj;  /* silence compiler warning about ignored parameter */

    switch (state) {
        case VP_LINE_STANDBY:
        case VP_LINE_TIP_OPEN:
        case VP_LINE_ACTIVE:
        case VP_LINE_ACTIVE_POLREV:
        case VP_LINE_TALK:
        case VP_LINE_TALK_POLREV:
        case VP_LINE_OHT:
        case VP_LINE_OHT_POLREV:
        case VP_LINE_DISCONNECT:
        case VP_LINE_RINGING:
        case VP_LINE_RINGING_POLREV:
        case VP_LINE_STANDBY_POLREV:
            /* Supported line states. */
            return FALSE;
        default:
            /* Line state not supported. */
            return TRUE;
    }
}

/**
 * This internal function unpacks an option value struct and adds it into the
 * specified command mailbox buffer.
 *
 * \param[in] pDevCtx   Device Context for the VCP2 device
 * \param[in] ppBuf     Pointer to the calling function's pBuf variable
 * \param[in] optionId  Option ID of the option value to be added
 * \param[in] value     Pointer to struct containing option-specific arguments
 */
static void
MailboxAddOptionValue(
    VpDevCtxType *pDevCtx,
    uint16 **ppBuf,
    VpOptionIdType optionId,
    void *value)
{
    uint16 *pBuf = *ppBuf;
    uint8 temp[3]; /* buffer for packing bytes */

    switch (optionId) {
        case VP_DEVICE_OPTION_ID_PULSE: {
            VpOptionPulseType *pulse = value;
            *pBuf++ = (uint16)pulse->makeMin;
            *pBuf++ = (uint16)pulse->makeMax;
            *pBuf++ = (uint16)pulse->breakMin;
            *pBuf++ = (uint16)pulse->breakMax;
            *pBuf++ = (uint16)pulse->interDigitMin;
            *pBuf++ = (uint16)pulse->flashMin;
            *pBuf++ = (uint16)pulse->flashMax;
            break;
        }
        case VP_DEVICE_OPTION_ID_CRITICAL_FLT: {
            VpOptionCriticalFltType *critical = value;
            temp[0] = (uint8)critical->acFltDiscEn;
            temp[1] = (uint8)critical->dcFltDiscEn;
            temp[2] = (uint8)critical->thermFltDiscEn;
            Vcp2MailboxAddBytes(&pBuf, temp, 3);
            break;
        }
        case VP_OPTION_ID_RING_CNTRL: {
            VpOptionRingControlType *ringControl = value;
            *pBuf++ = (uint16)ringControl->zeroCross;
            *pBuf++ = (uint16)ringControl->ringExitDbncDur;
            *pBuf++ = (uint16)ringControl->ringTripExitSt;
            break;
        }
        case VP_OPTION_ID_ZERO_CROSS: {
            VpOptionZeroCrossType *zeroCross = value;
            *pBuf++ = (uint16)*zeroCross;
            break;
        }
        case VP_OPTION_ID_PULSE_MODE: {
            VpOptionPulseModeType *pulseMode = value;
            *pBuf++ = (uint16)*pulseMode;
            break;
        }
        case VP_OPTION_ID_CODEC: {
            VpOptionCodecType *codec = value;
            *pBuf++ = (uint16)*codec;
            break;
        }
        case VP_OPTION_ID_PCM_HWY: {
            VpOptionPcmHwyType *pcmHwy = value;
            *pBuf++ = (uint16)*pcmHwy;
            break;
        }
        case VP_OPTION_ID_LOOPBACK: {
            VpOptionLoopbackType *loopback = value;
            *pBuf++ = (uint16)*loopback;
            break;
        }
        case VP_OPTION_ID_LINE_STATE: {
            VpOptionLineStateType *lineState = value;
            temp[0] = (uint8)lineState->battRev;
            temp[1] = (uint8)lineState->bat;
            Vcp2MailboxAddBytes(&pBuf, temp, 2);
            break;
        }
        case VP_OPTION_ID_EVENT_MASK: {
            VpOptionEventMaskType *eventMask = value;

            *pBuf++ = (uint16)eventMask->faults;
            *pBuf++ = (uint16)eventMask->signaling;

#if (VCP2_TIMESTAMP_SHIFT > 0)
            /* Never mask the TS_ROLLOVER event; the API needs it to maintain
               the high-order bits of the timestamp. */
            *(pBuf - 1) &= ~VP_DEV_EVID_TS_ROLLOVER;
#endif

            *pBuf++ = (uint16)eventMask->response;
            *pBuf++ = (uint16)eventMask->test;
            *pBuf++ = (uint16)eventMask->process;
            *pBuf++ = 0; /* eventMask->fxo; */
            *pBuf++ = 0; /* eventMask->packet; */
            break;
        }
        case VP_OPTION_ID_DTMF_MODE: {
            VpOptionDtmfModeType *pDtmfControl = value;
            *pBuf++ = (uint16)pDtmfControl->dtmfControlMode;
            *pBuf++ = (uint16)pDtmfControl->direction;
            break;
        }
        case VP_OPTION_ID_DTMF_SPEC: {
            VpOptionDtmfSpecType *dtmfSpec = value;
            *pBuf++ = (uint16)*dtmfSpec;
            break;
        }
        case VP_OPTION_ID_PCM_TXRX_CNTRL: {
            VpOptionPcmTxRxCntrlType *pPcmTxRxCntrl = value;
            *pBuf++ = (uint16)*pPcmTxRxCntrl;
            break;
        }
        case VP_DEVICE_OPTION_ID_DEVICE_IO: {
            VpOptionDeviceIoConfigType deviceIoConfig;

            /* Convert old-format struct into new format and call this function
               again. */
            MakeNewDevIoConfigStruct(pDevCtx, (VpOptionDeviceIoType *)value,
                &deviceIoConfig);
            MailboxAddOptionValue(pDevCtx, &pBuf, VP_DEVICE_OPTION_ID_DEV_IO_CFG,
                &deviceIoConfig);
            break;
        }
        case VP_DEVICE_OPTION_ID_RAMP2STBY:
            *pBuf++ = *((uint16 *)value);
            break;

        case VP_OPTION_ID_TIMESLOT: {
            VpOptionTimeslotType *pTimeSlot = value;
            *pBuf++ = (uint16)(pTimeSlot->tx);
            *pBuf++ = (uint16)(pTimeSlot->rx);
            break;
        }
        case VP_OPTION_ID_LINE_IO_CFG: {
            VpOptionLineIoConfigType *pLineIoConfig = value;
            *pBuf++ = ((uint16)pLineIoConfig->direction << 8)
                | pLineIoConfig->outputType;
            break;
        }
        case VP_DEVICE_OPTION_ID_DEV_IO_CFG: {
            VpOptionDeviceIoConfigType *pDeviceIoConfig = value;
            VpOptionLineIoConfigType *pLineIoConfig =
                &pDeviceIoConfig->lineIoConfig[0];
            uint16 temp = 0; /* silence compiler warning */
            int channelId;
            for (channelId = 0; channelId < VP_VCP2_MAX_NUM_LINES; channelId ++) {

                /* Pack each line's direction and output fields into a byte. */
                uint8 lineCfgByte = ((pLineIoConfig->direction & 0xF) << 4)
                    | (pLineIoConfig->outputType & 0xF);

                /* Pack two line bytes into each 16-bit word. */
                if (channelId % 2) {
                    *pBuf++ = (temp | lineCfgByte);
                } else {
                    temp = ((uint16)lineCfgByte << 8);
                }
                pLineIoConfig++;
            }
            break;
        }
        default:
            /* Error checking is elsewhere. */
            ;
    }

    /* Update the calling function's mailbox buffer pointer. */
    *ppBuf = pBuf;
}

/**
 * This internal function unpacks a signal arguments struct and adds it into the
 * specified command mailbox buffer.
 *
 * \param[in] ppBuf     Pointer to the calling function's pBuf variable
 * \param[in] sigType   Signal type
 * \param[in] pArgs     Pointer to struct containing signal-specific arguments
 */
static void
MailboxAddSignalArgs(
    uint16 **ppBuf,
    VpSendSignalType sigType,
    void *pArgs)
{
    uint16 *pBuf = *ppBuf;

    switch (sigType) {
        case VP_SENDSIG_MSG_WAIT_PULSE:
            if (pArgs == VP_NULL) {
                /* User wants to abort the signal. */
                *pBuf++ = 0;
                *pBuf++ = 0;
                *pBuf++ = 0;
                *pBuf++ = 0;
            } else {
                /* User wants to start sending a signal. */
                VpSendMsgWaitType *pSendMsg = pArgs;
                *pBuf++ = (uint16)pSendMsg->voltage;
                *pBuf++ = (uint16)pSendMsg->onTime;
                *pBuf++ = (uint16)pSendMsg->offTime;
                *pBuf++ = (uint16)pSendMsg->cycles;
            }
            break;
        case VP_SENDSIG_DTMF_DIGIT: {
            VpDigitType *pDigit = pArgs;
            *pBuf++ = (uint16)*pDigit;
            break;
        }
        case VP_SENDSIG_FWD_DISCONNECT:
            *pBuf++ = *(uint16 *)pArgs;
            break;
        default:
            /* Should never get here since args have already been checked in
               SignalNotSupported(). */
            break;
    }

     /* Update the calling function's mailbox buffer pointer. */
    *ppBuf = pBuf;
}

/*
 * This internal function copies a VpDeviceIoAccessDataType struct into an
 * equivalent VpDeviceIoAccessExtType struct.  It is used in converting a
 * VP_DEVICE_OPTION_ID_DEVICE_IO set-option command into a
 * VP_DEVICE_OPTION_ID_DEV_IO_CFG set-option command.
 *
 * \param[in]   pDevCtx         Device Context for the VCP2 device
 * \param[in]   pDeviceIoData   Old-format struct
 * \param[out]  pDeviceIoAccess New-format struct
 */
static void
MakeNewDevIoAccessStruct(
    VpDevCtxType *pDevCtx,
    VpDeviceIoAccessDataType *pDeviceIoData,
    VpDeviceIoAccessExtType *pDeviceIoAccess)
{
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpVcp2ProductType product = pDevObj->product;
    uint32 packedMask = pDeviceIoData->accessMask_31_0;
    uint32 packedData = pDeviceIoData->deviceIOData_31_0;
    uint8 channelId, availBits, availMask;

    /* VCP1-790 had one I/O bit per channel.  VCP1-880 had two. */
    if ((product & VP_VCP_PROD_SLAC_MASK) == VP_VCP_PROD_SLAC_VE880) {
        availBits = 2;  /* two GPIO bits per channel */
        availMask = 3;
    } else {
        availBits = 1;
        availMask = 1;
    }

    /* Copy data into VpDeviceIoAccessExtType struct. */
    for (channelId = 0; channelId < VP_VCP2_MAX_NUM_LINES; channelId++) {
        pDeviceIoAccess->lineIoBits[channelId].mask = packedMask & availMask;
        pDeviceIoAccess->lineIoBits[channelId].data = packedData & availMask;
        packedMask >>= availBits;
        packedData >>= availBits;

        /* Move to the next member of the VpDeviceIoAccessDataType struct? */
        if (channelId * availBits == 32) {
            packedMask = pDeviceIoData->accessMask_63_32;
            packedData = pDeviceIoData->deviceIOData_63_32;
        }
    }
    pDeviceIoAccess->direction = pDeviceIoData->accessType;
}

/*
 * This internal function copies a VpOptionDeviceIoType struct into an
 * equivalent VpOptionDeviceIoConfigType struct.  It is used in converting a
 * VpDeviceIoAccess() command into a VpDeviceIoAccessExt() command.
 *
 * \param[in]   pDevCtx         Device Context for the VCP2 device
 * \param[in]   pDeviceIo       Old-format struct
 * \param[out]  pDeviceIoConfig New-format struct
 */
static void
MakeNewDevIoConfigStruct(
    VpDevCtxType *pDevCtx,
    VpOptionDeviceIoType *pDeviceIo,
    VpOptionDeviceIoConfigType *pDeviceIoConfig)
{
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpVcp2ProductType product = pDevObj->product;
    uint32 packedDir = pDeviceIo->directionPins_31_0;
    uint32 packedType = pDeviceIo->outputTypePins_31_0;
    uint8 channelId, availBits, availMask;

    /* VCP1-790 had one I/O bit per channel.  VCP1-880 had two. */
    if ((product & VP_VCP_PROD_SLAC_MASK) == VP_VCP_PROD_SLAC_VE880) {
        availBits = 2;  /* two GPIO bits per channel */
        availMask = 3;
    } else {
        availBits = 1;
        availMask = 1;
    }

    /* Copy data into VpOptionDeviceIoConfigType struct. */
    for (channelId = 0; channelId < VP_VCP2_MAX_NUM_LINES; channelId++) {
        pDeviceIoConfig->lineIoConfig[channelId].direction = packedDir & availMask;
        pDeviceIoConfig->lineIoConfig[channelId].outputType = packedType & availMask;
        packedDir >>= availBits;
        packedType >>= availBits;

        /* Move to the next member of the VpOptionDeviceIoType struct? */
        if (channelId * availBits == 32) {
            packedDir = pDeviceIo->directionPins_63_32;
            packedType = pDeviceIo->outputTypePins_63_32;
        }
    }
}

/**
 * This internal function modifies the specified GPIO pin access mask to
 * conform with the line termination type selected for the specified line.
 *
 * \param[in]       pLineCtx    Line Context for a VCP2 line
 * \param[in,out]   pMask       Access mask
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_DEDICATED_PINS   The pMask argument was modified, because
 *                                      Some pins are controlled automatically
 *                                      according to the line termination type.
 */
static VpStatusType
MaskDedicatedPins(
    VpTermType termType,
    uint8 *pMask)
{
    uint8 requestedMask = *pMask;

    switch (termType) {
        case VP_TERM_FXS_75181:
            *pMask &= VCP2_GPIOMASK_FXS_75181;
            break;
        case VP_TERM_FXS_75282:
            *pMask &= VCP2_GPIOMASK_FXS_75282;
            break;
        case VP_TERM_FXS_TO_TL:
            *pMask &= VCP2_GPIOMASK_FXS_TO_TL;
            break;
        case VP_TERM_FXS_CO_TL:
            *pMask &= VCP2_GPIOMASK_FXS_CO_TL;
            break;
        case VP_TERM_FXS_GENERIC:
            *pMask &= VCP2_GPIOMASK_FXS_GENERIC;
            break;
        case VP_TERM_FXS_TITO_TL_R:
            *pMask &= VCP2_GPIOMASK_FXS_TITO_TL_R;
            break;
        case VP_TERM_FXS_RDT:
            *pMask &= VCP2_GPIOMASK_FXS_RDT;
            break;
        case VP_TERM_FXS_RR:
            *pMask &= VCP2_GPIOMASK_FXS_RR;
            break;
        default:
            /* Unknown termination type.  No restrictions. */
    }

    if (*pMask != requestedMask) {
        return VP_STATUS_DEDICATED_PINS;
    } else {
        return VP_STATUS_SUCCESS;
    }
}

/**
 * This internal function verifies that the option value passed is valid,
 * according to the specified option ID and VCP2 firmware product type.
 *
 * \param[in] pDevObj   Device Object of the VCP2 device
 * \param[in] optionId  Option ID of the supplied value
 * \param[in] value     Option value to be verified
 *
 * \retval ::TRUE   Option value is invalid or unsupported
 * \retval ::FALSE  Option value is OK
 */
static bool
OptionValueNotSupported(
    VpLineCtxType *pLineCtx,
    VpVcp2DeviceObjectType *pDevObj,
    VpOptionIdType optionId,
    void *value)
{
    switch (optionId) {
        case VP_DEVICE_OPTION_ID_PULSE: {
            VpOptionPulseType *pulse = value;
            if (
                (pulse->makeMin > pulse->makeMax) ||
                (pulse->breakMin > pulse->breakMax) ||
                (pulse->flashMin > pulse->flashMax)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_DEVICE_OPTION_ID_CRITICAL_FLT: {
            VpOptionCriticalFltType *critical = value;
            if ((pDevObj->product & VP_VCP_PROD_SLAC_MASK) == VP_VCP_PROD_SLAC_VE880) {
                /* VCP2-880 does not support state transition to disconnect
                 * upon fault detection. */
                if (
                    (critical->acFltDiscEn == TRUE) ||
                    (critical->dcFltDiscEn == TRUE)
                ) {
                    return TRUE;
                }
            }
            break;
        }
        case VP_OPTION_ID_RING_CNTRL: {
            VpOptionRingControlType *ringControl = value;
            if (LineStateNotSupported(pDevObj, ringControl->ringTripExitSt)) {
                return TRUE;
            }
            /* Intentional fall-through to check zero-cross option. */
        }
        case VP_OPTION_ID_ZERO_CROSS: {
            VpOptionZeroCrossType *zeroCross = value;
            if (*zeroCross < VP_OPTION_ZC_M4B) {
                return TRUE;
            }
            if ((pDevObj->product & VP_VCP_PROD_SLAC_MASK) == VP_VCP_PROD_SLAC_VE880) {
                if (*zeroCross > VP_OPTION_ZC_NONE) {
                    return TRUE;
                }
            } else if ((pDevObj->product & VP_VCP_PROD_SLAC_MASK) == VP_VCP_PROD_SLAC_VE790) {
                if (*zeroCross > VP_OPTION_ZC_B4M) {
                    return TRUE;
                }
            }
            break;
        }
        case VP_OPTION_ID_PULSE_MODE: {
            VpOptionPulseModeType *pulseMode = value;
            if (
                (*pulseMode != VP_OPTION_PULSE_DECODE_OFF) &&
                (*pulseMode != VP_OPTION_PULSE_DECODE_ON)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_OPTION_ID_CODEC: {
            VpOptionCodecType *codec = value;
            if (
                (*codec != VP_OPTION_ALAW) &&
                (*codec != VP_OPTION_MLAW) &&
                (*codec != VP_OPTION_LINEAR)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_OPTION_ID_PCM_HWY: {
            VpOptionPcmHwyType *pcmHwy = value;
            if (
                (*pcmHwy != VP_OPTION_HWY_A) &&
                (*pcmHwy != VP_OPTION_HWY_B)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_OPTION_ID_LOOPBACK: {
            VpOptionLoopbackType *loopback = value;
            if (
                (*loopback != VP_OPTION_LB_OFF) &&
                (*loopback != VP_OPTION_LB_TIMESLOT) &&
                (*loopback != VP_OPTION_LB_DIGITAL)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_OPTION_ID_LINE_STATE: {
            VpOptionLineStateType *lineState = value;
            if (lineState->bat >= VP_NUM_OPTION_BAT_IDS) {
                return TRUE;
            }
            break;
        }
        case VP_OPTION_ID_EVENT_MASK: {
            VpOptionEventMaskType *eventMask = value;
            /* Should return INVALID_ARG here if a nonmaskable event is masked,
               or an invalid event is unmasked.  However, to emulate VCP1
               behavior, we silently "correct" the event masks instead. */
            eventMask->faults    &= ~VP_API_NONMASK_FAULT_EVENTS;
            eventMask->signaling &= ~VP_API_NONMASK_SIGNALING_EVENTS;
            eventMask->response  &= ~VP_API_NONMASK_RESPONSE_EVENTS;
            eventMask->test      &= ~VP_API_NONMASK_TEST_EVENTS;
            eventMask->process   &= ~VP_API_NONMASK_PROCESS_EVENTS;
            eventMask->fxo       &= ~VP_API_NONMASK_FXO_EVENTS;
            eventMask->packet    &= ~VP_API_NONMASK_FXO_EVENTS;
            eventMask->faults    |= VCP2_INVALID_FAULT_EVENTS;
            eventMask->signaling |= VCP2_INVALID_SIGNALING_EVENTS;
            eventMask->response  |= VCP2_INVALID_RESPONSE_EVENTS;
            eventMask->test      |= VCP2_INVALID_TEST_EVENTS;
            eventMask->process   |= VCP2_INVALID_PROCESS_EVENTS;
            eventMask->fxo       |= VCP2_INVALID_FXO_EVENTS;
            eventMask->packet    |= VCP2_INVALID_PACKET_EVENTS;
            break;
        }
        case VP_OPTION_ID_DTMF_MODE: {
            VpOptionDtmfModeType *pDtmfControl = value;
            VpOptionDtmfModeControlType dtmfControlMode = pDtmfControl->dtmfControlMode;
            if (dtmfControlMode >= VP_NUM_OPTION_DTMF_IDS) {
                return TRUE;
            } else if (dtmfControlMode == VP_OPTION_DTMF_DECODE_ON) {
                if (pLineCtx == VP_NULL) {
                    /* Can't enable DTMF on all lines. */
                    return TRUE;
                } else if (
                    (pDtmfControl->direction != VP_DIRECTION_US) &&
                    ((pDevObj->product & VP_VCP_PROD_SLAC_MASK) == VP_VCP_PROD_SLAC_VE790)
                ) {
                    /* VCP2-790 supports only upstream DTMF detection. */
                    return TRUE;
                }
            }
            break;
        }
        case VP_OPTION_ID_DTMF_SPEC: {
            VpOptionDtmfSpecType *dtmfSpec = value;
            if (
                (*dtmfSpec != VP_OPTION_DTMF_SPEC_ATT) &&
                (*dtmfSpec != VP_OPTION_DTMF_SPEC_NTT) &&
                (*dtmfSpec != VP_OPTION_DTMF_SPEC_AUS) &&
                (*dtmfSpec != VP_OPTION_DTMF_SPEC_BRZL) &&
                (*dtmfSpec != VP_OPTION_DTMF_SPEC_ETSI)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_OPTION_ID_PCM_TXRX_CNTRL: {
            VpOptionPcmTxRxCntrlType *pPcmTxRxCntrl = value;
            if (
                (*pPcmTxRxCntrl < VP_OPTION_PCM_BOTH) ||
                (*pPcmTxRxCntrl > VP_OPTION_PCM_TX_ONLY)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_DEVICE_OPTION_ID_DEVICE_IO:
        case VP_OPTION_ID_LINE_IO_CFG:
        case VP_DEVICE_OPTION_ID_DEV_IO_CFG:
        case VP_DEVICE_OPTION_ID_RAMP2STBY:
        case VP_OPTION_ID_TIMESLOT:
            /* No argument checking for these options. */
            break;

        default:
            /* All other option IDs are invalid or not supported. */
            return TRUE;
    }

    /* Option value is OK. */
    return FALSE;
}

/**
 * This internal function performs additional processing required in
 * VpSetOption, other than sending the mailbox command.  This includes reading
 * device registers or updating the line or device object.  If no mailbox
 * command is required in addition to this special processing, this function
 * sets *pSendCommand = FALSE.
 *
 * \param[in]     pLineCtx      Line Context argument to VpSetOption (if any)
 * \param[in]     pDevObj       Device Object for the VCP2 device
 * \param[in]     optionId      optionId argument to VpSetOption
 * \param[in]     value         value argument to VpSetOption
 * \param[out]    pSendCommand  Whether the mailbox command should be sent
 */
static VpStatusType
ProcessOptionValue(
    VpLineCtxType *pLineCtx,
    VpVcp2DeviceObjectType *pDevObj,
    VpOptionIdType optionId,
    void *value,
    bool *pSendCommand)
{
    VpStatusType retval = VP_STATUS_SUCCESS;

    /* Update the status of the DTMF detectors. */
    if (optionId == VP_OPTION_ID_DTMF_MODE) {
        return UpdateDtmfStatusMembers(pLineCtx, pDevObj, value, pSendCommand);
    }

    /* Update the system interrupt mask register. */
    else if (optionId == VP_OPTION_ID_EVENT_MASK) {
        VpOptionEventMaskType *eventMask = value;
        uint16 sysIntMask = VP_SYS_FLT_VCP2_RSVD;

        if (eventMask->faults & VP_DEV_EVID_SYSTEM_FLT) {
            sysIntMask |= VP_SYS_FLT_VCP2_EVQ_OFL | VP_SYS_FLT_VCP2_WDOG |
                VP_SYS_FLT_VCP2_CFAIL;
        }

        if (VpHbiWr(pDevObj->deviceId, SYSINTMASK, &sysIntMask) == FALSE) {
            retval = VP_STATUS_ERR_HBI;
        }

#if (VCP2_TIMESTAMP_SHIFT > 0)
        /* The API never masks the timestamp rollover event internally.  But it
           needs to remember whether the user wants the event masked. */
        pDevObj->timeStampEventMasked =
            (eventMask->signaling & VP_DEV_EVID_TS_ROLLOVER) ? TRUE : FALSE;
#endif

    }

    return retval;
}

/**
 * This internal function verifies that the selected line termination type
 * supports the requested relay state.
 *
 * \param[in] rState    Relay state
 * \param[in] termType  Line termination type
 *
 * \retval ::TRUE   Relay state is not supported by termination type
 * \retval ::FALSE  Relay state is supported by termination type
 */
static bool
RelayStateNotSupported(
    VpRelayControlType rState,
    VpTermType termType)
{
    switch (rState) {
        case VP_RELAY_NORMAL:
        case VP_RELAY_RESET:
        case VP_RELAY_TESTOUT:
        case VP_RELAY_TALK:
        case VP_RELAY_RINGING:
        case VP_RELAY_TEST:
        case VP_RELAY_BRIDGED_TEST:
        case VP_RELAY_SPLIT_TEST:
        case VP_RELAY_DISCONNECT:
        case VP_RELAY_RINGING_NOLOAD:
        case VP_RELAY_RINGING_TEST:
            /* These states are supported by at least one termination type. */
            break;
        default:
            /* Relay state not supported by VCP2. */
            return TRUE;
    }

    /* Termination-type-specific checking for valid relay state: */
    switch (termType) {
        case VP_TERM_FXS_GENERIC:
            switch (rState) {
                case VP_RELAY_NORMAL:
                case VP_RELAY_TALK:
                case VP_RELAY_BRIDGED_TEST:
                    return FALSE;
                default:
                    return TRUE;
            }
        case VP_TERM_FXS_TITO_TL_R:
            switch (rState) {
                case VP_RELAY_NORMAL:
                case VP_RELAY_TALK:
                case VP_RELAY_RINGING:
                case VP_RELAY_TESTOUT:
                case VP_RELAY_DISCONNECT:
                case VP_RELAY_BRIDGED_TEST:
                case VP_RELAY_SPLIT_TEST:
                case VP_RELAY_RINGING_TEST:
                    return FALSE;
                default:
                    return TRUE;
            }
        case VP_TERM_FXS_TO_TL:
        case VP_TERM_FXS_CO_TL:
            switch (rState) {
                case VP_RELAY_NORMAL:
                case VP_RELAY_TALK:
                case VP_RELAY_TESTOUT:
                case VP_RELAY_DISCONNECT:
                case VP_RELAY_BRIDGED_TEST:
                case VP_RELAY_SPLIT_TEST:
                    return FALSE;
                default:
                    return TRUE;
            }
        case VP_TERM_FXS_75181:
            switch (rState) {
                case VP_RELAY_NORMAL:
                case VP_RELAY_TALK:
                case VP_RELAY_RINGING:
                    return FALSE;
                default:
                    return TRUE;
            }
        case VP_TERM_FXS_RDT:
            switch (rState) {
                case VP_RELAY_NORMAL:
                case VP_RELAY_TALK:
                case VP_RELAY_RINGING:
                case VP_RELAY_TESTOUT:
                case VP_RELAY_DISCONNECT:
                case VP_RELAY_BRIDGED_TEST:
                case VP_RELAY_SPLIT_TEST:
                    return FALSE;
                default:
                    return TRUE;
            }
        case VP_TERM_FXS_RR:
            switch (rState) {
                case VP_RELAY_NORMAL:
                case VP_RELAY_TALK:
                case VP_RELAY_RINGING:
                case VP_RELAY_BRIDGED_TEST:
                case VP_RELAY_RESET:
                    return FALSE;
                default:
                    return TRUE;
            }
        case VP_TERM_FXS_TI:
            switch (rState) {
                case VP_RELAY_NORMAL:
                case VP_RELAY_TALK:
                case VP_RELAY_TEST:
                case VP_RELAY_BRIDGED_TEST:
                case VP_RELAY_SPLIT_TEST:
                    return FALSE;
                default:
                    return TRUE;
            }
        case VP_TERM_FXS_75282:
            /* All relay states supported. */
        case VP_TERM_FXS_SPLITTER:
        case VP_TERM_FXS_ISOLATE:
            /* Not currently implemented.  No restrictions. */
        default:
            /* Unknown termination type.  Presume it supports all relay states. */
            return FALSE;
    }
}

/**
 * This internal function sets the base address register of the Code Load Page
 * (Page 255).
 *
 * \param[in] deviceId  Device ID of the VCP2 device
 * \param[in] ambaAddr  AMBA address to be assigned to the register
 *
 * \retval ::TRUE   Register set successfully.
 * \retval ::FALSE  HBI error.
 */
static bool
SetCodeLoadBase(
    VpDeviceIdType deviceId,
    uint32 ambaAddr)
{
    uint16 p255base[2];

    p255base[0] = (ambaAddr >> 16) & 0xFFFF;
    p255base[1] = ambaAddr & 0xFFFF;

    return VpHbiWr(deviceId, P255_BASE, p255base);
}

/**
 * This internal function verifies that the signal args passed are valid,
 * according to the specified signal type and VCP2 firmware product type.
 *
 * \param[in] pDevObj   Device Object of the VCP2 device
 * \param[in] sigType   Signal type of the supplied args
 * \param[in] pArgs     Signal args to be verified
 *
 * \retval ::TRUE   Signal/args are invalid or unsupported
 * \retval ::FALSE  Signal/args are OK
 */
static bool
SignalNotSupported(
    VpVcp2DeviceObjectType *pDevObj,
    VpSendSignalType sigType,
    void *pArgs)
{
    switch (sigType) {
        case VP_SENDSIG_MSG_WAIT_PULSE:
            if ((pDevObj->product & VP_VCP_PROD_SLAC_MASK) != VP_VCP_PROD_SLAC_VE792) {
                /* This is currently only supported by VCP2-792. */
                return TRUE;
            }
            break;
        case VP_SENDSIG_DTMF_DIGIT: {
            VpDigitType *pDigit = pArgs;
            if (
                (pDigit == VP_NULL) ||
                (!VpIsDigit(*pDigit))
            ) {
                return TRUE;
            }
            break;
        }
        case VP_SENDSIG_POLREV_PULSE:
        case VP_SENDSIG_FWD_DISCONNECT:
            if (pArgs == VP_NULL) {
                return TRUE;
            }
            break;
        default:
            /* Unknown signal type. */
            return TRUE;
    }

    /* Signal type and args are OK. */
    return FALSE;
}

/**
 * This internal function updates the dtmfResourcesRemaining and
 * dtmfDetectionEnabled members on the supplied pDtmfControl struct, depending
 * on the action (enable, disable or neither) and target (all lines, or a
 * single line).  It sets the pSendCommand variable to true or false to
 * indicate whether it is necessary to send an HBI mailbox command to the VCP2.
 *
 * \param[in]     pLineCtx      Line Context argument to VpSetOption
 * \param[in]     pDevObj       Device Object for the VCP2 device
 * \param[in,out] pDtmfControl  VP_OPTION_ID_DTMF_MODE arguments
 * \param[out]    pSendCommand  Whether the mailbox command should be sent
 */
static VpStatusType
UpdateDtmfStatusMembers(
    VpLineCtxType *pLineCtx,
    VpVcp2DeviceObjectType *pDevObj,
    VpOptionDtmfModeType *pDtmfControl,
    bool *pSendCommand)
{
    VpDeviceIdType deviceId = pDevObj->deviceId;
    VpOptionDtmfModeControlType dtmfControlMode = pDtmfControl->dtmfControlMode;
    uint8 *dtmfDetectionEnabled = pDtmfControl->dtmfDetectionEnabled;

#if (VCP2_LINE_FLAG_BYTES < VP_LINE_FLAG_BYTES)
    /* Clear all unused dtmfDetectionEnabled bits. */
    {
        int byte;
        for (byte = VCP2_LINE_FLAG_BYTES; byte < VP_LINE_FLAG_BYTES; byte++) {
            pDtmfMode->dtmfDetectionEnabled[byte] = 0;
        }
    }
#endif

    if (dtmfControlMode == VP_OPTION_DTMF_DECODE_OFF) {
        if (pLineCtx == VP_NULL) {

            /* DTMF detection is being disabled on all lines. */
            pDtmfControl->dtmfResourcesRemaining = GetTotalDtmfResources(pDevObj);
        } else {

            /* DTMF detection is being disabled on a single line. */
            VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
            int lineByte = pLineObj->channelId / 8;
            uint8 lineBit = 1 << (pLineObj->channelId % 8);

            /* Read the status registers from the VCP2. */
            VpStatusType retval = Vcp2ReadDtmfStatusRegs(deviceId, pDtmfControl, TRUE);
            if (retval != VP_STATUS_SUCCESS) {
                return retval;
            }

            /* Update the DTMF resources member only if not already disabled. */
            if (dtmfDetectionEnabled[lineByte] & lineBit) {
                pDtmfControl->dtmfResourcesRemaining++;
            }

            /* Clear the relevant dtmfDetectionEnabled bit. */
            dtmfDetectionEnabled[lineByte] &= ~lineBit;
        }
    } else {

        /* Read the status registers from the VCP2. */
        VpStatusType retval = Vcp2ReadDtmfStatusRegs(deviceId, pDtmfControl, TRUE);
        if (retval != VP_STATUS_SUCCESS) {
            return retval;
        }

        if (dtmfControlMode == VP_OPTION_DTMF_DECODE_ON) {

            /* DTMF detection is being enabled on a single line. */
            VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
            int lineByte = pLineObj->channelId / 8;
            uint8 lineBit = 1 << (pLineObj->channelId % 8);

            /* Update the DTMF resources member only if not already enabled. */
            if ((dtmfDetectionEnabled[lineByte] & lineBit) == 0) {
                pDtmfControl->dtmfResourcesRemaining--;
            }

            /* Set the relevant dtmfDetectionEnabled bit. */
            dtmfDetectionEnabled[lineByte] |= lineBit;

        } else {

            /* VP_OPTION_DTMF_DECODE_GET_STATUS */
            *pSendCommand = FALSE;
        }
    }

    /* Copy first 32 bits into deprecated member. */
    pDtmfControl->dtmfDetectionSetting = (uint32)dtmfDetectionEnabled[0]
        | (uint32)dtmfDetectionEnabled[1] << 8
        | (uint32)dtmfDetectionEnabled[2] << 16
        | (uint32)dtmfDetectionEnabled[3] << 24;

    return VP_STATUS_SUCCESS;
}

#endif /* VP_CC_VCP2_SERIES */
