/** \file vcp2_init.c
 * vcp2_init.c
 *
 * This file contains the implementation for the Initialization functions
 * in the VCP2 library.
 *
 * The functions herein are used internally by the API and should not be
 * called directly by the application.
 *
 * Copyright (c) 2008, Zarlink Semiconductor, Inc.
 */

/* Only compile this file if specified in vp_api_cfg.h: */
#include "vp_api_cfg.h"
#if defined (VP_CC_VCP2_SERIES)

#include "vcp2_api.h"
#include "vp_api_int.h"
#include "vcp2_api_int.h"
#include "boot_common.h"
#include "sys_service.h"
#include "vp_debug.h"


/* =================================
    Prototypes for Static Functions
   ================================= */

static VpStatusType
BootLoadCmdMboxWait(
    VpDeviceIdType deviceId);

static bool
HaltDevice(
    VpDeviceIdType deviceId);

static bool
HbiBootPrepare(
    VpDeviceIdType deviceId);

static VpStatusType
InitLineInt(
    VpLineCtxType *pLineCtx,
    VpProfilePtrType pAcProfile,
    VpProfilePtrType pDcProfile,
    VpProfilePtrType pRingProfile,
    MboxCmdIdType mailBoxCmd);

static bool
ResetDevice(
    VpDeviceIdType deviceId,
    uint16 resetType);

static bool
ValidateChecksum(
    VpDeviceIdType deviceId);


/* ==========================
    Initialization Functions
   ========================== */

/**
 * This function implements VpBootLoad() for VCP2 devices.  It resets the
 * VCP2 and loads a firmware image, either through the HBI, or from a Flash
 * device attached to the VCP2.  Refer to the VP-API Reference Guide for
 * more details about ::VpBootLoad().
 *
 * \note
 * This function must be called after ::VpMakeDeviceObject() and before
 * ::VpMakeLineObject().
 *
 * \param[in] pDevCtx       Pointer to the Device Context struct identifying
 *                          the device
 * \param[in] state         Boot state (for multi-block booting)
 * \param[in] pImageBuffer  Pointer to HBI firmware image (or NULL for Flash)
 * \param[in] bufferSize    Size in bytes of the HBI firmware image
 * \param[in] pScratchMem   Pointer to scratch memory (for decompression)
 * \param[in] validation    Whether the image checksum should be verified
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_FUNC_NOT_SUPPORTED   Flash booting is not yet implemented.
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2BootLoad(
    VpDevCtxType *pDevCtx,
    VpBootStateType state,
    VpImagePtrType pImageBuffer,
    uint32 bufferSize,
    VpScratchMemType *pScratchMem,
    VpBootModeType validation)
{
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16p pRecCnt = &pDevObj->recCnt;
    VpStatusType status = VP_STATUS_SUCCESS;
    uint16 vcp2ImageBlockOffset, vp792ImageBlockOffset;
    uint32 vcp2ImageByteLength, vp792ImageByteLength;
    uint16 vp792ImageBlocks, block;
    uint16 mbBuf[BUFSIZE_SLACBOOTLOAD];
    uint16 *pBuf = mbBuf;
    uint16 flag = VCP2_HBI_SWFLAG_CMD_MBOX;
    uint8 writeWords;
    
    if (pImageBuffer == VP_NULL) {
        /* No application image supplied; Flash boot requested. */
        /* \todo Flash booting not yet implemented. */
        VP_DOUT(VP_DBG_ERROR, ("Vcp2BootLoad(): %s.\n", "Flash booting not yet implemented"));
        return VP_STATUS_FUNC_NOT_SUPPORTED;
    }

    /* If the first 8 bytes spell out "ZARLINK", then we have a combined VCP2
     * and VP792 firmware image.  Otherwise, it's a normal VCP2 image. */
    if (!(pImageBuffer[0] == 'Z' &&
          pImageBuffer[1] == 'A' &&
          pImageBuffer[2] == 'R' &&
          pImageBuffer[3] == 'L' &&
          pImageBuffer[4] == 'I' &&
          pImageBuffer[5] == 'N' &&
          pImageBuffer[6] == 'K' &&
          pImageBuffer[7] == 0)) {
        VP_DOUT(DBG_INFO, ("Vcp2BootLoad(): Loading regular VCP2 image."));
        return VpBootLoadInternal(state, pImageBuffer, bufferSize,
            pScratchMem, validation, deviceId, pRecCnt, HbiBootPrepare,
            HaltDevice, ValidateChecksum);
    }

    VP_DOUT(DBG_INFO, ("Vcp2BootLoad(): Combined VCP2+VP792 firmware image detected."));

    /* Find the block offsets of the VCP2 and VP792 images in the combined image */
    vcp2ImageBlockOffset = pImageBuffer[VCP2_792_IMAGE_VCP2_START + 1];
    vcp2ImageBlockOffset |= pImageBuffer[VCP2_792_IMAGE_VCP2_START] << 8;

    vp792ImageBlockOffset = pImageBuffer[VCP2_792_IMAGE_792_START + 1];
    vp792ImageBlockOffset |= pImageBuffer[VCP2_792_IMAGE_792_START] << 8;

    /* Find the lengths of the VCP2 and VP792 images */
    vcp2ImageByteLength = pImageBuffer[VCP2_792_IMAGE_VCP2_LENGTH + 3];
    vcp2ImageByteLength |= pImageBuffer[VCP2_792_IMAGE_VCP2_LENGTH + 2] << 8;
    vcp2ImageByteLength |= pImageBuffer[VCP2_792_IMAGE_VCP2_LENGTH + 1] << 16;
    vcp2ImageByteLength |= pImageBuffer[VCP2_792_IMAGE_VCP2_LENGTH] << 24;

    vp792ImageByteLength = pImageBuffer[VCP2_792_IMAGE_792_LENGTH + 3];
    vp792ImageByteLength |= pImageBuffer[VCP2_792_IMAGE_792_LENGTH + 2] << 8;
    vp792ImageByteLength |= pImageBuffer[VCP2_792_IMAGE_792_LENGTH + 1] << 16;
    vp792ImageByteLength |= pImageBuffer[VCP2_792_IMAGE_792_LENGTH] << 24;
    vp792ImageBlocks = vp792ImageByteLength / 128;

    VP_DOUT(DBG_INFO, ("Vcp2BootLoad(): \n\tVCP2 block offset: %d\n\tVP792 block offset: %d\n\tVCP2 length (bytes): %ld\n\tVP792 image blocks: %d", vcp2ImageBlockOffset, vp792ImageBlockOffset, vcp2ImageByteLength, vp792ImageBlocks));

    /* Load the VCP2 firmware */
    status = VpBootLoadInternal(state,
        pImageBuffer + (vcp2ImageBlockOffset * 128), vcp2ImageByteLength,
        pScratchMem, validation, deviceId, pRecCnt, HbiBootPrepare,
        HaltDevice, ValidateChecksum);
    if (status != VP_STATUS_SUCCESS) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2BootLoad(): VCP2 bootload failed. status %d", status));
        return status;
    }
    VP_DOUT(DBG_INFO, ("Vcp2BootLoad(): VCP2 image loaded."));

    /* Wait for the command mailbox, indicating that the VCP2 is ready to start
     * the SLAC bootload */
    status = BootLoadCmdMboxWait(deviceId);
    if (status != VP_STATUS_SUCCESS) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2BootLoad(): Waiting for cmd mailbox failed. status %d", status));
        return status;
    }

    /* Send the mailbox command to begin the SLAC bootload */
    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_SLACBOOTLOAD, VP_ALL_LINES);
    /* Add the length argument. */
    *pBuf++ = (uint16)vp792ImageBlocks;
    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    status = Vcp2MailboxSend(deviceId, mbBuf, pBuf);
    if (status != VP_STATUS_SUCCESS) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2BootLoad(): Failed to send SLAC bootload mailbox cmd. status %d", status));
        return status;
    }

    /* We only need the VP792 image from now on, so set buffer pointer to the
     * start of the VP792 image */
    pImageBuffer += vp792ImageBlockOffset * 128;
    block = 0;
    while (block < vp792ImageBlocks) {
        /* If only one block remains, set the write length for one block,
         * otherwise two. */
        if (vp792ImageBlocks - block == 1) {
            writeWords = 63;
        } else {
            writeWords = 127;
        }

        status = BootLoadCmdMboxWait(deviceId);
        if (status != VP_STATUS_SUCCESS) {
            VP_DOUT(VP_DBG_ERROR, ("Vcp2BootLoad(): Waiting for cmd mailbox failed. status %d", status));
            return status;
        }

        /* Dump the block(s) to the cmd mailbox page.  Using separate VpHalHbiCmd
         * and VpHalHbiBootWr commands because VpHbiWr takes a buffer of words.
         * Ours is a buffer of bytes, and is not necessarily word-aligned. */
        if(!VpHalHbiCmd(deviceId, HBI_CMD_SELECT_PAGE(VCP2_HBI_CMD_MBOX_PAG))) {
            VP_DOUT(VP_DBG_ERROR, ("Vcp2BootLoad(): Select page failed"));
            return VP_STATUS_ERR_HBI;
        };
        if(!VpHalHbiCmd(deviceId, HBI_CMD_PAGED_WRITE(0, writeWords))) {
            VP_DOUT(VP_DBG_ERROR, ("Vcp2BootLoad(): Command write failed"));
            return VP_STATUS_ERR_HBI;
        };
        if(!VpHalHbiBootWr(deviceId, writeWords, pImageBuffer + block * 128)) {
            VP_DOUT(VP_DBG_ERROR, ("Vcp2BootLoad(): Block data write failed"));
            return VP_STATUS_ERR_HBI;
        };
        
        /* Release control of the mailbox to the VCP2. */
        if (VpHbiWr(deviceId, SW_FLAGS, &flag) == FALSE) {
            VP_DOUT(VP_DBG_ERROR, ("Vcp2BootLoad(): Kicking mailbox failed."));
            return VP_STATUS_ERR_HBI;
        }

        block += 2;
    }
    VP_DOUT(DBG_INFO, ("Vcp2BootLoad(): Finished successfully"));

    return status;

} /* Vcp2BootLoad */

/**
 * Starts the SLAC (CODEC) calibration sequence for one of the SLAC devices
 * controlled by a VCP2.  The VCP2 determines which SLAC to calibrate by
 * determining which SLAC the specified Line Context is associated with.
 * Refer to the VoicePath API User's Guide for details about ::VpCalCodec().
 *
 * \param[in]   pLineCtx    Line Context for any line controlled by the SLAC
 *                          to be calibrated
 * \param[in]   mode        SLAC calibration conditions
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG  Invalid mode
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2CalCodec(
    VpLineCtxType *pLineCtx,
    VpDeviceCalType mode)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_CALCODEC];
    uint16 *pBuf = buf;
    VpStatusType retval;

    if(pDevObj->devInit == FALSE) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2CalCodec(): %s.\n", "Device not initialized"));
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Check arguments. */
    if (mode >= VP_NUM_DEV_CAL_TYPES) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2CalCodec(): %s.\n", "Invalid mode argument"));
        return VP_STATUS_INVALID_ARG;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_CALCODEC, channelId);

    /* Add the "mode" arg to the mailbox buffer. */
    *pBuf++ = (uint16)mode;

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * Starts the SLIC calibration sequence for one of the SLIC devices
 * controlled by a VCP2.  Refer to the VoicePath API User's Guide for details
 * about ::VpCalSlic().
 *
 * \param[in]   pLineCtx    Line Context for the line to be calibrated
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2CalLine(
    VpLineCtxType *pLineCtx)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_CALLINE];
    uint16 *pBuf = buf;
    VpStatusType retval;

    if (pDevObj->devInit == FALSE) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2CalLine(): %s.\n", "Device not initialized"));
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_CALLINE, channelId);

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function applies the specified profiles to a line controlled by a
 * VCP2.  Refer to the VoicePath API User's Guide for details about
 * VpConfigLine().
 *
 * \param[in] pLineCtx      Line Context for the line to be configured
 * \param[in] pAcProfile    AC Profile to be applied to the line
 * \param[in] pDcProfile    DC Profile to be applied to the line
 * \param[in] pRingProfile  Ringing Profile to be applied to the line
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG          pLineCtx was NULL
 * \retval ::VP_STATUS_ERR_PROFILE
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2ConfigLine(
    VpLineCtxType *pLineCtx,
    VpProfilePtrType pAcProfile,
    VpProfilePtrType pDcProfile,
    VpProfilePtrType pRingProfile)
{
    return InitLineInt(pLineCtx, pAcProfile, pDcProfile, pRingProfile,
        VCP2_HBI_CMD_CONFIGLINE);
}

/**
 * This function initializes the Caller ID buffer (FSK data buffer) for the
 * specified line.  Refer to the VoicePath API User's Guide for details about
 * ::VpInitCid().
 *
 * \param[in]   pLineCtx        Line Context for the line to be configured
 * \param[in]   length          Amount of data to be put into the buffer
 * \param[in]   pCidData        Pointer to the data
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG  Length out of range, or pCidData == NULL
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2InitCid(
    VpLineCtxType *pLineCtx,
    uint8 length,
    uint8p pCidData)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_INITCID];
    uint16 *pBuf = buf;
    VpStatusType retval;

    if (pDevObj->devInit == FALSE) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2InitCid(): %s.\n", "Device not initialized"));
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Check arguments. */
    if ((length > MAX_CID_DATA_LENGTH) || (length == 0)) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2InitCid(): %s.\n", "Data length out of range"));
        return VP_STATUS_INVALID_ARG;
    }
    if (pCidData == VP_NULL) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2InitCid(): %s.\n", "Data pointer must not be NULL"));
        return VP_STATUS_INVALID_ARG;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_INITCID, channelId);

    /* Add length field to the mailbox buffer. */
    *pBuf++ = (uint16)length;

    /* Add the Caller ID data. */
    Vcp2MailboxAddBytes(&pBuf, pCidData, (int)length);

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/*
 * This function calibrates the device and initializes all lines on the device
 * using the supplied AC, DC, and Ringing parameters.  See the ::VpInitDevice()
 * discussion in the VP-API Reference Guide for more information.
 *
 * \note:
 * This function should only be called after VpBootLoad() and the subsequent
 * boot-complete event.
 *
 * \param[in] pDevCtx       Pointer to the Device Context struct identifying
 *                          the device to be initialized
 * \param[in] pDevProfile   Pointer to the Device Profile for the device
 * \param[in] pAcProfile    Pointer to an AC Profile to be applied to all
 *                          lines controlled by the device
 * \param[in] pDcProfile    Pointer to a DC Profile to be applied to all
 *                          lines controlled by the device
 * \param[in] pRingProfile  Pointer to a Ringing Profile to be applied to all
 *                          lines controlled by the device
 * \param[in] pRingProfile  Pointer to a Ringing Profile to be applied to all
 *                          lines controlled by the device
 * \param[in] pFxoAcProfile (unused)
 * \param[in] pFxoCfgProfile (unused)
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG  Device Profile pointer is NULL
 * \retval ::VP_STATUS_ERR_PROFILE
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 * \retval ::VP_STATUS_FUNC_NOT_SUPPORTED  FXO profile(s) non-NULL.
 */
VpStatusType
Vcp2InitDevice(
    VpDevCtxType *pDevCtx,
    VpProfilePtrType pDevProfile,
    VpProfilePtrType pAcProfile,
    VpProfilePtrType pDcProfile,
    VpProfilePtrType pRingProfile,
    VpProfilePtrType pFxoAcProfile,
    VpProfilePtrType pFxoCfgProfile)
{
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 mboxBuf[BUFSIZE_INITDEVICE];
    uint16 *pBuf = mboxBuf;
    VpStatusType retval = VP_STATUS_SUCCESS;

    /* Check arguments. */
    if (pDevProfile == VP_PTABLE_NULL) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2InitDevice(): %s.\n", "Device profile must not be NULL"));
        return VP_STATUS_INVALID_ARG;
    }
    if (
        (pFxoAcProfile != VP_PTABLE_NULL) ||
        (pFxoCfgProfile != VP_PTABLE_NULL)
    ) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2InitDevice(): %s.\n", "FXO profiles must be NULL"));
        return VP_STATUS_FUNC_NOT_SUPPORTED;
    }
    if (
        Vcp2BadProfileArg(pDevObj, VP_PROFILE_DEVICE, pDevProfile) ||
        Vcp2BadProfileArg(pDevObj, VP_PROFILE_AC, pAcProfile) ||
        Vcp2BadProfileArg(pDevObj, VP_PROFILE_DC, pDcProfile) ||
        Vcp2BadProfileArg(pDevObj, VP_PROFILE_RING, pRingProfile)
    ) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2InitDevice(): %s.\n", "Bad profile argument(s)"));
        return VP_STATUS_ERR_PROFILE;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_INITDEVICE, VP_ALL_LINES);

    /* Add the initial line state to the mailbox buffer. */
    *pBuf++ = (uint16)VP_LINE_DISCONNECT;

    /* Add the profiles to the mailbox buffer. */
    Vcp2MailboxAddProfile(&pBuf, VP_PROFILE_DEVICE, pDevProfile);
    Vcp2MailboxAddProfile(&pBuf, VP_PROFILE_AC, pAcProfile);
    Vcp2MailboxAddProfile(&pBuf, VP_PROFILE_DC, pDcProfile);
    Vcp2MailboxAddProfile(&pBuf, VP_PROFILE_RING, pRingProfile);

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    if (retval == VP_STATUS_SUCCESS) {
        retval = Vcp2MailboxSend(deviceId, mboxBuf, pBuf);
    }

    /* Note that the device is not initialized. */
    pDevObj->devInit = FALSE;

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function requests initialization of a line controlled by a VCP2.
 * The VCP2 applies the given profiles, places the line into a default state,
 * resets line-specific options to defaults, and generates an "initialization
 * complete" event.  Refer to the VoicePath API User's Guide for details about
 * ::VpInitLine().
 *
 * \param[in] pLineCtx      Line Context for the line to be initialized
 * \param[in] pAcProfile    AC Profile to be applied to the line
 * \param[in] pDcProfile    DC Profile to be applied to the line
 * \param[in] pRingProfile  Ringing Profile to be applied to the line
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG          pLineCtx was NULL
 * \retval ::VP_STATUS_ERR_PROFILE
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2InitLine(
    VpLineCtxType *pLineCtx,
    VpProfilePtrType pAcProfile,
    VpProfilePtrType pDcProfile,
    VpProfilePtrType pRingProfile)
{
    return InitLineInt(pLineCtx, pAcProfile, pDcProfile, pRingProfile,
        VCP2_HBI_CMD_INITLINE);
}

/**
 * This function initializes metering parameters for the specified line.
 * Refer to the VoicePath API User's Guide for details about
 * ::VpInitMeter().
 *
 * \param[in]   pLineCtx        Line Context for the line to be configured
 * \param[in]   pMeterProfile   Metering Profile to be applied to the line
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_ERR_PROFILE
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2InitMeter(
    VpLineCtxType *pLineCtx,
    VpProfilePtrType pMeterProfile)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_INITMETER];
    uint16 *pBuf = buf;
    VpStatusType retval;

    if (pDevObj->devInit == FALSE) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2InitMeter(): %s.\n", "Device not initialized"));
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Check arguments. */
    if (Vcp2BadProfileArg(pDevObj, VP_PROFILE_METER, pMeterProfile)) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2InitMeter(): %s.\n", "Bad profile argument"));
        return VP_STATUS_ERR_PROFILE;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_INITMETER, channelId);

    /* Add the Metering Profile. */
    Vcp2MailboxAddProfile(&pBuf, VP_PROFILE_METER, pMeterProfile);

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * Initializes the specified entry in the VCP2's Profile Table with the
 * specified profile.  Refer to the VoicePath API User's Guide for details
 * about ::VpInitProfile().
 *
 * \param[in]   pDevCtx         Device Context for the VCP2 device
 * \param[in]   profileType     Type of Profile Table entry to be initialized
 * \param[in]   pProfileIndex   Index of Profile Table entry to be initialized
 * \param[in]   pProfile        Pointer to profile data
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG  Unsupported profileType or pProfileIndex
 *                                  out of range
 * \retval ::VP_STATUS_ERR_PROFILE  Can't copy one Profile Table entry to
 *                                  another
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2InitProfile(
    VpDevCtxType *pDevCtx,
    VpProfileType profileType,
    VpProfilePtrType pProfileIndex,
    VpProfilePtrType pProfile)
{
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[MAX_PROFILE_LEN];
    uint16 *pBuf = buf;
    VpStatusType retval = VP_STATUS_SUCCESS;
    int idx = VpGetProfileIndex(pProfileIndex);
    uint16p profTabInit = pDevObj->profTabInit;

    /* Make sure profileType is a supported profile type. */
    if (profileType >= VP_VCP2_NUM_PROFILE_TYPES) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2InitProfile(): %s.\n", "Invalid profile type"));
        return VP_STATUS_INVALID_ARG;
    }

    /* pProfileIndex must be between VP_PTABLE_INDEX1 and VP_PTABLE_INDEX15. */
    if ((idx == -1) || (idx > Vcp2ProfileTableSize(profileType) - 1)) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2InitProfile(): %s.\n", "Profile index out of range"));
        return VP_STATUS_INVALID_ARG;
    }

    /* Cannot accept VP_PTABLE_INDEXx for pProfile. */
    if (VpGetProfileIndex(pProfile) != -1) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2InitProfile(): %s.\n", "Can't accept an index for profile pointer."));
        return VP_STATUS_ERR_PROFILE;
    }

    /* Verify that the profile is of the correct type and length. */
    if (Vcp2BadProfileArg(pDevObj, profileType, pProfile)) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2InitProfile(): %s.\n", "Profile too long or of wrong type."));
        return VP_STATUS_ERR_PROFILE;
    }

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    if (pProfile != VP_PTABLE_NULL) {
        int lenBytes, lenWords, page, offset;

        /* Extract length field from profile. */
        lenBytes = pProfile[VP_PROFILE_LENGTH] + 4;
        lenWords = (lenBytes + 1) / 2;

        /* Write profile data to stack buffer (word-aligned and big-endian). */
        Vcp2MailboxAddBytes(&pBuf, pProfile, lenBytes);

        /* Find out where this entry goes in the Profile Table. */
        Vcp2LocateProfileTableEntry(idx, profileType, &page, &offset);

        /* Send the buffer to the VCP2. */
        if (VpHbiWr(deviceId, page, offset, lenWords - 1, buf) == FALSE) {
            retval = VP_STATUS_ERR_HBI;
        }

        /* Mark the Profile Table entry as initialized. */
        profTabInit[profileType] |= (1 << idx);
    } else {

        /* If pProfile is NULL, mark the Profile Table entry as empty. */
        profTabInit[profileType] &= ~(1 << idx);
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * Initializes ring cadence and Caller ID options in preparation for
 * calling ::VpSetLineState(RINGING) on a particular line.  Refer to the VP
 * API User's Guide for details about ::VpInitRing().
 *
 * \param[in]   pLineCtx        Line Context for the line to be configured
 * \param[in]   pRingCadProfile Ring Cadence profile to be applied to the line
 * \param[in]   pCidProfile     Caller ID profile to be applied to the line
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_ERR_PROFILE
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2InitRing(
    VpLineCtxType *pLineCtx,
    VpProfilePtrType pRingCadProfile,
    VpProfilePtrType pCidProfile)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_INITRING];
    uint16 *pBuf = buf;
    int idx = VpGetProfileIndex(pCidProfile);
    VpStatusType retval;

    if (pDevObj->devInit == FALSE) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2InitRing(): %s.\n", "Device not initialized"));
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Check arguments. */
    if (
        Vcp2BadProfileArg(pDevObj, VP_PROFILE_RINGCAD, pRingCadProfile) ||
        Vcp2BadProfileArg(pDevObj, VP_PROFILE_CID, pCidProfile)
    ) {
        VP_DOUT(VP_DBG_ERROR, ("Vcp2InitRing(): %s.\n", "Bad profile argument(s)"));
        return VP_STATUS_ERR_PROFILE;
    }

    /* Currently, the VCP2 only supports passing a Profile Table index (not a
       pointer to a profile) for Caller ID profiles. */
    if ((pCidProfile != VP_PTABLE_NULL) && (idx < 0)) {
        return VP_STATUS_INVALID_ARG;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_INITRING, channelId);

    /* Add the Ring Cadence Profile. */
    Vcp2MailboxAddProfile(&pBuf, VP_PROFILE_RINGCAD, pRingCadProfile);

    /* Add the Caller ID Profile. */
    Vcp2MailboxAddProfile(&pBuf, VP_PROFILE_CID, pCidProfile);

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function enables or disables the use of the programmed battery voltages
 * used by the device for dc feed purposes.  Refer to the VP API User's Guide
 * for details about ::VpSetBatteries().
 *
 * \param[in] pLineCtx      Line Context struct for the line
 * \param[in] battMode      Control for device-measured or programmed batt mode
 * \param[in] pBatt         Battery voltage struct
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG          EN mode with NULL pBatt
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2SetBatteries(
    VpLineCtxType *pLineCtx,
    VpBatteryModeType battMode,
    VpBatteryValuesType *pBatt)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_SETBATTERIES];
    uint16 *pBuf = buf;
    VpStatusType retval;

    if (pDevObj->devInit == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Check arguments. */
    if ((battMode == VP_BATT_MODE_EN) && (pBatt == VP_NULL)) {
        return VP_STATUS_INVALID_ARG;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_SETBATTERIES, channelId);

    *pBuf++ = (uint16)battMode;
    if (battMode == VP_BATT_MODE_EN) {
        *pBuf++ = pBatt->batt1;
        *pBuf++ = pBatt->batt2;
        *pBuf++ = pBatt->batt3;
    }

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function resets the VCP2 without requiring that the firmware be
 * reloaded with VpBootLoad().  Refer to the VoicePath API User's Guide
 * for details about VpSoftReset().
 *
 * \param[in]   pDevCtx         Device Context for the VCP2 device
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2SoftReset(
    VpDevCtxType *pDevCtx)
{
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    if (ResetDevice(deviceId, VCP2_HBI_CLKGEN_STATUS_HI_SOFT_RESET) == FALSE) {
        return VP_STATUS_ERR_HBI;
    }

    return VP_STATUS_SUCCESS;
}


/* ==================
    Static Functions
   ================== */

/**
 *  This function ensures that the host has the right to write to the VCP2's
 *  command mailbox.  If the VCP2 owns the mailbox (i.e. is busy), this
 *  function returns VP_STATUS_MAILBOX_BUSY.
 *
 * \note If VP_VCP2_BOOTLOAD_MAILBOX_SPINWAIT is defined, this function
 *       continuously polls the VCP2 until the mailbox is available.  This
 *       function polls VP_VCP2_BOOTLOAD_MAILBOX_SPINWAIT times before giving
 *       up.
 *
 * \param[in] deviceId  Device ID identifying the VCP2 device.
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
static VpStatusType
BootLoadCmdMboxWait(
    VpDeviceIdType deviceId)
{
    uint16 flags;

#if (VP_VCP2_BOOTLOAD_MAILBOX_SPINWAIT > 0)
    uint32 i;
    for (i = VP_VCP2_BOOTLOAD_MAILBOX_SPINWAIT; i; i--) {
#endif

        /* Read mailbox flags register. */
        if (VpHbiRd(deviceId, SW_FLAGS, &flags) == FALSE) {
            return VP_STATUS_ERR_HBI;
        }

        /* Check command mailbox flag: 0 => host owns; 1=> VCP2 owns. */
        if ((flags & VCP2_HBI_SWFLAG_CMD_MBOX) == 0) {
            return VP_STATUS_SUCCESS;
        }

#if (VP_VCP2_BOOTLOAD_MAILBOX_SPINWAIT > 0)
        if (i % (VP_VCP2_BOOTLOAD_MAILBOX_SPINWAIT / 10) == 0) {
            VP_DOUT(DBG_WARNING, ("Waiting for VCP2 command mailbox"));
        }
    }
#endif

    return VP_STATUS_MAILBOX_BUSY;
}

/**
 * This internal function halts the VCP2 device.
 *
 * \param[in] deviceId  Device ID for the VCP2 device
 *
 * \retval ::TRUE   Device successfully halted.
 * \retval ::FALSE  Device could not be halted.
 */
static bool
HaltDevice(
    VpDeviceIdType deviceId)
{
    /* \todo This function is not yet implemented. */
    return TRUE;
}

/**
 * This internal function prepares a VCP2 device for HBI boot loading.  When
 * this function returns successfully, the VCP2 device is in a known state
 * allowing a firmware application to be reliably loaded into its memory.
 *
 * \param[in] deviceId  Device ID for the VCP2 device
 *
 * \retval ::TRUE   Device is ready to be boot-loaded.
 * \retval ::FALSE  Reset sequence did not complete successfully.
 */
static bool
HbiBootPrepare(
    VpDeviceIdType deviceId)
{
    uint16 temp[2];
    int i;

    /* Issue "HBI boot" reset command to the VCP2. */
    if (ResetDevice(deviceId, VCP2_HBI_CLKGEN_STATUS_HI_HBI_BOOT) == FALSE) {
        return FALSE;
    }

#ifdef VP_VCP2_BOARD_BOOT_DEBUG
    /* Make sure we can read and write to HBI registers. */
    temp[0] = 0x1234;
    temp[1] = 0x5678;
    if (VpHbiWr(deviceId, P255_BASE, temp) == FALSE) {
        return FALSE;
    }
    temp[0] = temp[1] = 0;
    if (VpHbiRd(deviceId, P255_BASE, temp) == FALSE) {
        return FALSE;
    }
    if ((temp[0] != 0x1234) || (temp[1] != 0x5600)) {
        VP_DOUT(VP_DBG_ERROR, ("HbiBootPrepare(): HBI access failed (%4.4X %4.4X != 1234 5600).\n",
            (int)temp[0], (int)temp[1]));
        return FALSE;
    }
#endif

#if (VP_VCP2_BOARD_CLKSRC == 0)
    /*
     * We don't want to program these registers unnecessarily, because there
     * would be a race condition unless it can be guaranteed that the PCLKSEL
     * register will be written less than 1.25 ms after the ResetDevice() call
     * (above).  Therefore, the following code is conditionally-compiled.
     * PCLK rates will be autodetected, unless the VCP2 is generating FS.
     *
     * The race condition would cause incorrect behavior in the following
     * scenario:
     *   0. ResetDevice() sends the "hardware reset" command and waits 2.5 ms.
     *   1. The VCP2 successfully autodetects PCLK.
     *   2. The VCP2 Boot Loader detects that CFAIL = 0 and starts the MBIST.
     *   3. The host initiates autodetection again by writing the PCLKSEL
     *      register.
     *   4. The PLL clock frequency decays during the MBIST (which is a no-no)
     *      while autodetection is ongoing.
     */
#if (VP_VCP2_BOARD_PCLKA_SEL & VCP2_HBI_PCLKSEL_FS_GEN_MASK)
    /* Configure PCLK A. */
    temp = VP_VCP2_BOARD_PCLKA_SEL;
    if (VpHbiWr(deviceId, PCLKSEL, &temp[0]) == FALSE) {
        return FALSE;
    }
#endif
#if (VP_VCP2_BOARD_PCLKB_SEL & VCP2_HBI_PCLKSEL_FS_GEN_MASK)
    /* Configure PCLK B. */
    temp = VP_VCP2_BOARD_PCLKA_SEL;
    if (VpHbiWr(deviceId, PCLKSEL, &temp[0]) == FALSE) {
        return FALSE;
    }
#endif
#endif /* (VP_VCP2_BOARD_CLKSRC == 0) */

#ifdef VP_VCP2_BOARD_BOOT_DEBUG
    /*
     * Wait up to 22 frames (2.75 ms) for clock synchronization.  (Worst case
     * would be if FS_A is absent: 11 frames to detect PCLKA failure, and 11
     * frames to detect PCLKB rate.)
     */
    for (i = 23; i > 0; i--) {
        if (VpHbiRd(deviceId, CLKGEN_STATUS_HI, &temp[0]) == FALSE) {
            return FALSE;
        }
        if (!(temp[0] & VCP2_HBI_CLKGEN_STATUS_HI_CFAIL_GLOBAL)) {
            break;
        }
        /* Wait one FS period before trying again. */
        VpSysWait(1);
    }
    if (temp[0] & VCP2_HBI_CLKGEN_STATUS_HI_CFAIL_GLOBAL) {
        VP_DOUT(VP_DBG_ERROR, ("HbiBootPrepare(): %s.\n", "Clock failure detected"));
        return FALSE;
    }
#endif

    /*
     * Here we must wait (polling every 30 ms) for memory self-test
     * completion, before we can load the application through HBI.
     */
    for (i = MBIST_POLL_ITERATIONS; i > 0; i--) {
        if (VpHbiRd(deviceId, SW_FLAGS, &temp[0]) == FALSE) {
            return FALSE;
        }
        if (((temp[0] & VCP2_HBI_SWFLAG_CMD_MBOX) == 0) || (i == 1)) {
            break;
        }
        /* Wait 30 ms before trying again. */
        VpSysWait(240);
    }
    if (temp[0] & VCP2_HBI_SWFLAG_CMD_MBOX) {
        VP_DOUT(VP_DBG_ERROR, ("HbiBootPrepare(): %s.\n", "MBIST not completed"));
        return FALSE;
    }

#ifdef VP_VCP2_BOARD_BOOT_DEBUG
    /* Verify self-test result. */
    if (VpHbiRd(deviceId, CLKGEN_STATUS_HI, &temp[0]) == FALSE) {
        return FALSE;
    }
    if (temp[0] & VCP2_HBI_CLKGEN_STATUS_HI_RESET_MASK) {
        VP_DOUT(VP_DBG_ERROR, ("HbiBootPrepare(): %s.\n", "MBIST failed"));
        return FALSE;
    }
#endif

    /* Ready to load HBI application. */
    return TRUE;
}

/**
 * This internal function initiates a VCP2 mailbox command for ::VpInitLine()
 * or ::VpConfigLine().
 *
 * \param[in] pLineCtx      Line Context for the line to be configured
 * \param[in] pAcProfile    AC Profile to be applied to the line
 * \param[in] pDcProfile    DC Profile to be applied to the line
 * \param[in] pRingProfile  Ringing Profile to be applied to the line
 * \param[in] mailBoxCmd    Mailbox command identifier (CMD_INITLINE or
 *                          CMD_CONFIGLINE) to be passed to the VCP2
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG          pLineCtx was NULL
 * \retval ::VP_STATUS_ERR_PROFILE
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
static VpStatusType
InitLineInt(
    VpLineCtxType *pLineCtx,
    VpProfilePtrType pAcProfile,
    VpProfilePtrType pDcProfile,
    VpProfilePtrType pRingProfile,
    MboxCmdIdType mailBoxCmd)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_INITLINE];
    uint16 *pBuf = buf;
    VpStatusType retval;

    if (pDevObj->devInit == FALSE) {
        VP_DOUT(VP_DBG_ERROR, ("InitLineInt(): %s.\n", "Device not initialized"));
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Check arguments. */
    if (
        Vcp2BadProfileArg(pDevObj, VP_PROFILE_AC, pAcProfile) ||
        Vcp2BadProfileArg(pDevObj, VP_PROFILE_DC, pDcProfile) ||
        Vcp2BadProfileArg(pDevObj, VP_PROFILE_RING, pRingProfile)
    ) {
        VP_DOUT(VP_DBG_ERROR, ("InitLineInt(): %s.\n", "Bad profile argument(s)"));
        return VP_STATUS_ERR_PROFILE;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, mailBoxCmd, channelId);

    /* Add the AC Profile. */
    Vcp2MailboxAddProfile(&pBuf, VP_PROFILE_AC, pAcProfile);

    /* Add the DC Profile. */
    Vcp2MailboxAddProfile(&pBuf, VP_PROFILE_DC, pDcProfile);

    /* Add the Ringing Profile. */
    Vcp2MailboxAddProfile(&pBuf, VP_PROFILE_RING, pRingProfile);

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This internal function resets a VCP2 device.  It calls VpHalHbiInit() to
 * initialize the HBI interface, and then issues a "hardware reset command" to
 * the VCP2 by writing the specified value into the upper word of the
 * CLKGEN_STATUS register.
 *
 * \param[in] deviceId  Device ID for the VCP2 device
 * \param[in] resetType Value to be written to CLKGEN_STATUS_HI
 *
 * \retval ::TRUE   Reset command was issued.
 * \retval ::FALSE  VpHalHbiInit() failed, or other HBI error.
 */
static bool
ResetDevice(
    VpDeviceIdType deviceId,
    uint16 resetType)
{
    /* Assume we've just powered on (worst case).  Wait for System Reset to
       complete. */
    VpSysWait(SYSTEM_RESET_TIME_FRAMES);

    /* Initialize the HBI interface on the VCP2. */
    if (VpHalHbiInit(deviceId) == FALSE) {
        VP_DOUT(VP_DBG_ERROR, ("ResetDevice(): %s.\n", "VpHalHbiInit() failed"));
        return FALSE;
    }

    /* Issue "hardware reset" command to the VCP2. */
    if (VpHbiWr(deviceId, CLKGEN_STATUS_HI, &resetType) == FALSE) {
        VP_DOUT(VP_DBG_ERROR, ("ResetDevice(): %s.\n", "HBI error"));
        return FALSE;
    }

    /* Wait for System Reset to complete. */
    VpSysWait(SYSTEM_RESET_TIME_FRAMES);

    return TRUE;
}

/**
 * This internal function validates the HBI boot checksum, which indicates
 * that a firmware image has been successfully loaded to the VCP2 through the
 * HBI.
 *
 * \param[in] deviceId  Device ID for the VCP2 device
 *
 * \retval ::TRUE   Checksum verified.
 * \retval ::FALSE  Checksum incorrect, or HBI error.
 */
static bool
ValidateChecksum(
    VpDeviceIdType deviceId)
{
    uint16 checksum[2];

    /* Read checksum. */
    if (VpHbiRd(deviceId, P255_CHECK, checksum) == FALSE) {
        return FALSE;
    }

    /* Verify checksum. */
    if ((checksum[0] != (uint16)0xAA55) || (checksum[1] != (uint16)0xAA55)) {
        VP_DOUT(VP_DBG_ERROR, ("ValidateChecksum(): %s.\n", "HBI boot checksum invalid"));
        return FALSE;
    }

    /* Checksum verified. */
    return TRUE;
}

#endif /* VP_CC_VCP2_SERIES */
