/** \file vcp2_query.c
 * vcp2_query.c
 *
 * This file contains the implementation for the Status and Query functions
 * in the VCP2 library.
 *
 * The functions herein are used internally by the API and should not be
 * called directly by the application.
 *
 * Copyright (c) 2008, Zarlink Semiconductor, Inc.
 */

/* Only compile this file if specified in vp_api_cfg.h: */
#include "vp_api_cfg.h"
#if defined (VP_CC_VCP2_SERIES)

#include "vcp2_api.h"
#include "vp_api_int.h"
#include "vcp2_api_int.h"


/* =================================
    Prototypes for Static Functions
   ================================= */

static bool
EventIsExternallyVisible(
    VpEventType *pEvent);

static void
FillEventStruct(
    VpDevCtxType *pDevCtx,
    VpEventType *pEvent,
    uint16 intparam);

static void
FixTipOpenBug(
    VpDevCtxType *pDevCtx,
    VpLineCtxType *pLineCtx);

static VpStatusType
HandleBootCmpEvent(
    VpEventType *pEvent);

static VpStatusType
HandleDevInitCmpEvent(
    VpEventType *pEvent);

static VpStatusType
HandleLineInitCmpEvent(
    VpEventType *pEvent);

#if (VCP2_TIMESTAMP_SHIFT > 0)
static void
HandleTimeStampScaling(
    VpEventType *pEvent);
#endif

static bool
IdentifyEvent(
    VpVcp2DeviceObjectType *pDevObj,
    uint16 eventWord,
    VpEventCategoryType *pEventCat,
    uint16 *pEventId,
    uint8 *pChannelId);

static VpStatusType
IdentifyResponse(
    VpDevCtxType *pDevCtx,
    VpEventType *pEvent);

static bool
IsDeviceEvent(
    VpEventType *pEvent);

static void
MakeOldDevIoAccessStruct(
    VpDevCtxType *pDevCtx,
    VpDeviceIoAccessExtType *pDeviceIoAccess,
    VpDeviceIoAccessDataType *pDeviceIoData);

static void
MakeOldDevIoConfigStruct(
    VpDevCtxType *pDevCtx,
    VpOptionDeviceIoConfigType *pDeviceIoConfig,
    VpOptionDeviceIoType *pDeviceIo);

static void
ParseDevIoConfigResponse(
    uint16 **ppBuf,
    VpOptionDeviceIoConfigType *pDeviceIoConfig);

static VpStatusType
ResetDeviceOptions(
    VpDevCtxType *pDevCtx);

static VpStatusType
ResetLineOptions(
    VpDevCtxType *pDevCtx,
    VpLineCtxType *pLineCtx);

static VpStatusType
ReadChecksumResponse(
    VpEventType *pEvent,
    void *pResults);

static VpStatusType
ReadGainCmpResponse(
    VpEventType *pEvent,
    void *pResults);

static VpStatusType
ReadDevIoRdResponse(
    VpEventType *pEvent,
    void *pResults);

static VpStatusType
ReadLineIoRdResponse(
    VpEventType *pEvent,
    void *pResults);

static VpStatusType
ReadLowLevCmdResponse(
    VpEventType *pEvent,
    void *pResults);

static VpStatusType
ReadRdLoopCondResponse(
    VpEventType *pEvent,
    void *pResults);

static VpStatusType
ReadRdOptionResponse(
    VpEventType *pEvent,
    void *pResults);

static VpStatusType
ReadResponse(
    uint16 *buf,
    int maxLen,
    VpEventType *pEvent);

static VpStatusType
ReadTestCmpResponse(
    VpEventType *pEvent,
    VpTestResultType *pResults);


/* ============================
    Status and Query Functions
   ============================ */

/**
 * This function clears an outstanding response (if any) in the response
 * mailbox.  Refer to the VoicePath API User's Guide for details about
 * VpClearResults().
 *
 * \param[in] pDevCtx   Device Context for the VCP2 device
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2ClearResults(
    VpDevCtxType *pDevCtx)
{
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 flags;
    VpStatusType retval = VP_STATUS_SUCCESS;

    if (pDevObj->devInit == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Ensure that the mailbox contains valid data */
    if (VpHbiRd(deviceId, HW_Reg_MBOXFLAG, &flags) == FALSE) {
        retval = VP_STATUS_ERR_HBI;
    } else {

        /* Reponse mailbox flag = 0 => host owns the mailbox. */
        if ((flags & VCP2_HBI_SWFLAG_RESP_MBOX) == 0) {

            /* Release the response mailbox. */
            flags = VCP2_HBI_SWFLAG_RESP_MBOX;
            if (VpHbiWr(deviceId, SW_FLAGS, &flags) == FALSE) {
                retval = VP_STATUS_ERR_HBI;
            }
        }
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function initiates a checksum calculation over the VCP2's code memory
 * and data initializers.  Refer to the VoicePath API User's Guide for details
 * about ::VpCodeChecksum().
 *
 * \param[in] pDevCtx   Device Context for the VCP2 device
 * \param[in] handle    Handle to be returned with the completion event
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2CodeCheckSum(
    VpDevCtxType *pDevCtx,
    uint16 handle)
{
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_CODECHECKSUM];
    uint16 *pBuf = buf;
    VpStatusType retval;

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_CODECHECKSUM, VP_ALL_LINES);

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function flushes all outstanding events in the VCP2's event queue.
 * Refer to the VoicePath API User's Guide for details about
 * ::VpFlushEvents().
 *
 * \param[in] pDevCtx   Device Context for the VCP2 device
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2FlushEvents(
    VpDevCtxType *pDevCtx)
{
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_FLUSHEVENTS];
    uint16 *pBuf = buf;
    VpStatusType retval;

    if (pDevObj->devInit == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_FLUSHEVENTS, VP_ALL_LINES);

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function gets an event (if available) from the VCP2.  Refer to the
 * VP-API User's Guide for events that can be obtained from ::VpGetEvent().
 *
 * \param[in]   pDevCtx     Device Context for the VCP2 device
 * \param[out]  pEvent      Calling function's event struct
 */
bool
Vcp2GetEvent(
    VpDevCtxType *pDevCtx,
    VpEventType *pEvent)
{
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[VCP2_HBI_INTPARAM_OFF + 1];
    VpStatusType retval = VP_STATUS_SUCCESS;
    int sanity = 256;

    do {

        /* Initialize pEvent struct: (eventId == 0) => no event detected. */
        pEvent->eventId = 0;

        /*
         * Note:  Critical section protection is not required if this function is
         * called from within an interrupt service routine (as is recommended for
         * VCP2).  However, there are at least two scenarios in which the
         * critical section is required:
         * 1. This function runs in a preemptible OS task that is launched by an
         *    ISR (so-called "deferred service routine").  If there is a
         *    possibility the task may be interrupted in order to run an HBI
         *    transaction (with the same chip-select), the critical section is
         *    required.
         * 2. This function runs within an ISR.  If there is a possibility that
         *    a higher-priority interrupt may occur and cause an HBI transaction,
         *    the critical section is required.
         */

        /* ----------------------------------------- Begin HBI Critical Section */
        VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

        /* Read the INTIND and INTPARAM registers. */
        if (VpHbiRd(deviceId, VCP2_HBI_INTIND_PAG, VCP2_HBI_INTIND_OFF,
            VCP2_HBI_INTPARAM_OFF, buf) == FALSE) {
            retval = VP_STATUS_ERR_HBI;
        }

        VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
        /* ------------------------------------------- End HBI Critical Section */

        /* Check for HBI error condition. */
        if (retval != VP_STATUS_SUCCESS) {
            pEvent->status = retval;
            return FALSE;
        }

        /* Check for empty event ID. */
        if (buf[VCP2_HBI_INTIND_OFF] == 0x0000) {

            /* No events pending. */
            pEvent->status = VP_STATUS_SUCCESS;
            return FALSE;
        }

        /* Parse the VCP2 interrupt type and parameter. */
        if (IdentifyEvent(pDevObj, buf[VCP2_HBI_INTIND_OFF], &pEvent->eventCategory,
            &pEvent->eventId, &pEvent->channelId) == FALSE) {

            /* Invalid event. */
            pEvent->status = VP_STATUS_FAILURE;
            return FALSE;
        }

        /* Fill in the calling function's pEvent struct. */
        FillEventStruct(pDevCtx, pEvent, buf[VCP2_HBI_INTPARAM_OFF]);

        /* The API needs to do extra processing for some events. */
        if (pEvent->eventCategory == VP_EVCAT_RESPONSE) {
            if (pEvent->eventId == VP_DEV_EVID_BOOT_CMP) {
                retval = HandleBootCmpEvent(pEvent);
            } else if (pEvent->eventId == VP_DEV_EVID_DEV_INIT_CMP) {
                retval = HandleDevInitCmpEvent(pEvent);
            } else if (pEvent->eventId == VP_LINE_EVID_LINE_INIT_CMP) {
                retval = HandleLineInitCmpEvent(pEvent);
            }
        }
    
#if (VCP2_TIMESTAMP_SHIFT > 0)
        /* Convert VCP2 timestamp granularity to API timestamp granularity. */
        else if (pEvent->eventCategory == VP_EVCAT_SIGNALING) {
            HandleTimeStampScaling(pEvent);
        }
#endif

        /* Event processing complete. */
        pEvent->status = retval;

        /* See if we need to report an event or an error. */
        if (EventIsExternallyVisible(pEvent) || (retval != VP_STATUS_SUCCESS)) {
            return TRUE;
        }

        /* Event was handled internally.  Before we return FALSE, read the
           event register again.  There may be another event pending. */
    } while (sanity--);

    /* Failed the sanity check -- for some reason, the event register is stuck. */
    pEvent->status = VP_STATUS_FAILURE;
    return FALSE;
}

/**
 * This function is returns the VP-API line state for the given line.  Refer
 * to the VoicePath API User's Guide for details about ::VpGetLineState().
 *
 * \param[in]   pLineCtx        Line Context for the line to be queried
 * \param[out]  pCurrentState   Pointer to the variable to contain the result
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2GetLineState(
    VpLineCtxType *pLineCtx,
    VpLineStateType *pCurrentState)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 lineState;
    VpStatusType retval;

    if (pDevObj->devInit == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    if (VpHbiRd(deviceId, LINESTATE(channelId), &lineState) == FALSE) {
        retval = VP_STATUS_ERR_HBI;
    } else {

        /* One 16-bit register contains two line states: high byte = even line number. */
        lineState = ((channelId & 1) == 0) ? (lineState >> 8) : (lineState & 0xFF);
        *pCurrentState = (VpLineStateType)lineState;

        retval = VP_STATUS_SUCCESS;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function returns the requested status information about a given line.
 * Refer to the VoicePath API User's Guide for details about
 * VpGetLineStatus().
 *
 * \param[in] pLineCtx      Line Context for the line to be queried
 * \param[in] input         Type of status information to be returned
 * \param[in] pStatus       Pointer to a variable to contain the results
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_INVALID_ARG      Input type not supported.
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2GetLineStatus(
    VpLineCtxType *pLineCtx,
    VpInputType input,
    bool *pStatus)
{
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDeviceStatusType deviceStatus;
    VpStatusType retval;

    deviceStatus.input = input;
    retval = Vcp2GetDeviceStatusExt(pDevCtx, &deviceStatus);

    if (retval == VP_STATUS_SUCCESS) {
        int index = channelId / 8;
        int bitpos = channelId % 8;
        *pStatus = (deviceStatus.status[index] >> bitpos) & 1;
    }

    return retval;
}

/**
 * This function returns the current loop conditions for the specified line.
 * Refer to the VoicePath API User's Guide for details about VpGetLoopCond().
 *
 * \param[in] pLineCtx  Line Context for the line to be queried
 * \param[in] handle    Handle to be returned with the completion event
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2GetLoopCond(
    VpLineCtxType *pLineCtx,
    uint16 handle)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_GETLOOPCOND];
    uint16 *pBuf = buf;
    VpStatusType retval;

    if (pDevObj->devInit == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_GETLOOPCOND, channelId);

    /* Add handle argument to the mailbox buffer. */
    *pBuf++ = (uint16)handle;

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function returns the status for all lines, multiplexed into an array
 * of 32 bit values.  Refer to the VoicePath API User's Guide for details
 * about this function.
 *
 * \param[in] pDevCtx       Device Context for the VCP2 device
 * \param[in] input         Type of status information to be returned
 * \param[in] pDevStatus    An array of two 32-bit integers in which to place
 *                          the results.
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_INVALID_ARG      Input type not supported.
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2GetDeviceStatus(
    VpDevCtxType *pDevCtx,
    VpInputType input,
    uint32 *pDevStatus)
{
    VpDeviceStatusType deviceStatus;
    uint8 *pStatus;
    VpStatusType retval = VP_STATUS_SUCCESS;

    /* Check arguments. */
    if (pDevStatus == VP_NULL) {
        return VP_STATUS_INVALID_ARG;
    }

    /* This is just a wrapper function for Vcp2GetDeviceStatusExt(). */
    deviceStatus.input = input;
    retval = Vcp2GetDeviceStatusExt(pDevCtx, &deviceStatus);

    /* Copy data into the calling function's variable. */
    pStatus = deviceStatus.status;
    *pDevStatus = *pStatus++;
    *pDevStatus |= (uint32)*pStatus++ << 8;
    *pDevStatus |= (uint32)*pStatus++ << 16;
    *pDevStatus |= (uint32)*pStatus << 24;

    return retval;
}

/**
 * This function returns the status for all lines, multiplexed into an array
 * of eight-bit values.  Refer to the VoicePath API User's Guide for details
 * about this function.
 *
 * \param[in] pDevCtx       Device Context for the VCP2 device
 * \param[in] pDevStatus    Struct containing parameters and results.
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_INVALID_ARG      Input type not supported.
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2GetDeviceStatusExt(
    VpDevCtxType *pDevCtx,
    VpDeviceStatusType *pDeviceStatus)
{
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 statusReg[VCP2_HBI_LINE_STATUS_LEN], *pStatusReg;
    VpInputType input = pDeviceStatus->input;
    uint8 *pStatus = pDeviceStatus->status;
    VpStatusType retval = VP_STATUS_SUCCESS;
    int i;

    if (pDevObj->devInit == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Check arguments. */
    if (input >= NUM_INPUT_TYPES) {
        return VP_STATUS_INVALID_ARG;
    }

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    if (VpHbiRd(deviceId, LINE_STATUS(input), statusReg) == FALSE) {
        retval = VP_STATUS_ERR_HBI;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    /* Copy results into struct. */
    pStatus = pDeviceStatus->status;
    pStatusReg = &statusReg[0];
    for (i = 0; i < VCP2_LINE_FLAG_BYTES; i++) {
        if (i % 2) {
            *pStatus++ = (uint8)(*pStatusReg++ >> 8);
        } else {
            *pStatus++ = (uint8)(*pStatusReg & 0xFF);
        }
    }

#if (VCP2_LINE_FLAG_BYTES < VP_LINE_FLAG_BYTES)
    /* Clear extra struct members. */
    for (i = VCP2_LINE_FLAG_BYTES; i < VP_LINE_FLAG_BYTES; i++) {
        *pStatus++ = 0;
    }
#endif

    return retval;
}

/**
 * This function initiates a VCP2 "read option" command.  Refer to the
 * VoicePath API User's Guide for details about VpGetLoopCond().
 *
 * \param[in]   pLineCtx        Line Context for line-specific options
 * \param[in]   pDevCtx         Device Context for the VCP2 device
 * \param[in]   optionId        Option ID to be queried
 * \param[in]   optionId        Handle to be returned with event
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG  Option ID invalid or not supported.
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2GetOption(
    VpLineCtxType *pLineCtx,
    VpDevCtxType *pDevCtx,
    VpOptionIdType optionId,
    uint16 handle)
{
    uint8 channelId;
    VpVcp2DeviceObjectType *pDevObj;
    VpDeviceIdType deviceId;
    uint16 buf[BUFSIZE_GETOPTION];
    uint16 *pBuf = buf;
    VpStatusType retval;

    /* Get Device Object, device ID, and channel ID. */
    if (pDevCtx != VP_NULL) {
        pDevObj = pDevCtx->pDevObj;
        channelId = VP_ALL_LINES;
    } else {
        VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
        pDevObj = pLineCtx->pDevCtx->pDevObj;
        channelId = pLineObj->channelId;
    }
    deviceId = pDevObj->deviceId;

    if (pDevObj->devInit == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Is the requested option supported by VCP2? */
    if (Vcp2DeviceOptionSupported(pDevObj, optionId) == TRUE) {
        /* For device options, must specify a Device Context. */
        if (pDevCtx == VP_NULL) {
            return VP_STATUS_INVALID_ARG;
        }
    } else if (Vcp2LineOptionSupported(pDevObj, optionId) == FALSE) {
        /* Option is not supported by this device. */
        return VP_STATUS_INVALID_ARG;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_GETOPTION, channelId);

    /* Add handle to the mailbox buffer. */
    *pBuf++ = handle;

    /* Add option ID to the mailbox buffer. */
    *pBuf++ = (uint16)optionId;

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This function reads the VCP2's response mailbox and places the results into
 * the provided struct.  Refer to the VP API User's Guide for details about
 * ::VpGetResults().
 *
 * \retval ::VP_STATUS_INVALID_ARG  pDevCtx = NULL in the pEvent struct, or
 *                                  the supplied event is not a response event
 * \retval ::VP_STATUS_MAILBOX_EMPTY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2GetResults(
    VpEventType *pEvent,
    void *pResults)
{
    VpDevCtxType *pDevCtx = pEvent->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj;
    VpDeviceIdType deviceId;
    uint16 flags;
    VpStatusType retval = VP_STATUS_SUCCESS;

    /* Check arguments. */
    if (pDevCtx == VP_NULL) {
        return VP_STATUS_INVALID_ARG;
    }

    /* Get Device Object and device ID. */
    pDevObj = pDevCtx->pDevObj;
    deviceId = pDevObj->deviceId;

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Find out whether the host owns the mailbox. */
    if (VpHbiRd(deviceId, SW_FLAGS, &flags) == FALSE) {
        retval = VP_STATUS_ERR_HBI;
    } else if (flags & VCP2_HBI_SWFLAG_RESP_MBOX) {
        retval = VP_STATUS_MAILBOX_EMPTY;
    } else {

        /* Host owns the mailbox. */
        if (pResults == VP_NULL) {

            /* User wants to know what kind of response is in the mailbox. */
            retval = IdentifyResponse(pDevCtx, pEvent);
        } else {

            /* User wants to copy the results into pResults struct. */
            retval = VP_STATUS_INVALID_ARG;
            if (pEvent->eventCategory == VP_EVCAT_RESPONSE) {
                switch (pEvent->eventId) {
                    case VP_DEV_EVID_BOOT_CMP:
                        retval = ReadChecksumResponse(pEvent, pResults);
                        break;
                    case VP_LINE_EVID_LLCMD_RX_CMP:
                        retval = ReadLowLevCmdResponse(pEvent, pResults);
                        break;
                    case VP_LINE_EVID_RD_LOOP:
                        retval = ReadRdLoopCondResponse(pEvent, pResults);
                        break;
                    case VP_LINE_EVID_GAIN_CMP:
                        retval = ReadGainCmpResponse(pEvent, pResults);
                        break;
                    case VP_LINE_EVID_RD_OPTION:
                        retval = ReadRdOptionResponse(pEvent, pResults);
                        break;
                    case VP_DEV_EVID_IO_ACCESS_CMP:
                        /* No results for "write-complete" events. */
                        if (pEvent->hasResults) {
                            retval = ReadDevIoRdResponse(pEvent, pResults);
                        }
                        break;
                    case VP_LINE_EVID_LINE_IO_RD_CMP:
                        retval = ReadLineIoRdResponse(pEvent, pResults);
                }
            } else if (pEvent->eventCategory == VP_EVCAT_TEST) {
                switch (pEvent->eventId) {
                    case VP_LINE_EVID_TEST_CMP:
                        retval = ReadTestCmpResponse(pEvent, pResults);
                        break;
                    case VP_DEV_EVID_CHKSUM:
                        retval = ReadChecksumResponse(pEvent, pResults);
                }
            }

            /* Relinquish control of the response mailbox to the VCP2. */
            if (retval == VP_STATUS_SUCCESS) {
                flags = VCP2_HBI_SWFLAG_RESP_MBOX;
                if (VpHbiWr(deviceId, SW_FLAGS, &flags) == FALSE) {
                    retval = VP_STATUS_ERR_HBI;
                }
            }
        }
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}


/* ==================
    Static Functions
   ================== */

/**
 * Some events are received from the VCP2, but not reported to the API, even
 * if the user has chosen to unmask them.  This function handles those events,
 * returning TRUE if the event needs to be reported, FALSE if it's handled
 * internally.
 *
 * \param[in]   pEvent  Pointer to the event struct
 *
 * \retval ::VP_STATUS_SUCCESS
 *
 * \retval ::TRUE   The event is visible to the application.
 * \retval ::FALSE  The event is handled internally by the API, and
 *                  VpGetEvent() returns false.
 */  
static bool
EventIsExternallyVisible(
    VpEventType *pEvent)
{

    if (pEvent->eventCategory == VP_EVCAT_SIGNALING) {

        if (pEvent->eventId &
            (VP_LINE_EVID_HOOK_OFF | VP_LINE_EVID_HOOK_ON | VP_LINE_EVID_STARTPULSE |
            VP_LINE_EVID_FLASH | VP_LINE_EVID_PULSE_DIG)) {
            VpVcp2LineObjectType *pLineObj = pEvent->pLineCtx->pLineObj;

            /* These events are internally masked when in the tip-open
               state, to cover a VCP2 firmware erratum. */
            if (pLineObj->tipOpenState) {
                return FALSE;
            }
        }

#if (VCP2_TIMESTAMP_SHIFT > 0)        
        else if (pEvent->eventId == VP_DEV_EVID_TS_ROLLOVER) {
            VpVcp2DeviceObjectType *pDevObj = pEvent->pDevCtx->pDevObj;
            uint16 bit = 1 << (16 - VCP2_TIMESTAMP_SHIFT);

            /* Increment locally-maintained high-order bits of the timestamp. */
            pDevObj->timeStampHiBits += bit;

            /* Check for timestamp rollover. */
            if ((pDevObj->timeStampHiBits & 0xFFFF) == 0) {

                /* Rollover occurred.  Generate event if not masked. */
                return ((pDevObj->timeStampEventMasked) == FALSE);
            } else {

                /* No timestamp rollover yet. */
                return FALSE;
            }
        }
#endif

    } else if (pEvent->eventCategory == VP_EVCAT_FAULT) {

        if (pEvent->eventId == VP_LINE_EVID_SYNC_FLT) {

            /* This event is being temporarily masked; it is not yet supported
               by the API. */
            return FALSE;
        }
    }

    /* Otherwise it's an externally-visible event. */
    return TRUE;
}

/**
 * This internal function sets the pLineCtx, pLineObj, lineId, parmHandle,
 * eventData, and hasResults fields in the supplied event struct, based on
 * the supplied VCP2 interrupt parameter value.
 *
 * \param[in]       pDevCtx     Device Context for VCP2 device
 * \param[in,out]   pEvent      Calling function's event struct
 * \param[in]       intparam    Value of VCP2's interrupt parameter
 */
static void
FillEventStruct(
    VpDevCtxType *pDevCtx,
    VpEventType *pEvent,
    uint16 intparam)
{
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    VpEventCategoryType eventCat = pEvent->eventCategory;
    uint8 channelId = pEvent->channelId;
    uint16 eventId = pEvent->eventId;

    /* Fill in deviceId and pDevCtx fields. */
    pEvent->deviceId = deviceId;
    pEvent->pDevCtx = pDevCtx;

    /* Fill in pLineCtx field. */
    if ((channelId == VP_ALL_LINES) || IsDeviceEvent(pEvent)) {
        pEvent->pLineCtx = VP_NULL;
        pEvent->channelId = 0;
    } else {
        pEvent->pLineCtx = pDevCtx->pLineCtx[channelId];
    }

    /* Fill in pLineObj and lineId fields (if available). */
    if (pEvent->pLineCtx != VP_NULL) {
        VpVcp2LineObjectType *pLineObj = pEvent->pLineCtx->pLineObj;
        pEvent->lineId = pLineObj->lineId;
    }

    /* Fill in the parmHandle and eventData fields.  Since the application
       will only use one or the other, it's safe to set them both to the same
       value... */
    pEvent->parmHandle = pEvent->eventData = intparam;

    /* ...except in the following cases: */
    if (eventCat == VP_EVCAT_SIGNALING) {
        if (eventId == VP_LINE_EVID_DTMF_DIG) {
            pEvent->eventData &= VCP2_HBI_INTIND_DTMF_DIG_MASK;
            pEvent->parmHandle &= VCP2_HBI_INTIND_DTMF_TIME_MASK;
        } else if (eventId == VP_LINE_EVID_PULSE_DIG) {
            pEvent->parmHandle = 0;
        }
    } else if ((eventId == VP_LINE_EVID_RD_OPTION) && (eventCat == VP_EVCAT_RESPONSE)) {
//      TO DO: get option ID from mailbox.  Needs critical section.
    }


    /* Fill in the hasResults field, based on the event type. */
    pEvent->hasResults = FALSE;
    if (eventCat == VP_EVCAT_RESPONSE) {
        switch (eventId) {
            case VP_DEV_EVID_BOOT_CMP:
            case VP_LINE_EVID_LLCMD_RX_CMP:
            case VP_LINE_EVID_RD_OPTION:
            case VP_LINE_EVID_RD_LOOP:
            case VP_LINE_EVID_GAIN_CMP:
            case VP_LINE_EVID_LINE_IO_RD_CMP:
                pEvent->hasResults = TRUE;
                break;
            case VP_DEV_EVID_IO_ACCESS_CMP:
                pEvent->hasResults = (pEvent->eventData == VP_IO_READ);
                break;
        }
    } else if (eventCat == VP_EVCAT_TEST) {
        switch (eventId) {
            case VP_LINE_EVID_TEST_CMP:
            case VP_DEV_EVID_CHKSUM:
                pEvent->hasResults = TRUE;
        }
    }
}

/*
 * This internal function is a work-around for a VCP2 firmware erratum:  We
 * have to keep track of whether we are in the tip-open line state, so we can
 * mask hook events in that state.  This function initializes the tipOpenState
 * flag in all line objects (if pLineCtx == VP_NULL) or a single line object.
 *
 * \param[in]       pDevCtx     Device Context for VCP2 device (or VP_NULL)
 * \param[in]       pLineCtx    Line Context for VCP2 device (or VP_NULL)
 */
static void
FixTipOpenBug(
    VpDevCtxType *pDevCtx,
    VpLineCtxType *pLineCtx)
{
    uint8 chan, lastChan;

    if (pLineCtx != VP_NULL) {
        VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
        pDevCtx = pLineCtx->pDevCtx;
        chan = lastChan = pLineObj->channelId;
    } else {
        VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
        chan = 0;
        lastChan = pDevObj->numChannels - 1;
    }

    for ( ; chan <= lastChan; chan++) {
        pLineCtx = pDevCtx->pLineCtx[chan];
        if (pLineCtx != VP_NULL) {
            VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
            pLineObj->tipOpenState = FALSE;
        }
    }
}

/**
 * This internal function responds to a boot-complete event by saving the
 * firmware product type into the Device Object and reinitializing the
 * function pointers in the Device Context.
 *
 * \param[in]   pEvent  Event struct for the boot-complete event
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG
 * \retval ::VP_STATUS_ERR_HBI
 */
static VpStatusType
HandleBootCmpEvent(
    VpEventType *pEvent)
{
    VpDevCtxType *pDevCtx = pEvent->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    VpChkSumType checksum;
    VpStatusType retval;

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    retval = ReadChecksumResponse(pEvent, &checksum);

    if (retval == VP_STATUS_SUCCESS) {

        /* Save the firmware product type for future conditionals. */
        pDevObj->product = checksum.vInfo.swProductId;

        /* Reassign function pointers based on the firmware product type. */
        Vcp2ResetDeviceContext(pDevCtx);
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}

/**
 * This internal function responds to a device-init-complete event by updating
 * Device Object variables for future argument checking, and applying default
 * option values for device-specific options.
 *
 * \param[in]   pEvent  Event struct for the device-init-complete event
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
static VpStatusType
HandleDevInitCmpEvent(
    VpEventType *pEvent)
{
    VpDevCtxType *pDevCtx = pEvent->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    VpStatusType retval;

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Enable argument checking for future VCP2-API calls. */
    pDevObj->devInit = TRUE;
    pDevObj->slacsFound = pEvent->eventData;

#if (VCP2_TIMESTAMP_SHIFT > 0)
    /* Reset the high-order bits of the device timestamp. */
    pDevObj->timeStampHiBits = 0;
    pDevObj->timeStampEventMasked = FALSE;
#endif

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    /* Initialize device-specific options to customer-defined defaults. */
    retval = ResetDeviceOptions(pDevCtx);

    /* Initialize line-specific options to customer-defined defaults. */
    if (retval == VP_STATUS_SUCCESS) {
        retval = ResetLineOptions(pDevCtx, VP_NULL);
    }

    return retval;
}

/**
 * This internal function responds to a line-init-complete event by applying
 * default option values for line-specific options.
 *
 * \param[in]   pEvent  Event struct for the device-init-complete event
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_INVALID_ARG
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
static VpStatusType
HandleLineInitCmpEvent(
    VpEventType *pEvent)
{
    VpLineCtxType *pLineCtx = pEvent->pLineCtx;
    VpStatusType retval;

    /* Could set a default line state here if needed. */

    /* Initialize line-specific options to customer-defined defaults. */
    retval = ResetLineOptions(VP_NULL, pLineCtx);

    return retval;
}

#if (VCP2_TIMESTAMP_SHIFT > 0)
/**
 * This internal function handles the conversion of VCP2 timestamps to API
 * timestamps.  The VCP2 firmware incorrectly scales the timestamp, such that
 * VCP2 timestamps need to be right-shifted by 2 before being reported by the
 * API.  The API also has to maintain an extra two high-order bits in the line
 * object.
 *
 * This function assumes the eventCategory for pEvent is VP_EVCAT_SIGNALING.
 *
 * \param[in,out]  pEvent   Pointer to the event struct
 */
static void
HandleTimeStampScaling(
    VpEventType *pEvent)
{
    /* The following events include a timestamp parameter. */
    if (pEvent->eventId &
        (VP_LINE_EVID_HOOK_OFF |
         VP_LINE_EVID_HOOK_ON |
         VP_LINE_EVID_STARTPULSE |
         VP_LINE_EVID_FLASH |
         VP_LINE_EVID_DTMF_DIG)) {

        VpVcp2DeviceObjectType *pDevObj = pEvent->pDevCtx->pDevObj;

        /* Rescale the timestamp and add the high-order bits. */
        pEvent->parmHandle >>= VCP2_TIMESTAMP_SHIFT;
        pEvent->parmHandle |= pDevObj->timeStampHiBits;
    }
}
#endif /* (VCP2_TIMESTAMP_SHIFT > 0) */

/**
 * This internal function identifies the category, event ID, and channel ID
 * of a VCP2 event, based on the supplied VCP2 intind parameter (eventWord).
 *
 * \param[in]   pDevObj     Device Object for the VCP2 device
 * \parma[in]   eventWord   Value obtained from the VCP2's intind register
 * \param[out]  pEventCat   Calling function's event category variable
 * \param[out]  pEventId    Calling function's event ID variable
 * \param[out]  pChannelId  Calling function's channel ID variable
 *
 * \retval  ::TRUE
 * \retval  ::FALSE     Failed to identify the event.
 */
static bool
IdentifyEvent(
    VpVcp2DeviceObjectType *pDevObj,
    uint16 eventWord,
    VpEventCategoryType *pEventCat,
    uint16 *pEventId,
    uint8 *pChannelId)
{
    uint16 eventCat;
    uint16 eventId;
    uint8 channelId;

    /* Handle VCP2 System Interrupts. */
    if (eventWord & VCP2_HBI_INTIND_SYSTEM_MASK) {
        *pEventCat = VP_EVCAT_FAULT;
        *pEventId = VP_DEV_EVID_SYSTEM_FLT;
        *pChannelId = VP_ALL_LINES;
        return TRUE;
    }

    /* Firmware event.  Extract the fields from the INTIND register value. */
    eventCat = (uint8)((eventWord & VCP2_HBI_INTIND_EVCAT_MASK) >>
        VCP2_HBI_INTIND_EVCAT_POS);
    eventId = (uint8)((eventWord & VCP2_HBI_INTIND_EVID_MASK) >>
        VCP2_HBI_INTIND_EVID_POS);
    channelId = (uint8)((eventWord & VCP2_HBI_INTIND_CHANNEL_MASK) >>
        VCP2_HBI_INTIND_CHANNEL_POS);

    /* Check arguments. */
    if (
        (channelId > pDevObj->numChannels) &&
        (channelId != VP_ALL_LINES)
    ) {
        /* Invalid channel ID. */
        return FALSE;
    }

    /* \todo Could do some event ID validity checking here. */

    /* Convert to VP-API event category and ID. */
    *pEventCat = (VpEventCategoryType)(eventCat - 1);
    *pEventId = 1 << eventId;

    /* Return channel ID. */
    *pChannelId = channelId;

    return TRUE;
}

/**
 * This internal function identifies the contents of the response mailbox, and
 * fills in the appropriate fields in the supplied pEvent struct accordingly.
 *
 * \param[in]   pDevCtx     Device Context for the VCP2 device
 * \param[out]  pEvent      Calling function's event struct
 *
 * \retval  ::VP_STATUS_SUCCESS
 * \retval  ::VP_STATUS_FAILURE Mailbox contains an invalid event ID
 * \retval  ::VP_STATUS_ERR_HBI
 */
static VpStatusType
IdentifyResponse(
    VpDevCtxType *pDevCtx,
    VpEventType *pEvent)
{
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[VCP2_HBI_MBOX_PAYLOAD + ARG_FLD];
    VpEventCategoryType eventCat;
    uint16 eventId;
    uint8 channelId;

    /* Discover the type of the response in the mailbox. */
    if (VpHbiRd(deviceId, VCP2_HBI_RESP_MBOX_PAG, VCP2_HBI_RESP_MBOX_OFF(0),
        VCP2_HBI_MBOX_PAYLOAD, buf) == FALSE) {
        return VP_STATUS_ERR_HBI;
    }
    if (IdentifyEvent(pDevObj, buf[VCP2_HBI_MBOX_ID_CHAN], &eventCat, &eventId,
        &channelId) == FALSE) {
        return VP_STATUS_FAILURE;
    }

    /* Fill in the calling function's VpEventType struct. */
    pEvent->eventCategory = eventCat;
    pEvent->eventId = eventId;
    pEvent->channelId = channelId;
    if (channelId == VP_ALL_LINES) {
        pEvent->pLineCtx = VP_NULL;
    } else {
        pEvent->pLineCtx = pDevCtx->pLineCtx[channelId];
    }
    pEvent->status = VP_STATUS_SUCCESS;
    pEvent->deviceId = deviceId;
    pEvent->eventData = buf[VCP2_HBI_MBOX_PAYLOAD]; /* option ID or test ID */
    pEvent->parmHandle = 0;
    pEvent->hasResults = TRUE;

    return VP_STATUS_SUCCESS;
}

/**
 * This internal function determines whether an event is line-specific or
 * device-specific.  It assumes the pDevCtx member of the passed pEvent struct
 * has already been initialized.  For the case of RD_OPTION events, it reads
 * the option ID from the response mailbox to determine whether the option is
 * line- or device-specific.
 *
 * \param[out]  pEvent      Event struct for the event to be examined
 *
 * \retval  ::TRUE          Event is device-specific
 * \retval  ::FALSE         Event is line-specific
 */
static bool
IsDeviceEvent(
    VpEventType *pEvent)
{
    VpVcp2DeviceObjectType *pDevObj = pEvent->pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 eventId = pEvent->eventId;

    /* Device-specific events always have channelId = 0. */
    if (pEvent->channelId != 0) {
        return FALSE;
    }

    switch (pEvent->eventCategory) {
        case VP_EVCAT_FAULT:
            return (eventId & VP_EVCAT_FAULT_DEV_EVENTS) != 0;
        case VP_EVCAT_SIGNALING:
            return (eventId & VP_EVCAT_SIGNALING_DEV_EVENTS) != 0;
        case VP_EVCAT_TEST:
            return (eventId & VP_EVCAT_TEST_DEV_EVENTS) != 0;
        case VP_EVCAT_PROCESS:
            return (eventId & VP_EVCAT_PROCESS_DEV_EVENTS) != 0;
        case VP_EVCAT_RESPONSE:
            if (eventId != VP_LINE_EVID_RD_OPTION) {
                return (eventId & VP_EVCAT_RESPONSE_DEV_EVENTS) != 0;
            } else {
                /* The RD_OPTION event is sometimes device-specific, other times
                   line-specific.  Find out based on which option ID is present in
                   the response mailbox. */
                uint16 optionId;

                /* Discover the type of the response in the mailbox.  Not
                   checking the error code here, because this function returns
                   bool, and FillEventStruct() returns void. */
                VpHbiRd(deviceId, VCP2_HBI_RESP_MBOX_PAG,
                    VCP2_HBI_RESP_MBOX_OFF(VCP2_HBI_MBOX_PAYLOAD), 0, &optionId);

                /* Identify the option ID as device-specific or line-specific. */
                return Vcp2DeviceOptionSupported(pDevObj, optionId);
            }

        default:
            /* Other categories are not currently supported by VCP2. */
            return FALSE;
    }
}

/*
 * This internal function cppies a VpDeviceIoAccessExtType struct into an
 * equivalent VpDeviceIoAccessDataType struct.  It is used in converting a
 * VP_DEVICE_OPTION_ID_DEV_IO_CFG set-option result into a
 * VP_DEVICE_OPTION_ID_DEVICE_IO set-option result.
 *
 * \param[in]   pDevCtx         Device Context for the VCP2 device
 * \param[out]  pDeviceIoAccess New-format struct
 * \param[in]   pDeviceIoData   Old-format struct
 */
static void
MakeOldDevIoAccessStruct(
    VpDevCtxType *pDevCtx,
    VpDeviceIoAccessExtType *pDeviceIoAccess,
    VpDeviceIoAccessDataType *pDeviceIoData)
{
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpVcp2ProductType product = pDevObj->product;
    uint32 packedMask = 0;
    uint32 packedData = 0;
    int channelId;
    uint8 availBits, availMask;

    /* VCP1-790 had one I/O bit per channel.  VCP1-880 had two. */
    if ((product & VP_VCP_PROD_SLAC_MASK) == VP_VCP_PROD_SLAC_VE880) {
        availBits = 2;  /* two GPIO bits per channel */
        availMask = 3;
    } else {
        availBits = 1;
        availMask = 1;
    }

    /* Copy data into VpOptionDeviceIoConfigType struct. */
    for (channelId = VP_VCP2_MAX_NUM_LINES - 1; channelId >= 0; channelId--) {
        packedMask <<= availBits;
        packedData <<= availBits;
        packedMask |= pDeviceIoAccess->lineIoBits[channelId].mask & availMask;
        packedData |= pDeviceIoAccess->lineIoBits[channelId].data & availMask;

        /* Move to the next member of the VpOptionDeviceIoType struct? */
        if (channelId * availBits == 32) {
            pDeviceIoData->accessMask_63_32 = packedMask;
            pDeviceIoData->deviceIOData_63_32 = packedData;
        }
    }
    pDeviceIoData->accessMask_31_0 = packedMask;
    pDeviceIoData->deviceIOData_31_0 = packedData;
    pDeviceIoData->accessType = pDeviceIoAccess->direction;
}

/*
 * This internal function copies a VpOptionDeviceIoConfigType struct into an
 * equivalent VpOptionDeviceIoType struct.  It is used in converting a
 * VpDeviceIoAccessExt() result into a VpDeviceIoAccess() result.
 *
 * \param[in]   pDevCtx         Device Context for the VCP2 device
 * \param[out]  pDeviceIoConfig New-format struct
 * \param[in]   pDeviceIo       Old-format struct
 */
static void
MakeOldDevIoConfigStruct(
    VpDevCtxType *pDevCtx,
    VpOptionDeviceIoConfigType *pDeviceIoConfig,
    VpOptionDeviceIoType *pDeviceIo)
{
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpVcp2ProductType product = pDevObj->product;
    uint32 packedDir = 0;
    uint32 packedType = 0;
    int channelId;
    uint8 availBits, availMask;

    /* VCP1-790 had one I/O bit per channel.  VCP1-880 had two. */
    if ((product & VP_VCP_PROD_SLAC_MASK) == VP_VCP_PROD_SLAC_VE880) {
        availBits = 2;  /* two GPIO bits per channel */
        availMask = 3;
    } else {
        availBits = 1;
        availMask = 1;
    }

    /* Copy data into VpOptionDeviceIoConfigType struct. */
    for (channelId = VP_VCP2_MAX_NUM_LINES - 1; channelId >= 0; channelId--) {
        packedDir <<= availBits;
        packedType <<= availBits;
        packedDir |= pDeviceIoConfig->lineIoConfig[channelId].direction & availMask;
        packedType |= pDeviceIoConfig->lineIoConfig[channelId].outputType & availMask;

        /* Move to the next member of the VpOptionDeviceIoType struct? */
        if (channelId * availBits == 32) {
            pDeviceIo->directionPins_63_32 = packedDir;
            pDeviceIo->outputTypePins_63_32 = packedType;
        }
    }
    pDeviceIo->directionPins_31_0 = packedDir;
    pDeviceIo->outputTypePins_31_0 = packedType;
}

/**
 * This internal function parses a device-io-access message in the
 * supplied response mailbox buffer.  It is used for the
 * VpDeviceIoAccess() and VpDeviceIoAccessExt() results.
 *
 * \param[in]   ppBuf           Mailbox cursor pointer
 * \param[out]  pDeviceIoAccess Struct to be filled with response data
 */
static void
ParseDevIoRdResponse(
    uint16 **ppBuf,
    VpDeviceIoAccessExtType *pDeviceIoAccess)
{
    VpLineIoBitsType *pLineIoBits = &pDeviceIoAccess->lineIoBits[0];
    uint16 *pBuf = *ppBuf;
    uint16 temp;
    int channelId;

    /* Read direction field from mailbox buffer. */
    temp = *pBuf++;
    pDeviceIoAccess->direction = (temp & 0xFF);

    /* Read channel-specific fields from mailbox buffer. */
    for (channelId = 0; channelId < VP_VCP2_MAX_NUM_LINES; channelId++) {
        if (channelId % 2) {
            temp <<= 8;
        } else {
            temp = *pBuf++;
        }
        pLineIoBits->mask = (temp >> 12) & 0xF;
        pLineIoBits->data = (temp >> 8) & 0xF;
        pLineIoBits++;
    }

    /* Update the calling function's mailbox cursor variable. */
    *ppBuf = pBuf;
}

/**
 * This internal function parses a read-option device-io-config message in the
 * supplied response mailbox buffer.  It is used for the
 * VP_DEVICE_OPTION_ID_DEVICE_IO and VP_OPTION_ID_DEV_IO_CFG options.
 *
 * \param[in]   ppBuf           Mailbox cursor pointer
 * \param[out]  pDeviceIoConfig Struct to be filled with response data
 */
static void
ParseDevIoConfigResponse(
    uint16 **ppBuf,
    VpOptionDeviceIoConfigType *pDeviceIoConfig)
{
    VpOptionLineIoConfigType *pLineIoConfig = &pDeviceIoConfig->lineIoConfig[0];
    uint16 *pBuf = *ppBuf;
    uint16 temp = 0; /* silence compiler warning */
    int channelId;

    /* Read channel-specific fields from mailbox buffer. */
    for (channelId = 0; channelId < VP_VCP2_MAX_NUM_LINES; channelId++) {
        if (channelId % 2) {
            temp >>= 8;
        } else {
            temp = *pBuf++;
        }
        pLineIoConfig->direction = (temp >> 4) & 0xF;
        pLineIoConfig->outputType = temp & 0xF;
        pLineIoConfig++;
    }

    /* Update the calling function's mailbox cursor variable. */
    *ppBuf = pBuf;
}

/**
 * This internal function reads a checksum-complete response from the VCP2's
 * response mailbox and fills in the supplied pResults struct accordingly.
 *
 * \param[in]   pEvent      Event struct identifying the response type
 * \param[out]  pResults    Calling function's results struct
 *
 * \retval  ::VP_STATUS_SUCCESS
 * \retval  ::VP_STATUS_ERR_MAILBOX_DATA    unexpected response type or size
 * \retval  ::VP_STATUS_ERR_HBI
 */
static VpStatusType
ReadChecksumResponse(
    VpEventType *pEvent,
    void *pResults)
{
    VpChkSumType *pChecksum = pResults;
    uint16 buf[BUFSIZE_CHECKSUM_RESP];
    uint16 *pBuf = &buf[VCP2_HBI_MBOX_PAYLOAD];
    VpStatusType retval;

    /* Read the response from the response mailbox. */
    retval = ReadResponse(buf, BUFSIZE_CHECKSUM_RESP, pEvent);
    if (retval != VP_STATUS_SUCCESS) {
        return retval;
    }

    /* Parse the mailbox contents. */
    pChecksum->loadChecksum      = ((uint32)*pBuf++ << 16);
    pChecksum->loadChecksum     |= (uint32)*pBuf++;
    pChecksum->vInfo.vtdRevCode  = (uint16)*pBuf++;
    pChecksum->vInfo.swProductId = (uint8)(*pBuf++ & 0xFF);
    pChecksum->vInfo.swVerMajor  = (uint8)(*pBuf >> 8);
    pChecksum->vInfo.swVerMinor  = (uint8)(*pBuf & 0xFF);

    /* To do: Add NGSLAC revision code handling here. */

    return VP_STATUS_SUCCESS;
}

/**
 * This internal function reads a device-io-access response from the
 * VCP2's response mailbox and fills in the supplied pResults struct
 * accordingly.
 *
 * \param[in]   pEvent      Event struct identifying the response type
 * \param[out]  pResults    Calling function's results struct
 *
 * \retval  ::VP_STATUS_SUCCESS
 * \retval  ::VP_STATUS_ERR_MAILBOX_DATA    unexpected response type or size
 * \retval  ::VP_STATUS_ERR_HBI
 */
static VpStatusType
ReadDevIoRdResponse(
    VpEventType *pEvent,
    void *pResults)
{
    VpDevCtxType *pDevCtx = pEvent->pDevCtx;
    uint16 buf[BUFSIZE_DEVIOACCESS_RESP];
    uint16 *pBuf = &buf[VCP2_HBI_MBOX_PAYLOAD];
    uint8 version;
    VpStatusType retval;

    /* Read the response from the response mailbox. */
    retval = ReadResponse(buf, BUFSIZE_DEVIOACCESS_RESP, pEvent);
    if (retval != VP_STATUS_SUCCESS) {
        return retval;
    }

    /* Parse the mailbox contents. */
    version = (uint8)(*pBuf >> 8);

    if (version == 1) {
        /* Vcp2DeviceIoAccess() was called.  Read the data into a new-format
           struct on the stack, then copy into pResults struct using old
           format. */
        VpDeviceIoAccessExtType deviceIoAccess;
        VpDeviceIoAccessDataType *pDeviceIoData = pResults;

        ParseDevIoRdResponse(&pBuf, &deviceIoAccess);
        MakeOldDevIoAccessStruct(pDevCtx, &deviceIoAccess, pDeviceIoData);
    } else {
        /* Vcp2DeviceIoAccessExt() was called.  Read the data into pResults
           struct using new format. */
        VpDeviceIoAccessExtType *pDeviceIoAccess = pResults;
        ParseDevIoRdResponse(&pBuf, pDeviceIoAccess);
    }

    return VP_STATUS_SUCCESS;
}

/**
 * This internal function reads a set-relative-gain response from the
 * VCP2's response mailbox and fills in the supplied pResults struct
 * accordingly.
 *
 * \param[in]   pEvent      Event struct identifying the response type
 * \param[out]  pResults    Calling function's results struct
 *
 * \retval  ::VP_STATUS_SUCCESS
 * \retval  ::VP_STATUS_ERR_MAILBOX_DATA    unexpected response type or size
 * \retval  ::VP_STATUS_ERR_HBI
 */
static VpStatusType
ReadGainCmpResponse(
    VpEventType *pEvent,
    void *pResults)
{
    VpRelGainResultsType *pRelGain = pResults;
    uint16 buf[BUFSIZE_GAINCMP_RESP];
    uint16 *pBuf = &buf[VCP2_HBI_MBOX_PAYLOAD];
    VpStatusType retval;

    /* Read the response from the response mailbox. */
    retval = ReadResponse(buf, BUFSIZE_GAINCMP_RESP, pEvent);
    if (retval != VP_STATUS_SUCCESS) {
        return retval;
    }

    /* Parse the mailbox contents. */
    pRelGain->gResult = (VpGainResultType)*pBuf++;
    pRelGain->gxValue = (uint16)*pBuf++;
    pRelGain->grValue = (uint16)*pBuf;

    return VP_STATUS_SUCCESS;
}

/**
 * This internal function reads a line-io-access response from the
 * VCP2's response mailbox and fills in the supplied pResults struct
 * accordingly.
 *
 * \param[in]   pEvent      Event struct identifying the response type
 * \param[out]  pResults    Calling function's results struct
 *
 * \retval  ::VP_STATUS_SUCCESS
 * \retval  ::VP_STATUS_ERR_MAILBOX_DATA    unexpected response type or size
 * \retval  ::VP_STATUS_ERR_HBI
 */
static VpStatusType
ReadLineIoRdResponse(
    VpEventType *pEvent,
    void *pResults)
{
    VpLineIoAccessType *pLineIoAccess = pResults;
    uint16 buf[BUFSIZE_LINEIOACCESS_RESP];
    uint16 *pBuf = &buf[VCP2_HBI_MBOX_PAYLOAD];
    VpStatusType retval;

    /* Read the response from the response mailbox. */
    retval = ReadResponse(buf, BUFSIZE_LINEIOACCESS_RESP, pEvent);
    if (retval != VP_STATUS_SUCCESS) {
        return retval;
    }

    /* Parse the mailbox contents. */
    pLineIoAccess->direction = *pBuf++ & 0xFF;
    pLineIoAccess->ioBits.mask = (uint8)(*pBuf >> 8);
    pLineIoAccess->ioBits.data = (uint8)(*pBuf & 0xFF);

    return VP_STATUS_SUCCESS;
}

/**
 * This internal function reads a lowlevel-command-complete response from the
 * VCP2's response mailbox and fills in the supplied pResults struct
 * accordingly.
 *
 * \param[in]   pEvent      Event struct identifying the response type
 * \param[out]  pResults    Calling function's results struct
 *
 * \retval  ::VP_STATUS_SUCCESS
 * \retval  ::VP_STATUS_ERR_MAILBOX_DATA    unexpected response type or size
 * \retval  ::VP_STATUS_ERR_HBI
 */
static VpStatusType
ReadLowLevCmdResponse(
    VpEventType *pEvent,
    void *pResults)
{
    uint8 *pData = pResults;
    uint16 buf[BUFSIZE_LOWLEVCMD_RESP];
    uint16 *pBuf = &buf[VCP2_HBI_MBOX_PAYLOAD];
    int lenBytes;
    VpStatusType retval;

    /* Read the response from the response mailbox. */
    retval = ReadResponse(buf, BUFSIZE_LOWLEVCMD_RESP, pEvent);
    if (retval != VP_STATUS_SUCCESS) {
        return retval;
    }

    /* Parse the mailbox contents. */
    lenBytes = (int)(*pBuf >> 8);

    while (lenBytes > 0) {
        *pData++ = (uint8)(*pBuf++ & 0xFF);
        if (--lenBytes > 0) {
            *pData++ = (uint8)(*pBuf >> 8);
        }
    }

    return VP_STATUS_SUCCESS;
}

/**
 * This internal function reads a read-loop-conditions response from the
 * VCP2's response mailbox and fills in the supplied pResults struct
 * accordingly.
 *
 * \param[in]   pEvent      Event struct identifying the response type
 * \param[out]  pResults    Calling function's results struct
 *
 * \retval  ::VP_STATUS_SUCCESS
 * \retval  ::VP_STATUS_ERR_MAILBOX_DATA    unexpected response type or size
 * \retval  ::VP_STATUS_ERR_HBI
 */
static VpStatusType
ReadRdLoopCondResponse(
    VpEventType *pEvent,
    void *pResults)
{
    VpLoopCondResultsType *pLoopCond = pResults;
    uint16 buf[BUFSIZE_RDLOOPCOND_RESP];
    uint16 *pBuf = &buf[VCP2_HBI_MBOX_PAYLOAD];
    VpStatusType retval;

    /* Read the response from the response mailbox. */
    retval = ReadResponse(buf, BUFSIZE_RDLOOPCOND_RESP, pEvent);
    if (retval != VP_STATUS_SUCCESS) {
        return retval;
    }

    /* Parse the mailbox contents. */
    pLoopCond->rloop       = (int16)*pBuf++;
    pLoopCond->ilg         = (int16)*pBuf++;
    pLoopCond->imt         = (int16)*pBuf++;
    pLoopCond->vsab        = (int16)*pBuf++;
    pLoopCond->vbat1       = (int16)*pBuf++;
    pLoopCond->vbat2       = (int16)*pBuf++;
    pLoopCond->vbat3       = (int16)*pBuf++;
    pLoopCond->mspl        = (int16)*pBuf++;
    pLoopCond->selectedBat = (VpBatteryType)*pBuf++;
    pLoopCond->dcFeedReg   = (VpDcFeedRegionType)*pBuf;

    return VP_STATUS_SUCCESS;
}

/**
 * This internal function reads a read-option response from the
 * VCP2's response mailbox and fills in the supplied pResults struct
 * accordingly.
 *
 * \param[in]   pEvent      Event struct identifying the response type
 * \param[out]  pResults    Calling function's results struct
 *
 * \retval  ::VP_STATUS_SUCCESS
 * \retval  ::VP_STATUS_ERR_MAILBOX_DATA    unexpected response type, size, or
 *                                          contents
 * \retval  ::VP_STATUS_ERR_HBI
 */
static VpStatusType
ReadRdOptionResponse(
    VpEventType *pEvent,
    void *pResults)
{
    VpDevCtxType *pDevCtx = pEvent->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    uint16 buf[BUFSIZE_RDOPTION_RESP];
    uint16 *pBuf = &buf[VCP2_HBI_MBOX_PAYLOAD];
    VpStatusType retval;
    VpOptionIdType optionId;

    /* Read the response from the response mailbox. */
    retval = ReadResponse(buf, BUFSIZE_RDOPTION_RESP, pEvent);
    if (retval != VP_STATUS_SUCCESS) {
        return retval;
    }

    /* Parse the mailbox contents. */
    optionId = (VpOptionIdType)*pBuf++;

    /* Parse the remainder of the response according to option ID. */
    switch (optionId) {
        case VP_DEVICE_OPTION_ID_PULSE:
        case VP_DEVICE_OPTION_ID_PULSE2: {
            VpOptionPulseType *pPulse = pResults;
            pPulse->makeMin     = (uint16)*pBuf++;
            pPulse->makeMax     = (uint16)*pBuf++;
            pPulse->breakMin    = (uint16)*pBuf++;
            pPulse->breakMax    = (uint16)*pBuf++;
            pPulse->interDigitMin = (uint16)*pBuf++;
            pPulse->flashMin    = (uint16)*pBuf++;
            pPulse->flashMax    = (uint16)*pBuf;
            break;
        }
        case VP_DEVICE_OPTION_ID_CRITICAL_FLT: {
            VpOptionCriticalFltType *pCflt = pResults;
            pCflt->acFltDiscEn = (bool)*pBuf++;
            pCflt->dcFltDiscEn = (bool)*pBuf++;
            pCflt->thermFltDiscEn = (bool)*pBuf;
            break;
        }
        case VP_OPTION_ID_ZERO_CROSS: {
            VpOptionZeroCrossType *pZero = pResults;
            *pZero = (VpOptionZeroCrossType)*pBuf;
            break;
        }
        case VP_DEVICE_OPTION_ID_RAMP2STBY: {
            uint16 *pRamp = pResults;
            *pRamp = (uint16)*pBuf;
            break;
        }
        case VP_OPTION_ID_PULSE_MODE: {
            VpOptionPulseModeType *pPulse = pResults;
            *pPulse = (VpOptionPulseModeType)*pBuf;
            break;
        }
        case VP_OPTION_ID_TIMESLOT: {
            VpOptionTimeslotType *pTimeslot = pResults;
            pTimeslot->tx = (uint8)*pBuf++;
            pTimeslot->rx = (uint8)*pBuf;
            break;
        }
        case VP_OPTION_ID_CODEC: {
            VpOptionPcmHwyType *pCodec = pResults;
            *pCodec = (VpOptionPcmHwyType)*pBuf;
            break;
        }
        case VP_OPTION_ID_PCM_HWY: {
            VpOptionPcmHwyType *pPcmHwy = pResults;
            *pPcmHwy = (VpOptionPcmHwyType)*pBuf;
            break;
        }
        case VP_OPTION_ID_LOOPBACK: {
            VpOptionLoopbackType *pLoopback = pResults;
            *pLoopback = (VpOptionLoopbackType)*pBuf;
            break;
        }
        case VP_OPTION_ID_LINE_STATE: {
            VpOptionLineStateType *pLineState = pResults;
            pLineState->battRev = (bool)( ((uint16)*pBuf & 0x0100) >> 8);
            pLineState->bat     = (VpOptionBatType)((uint16)*pBuf & 0x00FF);
            break;
        }
        case VP_OPTION_ID_EVENT_MASK: {
            VpOptionEventMaskType *pEventMask = pResults;
            pEventMask->faults      = (uint16)*pBuf++;
            pEventMask->signaling   = (uint16)*pBuf++;

#if (VCP2_TIMESTAMP_SHIFT > 0)
            /* The TS_ROLLOVER event is never masked internally; the API needs
               it to maintain the high-order bits of the timestamp. */
            if (pDevObj->timeStampEventMasked) {
                pEventMask->signaling |= VP_DEV_EVID_TS_ROLLOVER;
            }
#endif

            pEventMask->response    = (uint16)*pBuf++;
            pEventMask->test        = (uint16)*pBuf++;
            pEventMask->process     = (uint16)*pBuf;
            pEventMask->fxo         = 0xFFFF;
            pEventMask->packet      = 0xFFFF;
            break;
        }
        case VP_OPTION_ID_RING_CNTRL: {
            VpOptionRingControlType *pRingCtrl = pResults;
            pRingCtrl->zeroCross        = (VpOptionZeroCrossType)*pBuf++;
            pRingCtrl->ringExitDbncDur  = (uint16)*pBuf++;
            pRingCtrl->ringTripExitSt   = (VpLineStateType)*pBuf;
            break;
        }
        case VP_OPTION_ID_DTMF_MODE: {
            VpOptionDtmfModeType *pDtmfMode = pResults;
            uint8 *dtmfDetectionEnabled = &pDtmfMode->dtmfDetectionEnabled[0];

            pDtmfMode->dtmfControlMode      = (VpOptionDtmfModeControlType)*pBuf++;
            pDtmfMode->direction            = (VpDirectionType)*pBuf;

#if (VCP2_LINE_FLAG_BYTES < VP_LINE_FLAG_BYTES)
            /* Clear all unused dtmfDetectionEnabled bits. */
            {
                int byte;
                for (byte = VCP2_LINE_FLAG_BYTES; byte < VP_LINE_FLAG_BYTES; byte++) {
                    pDtmfMode->dtmfDetectionEnabled[byte] = 0;
                }
            }
#endif

            /* Read DTMF status registers from the VCP2. */
            retval = Vcp2ReadDtmfStatusRegs(pDevObj->deviceId, pDtmfMode, FALSE);

            /* Copy first 32 bits into deprecated member. */
            pDtmfMode->dtmfDetectionSetting = (uint32)dtmfDetectionEnabled[0]
                | (uint32)dtmfDetectionEnabled[1] << 8
                | (uint32)dtmfDetectionEnabled[2] << 16
                | (uint32)dtmfDetectionEnabled[3] << 24;

            break;
        }
        case VP_OPTION_ID_DTMF_SPEC: {
            VpOptionDtmfSpecType *pDtmfSpec = pResults;
            *pDtmfSpec = (VpOptionDtmfSpecType)*pBuf;
            break;
        }
        case VP_DEVICE_OPTION_ID_DEVICE_IO: {
            VpOptionDeviceIoType *pDeviceIo = pResults;
            VpOptionDeviceIoConfigType deviceIoConfig;

            ParseDevIoConfigResponse(&pBuf, &deviceIoConfig);
            MakeOldDevIoConfigStruct(pDevCtx, &deviceIoConfig, pDeviceIo);
            break;
        }
        case VP_DEVICE_OPTION_ID_DEV_IO_CFG: {
            VpOptionDeviceIoConfigType *pDeviceIoConfig = pResults;
            ParseDevIoConfigResponse(&pBuf, pDeviceIoConfig);
            break;
        }
        case VP_OPTION_ID_LINE_IO_CFG: {
            VpOptionLineIoConfigType *pLineIoConfig = pResults;
            pLineIoConfig->direction = (*pBuf >> 8) & 0xF;
            pLineIoConfig->outputType = *pBuf & 0xF;
            break;
        }
        case VP_OPTION_ID_PCM_TXRX_CNTRL: {
            VpOptionPcmTxRxCntrlType *pPcmTxRxCtrl = pResults;
            *pPcmTxRxCtrl = (VpOptionPcmTxRxCntrlType)*pBuf;
            break;
        }
        default:
            /* Response mailbox contains unsupported option. */
            retval = VP_STATUS_ERR_MAILBOX_DATA;
            break;
    }
    return retval;
}

/**
 * This internal function reads a response from the VCP2's response mailbox
 * and verifies that the mailbox contents are of the expected type and length.
 *
 * \param[out]  buf         Buffer to read the data into
 * \param[in]   maxLen      Maximum expected data length
 * \param[in]   pEvent      Event struct identifying the response type
 *
 * \retval  ::VP_STATUS_SUCCESS
 * \retval  ::VP_STATUS_ERR_MAILBOX_DATA    unexpected response type or size
 * \retval  ::VP_STATUS_ERR_HBI
 */
static VpStatusType
ReadResponse(
    uint16 *buf,
    int maxLen,
    VpEventType *pEvent)
{
    VpVcp2DeviceObjectType *pDevObj = pEvent->pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    VpEventCategoryType mboxEventCat;
    uint16 mboxEventId;
    uint8 mboxChannelId;

    /* Read the mailbox contents into the supplied buffer. */
    if (VpHbiRd(deviceId, VCP2_HBI_RESP_MBOX_PAG, VCP2_HBI_RESP_MBOX_OFF(0),
        maxLen - 1, buf) == FALSE) {
        return VP_STATUS_ERR_HBI;
    }

    /* Verify that the response is of the expected type. */
    if (
        (IdentifyEvent(pDevObj, buf[VCP2_HBI_MBOX_ID_CHAN], &mboxEventCat,
            &mboxEventId, &mboxChannelId) == FALSE) ||
        (pEvent->eventCategory != mboxEventCat) ||
        (pEvent->eventId != mboxEventId) ||
        (pEvent->channelId != mboxChannelId)
    ) {
        return VP_STATUS_ERR_MAILBOX_DATA;
    }

    return VP_STATUS_SUCCESS;
}

/**
 * This internal function reads a test-complete response from the
 * VCP2's response mailbox and fills in the supplied pResults struct
 * accordingly.
 *
 * \param[in]   pEvent      Event struct identifying the response type
 * \param[out]  pResults    Calling function's results struct
 *
 * \retval  ::VP_STATUS_SUCCESS
 * \retval  ::VP_STATUS_ERR_MAILBOX_DATA    unexpected response type or size
 * \retval  ::VP_STATUS_ERR_HBI
 */
static VpStatusType
ReadTestCmpResponse(
    VpEventType *pEvent,
    VpTestResultType *pResults)
{
    VpTestResultType *pTestResult = pResults;
    uint16 buf[BUFSIZE_TESTCMP_RESP];
    uint16 *pBuf = &buf[VCP2_HBI_MBOX_PAYLOAD];
    VpStatusType retval;

    /* Read the response from the response mailbox. */
    retval = ReadResponse(buf, BUFSIZE_TESTCMP_RESP, pEvent);
    if (retval != VP_STATUS_SUCCESS) {
        return retval;
    }

    /* Parse the mailbox contents. */
    pTestResult->testId = (VpTestIdType)*pBuf++;
    pTestResult->errorCode = (VpTestStatusType)*pBuf++;

    if (pTestResult->testId == VCP2_TEST_ID_AC_RLOOP_PHASE) {
        pTestResult->testId = VP_TEST_ID_AC_RLOOP_PHASE;
    }

    /* Parse the remainder of the response according to test type. */
    switch (pResults->testId) {
        case VP_TEST_ID_PREPARE:
        case VP_TEST_ID_CONCLUDE:
            /* No result for these two "tests." */
        case VP_TEST_ID_TONE_GEN:
            /* No other data for this test, just continue. */
            break;
        case VP_TEST_ID_ARB_1TONE:
            pTestResult->result.tone.highFreq = (uint16)*pBuf++;
            pTestResult->result.tone.highAmp  = (uint16)*pBuf;
            break;
        case VP_TEST_ID_DIALTONE:
            pTestResult->result.tone.rmsPwr   = (uint16)*pBuf++;
            pTestResult->result.tone.highFreq = (uint16)*pBuf++;
            pTestResult->result.tone.highAmp  = (uint16)*pBuf++;
            pTestResult->result.tone.secFreq  = (uint16)*pBuf++;
            pTestResult->result.tone.secAmp   = (uint16)*pBuf++;
            pTestResult->result.tone.thdFreq  = (uint16)*pBuf++;
            pTestResult->result.tone.thdAmp   = (uint16)*pBuf;
            break;
        case VP_TEST_ID_KEYPAD:
            pTestResult->result.keypad.kTestResult       = (VpKeyTestResultType)*pBuf++;
            pTestResult->result.keypad.digit             = (VpDigitType)*pBuf++;
            pTestResult->result.keypad.lowFreq           = (uint16)*pBuf++;
            pTestResult->result.keypad.lowAmp            = (uint16)*pBuf++;
            pTestResult->result.keypad.highFreq          = (uint16)*pBuf++;
            pTestResult->result.keypad.highAmp           = (uint16)*pBuf++;
            pTestResult->result.keypad.fftSize           = (uint16)*pBuf++;
            pTestResult->result.keypad.minBreak          = (uint16)*pBuf++;
            pTestResult->result.keypad.maxBreak          = (uint16)*pBuf++;
            pTestResult->result.keypad.aveBreak          = (uint16)*pBuf++;
            pTestResult->result.keypad.pulsePerSec       = (uint16)*pBuf++;
            pTestResult->result.keypad.hookFlashDet      = (bool)*pBuf++;
            pTestResult->result.keypad.hookFlashDuration = (uint16)*pBuf++;
            pTestResult->result.keypad.disconnectDet     = (bool)*pBuf;
            break;
        case VP_TEST_ID_OPEN_VDC:
            pTestResult->result.sVout = (int16)*pBuf;
            break;
        case VP_TEST_ID_OPEN_VAC:
        case VP_TEST_ID_DIFF_VAC:
        case VP_TEST_ID_MONITOR_IV:
            pTestResult->result.uVout = (uint16)*pBuf;
            break;
        case VP_TEST_ID_NOISE:
        case VP_TEST_ID_HYBRID_LOSS:
        case VP_TEST_ID_AC_TRANS:
            pTestResult->result.vrms = (uint16)*pBuf;
            break;
        case VP_TEST_ID_SNR_QDIST:
            pTestResult->result.snrqd.vrmsNoise  = (uint16)*pBuf++;
            pTestResult->result.snrqd.vrmsSignal = (uint16)*pBuf;
            break;
        case VP_TEST_ID_SET_SENSE_GAIN:
            /* No other data for this test, just continue. */
            break;
        case VP_TEST_ID_SET_BATTERIES:
            /* No other data for this test, just continue. */
            break;
        case VP_TEST_ID_DC_RLOOP:
            pTestResult->result.dcvab.vab = (int16)*pBuf++;
            pTestResult->result.dcvab.ilg = (int16)*pBuf;
            break;
        case VP_TEST_ID_AC_RLOOP:
            pTestResult->result.acimt.imt = (int16)*pBuf++;
            pTestResult->result.acimt.ilg = (int16)*pBuf;
            break;
        case VP_TEST_ID_AC_RLOOP_PHASE:
            pTestResult->result.acimtp.imt   = (int16)*pBuf++;
            pTestResult->result.acimtp.ilg   = (int16)*pBuf++;
            pTestResult->result.acimtp.phase = (int16)*pBuf;
            break;
        case VP_TEST_ID_NULL_IGND:
            pTestResult->result.vcm = (int16)*pBuf;
            break;
        case VP_TEST_ID_3ELE_CAP:
            pTestResult->result.cap.iahl  = (int16)*pBuf++;
            pTestResult->result.cap.iahh  = (int16)*pBuf++;
            pTestResult->result.cap.ibhh  = (int16)*pBuf++;
            pTestResult->result.cap.iall  = (int16)*pBuf++;
            pTestResult->result.cap.ibll  = (int16)*pBuf++;
            pTestResult->result.cap.vabhl = (int16)*pBuf;
            break;
        case VP_TEST_ID_3ELE_RES:
        case VP_TEST_ID_3ELE_RES_VMID:
            pTestResult->result.res.iahl  = (int16)*pBuf++;
            pTestResult->result.res.ibhl  = (int16)*pBuf++;
            pTestResult->result.res.iahh  = (int16)*pBuf++;
            pTestResult->result.res.ibhh  = (int16)*pBuf++;
            pTestResult->result.res.iall  = (int16)*pBuf++;
            pTestResult->result.res.ibll  = (int16)*pBuf++;
            pTestResult->result.res.vabhl = (int16)*pBuf;
            break;
        case VP_TEST_ID_IDC:
        case VP_TEST_ID_IAC:
            pTestResult->result.vix.ia = (int16)*pBuf++;
            pTestResult->result.vix.ib = (int16)*pBuf;
            break;
        case VP_TEST_ID_ALT_REN:
            pTestResult->result.ren.imt   = (int16)*pBuf++;
            pTestResult->result.ren.imtDc = (int16)*pBuf;
            break;
        case VP_TEST_ID_HOOK_STATUS:
            pTestResult->result.hstatus = (bool)*pBuf;
            break;
        case VP_TEST_ID_UNBAL_TONE:
            /* No other data for this test, just continue. */
            break;
        default:
            if (
                (pTestResult->testId >= VP_TEST_ID_RSVD_MIN) &&
                (pTestResult->testId <= VP_TEST_ID_RSVD_MAX)
            ) {
                /* Handle "reserved" test IDs in a generic fashion:
                 * Copy words from upstream mailbox into 'pTestResult->result'
                 * buffer.  Number of words to copy is the next value in
                 * the mailbox. */
                uint16 numResults = *pBuf++;
                uint16 *pGenericResults = (uint16 *)&pTestResult->result;
                while (numResults--) {
                    *pGenericResults++ = *pBuf++;
                }
            } else {
                /* Response mailbox contains unrecognized test ID. */
                return VP_STATUS_ERR_MAILBOX_DATA;
            }
    }

    return VP_STATUS_SUCCESS;
}

/**
 * This internal function resets all device-specific options on the specified
 * VCP2 device to their default values as specified in vp_api_cfg.h.
 *
 * \param[in]   pDevCtx     Device Context for the VCP2 device
 *
 * \retval      VP_STATUS_SUCCESS
 */
static VpStatusType
ResetDeviceOptions(
    VpDevCtxType *pDevCtx)
{
    VpStatusType retval = VP_STATUS_SUCCESS;

#if (CUSTOM_EVENT_MASK_DEFAULT)
    VpOptionEventMaskType eventMask;

    eventMask.faults = VCP2_DEFAULT_MASK_FAULT;
    eventMask.fxo = VCP2_DEFAULT_MASK_FXO;
    eventMask.signaling = VCP2_DEFAULT_MASK_SIGNALING;
    eventMask.response = VCP2_DEFAULT_MASK_RESPONSE;
    eventMask.test = VCP2_DEFAULT_MASK_TEST;
    eventMask.packet = VCP2_DEFAULT_MASK_PACKET;

    retval = Vcp2SetOption(VP_NULL, pDevCtx, VP_OPTION_ID_EVENT_MASK, &eventMask);
#endif

#if (CUSTOM_PULSE_DEFAULT)
    if (retval == VP_STATUS_SUCCESS) {
        VpOptionPulseType value = {
            VP_OPTION_DEFAULT_DP_BREAK_MIN,
            VP_OPTION_DEFAULT_DP_BREAK_MAX,
            VP_OPTION_DEFAULT_DP_MAKE_MIN,
            VP_OPTION_DEFAULT_DP_MAKE_MAX,
            VP_OPTION_DEFAULT_DP_INTER_DIG_MIN,
            VP_OPTION_DEFAULT_DP_FLASH_MIN,
            VP_OPTION_DEFAULT_DP_FLASH_MAX
        };
        retval = Vcp2SetOption(VP_NULL, pDevCtx,
            VP_DEVICE_OPTION_ID_PULSE, &value);
    }
#endif

#if (CUSTOM_CRITICAL_FLT_DEFAULT)
    if (retval == VP_STATUS_SUCCESS) {
        VpOptionCriticalFltType value = {
            VP_OPTION_DEFAULT_CF_AC_DIS_EN,
            VP_OPTION_DEFAULT_CF_DC_DIS_EN,
            VP_OPTION_DEFAULT_CF_THERMAL_DIS_EN
        };
        retval = Vcp2SetOption(VP_NULL, pDevCtx,
            VP_DEVICE_OPTION_ID_CRITICAL_FLT, &value);
    }
#endif

#if (CUSTOM_RAMP2STBY_DEFAULT)
    if (retval == VP_STATUS_SUCCESS) {
        uint16 value = VP_OPTION_DEFAULT_RAMP2STBY;
        retval = Vcp2SetOption(VP_NULL, pDevCtx, VP_DEVICE_OPTION_ID_RAMP2STBY,
            &value);
    }
#endif

    return retval;
}

/**
 * This internal function resets all line-specific options on the specified
 * VCP2 line to their default values as specified in vp_api_cfg.h.
 *
 * \param[in]   pDevCtx      Device Context for the VCP2 device
 * \param[in]   pLineCtx     Line Context for the VCP2 line (or VP_NULL)
 *
 * \retval      ::VP_STATUS_SUCCESS
 */
static VpStatusType
ResetLineOptions(
    VpDevCtxType *pDevCtx,
    VpLineCtxType *pLineCtx)
{
    VpStatusType retval = VP_STATUS_SUCCESS;
    
    FixTipOpenBug(pDevCtx, pLineCtx);

#if (CUSTOM_ZERO_CROSS_DEFAULT)
    if (retval == VP_STATUS_SUCCESS) {
        VpOptionZeroCrossType value = VP_OPTION_DEFAULT_ZERO_CROSS;
        retval = Vcp2SetOption(pLineCtx, pDevCtx,
            VP_OPTION_ID_ZERO_CROSS, &value);
    }
#endif

#if (CUSTOM_PULSE_MODE_DEFAULT)
    if (retval == VP_STATUS_SUCCESS) {
        VpOptionPulseModeType value = VP_OPTION_DEFAULT_PULSE_MODE;
        retval = Vcp2SetOption(pLineCtx, pDevCtx, VP_OPTION_ID_PULSE_MODE,
            &value);
    }
#endif

#if (CUSTOM_CODEC_DEFAULT)
    if (retval == VP_STATUS_SUCCESS) {
        VpOptionCodecType value = VP_OPTION_DEFAULT_CODEC_MODE;
        retval = Vcp2SetOption(pLineCtx, pDevCtx, VP_OPTION_ID_CODEC,
            &value);
    }
#endif

#if (CUSTOM_PCM_HWY_DEFAULT)
    if (retval == VP_STATUS_SUCCESS) {
        VpOptionPcmHwyType value = VP_OPTION_DEFAULT_PCM_HWY;
        retval = Vcp2SetOption(pLineCtx, pDevCtx, VP_OPTION_ID_PCM_HWY,
            &value);
    }
#endif

#if (CUSTOM_LOOPBACK_DEFAULT)
    if (retval == VP_STATUS_SUCCESS) {
        VpOptionLoopbackType value = VP_OPTION_DEFAULT_LOOP_BACK;
        retval = Vcp2SetOption(pLineCtx, pDevCtx, VP_OPTION_ID_LOOPBACK,
            &value);
    }
#endif

#if (CUSTOM_LINE_STATE_DEFAULT)
    if (retval == VP_STATUS_SUCCESS) {
        VpOptionLineStateType value = {
            VP_OPTION_DEFAULT_LS_BAT_REV,
            VP_OPTION_DEFAULT_LS_BAT
        };
        retval = Vcp2SetOption(pLineCtx, pDevCtx, VP_OPTION_ID_LINE_STATE,
            &value);
    }
#endif

#if (CUSTOM_EVENT_MASK_DEFAULT)
    if (retval == VP_STATUS_SUCCESS) {
        VpOptionEventMaskType value = {
            VP_OPTION_DEFAULT_FAULT_EVENT_MASK,
            VP_OPTION_DEFAULT_SIGNALING_EVENT_MASK,
            VP_OPTION_DEFAULT_RESPONSE_EVENT_MASK,
            VP_OPTION_DEFAULT_TEST_EVENT_MASK,
            VP_OPTION_DEFAULT_PROCESS_EVENT_MASK,
            VP_OPTION_DEFAULT_FXO_EVENT_MASK,
            VP_OPTION_DEFAULT_PACKET_EVENT_MASK
        };
        retval = Vcp2SetOption(pLineCtx, pDevCtx, VP_OPTION_ID_EVENT_MASK,
            &value);
    }
#endif

#if (CUSTOM_RING_CNTRL_DEFAULT)
    if (retval == VP_STATUS_SUCCESS) {
        VpOptionRingControlType value = {
            VP_OPTION_DEFAULT_RC_ZERO_CROSS,
            VP_OPTION_DEFAULT_RC_RING_EXIT_DBNC_VAL,
            VP_OPTION_DEFAULT_RC_RING_EXIT_STATE
        };
        retval = Vcp2SetOption(pLineCtx, pDevCtx, VP_OPTION_ID_RING_CNTRL,
            &value);
    }
#endif

#if (CUSTOM_DTMF_MODE_DEFAULT)
    if (retval == VP_STATUS_SUCCESS) {
        VpOptionDtmfModeType value = {
            VP_OPTION_DEFAULT_DM_DTMF_CONTROL_MODE,
            VP_OPTION_DEFAULT_DM_DIRECTION,
            0,
            0
        };
        retval = Vcp2SetOption(pLineCtx, pDevCtx, VP_OPTION_ID_DTMF_MODE,
            &value);
    }
#endif

#if (CUSTOM_PCM_TXRX_CNTRL_DEFAULT)
    if (retval == VP_STATUS_SUCCESS) {
        VpOptionPcmTxRxCntrlType value = VP_OPTION_DEFAULT_PCM_TXRX_CNTRL;
        retval = Vcp2SetOption(pLineCtx, pDevCtx, VP_OPTION_ID_PCM_TXRX_CNTRL,
            &value);
    }
#endif

    return retval;
}

#endif /* VP_CC_VCP2_SERIES */
