/** \file vcp2_test.c
 * vcp2_test.c
 *
 * This file contains the implementation for the Line Testing functions
 * in the VCP2 library.
 *
 * The functions herein are used internally by the API and should not be
 * called directly by the application.
 *
 * Copyright (c) 2008, Zarlink Semiconductor, Inc.
 */

/* Only compile this file if specified in vp_api_cfg.h: */
#include "vp_api_cfg.h"
#if defined (VP_CC_VCP2_SERIES)

#include "vcp2_api.h"
#include "vp_api_int.h"
#include "vcp2_api_int.h"


/* =================================
    Prototypes for Static Functions
   ================================= */

static bool
InvalidFftSize(
    uint16 fftSize);

static void
MailboxAddTestArgs(
    uint16 **ppBuf,
    VpTestIdType testId,
    const void *pArgs);

static bool
TestArgsNotSupported(
    VpVcp2DeviceObjectType *pDevObj,
    VpTestIdType testId,
    const void *pArgs);

static bool
TestNotSupported(
    VpVcp2DeviceObjectType *pDevObj,
    VpTestIdType testId);


/* ========================
    Line Testing Functions
   ======================== */

/**
 * This function implements ::VpTestLine() for VCP2 devices.  See the VoicePath
 * API Reference Guide for details.
 *
 * \param[in] pLineCtx      Line Context for the line under test
 * \param[in] testId        Identifies which line test is to be run
 * \param[in] pArgs         Pointer to test-specific arguments struct
 * \param[in] handle        Handle to be returned with "test complete" event
 *
 * \retval ::VP_STATUS_SUCCESS
 * \retval ::VP_STATUS_DEV_NOT_INITIALIZED
 * \retval ::VP_STATUS_FUNC_NOT_SUPPORTED   Invalid or unsupported test ID
 * \retval ::VP_STATUS_INVALID_ARG          Invalid test argument(s)
 * \retval ::VP_STATUS_MAILBOX_BUSY
 * \retval ::VP_STATUS_ERR_HBI
 */
VpStatusType
Vcp2TestLine(
    VpLineCtxType *pLineCtx,
    VpTestIdType testId,
    const void *pArgs,
    uint16 handle)
{
    VpVcp2LineObjectType *pLineObj = pLineCtx->pLineObj;
    uint8 channelId = pLineObj->channelId;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcp2DeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 buf[BUFSIZE_TESTLINE];
    uint16 *pBuf = buf;
    VpStatusType retval;

    if (pDevObj->devInit == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    if (TestNotSupported(pDevObj, testId)) {
        return VP_STATUS_FUNC_NOT_SUPPORTED;
    }

    if (TestArgsNotSupported(pDevObj, testId, pArgs)) {
        return VP_STATUS_INVALID_ARG;
    }

    /* Add the standard header information to the mailbox buffer. */
    Vcp2MailboxInit(&pBuf, VCP2_HBI_CMD_TESTLINE, channelId);

    /* Add the handle field to the mailbox buffer. */
    *pBuf++ = (uint16)handle;

    /* Add the Test ID field. */
    if (testId == VP_TEST_ID_AC_RLOOP_PHASE) {
        *pBuf++ = VCP2_TEST_ID_AC_RLOOP_PHASE;
    } else {
        *pBuf++ = (uint16)testId;
    }

    /* Add the test-specific arguments. */
    MailboxAddTestArgs(&pBuf, testId, pArgs);

    /* ----------------------------------------- Begin HBI Critical Section */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);

    /* Stuff the data into the mailbox and release the mailbox to the VCP2. */
    retval = Vcp2MailboxSend(deviceId, buf, pBuf);

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);
    /* ------------------------------------------- End HBI Critical Section */

    return retval;
}


/* ==================
    Static Functions
   ================== */

/**
 * This internal function determines whether the argument is a valid FFT
 * size (i.e. supported by the VCP2 firmware).
 *
 * \param[in] fftSize   FFT size to be evaluated for validity
 *
 * \retval ::TRUE   FFT size is invalid
 * \retval ::FALSE  FFT size is valid
 */
static bool
InvalidFftSize(
    uint16 fftSize)
{
    /* Zero is valid. */
    if (fftSize == 0) {
        return FALSE;
    }

    /* If not a power of two, then invalid. */
    if ((fftSize - 1) & fftSize) {
        return TRUE;
    }

    /* Is it within the accepted limits for nonzero values? */
    if (fftSize & 0x3F80) {
        return FALSE;
    }

    return TRUE;
}

/**
 * This internal function unpacks a test args struct and adds it into the
 * specified command mailbox buffer.
 *
 * \param[in] ppBuf     Pointer to the calling function's pBuf variable
 * \param[in] testId    Test ID of the test argument(s) to be added
 * \param[in] value     Pointer to struct containing test-specific arguments
 */
static void
MailboxAddTestArgs(
    uint16 **ppBuf,
    VpTestIdType testId,
    const void *pArgs)
{
    uint16 *pBuf = *ppBuf;

    /* Add the test arguments to the mailbox */
    switch (testId) {
        case VP_TEST_ID_PREPARE:
            /* No parameters for this test type. */
            break;

        case VP_TEST_ID_CONCLUDE: {
            const VpTestConcludeType *conclude = pArgs;
            *pBuf++ = (uint16)conclude->abortTest;
            break;
        }
        case VP_TEST_ID_TONE_GEN: {
            const VpTestToneGenType *toneGen = pArgs;
            *pBuf++ = (uint16)toneGen->duration;
            *pBuf++ = (uint16)toneGen->freq1;
            *pBuf++ = (uint16)toneGen->amp1;
            *pBuf++ = (uint16)toneGen->freq2;
            *pBuf++ = (uint16)toneGen->amp2;
            *pBuf++ = (uint16)toneGen->freq3;
            *pBuf++ = (uint16)toneGen->amp3;
            *pBuf++ = (uint16)toneGen->freq4;
            *pBuf++ = (uint16)toneGen->amp4;
            break;
        }
        case VP_TEST_ID_ARB_1TONE: {
            const VpTestArb1ToneType *arb1Tone = pArgs;
            *pBuf++ = (uint16)arb1Tone->timeout;
            *pBuf++ = (uint16)arb1Tone->fftSize;
            *pBuf++ = (uint16)arb1Tone->threshold;
            *pBuf++ = (uint16)arb1Tone->useBuffer;
            break;
        }
        case VP_TEST_ID_KEYPAD: {
            const VpTestKeypadType *keypad = pArgs;
            *pBuf++ = (uint16)keypad->kTest;
            *pBuf++ = (uint16)keypad->timeout;
            *pBuf++ = (uint16)keypad->freq;
            *pBuf++ = (uint16)keypad->amp;
            *pBuf++ = (uint16)keypad->fftSize;
            *pBuf++ = (uint16)keypad->threshold;
            *pBuf++ = (uint16)keypad->pulseOpt.makeMin;
            *pBuf++ = (uint16)keypad->pulseOpt.makeMax;
            *pBuf++ = (uint16)keypad->pulseOpt.breakMin;
            *pBuf++ = (uint16)keypad->pulseOpt.breakMax;
            *pBuf++ = (uint16)keypad->pulseOpt.interDigitMin;
            *pBuf++ = (uint16)keypad->pulseOpt.flashMin;
            *pBuf++ = (uint16)keypad->pulseOpt.flashMax;
            *pBuf++ = (uint16)keypad->useBuffer;
            break;
        }
        case VP_TEST_ID_OPEN_VDC:
        case VP_TEST_ID_OPEN_VAC: {
            const VpTestOpenVType *openV = pArgs;
            *pBuf++ = (uint16)openV->calMode;   /* ignored for open VAC */
            *pBuf++ = (uint16)openV->integrateTime;
            *pBuf++ = (uint16)openV->settlingTime;
            *pBuf++ = (uint16)openV->tip;
            break;
        }
        case VP_TEST_ID_DIFF_VAC: {
            const VpTestDiffVacType *diffVac = pArgs;
            *pBuf++ = (uint16)diffVac->integrateTime;
            *pBuf++ = (uint16)diffVac->gvsaCal;
            *pBuf++ = (uint16)diffVac->gvsbCal;
            *pBuf++ = (uint16)diffVac->settlingTime;
            break;
        }
        case VP_TEST_ID_NOISE: {
            const VpTestNoiseType *noise = pArgs;
            *pBuf++ = (uint16)noise->integrateTime;
            *pBuf++ = (uint16)noise->filter;
            *pBuf++ = (uint16)noise->amp;
            *pBuf++ = (uint16)noise->useBuffer;
            break;
        }
        case VP_TEST_ID_DIALTONE: {
            const VpTestDialToneType *dialTone = pArgs;
            *pBuf++ = (uint16)dialTone->inward;
            *pBuf++ = (uint16)dialTone->timeout;
            *pBuf++ = (uint16)dialTone->detectGran;
            *pBuf++ = (uint16)dialTone->threshold;
            *pBuf++ = (uint16)dialTone->fftSize;
            *pBuf++ = (uint16)dialTone->duration;
            *pBuf++ = (uint16)dialTone->useBuffer;
            break;
        }
        case VP_TEST_ID_SNR_QDIST: {
            const VpTestSnrQDistorType *snrqd = pArgs;
            *pBuf++ = (uint16)snrqd->integrateTime;
            *pBuf++ = (uint16)snrqd->amp;
            *pBuf++ = (uint16)snrqd->useBuffer;
            break;
        }
        case VP_TEST_ID_AC_TRANS:
        case VP_TEST_ID_HYBRID_LOSS: {
            const VpTestTransHybridLossType *hybridLoss = pArgs;
            *pBuf++ = (uint16)hybridLoss->integrateTime;
            *pBuf++ = (uint16)hybridLoss->freq;
            *pBuf++ = (uint16)hybridLoss->amp;
            *pBuf++ = (uint16)hybridLoss->useBuffer;
            break;
        }
        case VP_TEST_ID_3ELE_RES: {
            const VpTestResType *res = pArgs;
            *pBuf++ = (uint16)res->calMode;
            *pBuf++ = (uint16)res->vCm;
            *pBuf++ = (uint16)res->vDiff;
            *pBuf++ = (uint16)res->integrateTime;
            *pBuf++ = (uint16)res->settlingTime;
            break;
        }
        case VP_TEST_ID_3ELE_RES_VMID: {
            const VpTest3EleResVmidType *resVmid = pArgs;
            *pBuf++ = (uint16)resVmid->calMode;
            *pBuf++ = (uint16)resVmid->vCm;
            *pBuf++ = (uint16)resVmid->vDiff;
            *pBuf++ = (uint16)resVmid->integrateTime;
            *pBuf++ = (uint16)resVmid->settlingTime;
            *pBuf++ = (uint16)resVmid->vMid;
            break;
        }
        case VP_TEST_ID_3ELE_CAP: {
            const VpTestCapType *cap = pArgs;
            *pBuf++ = (uint16)cap->calMode;
            *pBuf++ = (uint16)cap->vCm;
            *pBuf++ = (uint16)cap->vDiff;
            *pBuf++ = (uint16)cap->freq;
            *pBuf++ = (uint16)cap->gimtCal;
            *pBuf++ = (uint16)cap->gilgCal;
            *pBuf++ = (uint16)cap->integrateTime;
            *pBuf++ = (uint16)cap->settlingTime;
            break;
        }
        case VP_TEST_ID_SET_SENSE_GAIN: {
            const VpTestSetSenseGainType *senseGain = pArgs;
            *pBuf++ = (uint16)senseGain->imt;
            *pBuf++ = (uint16)senseGain->ilg;
            break;
        }
        case VP_TEST_ID_SET_BATTERIES: {
            const VpTestSetBatteriesType *batteries = pArgs;
            *pBuf++ = (uint16)batteries->bs0;
            *pBuf++ = (uint16)batteries->bs1;
            break;
        }
        case VP_TEST_ID_DC_RLOOP: {
            const VpTestDcRLoopType *dcRLoop = pArgs;
            *pBuf++ = (uint16)dcRLoop->calMode;
            *pBuf++ = (uint16)dcRLoop->iTestLevel;
            *pBuf++ = (uint16)dcRLoop->integrateTime;
            *pBuf++ = (uint16)dcRLoop->settlingTime;
            break;
        }
        case VP_TEST_ID_AC_RLOOP:
        case VP_TEST_ID_AC_RLOOP_PHASE: {
            const VpTestAcRLoopType *acRLoop = pArgs;
            *pBuf++ = (uint16)acRLoop->calMode;
            *pBuf++ = (uint16)acRLoop->freq;
            *pBuf++ = (uint16)acRLoop->vTestLevel;
            *pBuf++ = (uint16)acRLoop->vBias;
            *pBuf++ = (uint16)acRLoop->integrateTime;
            *pBuf++ = (uint16)acRLoop->settlingTime;
            break;
        }
        case VP_TEST_ID_NULL_IGND: {
            const VpTestNullIGndType *nullIGnd = pArgs;
            *pBuf++ = (uint16)nullIGnd->vTestLevel;
            *pBuf++ = (uint16)nullIGnd->iTarget;
            *pBuf++ = (uint16)nullIGnd->tip;
            *pBuf++ = (uint16)nullIGnd->settlingTime;
            break;
        }
        case VP_TEST_ID_IDC: {
            const VpTestIdcType *idc = pArgs;
            *pBuf++ = (uint16)idc->calMode;
            *pBuf++ = (uint16)idc->vCm;
            *pBuf++ = (uint16)idc->integrateTime;
            *pBuf++ = (uint16)idc->settlingTime;
            break;
        }
        case VP_TEST_ID_IAC: {
            const VpTestIacType *iac = pArgs;
            *pBuf++ = (uint16)iac->gimtCal;
            *pBuf++ = (uint16)iac->gilgCal;
            *pBuf++ = (uint16)iac->vCm;
            *pBuf++ = (uint16)iac->integrateTime;
            *pBuf++ = (uint16)iac->settlingTime;
            break;
        }
        case VP_TEST_ID_ALT_REN: {
            const VpTestRenType *ren = pArgs;
            *pBuf++ = (uint16)ren->integrateTime;
            *pBuf++ = (uint16)ren->vDiff;
            break;
        }
        case VP_TEST_ID_HOOK_STATUS:
            /* No parameters for this test type. */
            break;
        case VP_TEST_ID_MONITOR_IV: {
            const VpTestMonitorIvType *monitorIv = pArgs;
            *pBuf++ = (uint16)monitorIv->integrateTime;
            *pBuf++ = (uint16)monitorIv->settlingTime;
            *pBuf++ = (uint16)monitorIv->type;
            *pBuf++ = (uint16)monitorIv->measurement;
            break;
        }
        case VP_TEST_ID_UNBAL_TONE: {
            const VpTestUnbalToneType *uTone = pArgs;
            *pBuf++ = (uint16)uTone->freq;
            *pBuf++ = (uint16)uTone->vTestlevel;
            *pBuf++ = (uint16)uTone->vTipBias;
            *pBuf++ = (uint16)uTone->vRingBias;
            *pBuf++ = (uint16)uTone->duration;
            *pBuf++ = (uint16)uTone->tip;
        }
        default:
            if (
                (testId >= VP_TEST_ID_RSVD_MIN) &&
                (testId <= VP_TEST_ID_RSVD_MAX)
            ) {
                /* Handle "reserved" test IDs in a generic fashion:
                 * Copy words from '(uint16p)pArgs' into downstream
                 * mailbox.  Number of arguments is contained in the
                 * first element of the '(uint16p)pArgs' array. */
                const uint16 *pArg = pArgs;
                int numArgs = pArg[0];
                while (numArgs--) {
                    *pBuf++ = *pArg++;
                }
            }
    }

    /* Update the calling function's mailbox buffer pointer. */
    *ppBuf = pBuf;
}

/**
 * This internal function verifies that the test-specific arguments are valid
 * for the specified test type, based on the firmware product type.
 *
 * \param[in] pDevObj   Device Object of the VCP2 device
 * \param[in] testId    Test ID to be verified
 * \param[in] pArgs     Arguments to be evaluated for validity
 *
 * \retval ::TRUE   Test argument value(s) are not supported
 * \retval ::FALSE  Test argument values are supported
 */
static bool
TestArgsNotSupported(
    VpVcp2DeviceObjectType *pDevObj,
    VpTestIdType testId,
    const void *pArgs)
{
    switch (testId) {
        case VP_TEST_ID_PREPARE:
            /* No parameters for this test type. */
            break;

        case VP_TEST_ID_CONCLUDE: {
            const VpTestConcludeType *conclude = pArgs;
            if (
                (conclude->abortTest != TRUE) &&
                (conclude->abortTest != FALSE)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_TEST_ID_TONE_GEN:
            /* No error checking of these parameters required. */
            break;

        case VP_TEST_ID_ARB_1TONE: {
            const VpTestArb1ToneType *arb1Tone = pArgs;
            if (InvalidFftSize(arb1Tone->fftSize)) {
                return TRUE;
            }
            if (
                (arb1Tone->useBuffer != TRUE) &&
                (arb1Tone->useBuffer != FALSE)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_TEST_ID_KEYPAD: {
            const VpTestKeypadType *keypad = pArgs;
            if (
                (keypad->kTest >= VP_NUM_KEY_TEST_TYPES) ||
                (InvalidFftSize(keypad->fftSize))
            ) {
                return TRUE;
            }
            if (
                (keypad->kTest == VP_KEY_TEST_PULSE) ||
                (keypad->kTest == VP_KEY_TEST_BOTH)
            ) {
                /* Verify valid pulse detection options. */
                const VpOptionPulseType *pulseOpt = &keypad->pulseOpt;
                if (
                    (pulseOpt->makeMin > pulseOpt->makeMax) ||
                    (pulseOpt->breakMin > pulseOpt->breakMax) ||
                    (pulseOpt->flashMin > pulseOpt->flashMax)
                ) {
                    return TRUE;
                }
            }
            if (
                (keypad->useBuffer != TRUE) &&
                (keypad->useBuffer != FALSE)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_TEST_ID_OPEN_VDC:
        case VP_TEST_ID_OPEN_VAC: {
            const VpTestOpenVType *openV = pArgs;
            if (
                (openV->tip != VP_TEST_TIP) &&
                (openV->tip != VP_TEST_RING)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_TEST_ID_DIFF_VAC:
            /* No error checking of these parameters required. */
            break;

        case VP_TEST_ID_NOISE: {
            const VpTestNoiseType *noise = pArgs;
            if (
                (noise->filter < 0) ||
                (noise->filter >= VP_NUM_TEST_FILTER_TYPES)
            ) {
                return TRUE;
            }
            if (
                (noise->useBuffer != TRUE) &&
                (noise->useBuffer != FALSE)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_TEST_ID_DIALTONE: {
            const VpTestDialToneType *dialTone = pArgs;
            if (
                (dialTone->inward != TRUE) &&
                (dialTone->inward != FALSE)
            ) {
                return TRUE;
            }
            if (InvalidFftSize(dialTone->fftSize)) {
                return TRUE;
            }
            if (
                (dialTone->useBuffer != TRUE) &&
                (dialTone->useBuffer != FALSE)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_TEST_ID_SNR_QDIST: {
            const VpTestSnrQDistorType *snrqd = pArgs;
            if (
                (snrqd->useBuffer != TRUE) &&
                (snrqd->useBuffer != FALSE)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_TEST_ID_AC_TRANS:
        case VP_TEST_ID_HYBRID_LOSS: {
            const VpTestTransHybridLossType *hybridLoss = pArgs;
            if (
                (hybridLoss->useBuffer != TRUE) &&
                (hybridLoss->useBuffer != FALSE)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_TEST_ID_3ELE_RES: {
            const VpTestResType *res = pArgs;
            if (
                (res->calMode != TRUE) &&
                (res->calMode != FALSE)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_TEST_ID_3ELE_RES_VMID: {
            const VpTest3EleResVmidType *resVmid = pArgs;
            if (
                (resVmid->calMode != TRUE) &&
                (resVmid->calMode != FALSE)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_TEST_ID_3ELE_CAP: {
            const VpTestCapType *cap = pArgs;
            if (
                (cap->calMode != TRUE) &&
                (cap->calMode != FALSE)
            ) {
                return TRUE;
            }
            if (cap->freq > 4000) {
                return TRUE;
            }
            break;
        }
        case VP_TEST_ID_SET_SENSE_GAIN: {
            const VpTestSetSenseGainType *senseGain = pArgs;
            if (
                (senseGain->imt != VP_SENSE_GAIN_HIGH) &&
                (senseGain->imt != VP_SENSE_GAIN_LOW)
            ) {
                return TRUE;
            }
            if (
                (senseGain->ilg != VP_SENSE_GAIN_HIGH) &&
                (senseGain->ilg != VP_SENSE_GAIN_LOW)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_TEST_ID_SET_BATTERIES: {
            const VpTestSetBatteriesType *batteries = pArgs;
            if (
                (batteries->bs0 != VP_SENSE_GAIN_HIGH) &&
                (batteries->bs0 != VP_SENSE_GAIN_LOW)
            ) {
                return TRUE;
            }
            if (
                (batteries->bs1 != VP_SENSE_GAIN_HIGH) &&
                (batteries->bs1 != VP_SENSE_GAIN_LOW)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_TEST_ID_DC_RLOOP: {
            const VpTestDcRLoopType *dcRLoop = pArgs;
            if (
                (dcRLoop->calMode != TRUE) &&
                (dcRLoop->calMode != FALSE)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_TEST_ID_AC_RLOOP:
        case VP_TEST_ID_AC_RLOOP_PHASE: {
            const VpTestAcRLoopType *acRLoop = pArgs;
            if (
                (acRLoop->calMode != TRUE) &&
                (acRLoop->calMode != FALSE)
            ) {
                return TRUE;
            }
            if (acRLoop->freq > 4000) {
                return TRUE;
            }
            break;
        }
        case VP_TEST_ID_NULL_IGND: {
            const VpTestNullIGndType *nullIGnd = pArgs;
            if (
                (nullIGnd->tip != VP_TEST_TIP) &&
                (nullIGnd->tip != VP_TEST_RING)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_TEST_ID_IDC: {
            const VpTestIdcType *idc = pArgs;
            if (
                (idc->calMode != TRUE) &&
                (idc->calMode != FALSE)
            ) {
                return TRUE;
            }
            break;
        }
        case VP_TEST_ID_IAC:
            /* No error checking of these parameters required. */
            break;

        case VP_TEST_ID_ALT_REN:
            /* No error checking of these parameters required. */
            break;

        case VP_TEST_ID_HOOK_STATUS:
            /* No parameters for this test type. */
            break;
            
        case VP_TEST_ID_MONITOR_IV: {
            const VpTestMonitorIvType *monIv = pArgs;
            if (
                (monIv->type != VP_TEST_TYPE_DC) &&
                (monIv->type != VP_TEST_TYPE_AC)
            ) {
                return TRUE;
            }
            switch (monIv->measurement) {
                case VP_TEST_MEAS_VA:
                case VP_TEST_MEAS_VB:
                case VP_TEST_MEAS_VAB:
                case VP_TEST_MEAS_IMT:
                case VP_TEST_MEAS_ILG:
                    break;
                default:
                    return TRUE;
            }
            break;
        }
        case VP_TEST_ID_UNBAL_TONE: {
            const VpTestUnbalToneType *uTone = pArgs;
            if (
                (uTone->tip != VP_TEST_TIP) &&
                (uTone->tip != VP_TEST_RING)
            ) {
                return TRUE;
            }
            if (uTone->freq > 4000) {
                return TRUE;
            }
            break;
        }
        default:
            if (
                (testId >= VP_TEST_ID_RSVD_MIN) &&
                (testId <= VP_TEST_ID_RSVD_MAX)
            ) {
                /* Customer-specific tests.  Verify that the arguments will
                 * fit in the mailbox buffer. */
                const uint16 *pArg = pArgs;
                int numArgs = pArg[0];
                if (numArgs > (BUFSIZE_TESTLINE - 1)) {
                    return TRUE;
                }
            }
            break;
    }

    /* All arguments are valid. */
    return FALSE;
}

/**
 * This internal function verifies that the selected test ID is supported
 * by the specified VCP2 device, based on the firmware product type.
 *
 * \param[in] pDevObj   Device Object of the VCP2 device
 * \param[in] testId    Test ID to be verified
 *
 * \retval ::TRUE   Test is not supported by firmware product type
 * \retval ::FALSE  Test is supported.
 */
static bool
TestNotSupported(
    VpVcp2DeviceObjectType *pDevObj,
    VpTestIdType testId)
{

    /* Determine the capabilities of the detected VCP2 firmware product. */
    int features = pDevObj->product & VP_VCP_PROD_FEAT_MASK;
    bool noAdvancedTestSupport = (features != VP_VCP_PROD_FEAT_AT)
        && (features != VP_VCP_PROD_FEAT_DEMO);
    bool noBasicTestSupport = noAdvancedTestSupport
        && (features != VP_VCP_PROD_FEAT_BT);

    /* Standard VCP2-790 line tests: */
    switch (testId) {

        /* Basic test package: */
        case VP_TEST_ID_PREPARE:
        case VP_TEST_ID_CONCLUDE:
        case VP_TEST_ID_SET_SENSE_GAIN:
        case VP_TEST_ID_OPEN_VDC:
        case VP_TEST_ID_OPEN_VAC:
        case VP_TEST_ID_DIFF_VAC:
        case VP_TEST_ID_IDC:
        case VP_TEST_ID_DC_RLOOP:
        case VP_TEST_ID_AC_RLOOP:
        case VP_TEST_ID_HOOK_STATUS:
        case VP_TEST_ID_ALT_REN:
        case VP_TEST_ID_3ELE_RES:
        case VP_TEST_ID_3ELE_RES_VMID:
        case VP_TEST_ID_NOISE:
        case VP_TEST_ID_HYBRID_LOSS:
        case VP_TEST_ID_AC_TRANS:
        case VP_TEST_ID_SET_BATTERIES:
            return noBasicTestSupport;

        /* Advanced test package: */
        case VP_TEST_ID_IAC:
        case VP_TEST_ID_TONE_GEN:
        case VP_TEST_ID_3ELE_CAP:
        case VP_TEST_ID_NULL_IGND:
        case VP_TEST_ID_KEYPAD:
        case VP_TEST_ID_SNR_QDIST:
        case VP_TEST_ID_ARB_1TONE:
        case VP_TEST_ID_DIALTONE:
        case VP_TEST_ID_AC_RLOOP_PHASE:
        case VP_TEST_ID_MONITOR_IV:
        case VP_TEST_ID_PCM_COLLECT:
        case VP_TEST_ID_UNBAL_TONE:
            return noAdvancedTestSupport;

        default:
            /* Silence the compiler warning message about unhandled values. */
            ;
    }

    /* Customer-specific line tests: */
    if (
        (testId >= VP_TEST_ID_RSVD_MIN) &&
        (testId <= VP_TEST_ID_RSVD_MAX)
    ) {
        return FALSE;
    }

    /* Otherwise, invalid test ID. */
    return TRUE;
}

#endif /* VP_CC_VCP2_SERIES */
