/** \file vp_kernel.c
 * vp_kernel.c
 *
 *  This file contains the implementation of the VP-API to the VP-API Kernel
 * driver.
 */

/* INCLUDES */
#include    "vp_api.h"

#if defined (VP_CC_KWRAP)  /* Compile only if required */

#include    "leg_vp_dvr.h"
#include    "vp_kernel.h"
#include    "vp_api_int.h"
#include    "boot_common.h"


/*******************************************************************************
 *              Linux Kwrap specific API function prototypes
 ******************************************************************************/
/* Initialization functions */
static VpStatusType KWrapInitDevice(
    VpDevCtxType        *pDevCtx,
    VpProfilePtrType    pDevProfile,
    VpProfilePtrType    pAcProfile,
    VpProfilePtrType    pDcProfile,
    VpProfilePtrType    pRingProfile,
    VpProfilePtrType    pFxoAcProfile,
    VpProfilePtrType    pFxoCfgProfile);

static VpStatusType KWrapInitLine(
    VpLineCtxType       *pLineCtx,
    VpProfilePtrType    pAcProfile,
    VpProfilePtrType    pDcProfile,
    VpProfilePtrType    pRingProfile);

static VpStatusType KWrapConfigLine(
    VpLineCtxType       *pLineCtx,
    VpProfilePtrType    pAcProfile,
    VpProfilePtrType    pDcProfile,
    VpProfilePtrType    pRingProfile);

static VpStatusType KWrapCalCodec(
    VpLineCtxType       *pLineCtx,
    VpDeviceCalType     mode);

static VpStatusType KWrapCalLn(
    VpLineCtxType       *pLineCtx);

static VpStatusType KWrapInitRing(
    VpLineCtxType       *pLineCtx,
    VpProfilePtrType    pCadProfile,
    VpProfilePtrType    pCidProfile);

static VpStatusType KWrapInitCid(
    VpLineCtxType       *pLineCtx,
    uint8               length,
    uint8p              pCidData);

static VpStatusType KWrapInitMeter(
    VpLineCtxType       *pLineCtx,
    VpProfilePtrType    pMeterProfile);

static VpStatusType KWrapInitProf(
    VpDevCtxType        *pDevCtx,
    VpProfileType       type,
    VpProfilePtrType    pProfileIndex,
    VpProfilePtrType    pProfile);

static VpStatusType KWrapSoftReset(
    VpDevCtxType        *pDevCtx);

/* VP-API Control Functions */
static VpStatusType KWrapSetLineState(
    VpLineCtxType       *pLineCtx,
    VpLineStateType     state);

static VpStatusType KWrapSetLineTone(
    VpLineCtxType       *pLineCtx,
    VpProfilePtrType    pToneProfile,
    VpProfilePtrType    pCadProfile,
    VpDtmfToneGenType   *pDtmfControl);

static VpStatusType KWrapSetRelayState(
    VpLineCtxType       *pLineCtx,
    VpRelayControlType  rState);

static VpStatusType KWrapSetRelGain(
    VpLineCtxType       *pLineCtx,
    uint16              txLevel,
    uint16              rxLevel,
    uint16              handle);

static VpStatusType KWrapSendSignal(
    VpLineCtxType       *pLineCtx,
    VpSendSignalType    signalType,
    void                *pSignalData);

static VpStatusType KWrapSendCid(
    VpLineCtxType       *pLineCtx,
    uint8               length,
    VpProfilePtrType    pCidProfile,
    uint8p              pCidData);

static VpStatusType KWrapContinueCid(
    VpLineCtxType       *pLineCtx,
    uint8               length,
    uint8p              pCidData);

static VpStatusType KWrapStartMeter(
    VpLineCtxType       *pLineCtx,
    uint16              onTime,
    uint16              offTime,
    uint16              numMeters);

static VpStatusType KWrapSetOption(
    VpLineCtxType       *pLineCtx,
    VpDevCtxType        *pDevCtx,
    VpOptionIdType      option,
    void                *pValue);

static VpStatusType KWrapDeviceIoAccess(
    VpDevCtxType        *pDevCtx,
    VpDeviceIoAccessDataType *pDeviceIoData);

static VpStatusType KWrapSelfTest(
    VpLineCtxType       *pLineCtx);

static VpStatusType KWrapFillTestBuf(
    VpLineCtxType       *pLineCtx,
    uint16              length,
    VpVectorPtrType     pData);

static VpStatusType KWrapLowLevelCmd(
    VpLineCtxType       *pLineCtx,
    uint8p              cmd,
    uint8               len,
    uint16              handle);

/* VP-API Status and Query Functions */
static VpStatusType KWrapGetLineStatus(
    VpLineCtxType       *pLineCtx,
    VpInputType         input,
    bool                *pStatus);

static VpStatusType KWrapGetDeviceStatus(
    VpDevCtxType        *pDevCtx,
    VpInputType         input,
    uint32              *pDeviceStatus);

static VpStatusType KWrapGetLoopCond(
    VpLineCtxType       *pLineCtx,
    uint16              handle);

static VpStatusType KWrapGetOption(
    VpLineCtxType       *pLineCtx,
    VpDevCtxType        *pDevCtx,
    VpOptionIdType      option,
    uint16              handle);

static VpStatusType KWrapGetLineState(
    VpLineCtxType       *pLineCtx,
    VpLineStateType     *pCurrentState);

static VpStatusType KWrapFlushEvents(
    VpDevCtxType        *pDevCtx);

static VpStatusType KWrapGetResults(
    VpEventType         *pEvent,
    void                *pResults);

static VpStatusType KWrapClearResults(
    VpDevCtxType        *pDevCtx);

static VpStatusType KWrapCodeCheckSum(
    VpDevCtxType        *pDevCtx,
    uint16              handle);

/* VP-API Line Tests Structs */
static VpStatusType KWrapTestLine(
    VpLineCtxType       *pLineCtx,
    VpTestIdType        test,
    const void          *pArgs,
    uint16              handle);

/*******************************************************************************
 * VpMakeVcpDeviceCtx()
 * This function
 *
 * Arguments:
 *
 *  pDevCtx         -   Device Context to be initialized by other input
 *  pDevObj         -   Device Object to be pointed to by device context
 *
 * Preconditions:
 * This function should be called after initializing the device object. This
 * function can be called more than once without modifying the contents of the
 * device object.
 *
 * Postconditions:
 * Initializes device context to be able to handle VCP functionality.
 ******************************************************************************/

VpStatusType
VpMakeKWrapDeviceObject(
    VpDevCtxType            *pDevCtx,
    VpKWrapDeviceObjectType *pDevObj)
{
    LegVpDvrMkDevObjType    mkDevObjInfo;
    char                    *nodeStr = VP_GET_DEV_NODE(pDevObj->deviceId);

    /* Open the device driver*/
    pDevObj->fileDes = open(nodeStr, O_RDWR);
    if (-1 == pDevObj->fileDes) {
        return VP_STATUS_FAILURE;
    }

    /* issue the corresponding ioctl */
    mkDevObjInfo.deviceId = pDevObj->deviceId;
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_MK_DEV_OBJ, &mkDevObjInfo)) {
        return VP_STATUS_FAILURE;
    }

    /* check status of VP-API command */
    if (VP_STATUS_SUCCESS != mkDevObjInfo.status) {
        return mkDevObjInfo.status;
    }

    pDevObj->maxChannels = VP_KERNEL_MAX_NUM_LINES;

    if (pDevCtx != VP_NULL) {
        return VpMakeKWrapDeviceCtx(pDevCtx, pDevObj);
    } else {
        return VP_STATUS_SUCCESS;
    }
} /* VpMakeKwrapDeviceObject() */

/*******************************************************************************
 * VpMakeVcpDeviceCtx()
 * This function initializes the device context to handle VCP functionality.
 *
 * Arguments:
 *
 *  pDevCtx         -   Device Context to be initialized by other input
 *  pDevObj         -   Device Object to be pointed to by device context
 *
 * Preconditions:
 * This function should be called after initializing the device object. This
 * function can be called more than once without modifying the contents of the
 * device object.
 *
 * Postconditions:
 * Initializes device context to be able to handle VCP functionality.
 ******************************************************************************/
VpStatusType
VpMakeKWrapDeviceCtx(
    VpDevCtxType *pDevCtx,
    VpKWrapDeviceObjectType *pDevObj)
{
    /* redundant checking of the object and context */
    if((pDevCtx == VP_NULL) || (pDevObj == VP_NULL)) {
        return VP_STATUS_INVALID_ARG;
    }
    /* Initialize the members of device context */
    pDevCtx->pDevObj    = pDevObj;
    pDevCtx->deviceType = VP_DEV_KWRAP;

    /* System Configuration function pointers */
    pDevCtx->funPtrsToApiFuncs.MakeLineObject   = KwapMakeLineObject;

    /* Initialization function pointers */
    pDevCtx->funPtrsToApiFuncs.BootLoad         = BootLoadImage;
    pDevCtx->funPtrsToApiFuncs.InitDevice       = KWrapInitDevice;
    pDevCtx->funPtrsToApiFuncs.InitLine         = KWrapInitLine;
    pDevCtx->funPtrsToApiFuncs.ConfigLine       = KWrapConfigLine;
    pDevCtx->funPtrsToApiFuncs.CalCodec         = KWrapCalCodec;
    pDevCtx->funPtrsToApiFuncs.CalLine          = KWrapCalLn;
    pDevCtx->funPtrsToApiFuncs.InitRing         = KWrapInitRing;
    pDevCtx->funPtrsToApiFuncs.InitCid          = KWrapInitCid;
    pDevCtx->funPtrsToApiFuncs.InitMeter        = KWrapInitMeter;
    pDevCtx->funPtrsToApiFuncs.InitProfile      = KWrapInitProf;
    pDevCtx->funPtrsToApiFuncs.SoftReset        = KWrapSoftReset;

    /* Control function pointers */
    pDevCtx->funPtrsToApiFuncs.SetLineState     = KWrapSetLineState;
    pDevCtx->funPtrsToApiFuncs.SetLineTone      = KWrapSetLineTone;
    pDevCtx->funPtrsToApiFuncs.SetRelayState    = KWrapSetRelayState;
    pDevCtx->funPtrsToApiFuncs.SetRelGain       = KWrapSetRelGain;
    pDevCtx->funPtrsToApiFuncs.SendSignal       = KWrapSendSignal;
    pDevCtx->funPtrsToApiFuncs.SendCid          = KWrapSendCid;
    pDevCtx->funPtrsToApiFuncs.ContinueCid      = KWrapContinueCid;
    pDevCtx->funPtrsToApiFuncs.StartMeter       = KWrapStartMeter;
    pDevCtx->funPtrsToApiFuncs.SetOption        = KWrapSetOption;
    pDevCtx->funPtrsToApiFuncs.DeviceIoAccess   = KWrapDeviceIoAccess;
    pDevCtx->funPtrsToApiFuncs.SelfTest         = KWrapSelfTest;
    pDevCtx->funPtrsToApiFuncs.FillTestBuf      = KWrapFillTestBuf;
    pDevCtx->funPtrsToApiFuncs.LowLevelCmd      = KWrapLowLevelCmd;

    /* Status and Query function pointers */
    pDevCtx->funPtrsToApiFuncs.GetLineStatus    = KWrapGetLineStatus;
    pDevCtx->funPtrsToApiFuncs.GetDeviceStatus  = KWrapGetDeviceStatus;
    pDevCtx->funPtrsToApiFuncs.GetLoopCond      = KWrapGetLoopCond;
    pDevCtx->funPtrsToApiFuncs.GetOption        = KWrapGetOption;
    pDevCtx->funPtrsToApiFuncs.GetLineState     = KWrapGetLineState;
    pDevCtx->funPtrsToApiFuncs.FlushEvents      = KWrapFlushEvents;
    pDevCtx->funPtrsToApiFuncs.GetResults       = KWrapGetResults;
    pDevCtx->funPtrsToApiFuncs.ClearResults     = KWrapClearResults;
    pDevCtx->funPtrsToApiFuncs.CodeCheckSum     = KWrapCodeCheckSum;

    pDevCtx->funPtrsToApiFuncs.TestLine         = KWrapTestLine;

    return VP_STATUS_SUCCESS;
} /* VpMakeKWrapDeviceCtx() */


/*******************************************************************************
 * VpMakeKwrapLineObject()
 * This function
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
VpStatusType
KWrapMakeLineObject(
    VpTermType              termType,
    uint8                   channelId,
    VpLineCtxType           *pLineCtx,
    VpKWrapLineObjectType   *pLineObj,
    VpDevCtxType            *pDevCtx)
{
    LegVpDvrMkLnObjType         mkLnObjInfo;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* Basic error checking */
    if (channelId >= pDevObj->maxChannels) {
        return VP_STATUS_INVALID_ARG;
    }

    /* Convert API Line termination type to VCP specific termination type */
    if (VP_NUM_TERM_TYPES < termType) {
        return  VP_STATUS_ERR_VTD_CODE;
    }

    /* load up the ioctl struct */
    mkLnObjInfo.channelId   = channelId;
    mkLnObjInfo.termType    = termType;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_MK_LN_OBJ, &mkLnObjInfo)) {
        return VP_STATUS_FAILURE;
    }

    /* check status of VP-API command */
    if (VP_STATUS_SUCCESS != mkLnObjInfo.status) {
        return mkLnObjInfo.status;
    }

    /* Initialize line context */
    pLineCtx->pLineObj  = pLineObj;
    pLineCtx->pDevCtx   = pDevCtx;

    /* Establish the link between device context to line context */
    pDevCtx->pLineCtx[channelId] = pLineCtx;

    /* Initialize line object */
    pLineObj->channelId = channelId;
    pLineObj->termType  = termType;

    /* Note that we are not initializing the 'lineId' member because we do not
     * know what it should be initialized (that information is not provided
     * through the VpMakeLineObj() function) and the basic type of this member
     * is defined by the customer and hence this variable could not have one
     * default value. */

    return mkLnObjInfo.status;
}  /* VpMakeKwrapLineObject() */



/******************************************************************************
 *                   VCP Specific Initialization Functions                    *
 ******************************************************************************/



/******************************************************************************
 * KWrapInitDevice()
 * This function calibrates the device and initializes all lines on the device
 * with the AC, DC, and Ringing parameters passed. See VP-API reference guide
 * for more information.
 *
 * Preconditions:
 * This function should be called only after creating and initializing the
 * device context.
 *
 * Postconditions:
 * Device is calibrated and all lines associated with this device are
 * initialized with the AC, DC, and Ringing Paramaters passed (DC and Ringing
 * apply to FXS type lines only).
 ******************************************************************************/
static VpStatusType
KWrapInitDevice(
    VpDevCtxType        *pDevCtx,       /**< Pointer to device context for the
                                        * device that will be initialized */
    VpProfilePtrType    pDevProfile,    /**< Pointer to Device Profile */
    VpProfilePtrType    pAcProfile,     /**< Pointer to AC Profile that is applied
                                        * to all FXS lines on this device */
    VpProfilePtrType    pDcProfile,     /**< Pointer to DC Profile that is applied
                                        * to all FXS lines on this device */
    VpProfilePtrType    pRingProfile,   /**< Pointer to Ringing Profile that is
                                        * applied to all FXS lines on this device*/
    VpProfilePtrType    pFxoAcProfile,  /**< Pointer to AC Profile that is applied
                                        * to all FXO lines on this device */
    VpProfilePtrType    pFxoCfgProfile) /**< Pointer to Config Profile that is
                                        * applied to all FXO lines on this device*/
{
    LegVpDvrInitDevType         initDevInfo;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    initDevInfo.pDevProfile     = pDevProfile;
    initDevInfo.pAcProfile      = pAcProfile;
    initDevInfo.pDcProfile      = pDcProfile;
    initDevInfo.pRIngProfile    = pRingProfile;
    initDevInfo.pFxoAcProfile   = pFxoAcProfile;
    initDevInfo.pFxoCfgProfile  = pFxoCfgProfile;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_INIT_DEV, &initDevInfo)) {
        return VP_STATUS_FAILURE;
    }

    return initDevInfo.status;
}  /* KWrapInitDevice() */

/*******************************************************************************
 * KWrapInitLine()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapInitLine(
    VpLineCtxType       *pLineCtx,
    VpProfilePtrType    pAcProfile,        /* AC Profile Selector     */
    VpProfilePtrType    pDcProfile,        /* DC Profile Selector     */
    VpProfilePtrType    pRingProfile)      /* Ring Profile Selector   */
{
    LegVpDvrInitLnType          initLnInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    initLnInfo.channelId        = pLineObj->channelId;
    initLnInfo.pAcProfile       = pAcProfile;
    initLnInfo.pDcOrFxoProfile  = pDcProfile;
    initLnInfo.pRIngProfile     = pRingProfile;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_INIT_LN, &initLnInfo)) {
        return VP_STATUS_FAILURE;
    }

    return initLnInfo.status;

} /* KWrapInitLine() */


/*******************************************************************************
 * KWrapConfigLine()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapConfigLine(
    VpLineCtxType       *pLineCtx,
    VpProfilePtrType    pAcProfile,        /* AC Profile Selector     */
    VpProfilePtrType    pDcProfile,        /* DC Profile Selector     */
    VpProfilePtrType    pRingProfile)      /* Ring Profile Selector   */
{
    LegVpDvrCfgLnType           cfgLnInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    cfgLnInfo.channelId        = pLineObj->channelId;
    cfgLnInfo.pAcProfile       = pAcProfile;
    cfgLnInfo.pDcOrFxoProfile  = pDcProfile;
    cfgLnInfo.pRIngProfile     = pRingProfile;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_CFG_LN, &cfgLnInfo)) {
        return VP_STATUS_FAILURE;
    }

    return cfgLnInfo.status;

} /* KWrapInitLine() */

/*******************************************************************************
 * KWrapCalCodec()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapCalCodec(
    VpLineCtxType       *pLineCtx,    /* Selects the target line number */
    VpDeviceCalType     mode)       /* Specifies the method for calibrating a SLAC */
{
    LegVpDvrCalCodecType        calCodecInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    calCodecInfo.channelId      = pLineObj->channelId;
    calCodecInfo.mode           = mode;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_CAL_CODEC, &calCodecInfo)) {
        return VP_STATUS_FAILURE;
    }

    return calCodecInfo.status;

} /* KWrapCalCodec() */

/*******************************************************************************
 * KWrapCalLn()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapCalLn(VpLineCtxType    *pLineCtx)
{

    LegVpDvrCalLnType           calLnInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    calLnInfo.channelId         = pLineObj->channelId;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_CAL_LN, &calLnInfo)) {
        return VP_STATUS_FAILURE;
    }

    return calLnInfo.status;
} /* KWrapCalLn() */

/*******************************************************************************
 * KWrapInitRing()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapInitRing(
    VpLineCtxType       *pLineCtx,         /* Pointer to line context  */
    VpProfilePtrType    pCadProfile,       /* Cadence profile selector */
    VpProfilePtrType    pCidProfile)       /* Caller ID profile selector */
{
    LegVpDvrInitRingType        initRingInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    initRingInfo.channelId      = pLineObj->channelId;
    initRingInfo.pCadProfile    = pCadProfile;
    initRingInfo.pCidProfile    = pCidProfile;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_INIT_RING, &initRingInfo)) {
        return VP_STATUS_FAILURE;
    }

    return initRingInfo.status;
} /* KWrapInitRing() */

/*******************************************************************************
 * KWrapInitCid()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapInitCid(
    VpLineCtxType       *pLineCtx,
    uint8               length,
    uint8p              pCidData)
{
    LegVpDvrInitCidType         initCidInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    initCidInfo.channelId       = pLineObj->channelId;
    initCidInfo.length          = length;
    initCidInfo.pCidData        = pCidData;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_INIT_CID, &initCidInfo)) {
        return VP_STATUS_FAILURE;
    }

    return initCidInfo.status;
} /* KWrapInitCid() */


/*******************************************************************************
 * KWrapInitMeter()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapInitMeter(
    VpLineCtxType       *pLineCtx,
    VpProfilePtrType    pMeterProfile)
{
    LegVpDvrInitMtrType         initMtrInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    initMtrInfo.channelId       = pLineObj->channelId;
    initMtrInfo.pMeterProfile   = pMeterProfile;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_INIT_MTR, &initMtrInfo)) {
        return VP_STATUS_FAILURE;
    }

    return initMtrInfo.status;
} /* KWrapInitMeter() */

/*******************************************************************************
 * KWrapInitProf()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapInitProf(
    VpDevCtxType        *pDevCtx,
    VpProfileType       type,
    VpProfilePtrType    pProfileIndex,
    VpProfilePtrType    pProfile)
{
    LegVpDvrInitProfType        initProfInfo;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    initProfInfo.type           = type;
    initProfInfo.pProfileIndex  = pProfileIndex;
    initProfInfo.pProfile       = pProfile;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_INIT_PROF, &initProfInfo)) {
        return VP_STATUS_FAILURE;
    }

    return initProfInfo.status;
} /* KWrapInitProf() */


/*******************************************************************************
 * KWrapInitRing()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapSoftReset(
    VpDevCtxType        *pDevCtx)
{
    LegVpDvrSftRstType          softResetInfo;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_SFT_RST, &softResetInfo)) {
        return VP_STATUS_FAILURE;
    }

    return softResetInfo.status;
} /* KWrapSoftReset() */

/*******************************************************************************
 * KWrapSetLineState()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapSetLineState(
    VpLineCtxType       *pLineCtx,
    VpLineStateType     state)
{
    LegVpDvrSetLnStType         setLnStateInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    setLnStateInfo.channelId    = pLineObj->channelId;
    setLnStateInfo.state        = state;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_SET_LN_ST, &setLnStateInfo)) {
        return VP_STATUS_FAILURE;
    }

    return setLnStateInfo.status;
} /* KWrapSetLineState() */

/*******************************************************************************
 * KWrapSetLineTone()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapSetLineTone(
    VpLineCtxType       *pLineCtx,
    VpProfilePtrType    pToneProfile,
    VpProfilePtrType    pCadProfile,
    VpDtmfToneGenType   *pDtmfControl)
{
    LegVpDvrSetLnTnType         setLnToneInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    setLnToneInfo.channelId     = pLineObj->channelId;
    setLnToneInfo.pToneProfile  = pToneProfile;
    setLnToneInfo.pCadProfile   = pCadProfile;
    setLnToneInfo.pDtmfControl  = pDtmfControl;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_SET_LN_TN, &setLnToneInfo)) {
        return VP_STATUS_FAILURE;
    }

    return setLnToneInfo.status;
} /* KWrapSetLineTone() */

/*******************************************************************************
 * KWrapSetRelayState()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapSetRelayState(
    VpLineCtxType       *pLineCtx,
    VpRelayControlType  rState)
{
    LegVpDvrSetRlyStType        setRlyStInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    setRlyStInfo.channelId      = pLineObj->channelId;
    setRlyStInfo.rState         = rState;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_SET_RLY_ST, &setRlyStInfo)) {
        return VP_STATUS_FAILURE;
    }
    return setRlyStInfo.status;
} /* KWrapSetRelayState() */

/*******************************************************************************
 * KWrapSetRelGain()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapSetRelGain(
    VpLineCtxType       *pLineCtx,
    uint16              txLevel,
    uint16              rxLevel,
    uint16              handle)
{
    LegVpDvrSetRelGnType        setRelGnInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    setRelGnInfo.channelId      = pLineObj->channelId;
    setRelGnInfo.txLevel        = txLevel;
    setRelGnInfo.rxLevel        = rxLevel;
    setRelGnInfo.handle         = handle;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_SET_REL_GN, &setRelGnInfo)) {
        return VP_STATUS_FAILURE;
    }
    return setRelGnInfo.status;
} /* KWrapSetRelGain() */

/*******************************************************************************
 * KWrapSendSignal()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapSendSignal(
    VpLineCtxType       *pLineCtx,
    VpSendSignalType    signalType,
    void                *pSignalData)
{
    LegVpDvrSendSigType         sendSigInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    sendSigInfo.channelId       = pLineObj->channelId;
    sendSigInfo.signalType      = signalType;
    sendSigInfo.pSignalData     = pSignalData;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_SEND_SIG, &sendSigInfo)) {
        return VP_STATUS_FAILURE;
    }
    return sendSigInfo.status;
} /* KWrapSendSignal() */

/*******************************************************************************
 * KWrapSendCid()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapSendCid(
    VpLineCtxType       *pLineCtx,
    uint8               length,
    VpProfilePtrType    pCidProfile,
    uint8p              pCidData)
{
    LegVpDvrSendCidType         sendCidInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    sendCidInfo.channelId       = pLineObj->channelId;
    sendCidInfo.length          = length;
    sendCidInfo.pCidProfile     = pCidProfile;
    sendCidInfo.pCidData        = pCidData;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_SEND_CID, &sendCidInfo)) {
        return VP_STATUS_FAILURE;
    }
    return sendCidInfo.status;
} /* KWrapSendCid() */

/*******************************************************************************
 * KWrapContinueCid()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapContinueCid(
    VpLineCtxType       *pLineCtx,
    uint8               length,
    uint8p              pCidData)
{
    LegVpDvrCntCidType          cntCidInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    cntCidInfo.channelId       = pLineObj->channelId;
    cntCidInfo.length          = length;
    cntCidInfo.pCidData        = pCidData;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_CNT_CID, &cntCidInfo)) {
        return VP_STATUS_FAILURE;
    }
    return cntCidInfo.status;
} /* KWrapContinueCid() */

/*******************************************************************************
 * KWrapStartMeter()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapStartMeter(
    VpLineCtxType       *pLineCtx,
    uint16              onTime,
    uint16              offTime,
    uint16              numMeters)
{
    LegVpDvrStartMtrType        startMtrInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    startMtrInfo.channelId       = pLineObj->channelId;
    startMtrInfo.onTime          = onTime;
    startMtrInfo.offTime         = offTime;
    startMtrInfo.numMeters       = numMeters;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_START_MTR, &startMtrInfo)) {
        return VP_STATUS_FAILURE;
    }
    return startMtrInfo.status;
} /* KWrapStartMeter() */

/*******************************************************************************
 * KWrapSetOption()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapSetOption(
    VpLineCtxType       *pLineCtx,
    VpDevCtxType        *pDevCtx,
    VpOptionIdType      option,
    void                *pValue)
{
    LegVpDvrSetOptionType       setOptionInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    setOptionInfo.channelId       = pLineObj->channelId;
    setOptionInfo.option          = option;
    setOptionInfo.pValue          = pValue;

    /* figure out if this is a device or line setting */
    if (pLineCtx != VP_NULL) {
        setOptionInfo.lnRequest   = TRUE;
    } else {
        setOptionInfo.lnRequest   = FALSE;
    }

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_SET_OPTION, &setOptionInfo)) {
        return VP_STATUS_FAILURE;
    }
    return setOptionInfo.status;
} /* KWrapSetOption() */

/*******************************************************************************
 * KWrapDeviceIoAccess()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapDeviceIoAccess(
    VpDevCtxType        *pDevCtx,
    VpDeviceIoAccessDataType *pDeviceIoData)
{
    LegVpDvrDevIoAccType        devIoAccInfo;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    devIoAccInfo.pDeviceIoData  = pDeviceIoData;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_DEV_IO_ACC, &devIoAccInfo)) {
        return VP_STATUS_FAILURE;
    }
    return devIoAccInfo.status;
} /* KWrapDeviceIoAccess() */

/*******************************************************************************
 * KWrapFillTestBuf()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapSelfTest(
    VpLineCtxType       *pLineCtx)
{
    LegVpDvrSelfTestType        selfTestAccInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    selfTestAccInfo.channelId   = pLineObj->channelId;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_DEV_IO_ACC, &selfTestAccInfo)) {
        return VP_STATUS_FAILURE;
    }
    return selfTestAccInfo.status;
} /* KWrapSelfTest() */


/*******************************************************************************
 * KWrapFillTestBuf()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapFillTestBuf(
    VpLineCtxType       *pLineCtx,
    uint16              length,
    VpVectorPtrType     pData)
{
    LegVpDvrFillBuffType        fillTestBuffInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    fillTestBuffInfo.channelId  = pLineObj->channelId;
    fillTestBuffInfo.length     = length;
    fillTestBuffInfo.pData      = pData;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_FILL_BUFF, &fillTestBuffInfo)) {
        return VP_STATUS_FAILURE;
    }
    return fillTestBuffInfo.status;
} /* KWrapFillTestBuf() */

/*******************************************************************************
 * KWrapLowLevelCmd()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapLowLevelCmd(
    VpLineCtxType       *pLineCtx,
    uint8               *pCmdData,
    uint8               len,
    uint16              handle)
{
    LegVpDvrLowLvlCmdType       lowLevCmdInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    lowLevCmdInfo.channelId     = pLineObj->channelId;
    lowLevCmdInfo.pCmdData      = pCmdData;
    lowLevCmdInfo.len           = len;
    lowLevCmdInfo.handle        = handle;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_LOW_LVL_CMD, &lowLevCmdInfo)) {
        return VP_STATUS_FAILURE;
    }
    return lowLevCmdInfo.status;
} /* KWrapLowLevelCmd() */

/*******************************************************************************
 * KWrapGetLineStatus()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapGetLineStatus(
    VpLineCtxType       *pLineCtx,
    VpInputType         input,
    bool                *pStatus)
{
    LegVpDvrGetLnStatusType     getLnStatusInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    getLnStatusInfo.channelId   = pLineObj->channelId;
    getLnStatusInfo.input       = input;
    getLnStatusInfo.pStatus     = pStatus;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_GET_LN_STATUS, &getLnStatusInfo)) {
        return VP_STATUS_FAILURE;
    }

    return getLnStatusInfo.status;
} /* KWrapGetLineStatus() */

/*******************************************************************************
 * KWrapGetDeviceStatus()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapGetDeviceStatus(
    VpDevCtxType        *pDevCtx,
    VpInputType         input,
    uint32              *pDeviceStatus)
{
    LegVpDvrGetDevStatusType    getDevStatusInfo;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    getDevStatusInfo.input          = input;
    getDevStatusInfo.pDeviceStatus  = pDeviceStatus;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_GET_DEV_STATUS, &getDevStatusInfo)) {
        return VP_STATUS_FAILURE;
    }

    return getDevStatusInfo.status;
} /* KWrapGetDeviceStatus() */


/*******************************************************************************
 * KWrapGetLoopCond()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapGetLoopCond(
    VpLineCtxType       *pLineCtx,
    uint16              handle)
{
    LegVpDvrGetLpCondType       getLpCondInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    getLpCondInfo.channelId     = pLineObj->channelId;
    getLpCondInfo.handle        = handle;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_GET_LP_COND, &getLpCondInfo)) {
        return VP_STATUS_FAILURE;
    }

    return getLpCondInfo.status;
} /* KWrapGetLoopCond() */


/*******************************************************************************
 * KWrapGetOption()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapGetOption(
    VpLineCtxType       *pLineCtx,
    VpDevCtxType        *pDevCtx,
    VpOptionIdType      option,
    uint16              handle)
{
    LegVpDvrGetOptionType       getOptionInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    getOptionInfo.channelId     = pLineObj->channelId;
    getOptionInfo.option        = option;
    getOptionInfo.handle        = handle;

    /* figure out if this is a device or line setting */
    if (pLineCtx != VP_NULL) {
        getOptionInfo.lnRequest   = TRUE;
    } else {
        getOptionInfo.lnRequest   = FALSE;
    }

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_GET_OPTION, &getOptionInfo)) {
        return VP_STATUS_FAILURE;
    }

    return getOptionInfo.status;
} /* KWrapGetOption() */


/*******************************************************************************
 * KWrapGetLineState()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapGetLineState(
    VpLineCtxType       *pLineCtx,
    VpLineStateType     *pCurrentState)
{
    LegVpDvrGetLnStType         getLnStInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    getLnStInfo.channelId       = pLineObj->channelId;
    getLnStInfo.pCurrentState   = pCurrentState;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_GET_LN_ST, &getLnStInfo)) {
        return VP_STATUS_FAILURE;
    }

    return getLnStInfo.status;
} /* KWrapGetLineState() */


/*******************************************************************************
 * KWrapFlushEvents()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapFlushEvents(
    VpDevCtxType        *pDevCtx)
{
    LegVpDvrFlshEvntType        flshEvntInfo;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_FLSH_EVNT, &flshEvntInfo)) {
        return VP_STATUS_FAILURE;
    }

    return flshEvntInfo.status;
} /* KWrapFlushEvents() */


/*******************************************************************************
 * KWrapGetResults()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapGetResults(
    VpEventType         *pEvent,
    void                *pResults)
{
    LegVpDvrGetRsltType         getRsltInfo;
    VpDevCtxType                *pDevCtx = pEvent->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    getRsltInfo.pEvent          = pEvent;
    getRsltInfo.pResults        = pResults;
    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_GET_RSLT, &getRsltInfo)) {
        return VP_STATUS_FAILURE;
    }

    return getRsltInfo.status;
} /* KWrapGetResults() */


/*******************************************************************************
 * KWrapClearResults()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapClearResults(
    VpDevCtxType        *pDevCtx)
{
    LegVpDvrClrRsltType         clrRsltInfo;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_CLR_RSLT, &clrRsltInfo)) {
        return VP_STATUS_FAILURE;
    }

    return clrRsltInfo.status;
} /* KWrapClearResults() */


/*******************************************************************************
 * KWrapCodeCheckSum()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapCodeCheckSum(
    VpDevCtxType        *pDevCtx,
    uint16              handle)
{
    LegVpDvrCodeCkSumType       codeCkSumInfo;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    codeCkSumInfo.handle        = handle;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_CODE_CK_SUM, &codeCkSumInfo)) {
        return VP_STATUS_FAILURE;
    }

    return codeCkSumInfo.status;
} /* KWrapCodeCheckSum() */


/*******************************************************************************
 * KWrapCodeCheckSum()
 * This function ...
 *
 * Arguments:
 *
 * Preconditions:
 *
 * Postconditions:
 ******************************************************************************/
static VpStatusType
KWrapTestLine(
    VpLineCtxType       *pLineCtx,
    VpTestIdType        test,
    const void          *pArgs,
    uint16              handle)
{
    LegVpDvrTestLnType          testLnInfo;
    VpKWrapLineObjectType       *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType                *pDevCtx = pLineCtx->pDevCtx;
    VpKWrapDeviceObjectType     *pDevObj = pDevCtx->pDevObj;

    /* load up the ioctl struct */
    testLnInfo.channelId        = pLineObj->channelId;
    testLnInfo.test             = test;
    testLnInfo.pArgs            = pArgs;
    testLnInfo.handle           = handle;

    /* issue the corresponding ioctl */
    if (0 < ioctl(pDevObj->fileDes, LEG_VP_DVR_IOCTL_TEST_LN, &testLnInfo)) {
        return VP_STATUS_FAILURE;
    }

    return testLnInfo.status;
} /* KWrapTestLine() */


#endif /* VP_DEV_KWRAP */
