/** \file quick_start_example.c
 * quick_start_example.c
 *
 * This file provides a simple quick start application built around
 * VP-API.
 *
 * Copyright (c) 2008, Zarlink Semiconductor, Inc.
 */

/* INCLUDES */
#include "app_des.h"    /* Provides Description of hardware and VCP-firmware
                         * product type*/
#ifdef USING_QUICK_START_APP  /* Conditionally compiled file */
#include "vp_hal.h"
#include "vp_api.h"
#include <stdlib.h>

#include <stdio.h>
#ifndef term_printf
/* Define the following as necessary to redirect the debug output */
#define term_printf printf
#endif /* term_printf */

/*******************************************************************************
 * Conditional Compile time options; Change these based on your software       *
 * architecture and other needs.                                               *
 ******************************************************************************/
#undef EXIT_ON_ERROR       /* Define this to exit on error */
#define TEST_ON_LEG_BOARD   /* This is to Test on Zarlink Semiconductor development board */

#define USE_DUMMY_FRAMER    /* Makes use of dummy framer implemenation. Once
                             * actual framer driver is developed, remove this
                             * conditional compile and use correct files in the
                             * project.     */
#ifdef USE_DUMMY_FRAMER
#include "framer_dummy.h" /* Include this file for bringing in Framer
                           * functionality. Presently the inlcuded file is a
                           * dummy implementation. */
#endif /* USE_DUMMY_FRAMER */

#define BOOTLOADING_VCP /* Define this if you would like QuickStartAppMain() to
                         * perform bootloading of the VCP. */

#define POLL_VCP_EVENTS /* Define this if you would like QuickStartAppMain() to
                         * poll for VCP events and handle them. If VCP events
                         * are handled through a interrupt service routine (in
                         * an interrupt context) you do not need
                         * to define this. */

/*******************************************************************************
 * VCP-Ve790 Product specific quick start demonstration features               *
 ******************************************************************************/
#ifdef VP_CC_VCP_790_SERIES
#undef TEST_LOW_LEVEL_CMD          /* Define this to test VpLowLevelCmd()
                                     * write/read operation. */
#define DEMONSTRATE_TEST            /* Define this to demonstrate Test
                                     * functions */
#endif /* VP_CC_VCP_790_SERIES */

/*******************************************************************************
 * VCP-Ve880 Product specific quick start demonstration features               *
 ******************************************************************************/
#ifdef VP_CC_VCP_880_SERIES
#undef DEMONSTRATE_LINE_INIT        /* Define this to demonstrate Line
                                     * initialization function. */
#endif /* VP_CC_VCP_880_SERIES */

/*******************************************************************************
 * Common VCP quick start demonstration features                               *
 ******************************************************************************/
#define DEMONSTRATE_DTMF_DECODE     /* Demonstrate DTMF decoding */
#define DEMONSTRATE_METERING        /* Demonstrate metering */
#define HANDLE_DTMF_CMDS            /* Define this to enable processing
                                     * DTMF commands */
#define DEMONSTRATE_GET_LINE_STATE  /* Demonstrate reading line state */
#define DEMONSTRATE_CID             /* Define this to demonstrate CID
                                     * functions */
#define DEMONSTRATE_GET_LOOP_COND   /* Demonstrate get loop condition command */

#define DEMONSTRATE_HOWLER_TONE     /* Demonstrate howler tone generation */

#if defined(DEMONSTRATE_GET_LOOP_COND) && defined(TEST_LOW_LEVEL_CMD)
#error "Please enable only DEMONSTRATE_GET_LOOP_COND or TEST_LOW_LEVEL_CMD"
#endif
/*******************************************************************************
 * Feature demonstration specific options                                      *
 ******************************************************************************/
#ifdef DEMONSTRATE_CID
#undef CID_DEMO_USE_SENDCID             /* Define to use VpSendCid()
                                         * (vs. VpInitCid) */
#define RING_CAD_STD RING_CAD_US_CID    /* Macro to select the country selection
                                         * for the CID */
#endif /* DEMONSTRATE_CID */

#ifdef DEMONSTRATE_TEST
static VpStatusType LineTester(bool startFromBeg);
static VpStatusType TestEventHandler(VpEventType *pEvent);

uint8 testLineId = 0;       /* Line for which test needs to be performed */
bool testOnGoing = FALSE;   /* indicates whether there is a on-going test */
#endif /* DEMONSTRATE_TEST */

#ifdef DEMONSTRATE_METERING
bool meteringGoingOn[MAX_LINES_ON_BOARD];
#endif /* DEMONSTRATE_METERING */

/* Create an instance of VCP device context and Device Object
 * Note that memory for these instances can be obtained from memory management
 * functions (like malloc). */
VpVcpDeviceObjectType devObj[MAX_DEVICES_ON_BOARD];
VpDevCtxType devCtx[MAX_DEVICES_ON_BOARD];

/* Create instance of VCP line context and line Object
 * Note that memory for these instances can be obtained from memory management
 * functions (like malloc). */
VpLineCtxType lineCtx[MAX_LINES_ON_BOARD];
VpVcpLineObjectType lineObj[MAX_LINES_ON_BOARD];

/* It is assumed here that boot load image is stored in the global memory.
 * Notes about storing boot image:
 * Every system is different in its mechanisms in storing the data in FLASH
 * Some systems might just make use FLASH as just an extension of their overall
 * memory (either directly or through mapping appropriate page). In such systems
 * obtaining boot image is nothing but obtaining right address (and page) where
 * image is stored.
 * Other systems might make use of FLASH file system. In such systems obtaining
 * boot image is like reading a file. In this case the boot image file would
 * have to be read from the FLASH. Flash file system would provide appprpriate
 * calls to perform such read accesses.
 *
 * Other notes about length of boot image that can be passed to VpBootLoad:
 * VpBootImage function can take the entire length of the boot image or it can
 * be called many times with smaller block sizes. This is described in the
 * VP-API specification.
 */
extern VpImagePtrType bootImage; /* pointer to the VCP firmware image in RAM */
extern unsigned long dspBootStreamLen;  /* size of the VCP firmware image */

/* If using a compressed VCP boot image, then VpBootLoad() requires some
 * temporary scratch memory to do the decompression.  Here we allocate it
 * statically, for simplicity. */
static VpScratchMemType scratchMem;

/* Include the appropriate profiles */
#ifdef VP_CC_VCP_790_SERIES
#include "profiles8_790.h"
#else
#include "profiles8_880.h"
#endif /* VP_CC_VCP_790_SERIES */

/*******************************************************************************
 *                      Low level command debug interface                      *
 ******************************************************************************/
#define LOW_LEVEL_CMD_MAX_LEN   30  /* Maximum length of command to be
                                     * excercised */

/* Example low level commands */
#define LOOP_SUPER_VISION_RD    0xc3
#define LOOP_SUPER_VISION_LEN   8

#define DC_FEED_RD              0xc7
#define DC_FEED_LEN             14

#define READ_VERSION            0x73
#define READ_VERSION_LEN        1


static int presentCommandLength;  /* Indicates the length of the command */

/*******************************************************************************
 *                      Quick Start Application related variables              *
 ******************************************************************************/
typedef enum {
    ONHOOK_NO_RING = 0,
    ONHOOK_RING,
    APP_LINEST_ENUM_SIZE = FORCE_STANDARD_C_ENUM_SIZE /* Portability Req. */
} AppLineStateType;

AppLineStateType lineState[MAX_LINES_ON_BOARD];     /* Application current
                                                     * line state */
uint32 timeStampCounter = 0;
/*******************************************************************************
 *                      Function Declarations                                  *
 ******************************************************************************/
static VpStatusType InitAfterBoot(void);
static void PrintEvent(VpEventType *pEvent);
static VpStatusType PrintGetOptionResults(VpEventType *pEvent);
static VpStatusType PrintGetLoopCondResults(VpEventType *pEvent);
static VpStatusType ProcessDtmfDigits(uint8 channelId, uint8 digit);
static void PrintHelp(void);
static void ResetCommandState(void);
#ifdef DEMONSTRATE_CID
#include "cli_utilities.h"
static uint8* GetCallerDataPointer(bool startFromBegining, uint8 *messageSize);
#endif /* DEMONSTRATE_CID */

/* strings corresponding to VpStatusType */
char *retstring[] = {
    "VP_STATUS_SUCCESS",
    "VP_STATUS_FAILURE",
    "VP_STATUS_FUNC_NOT_SUPPORTED",
    "VP_STATUS_INVALID_ARG",
    "VP_STATUS_MAILBOX_BUSY",
    "VP_STATUS_ERR_VTD_CODE",
    "VP_STATUS_OPTION_NOT_SUPPORTED",
    "VP_STATUS_ERR_VERIFY",
    "VP_STATUS_DEVICE_BUSY",
    "VP_STATUS_MAILBOX_EMPTY",
    "VP_STATUS_ERR_MAILBOX_DATA",
    "VP_STATUS_ERR_HBI",
    "VP_STATUS_ERR_IMAGE",
    "VP_STATUS_IN_CRTCL_SECTN",
    "VP_STATUS_DEV_NOT_INITIALIZED",
    "VP_STATUS_ERR_PROFILE",
    "VP_STATUS_INVALID_VOICE_STREAM"};

/* strings corresponding to VpLineStateType */
char *statestring[] = {
    "STANDBY",
    "TIP_OPEN",
    "ACTIVE",
    "ACTIVE_POLREV",
    "TALK",
    "TALK_POLREV",
    "OHT",
    "OHT_POLREV",
    "DISCONNECT",
    "RINGING",
    "RINGING_POLREV",
    "FXO_OHT",
    "FXO_LOOP_OPEN",
    "FXO_LOOP_CLOSE",
    "FXO_TALK",
    "FXO_RING_GND"
};

/*
 * QuickStartAppMain()
 * This function is the main function for VP-API quick start application.
 * This function could be called from the "main()" after the necessary board
 * specific initialization (based on the board). If there is no such board set
 * up requied, then this function could be renamed as "main" itself.
 *
 * This function implements a simple demo application using the VP-API. It
 * implements the following high level state machine.
 *
 * Hook State:
 * State 1, Reset on hook:      OHT state
 * State 2, First off hook:     Dial tone
 * State 3, First on hook:      Ringing
 * State 4, Second off hook:    Morse code message "CQ...."
 * State 5, Second on hook:     Go back state 1
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * See introduction.
 */
VpStatusType QuickStartAppMain(void)
{
    VpDeviceIdType deviceId = USE_CHIP_SELECT_ID; /* Chip select ID of the VCP */
    VpStatusType status;
    VpEventType event;
    bool eventPresent = FALSE;
    int i;

    /* term_printf("\n DeviceOject = %d, Line Object = %d\n",
                sizeof(VpDevCtxType) + sizeof(VpVcpDeviceObjectType),
                sizeof(VpLineCtxType) + sizeof(VpVcpLineObjectType)); */

    term_printf("\n\nWelcome to the Zarlink Semiconductor VP-API quick start demo"
                " application.\n\n");

    /* Initialize application variables */
    for (i = 0; i < MAX_LINES_ON_BOARD; i++)  {
      lineState[i] = ONHOOK_NO_RING;
#ifdef DEMONSTRATE_METERING
      meteringGoingOn[i] = FALSE;
#endif /* DEMONSTRATE_METERING */
    }

    /* Reset the DTMF command decoder state machine */
    ResetCommandState();

#ifdef USE_DUMMY_FRAMER
    /* Initialize and configure the framer; This step may be required if clocks
     * provided to the SLAC are derived from the framer device. */
    FliuInit();
#endif /* USE_DUMMY_FRAMER */

#ifdef BOOTLOADING_VCP
    /* Create and initialize VCP device context */
    term_printf("Creating Device Object...");
    status = VpMakeDeviceObject(VP_DEV_VCP_SERIES, deviceId, &devCtx[0],
                                &devObj[0]);
    if(status != VP_STATUS_SUCCESS){ return status; /* Something is wrong */ }
    term_printf("done\n");

    /* Bootload the VCP */
    term_printf("Downloading VCP boot image...");
    status = VpBootLoad(&devCtx[0], VP_BOOT_STATE_FIRSTLAST, bootImage,
                        dspBootStreamLen, &scratchMem, VP_BOOT_MODE_VERIFY);
    if(status != VP_STATUS_SUCCESS){ return status; /* Something is wrong */ }
    term_printf("done\n");

#else /* BOOTLOADING_VCP */
    /* Bootloading performed else where so go ahead and create the line objects
     * and initialize the device */
    status = InitAfterBoot();
    if(status != VP_STATUS_SUCCESS){ return status; /* Something is wrong */ }
#endif /* BOOTLOADING_VCP */


#ifdef POLL_VCP_EVENTS
    /* An eternal loop for Quick Start Application */
    while(1){
        eventPresent = VpGetEvent(&devCtx[0], &event);
        if(event.status != VP_STATUS_SUCCESS) {
            if(event.status < VP_STATUS_NUM_TYPES) {
                term_printf("ERROR: VpGetEvent = %s.\n\n",
                                retstring[event.status]);
            } else {
                term_printf("ERROR: VpGetEvent = %d.\n\n", event.status);
            }
        } else if(eventPresent == TRUE) {
            status = VtdEventHandler(&event);
            if(status != VP_STATUS_SUCCESS) {
#ifdef EXIT_ON_ERROR
                return status; /* Something is wrong */
#else
                if(event.status < VP_STATUS_NUM_TYPES) {
                    term_printf("ERROR: VtdEventHandler = %s.\n\n",
                                    retstring[status]);
                } else {
                    term_printf("ERROR: VtdEventHandler = %d.\n\n", status);
                }
#endif
            }
        }
    }
#else
    return VP_STATUS_SUCCESS;
#endif /* POLL_VCP_EVENTS */
} /* QuickStartAppMain() */


/*
 * InitAfterBoot()
 * This function creats line contexts and also initiates device initialization
 * functions.
 *
 * Preconditions:
 * Assumes the device is boot loaded and boot complete event has been received.
 *
 * Postconditions:
 * Calls VpDeviceInit().
 */
static VpStatusType InitAfterBoot(void)
{
    VpStatusType status;
    int i;
    VpTermType termType;

    /* Create all the line contexts */
    for (i = 0; i < MAX_LINES_ON_BOARD; i++)  {
        term_printf("creating Line Object %d, TermType = ", i);
#ifdef VP_CC_VCP_790_SERIES
        termType = VP_TERM_FXS_TO_TL;
        term_printf("VP_TERM_FXS_TO_TL...");
#else
        termType = VP_TERM_FXS_GENERIC;
        term_printf("VP_TERM_FXS_GENERIC...");
#endif /* VP_CC_VCP_790_SERIES */

        /* Ignore the termination type for now. This functionality will
         * be supported later */
        status = VpMakeLineObject(termType, i, &lineCtx[i], &lineObj[i],
                                    &devCtx[0]);
        if(status != VP_STATUS_SUCCESS){
            return status; /* Something is wrong */
        }
        term_printf("done\n");
    }

#ifdef DEMONSTRATE_CID
    term_printf("Initializing CID Profile table\n");
    status = VpInitProfile(&devCtx[0], VP_PROFILE_CID, VP_PTABLE_INDEX1,
                            &CLI_TYPE1_US[0]);

    if(status != VP_STATUS_SUCCESS){ return status; /* Something is wrong */ }
#endif /* DEMONSTRATE_CID */

    /* Initialize the device */
    term_printf("Initializing VCP...");
    status = VpInitDevice(&devCtx[0], &DEV_PROFILE[0], &AC_600R_FXS[0],
                            &DC_25MA_CC[0], &RING_20HZ_SINE[0], VP_PTABLE_NULL,
                            VP_PTABLE_NULL);
    if(status != VP_STATUS_SUCCESS){ return status; /* Something is wrong */ }
    term_printf("done\n");
    return status;
}

/*
 * VtdEventHandler()
 * This function handles the events from the VCP and takes necessary action
 * based on those events.
 *
 * Preconditions:
 * This function could be called from either the interrupt context or it could
 * be called from a routine that is polling for events. In either case this
 * function SHOULD be passed a pointer to an event that was filled by the
 * VpGetEvent().
 *
 * Postconditions:
 * This line state/device states are modified by this function.
 * This function returns VP_STATUS_SUCCESS if everything went right
 */
static bool applyCadence = FALSE;
VpStatusType VtdEventHandler(VpEventType *pEvent)
{
    VpStatusType status = VP_STATUS_FAILURE;
    VpChkSumType chkSumInfo;
    int i;
    uint8 chanId = pEvent->channelId;
    VpOptionTimeslotType pcmTimeSlot;
    VpOptionCodecType temp = VP_OPTION_MLAW;
    uint8 lowLevelCmdBuf[LOW_LEVEL_CMD_MAX_LEN];
    VpOptionEventMaskType eventMask;
    VpOptionDtmfModeType dtmfMode;
    uint8 dtmfDigit;
    bool leadingEdge = FALSE;
    int timeslot;
#ifdef DEMONSTRATE_GET_LINE_STATE
    VpLineStateType getLineStateVal;
#endif /* DEMONSTRATE_GET_LINE_STATE */

#ifdef DEMONSTRATE_CID
    uint8 callerIdDataSize, *cidDataPtr;
#ifdef CID_DEMO_USE_SENDCID
    static bool firstBreakInterval = FALSE;
#endif
#endif /* DEMONSTRATE_CID */

    dtmfMode.dtmfControlMode = VP_OPTION_DTMF_DECODE_ON;
    dtmfMode.direction = VP_DIRECTION_US;

    /* Take necessary action based on the event */
    if(pEvent->eventCategory == VP_EVCAT_RESPONSE) {
        switch  (pEvent->eventId) {
            case VP_DEV_EVID_BOOT_CMP: {
                /* Read the results associated with this event */
                term_printf("Received event: VP_EVID_BOOT_CMP\n");
                term_printf("Reading Checksum...");
#define  TEST_POLL_RESPONSE_EVENT
#ifdef  TEST_POLL_RESPONSE_EVENT
                pEvent->eventId = 0;
                status = VpGetResults(pEvent, VP_NULL);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                PrintEvent(pEvent);
#endif /* TEST_POLL_RESPONSE_EVENT */

                status = VpGetResults(pEvent, &chkSumInfo);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                term_printf("done. Checksum Data:\n");
                term_printf("VpChkSumType.loadChkSum = 0x%lx\n",
                                chkSumInfo.loadChecksum);
                term_printf("VpChkSumType.vInfo.vtdRevCode = 0x%x\n",
                                chkSumInfo.vInfo.vtdRevCode);
                term_printf("VpChkSumType.vInfo.swProductId = 0x%x\n",
                                chkSumInfo.vInfo.swProductId);
                term_printf("VpChkSumType.vInfo.swVerMajor = 0x%x\n",
                                chkSumInfo.vInfo.swVerMajor);
                term_printf("VpChkSumType.vInfo.swVerMinor = 0x%x\n",
                                chkSumInfo.vInfo.swVerMinor);

                /* Create Line Objects and Initialize the device */
                status = InitAfterBoot();
                break;
            }
            case VP_DEV_EVID_DEV_INIT_CMP: {
                term_printf("Received event: VP_DEV_EVID_DEV_INIT_CMP(%d)\n",
                                chanId);

                /* Veryify Calibration was successful */
                term_printf("Device initialization complete Event Data: 0x%x\n",
                                pEvent->eventData);

                if(pEvent->eventData == 0) {
                    term_printf("ERROR: Device initialization Failed; "
                                "NO SLAC found; Exiting...\n");

                    status = VP_STATUS_FAILURE;
                    break;
                }

                /* Setup the event mask that is different compared to default
                 * masks; API masks all event (other than non-maskable). So
                 * Unmask all necessary events */
                eventMask.faults = VP_EVCAT_FAULT_UNMASK_ALL;
                eventMask.signaling = VP_EVCAT_SIGNALING_UNMASK_ALL;
                eventMask.response = VP_EVCAT_RESPONSE_UNMASK_ALL;
                eventMask.test = VP_EVCAT_TEST_UNMASK_ALL;
                eventMask.process = VP_EVCAT_PROCESS_UNMASK_ALL;
                eventMask.fxo = VP_EVENT_MASK_ALL; /* Not applicable to VCP */
                eventMask.packet = VP_EVENT_MASK_ALL;  /* Not applicable */

                term_printf("Setting up events masks...\n");
                status = VpSetOption(VP_NULL, pEvent->pDevCtx,
                                      VP_OPTION_ID_EVENT_MASK, &eventMask);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }

#ifdef DEMONSTRATE_LINE_INIT
                /* Initialize a line; Note that this application is not written
                 * to handle event generated as a consequence of this function
                 * call properly. */
                i = 1;
                term_printf("Initialize a line %d", i);
                status = VpInitLine(&lineCtx[i], &AC_600R_FXS[0],
                                    &DC_25MA_CC[0], &RING_20HZ_SINE[0]);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                term_printf("done\n");
#endif  /* DEMONSTRATE_LINE_INIT */

                /* Configure a line differently */
                i = 1;
                term_printf("Configuring line %d for different DC feed, "
                                "Ringing Profile...", i);
                status = VpConfigLine(&lineCtx[i], &AC_600R_FXS[0],
                                    &DC_25MA_CC[0], &RING_20HZ_SINE[0]);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                term_printf("done\n");

                /* Initialize all the lines now ... */
                for(i=0; i< MAX_LINES_ON_BOARD; i++){

                    term_printf("Line %d ", i);

                    /* Initialize Ringing parameters */
                    term_printf("Ring Init, ");
#ifdef DEMONSTRATE_CID
#ifdef CID_DEMO_USE_SENDCID
                    status = VpInitRing(&lineCtx[i], &RING_CAD_NOCID[0],
                                            VP_PTABLE_NULL);
#else /* !CID_DEMO_USE_SENDCID */
                    status = VpInitRing(&lineCtx[i], &RING_CAD_STD[0],
                                            VP_PTABLE_INDEX1);
#endif /* !CID_DEMO_USE_SENDCID */
#else /* !DEMONSTRATE_CID */
                    status = VpInitRing(&lineCtx[i], &RING_CAD_NOCID[0],
                                            VP_PTABLE_NULL);
#endif /* DEMONSTRATE_CID */
                    if(status != VP_STATUS_SUCCESS){
                        break; /* Something is wrong */
                    }

                    if(i >= LINES_FROM_WHICH_LCAS_IS_PRESENT) {
                        /* Place the LCAS in normal state */
                        term_printf("LCAS State->Normal, ");
                        status = VpSetRelayState(&lineCtx[i], VP_RELAY_NORMAL);
                        if(status != VP_STATUS_SUCCESS){
                            break; /* Something is wrong */
                        }
                    }

                    /* Configure the PCM time slot; Presently there is one to
                     * one mapping of channels and time slots used by them. This
                     * could be modified(if needed) as appropriate. */
                    timeslot = 2 * i;   /* Leave the slots in between for
                                         * Testing */
                    pcmTimeSlot.rx = timeslot;
                    pcmTimeSlot.tx = timeslot;
                    term_printf("TimeSlot->%d, ", timeslot);
                    status = VpSetOption(&lineCtx[i], VP_NULL,
                                          VP_OPTION_ID_TIMESLOT, &pcmTimeSlot);
                    if(status != VP_STATUS_SUCCESS){
                        break; /* Something is wrong */
                    }

                    /* Configure CODEC type */
                    term_printf("CODEC->Mu-Law, ");
                    status = VpSetOption(&lineCtx[i], VP_NULL,
                                            VP_OPTION_ID_CODEC, &temp);
                    if(status != VP_STATUS_SUCCESS){
                        break; /* Something is wrong */
                    }

#ifdef DEMONSTRATE_DTMF_DECODE
                    if(i<MAX_DTMF_DETECTION_LINES){
                        /* Enable DTMF detection for certain lines */
                        term_printf("DTMF->ON, ");
                        status = VpSetOption(&lineCtx[i], VP_NULL,
                                        VP_OPTION_ID_DTMF_MODE, &dtmfMode);
                        term_printf("DTMF Status=0x%.8lx, ",
                                        dtmfMode.dtmfDetectionSetting);
                        term_printf("DTMF Remaing Resources=%d, ",
                                        dtmfMode.dtmfResourcesRemaining);
                        if(status != VP_STATUS_SUCCESS){
                            break; /* Something is wrong */
                        }
                    }

                    if(i== 1){
                        /* Enable DTMF detection for certain lines */
                        term_printf("\nDTMF->OFF, ");
                        dtmfMode.dtmfControlMode = VP_OPTION_DTMF_DECODE_OFF;
                        status = VpSetOption(&lineCtx[i], VP_NULL,
                                        VP_OPTION_ID_DTMF_MODE, &dtmfMode);
                        term_printf("DTMF Status=0x%.8lx, ",
                                        dtmfMode.dtmfDetectionSetting);
                        term_printf("DTMF Remaing Resources=%d, \n",
                                        dtmfMode.dtmfResourcesRemaining);
                        if(status != VP_STATUS_SUCCESS){
                            break; /* Something is wrong */
                        }
                        dtmfMode.dtmfControlMode = VP_OPTION_DTMF_DECODE_ON;
                    }
#endif /* DEMONSTRATE_DTMF_DECODE */
#ifdef DEMONSTRATE_GET_LINE_STATE
                    /* Get current line state (should be disconnect) */
                    status = VpGetLineState(&lineCtx[i], &getLineStateVal);
                    if(status != VP_STATUS_SUCCESS){
                        break; /* Something is wrong */
                    }

                    /* Place the Line in OHT state */
                    term_printf("Line State: %s -> OHT\n",
                                    statestring[getLineStateVal]);
                    status = VpSetLineState(&lineCtx[i], VP_LINE_OHT);
#else
                    /* Place the Line in OHT state */
                    term_printf("Line State: OHT\n");
                    status = VpSetLineState(&lineCtx[i], VP_LINE_OHT);
#endif /* DEMONSTRATE_GET_LINE_STATE */
                    if(status != VP_STATUS_SUCCESS){
                        break; /* Something is wrong */
                    }
                }

                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }

#ifdef TEST_LOW_LEVEL_CMD
                /*Issue a low level command */
                term_printf("Issuing Low level command...");
                presentCommandLength = LOOP_SUPER_VISION_LEN;
                lowLevelCmdBuf[0] = LOOP_SUPER_VISION_RD;
                status = VpLowLevelCmd(&lineCtx[0], &lowLevelCmdBuf[0],
                                           presentCommandLength, 0);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                term_printf("done\n");
#endif  /* TEST_LOW_LEVEL_CMD */

#ifdef DEMONSTRATE_GET_LOOP_COND
                /*Issue a low level command */
                term_printf("Issuing Get loop condition command...");
                status = VpGetLoopCond(&lineCtx[0], 0);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                term_printf("done\n");
#endif  /* DEMONSTRATE_GET_LOOP_COND */

                break;
            }
            case VP_LINE_EVID_LINE_INIT_CMP: {
                term_printf("Received event: VP_DEV_EVID_LINE_INIT_CMP(%d)\n",
                                chanId);
                status = VP_STATUS_SUCCESS;
                break;
            }
            case VP_LINE_EVID_LLCMD_RX_CMP: {
                term_printf("Received event: VP_EVID_LLCMD_RX_CMP(%d)\n",
                            chanId);
                /* Reading Response to last Low level command */
                term_printf("Reading Response for line = %x\n",
                                pEvent->channelId);

                /* Read the results from the low level command and print */
                status = VpGetResults(pEvent, &lowLevelCmdBuf[0]);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                term_printf("Data = ");
                for(i = 0; i<presentCommandLength; i++){
                    term_printf("%2.2hX ", lowLevelCmdBuf[i]);
                }
                term_printf("\n");
                break;
            }
            case VP_LINE_EVID_LLCMD_TX_CMP: {
                term_printf("Received event: VP_EVID_LLCMD_TX_CMP(%d)\n",
                                chanId);
                status = VP_STATUS_SUCCESS;
                break;
            }
            case VP_DEV_EVID_DNSTR_MBOX:    {
                term_printf("Received event: VP_DEV_EVID_DNSTR_MBOX(%d)\n",
                    pEvent->eventData);
                status = VP_STATUS_SUCCESS;
                break;
            }
            case VP_LINE_EVID_RD_OPTION: {
                /* Print the results */
                status = PrintGetOptionResults(pEvent);
                break;
            }
            case VP_LINE_EVID_RD_LOOP: {
                /* Print the results */
                status = PrintGetLoopCondResults(pEvent);
                break;
            }
        }
    }else if(pEvent->eventCategory == VP_EVCAT_SIGNALING) {
        switch  (pEvent->eventId) {
            case VP_LINE_EVID_HOOK_ON: {
                term_printf("Received event: VP_EVID_HOOK_ON(%d); TS = %u\n",
                            chanId, pEvent->parmHandle);
#ifdef DEMONSTRATE_GET_LINE_STATE
                /* Get current line state */
                status = VpGetLineState(pEvent->pLineCtx, &getLineStateVal);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
#endif /* DEMONSTRATE_GET_LINE_STATE */

                /* Toggle between on-hook ring and on-hook no ring state */
                if(lineState[chanId] == ONHOOK_NO_RING){

#if defined(DEMONSTRATE_CID) && !defined(CID_DEMO_USE_SENDCID)
                    /* Initialize CallerID data */
                    cidDataPtr = GetCallerDataPointer(TRUE, &callerIdDataSize);
                    if((callerIdDataSize > 0) && (cidDataPtr != VP_NULL)) {
                        term_printf("Writing CID data (VpInitCid()) "
                                        "size = %d\n", callerIdDataSize);
                        status = VpInitCid(pEvent->pLineCtx, callerIdDataSize,
                            cidDataPtr);

                        if(status != VP_STATUS_SUCCESS){
                            break; /* Something is wrong */
                        }
                    }
#endif
                    lineState[chanId] = ONHOOK_RING;
#ifdef DEMONSTRATE_GET_LINE_STATE
                    term_printf("Line State: %s -> Ringing\n",
                                    statestring[getLineStateVal]);
#else
                    term_printf("Line State: Ringing\n");
#endif /* DEMONSTRATE_GET_LINE_STATE */
#if defined(DEMONSTRATE_CID) && defined(CID_DEMO_USE_SENDCID)
                    firstBreakInterval = TRUE;
#endif
                    status = VpSetLineState(pEvent->pLineCtx, VP_LINE_RINGING);
                    if(status != VP_STATUS_SUCCESS){
                        break; /* Something is wrong */
                    }

                } else {
                    lineState[chanId] = ONHOOK_NO_RING;
#ifdef DEMONSTRATE_GET_LINE_STATE
                    term_printf("Line State: %s -> OHT\n",
                                    statestring[getLineStateVal]);
#else
                    term_printf("Line State: OHT\n");
#endif /* DEMONSTRATE_GET_LINE_STATE */
                    status = VpSetLineState(pEvent->pLineCtx, VP_LINE_OHT);
                    if(status != VP_STATUS_SUCCESS){
                        break; /* Something is wrong */
                    }
                }
                break;
            }
            case VP_LINE_EVID_HOOK_OFF: {
                term_printf("Received event: VP_EVID_HOOK_OFF(%d); TS = %u\n",
                            chanId, pEvent->parmHandle);

#ifdef DEMONSTRATE_GET_LINE_STATE
                /* Get current line state */
                status = VpGetLineState(pEvent->pLineCtx, &getLineStateVal);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }

                term_printf("Line State: %s -> Talk, ",
                                statestring[getLineStateVal]);
#else
                term_printf("Line State: Talk, ");
#endif /* DEMONSTRATE_GET_LINE_STATE */
                status = VpSetLineState(pEvent->pLineCtx, VP_LINE_TALK);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }

                if(applyCadence == TRUE) {
#ifdef VP_CC_VCP_790_SERIES
                    term_printf("Feeding message\n");
                    applyCadence = FALSE;
                    /* User picks up in response to ringing; Generate a tone
                     * message instead of dial tone */
                    status = VpSetLineTone(pEvent->pLineCtx, &TONE_ONEKHZ_L[0],
                                            &TONE_CAD_SIT[0], VP_NULL);
#else
                    term_printf("Feeding ringback\n");
                    applyCadence = FALSE;
                    /* User picks up in response to ringing; Generate a tone
                     * message instead of dial tone */
                    status = VpSetLineTone(pEvent->pLineCtx, &TONE_RINGBACK[0],
                                            TONE_CAD_RNGBACK, VP_NULL);
#endif
                } else {
                    term_printf("Feeding Dial Tone\n");
                    applyCadence = TRUE;
                    status = VpSetLineTone(pEvent->pLineCtx, &TONE_DIAL[0],
                                            VP_PTABLE_NULL, VP_NULL);
                }
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                break;

            }
            case VP_LINE_EVID_DTMF_DIG: {
                /* DTMF digit detected; Print the digit */
                dtmfDigit = pEvent->eventData & 0x0f;
                if((pEvent->eventData & 0x10) == 0) {
                    leadingEdge = FALSE;
                } else {
                    leadingEdge = TRUE;
                }
#ifdef HANDLE_DTMF_CMDS
                term_printf("Line: %d, DTMF Digit: %d, Timestamp %u: "
                            "Edge: %s\n",
                            pEvent->channelId, dtmfDigit,
                            pEvent->parmHandle,
                            leadingEdge ? "Leading":"Trailing");
                if(leadingEdge == TRUE) {
                    status = ProcessDtmfDigits(pEvent->channelId, dtmfDigit);
                } else {
                    status = VP_STATUS_SUCCESS; /* Nothing to do */
                }
#else
                term_printf("Line: %d, DTMF Digit: %d, Timestamp %u: "
                            "Edge: %s\n",
                            pEvent->channelId, dtmfDigit,
                            pEvent->parmHandle,
                            leadingEdge ? "Leading":"Trailing");
                status = VP_STATUS_SUCCESS; /* Event is consumed */
#endif /* HANDLE_DTMF_CMDS */
                break;
            }
            case VP_DEV_EVID_TS_ROLLOVER: {
                timeStampCounter++;
                /* Uncomment if necessary */
                term_printf("Received event: VP_DEV_EVID_TS_ROLLOVER; Count: "
                                "%lu \n", timeStampCounter);
#ifdef DEMONSTRATE_TEST
                /* Initiate testing */
                if(timeStampCounter == 1) {
                    status = LineTester(TRUE);
                    break;
                } else {
                    status = VP_STATUS_SUCCESS; /* Event is consumed */
                    break;
                }
#else
                status = VP_STATUS_SUCCESS; /* Event is consumed */
                break;
#endif /* DEMONSTRATE_TEST */
            }
        }
    } else if(pEvent->eventCategory == VP_EVCAT_PROCESS) {
        switch  (pEvent->eventId) {
            case VP_LINE_EVID_RING_CAD: {
                term_printf("Received event: VP_LINE_EVID_RING_CAD(%d)",
                                chanId);
                status = VP_STATUS_SUCCESS;
                if (pEvent->eventData == VP_RING_CAD_MAKE) {
                    term_printf("make \n");
                } else if (pEvent->eventData == VP_RING_CAD_BREAK) {
                    term_printf("break \n");
#if defined(DEMONSTRATE_CID) && defined(CID_DEMO_USE_SENDCID)
                    /* Demonstrating U.S. Type I caller ID; wait for the first
                     * off interval in the ringing cadence.  Do not send the
                     * data again in subsequent off intervals. */
                    if (!firstBreakInterval) {
                        break;
                    }
                    firstBreakInterval = FALSE;
                    /* Send first chunk of CallerID data */
                    cidDataPtr = GetCallerDataPointer(TRUE, &callerIdDataSize);
                    if((callerIdDataSize > 0) && (cidDataPtr != VP_NULL)) {
                        term_printf("Writing CID data (VpSendCid()) "
                                        "size = %d\n", callerIdDataSize);
                        status = VpSendCid(pEvent->pLineCtx, callerIdDataSize,
                            VP_PTABLE_INDEX1, cidDataPtr);
                    }
#endif
                } else {
                    term_printf("done; ");
                }
                break;
            }
#ifdef DEMONSTRATE_CID
            case VP_LINE_EVID_CID_DATA: {
                term_printf("Received event: VP_LINE_EVID_CID_DATA(%d); - %s\n",
                                chanId,
                                (pEvent->eventData == VP_CID_DATA_TX_DONE) ?
                                 "CID Complete":"Is there any more data ?");

                    /* Get the subsequent CID data blocks */
                    cidDataPtr = GetCallerDataPointer(FALSE, &callerIdDataSize);
                    if((callerIdDataSize > 0) && (cidDataPtr != VP_NULL)) {

                        if(pEvent->eventData == VP_CID_DATA_TX_DONE) {
                            term_printf("Abroting CID due to OFF-Hook\n");
                            status = VP_STATUS_SUCCESS;
                            break;
                        }

                        term_printf("Writing CID data (VpContinueCid()) "
                                        "size=%d\n", callerIdDataSize);
                        status = VpContinueCid(pEvent->pLineCtx,
                                        callerIdDataSize, cidDataPtr);
                        if(status != VP_STATUS_SUCCESS){
                            break; /* Something is wrong */
                        }
                    }

                status = VP_STATUS_SUCCESS;
                break;
            }
#endif /* DEMONSTRATE_CID */
            case VP_LINE_EVID_MTR_CMP: {
                term_printf("Received event: VP_LINE_EVID_MTR_CMP(%d)\n",
                                chanId);
                meteringGoingOn[chanId] = FALSE;
                status = VP_STATUS_SUCCESS;
                break;
            }
            case VP_LINE_EVID_MTR_ABORT: {
                term_printf("Received event: VP_LINE_EVID_MTR_ABORT(%d) "
                    "Number of pulses sent = %d\n", chanId, pEvent->eventData);
                meteringGoingOn[chanId] = FALSE;/* Aborted for whatever reason*/
                status = VP_STATUS_SUCCESS;
                break;
            }

            case VP_LINE_EVID_TONE_CAD: {
                term_printf("Received event: VP_LINE_EVID_TONE_CAD(%d)"
                    "Event Data = %d\n", chanId, pEvent->eventData);
                status = VP_STATUS_SUCCESS;
                break;
            }
        }
    }
#ifdef DEMONSTRATE_TEST
     else if(pEvent->eventCategory == VP_EVCAT_TEST) {
        status = TestEventHandler(pEvent);
    }
#endif /* DEMONSTRATE_TEST */

    if(status != VP_STATUS_SUCCESS){

        /* Either the event was not handled or there is error in handling the
         * event. Print the event. */
        term_printf("ERROR: Unknown event or Error while handling the event; "
                    "Printing the event...");
        PrintEvent(pEvent);

        /* If Event has a result and it has not been handled until now,
         * Clear it */
        if(pEvent->hasResults == TRUE) {
            term_printf("WARNING: Unhandled results, Clearing it...\n");
            VpClearResults(pEvent->pDevCtx);
        }
    }

    return status;
} /* VtdEventHandler() */


/*
 * PrintEvent()
 * This function an un-handled event.
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * None
 */
static void PrintEvent(
    VpEventType *pEvent)
{
    char *eventCategory[] = {
                                "VP_EVCAT_FAULT",
                                "VP_EVCAT_SIGNALING",
                                "VP_EVCAT_RESPONSE",
                                "VP_EVCAT_TEST",
                                "VP_EVCAT_PROCESS",
                                "VP_EVCAT_FXO",
                                "VP_EVCAT_PACKET",
    };

    if(pEvent->eventCategory < VP_NUM_EVCATS) {
        term_printf("VpEvent Cat:Id = %s:",
                        eventCategory[pEvent->eventCategory]);
    } else {
        term_printf("VpEvent Cat:Id = 0x%x:",
                        pEvent->eventCategory);
    }

    if(pEvent->eventCategory == VP_EVCAT_FAULT) {
        switch (pEvent->eventId) {
            case VP_DEV_EVID_BAT_FLT:   {
                term_printf("VP_DEV_EVID_BAT_FLT\n"); break;
            }
            case VP_DEV_EVID_CLK_FLT:   {
                term_printf("VP_DEV_EVID_CLK_FLT\n"); break;
            }
            case VP_LINE_EVID_THERM_FLT:{
                term_printf("VP_LINE_EVID_THERM_FLT\n"); break;
            }
            case VP_LINE_EVID_DC_FLT:   {
                term_printf("VP_DEVP_LINE_EVID_DC_FLT\n"); break;
            }
            case VP_LINE_EVID_AC_FLT:   {
                term_printf("VP_LINE_EVID_AC_FLT\n"); break;
            }
            case VP_DEV_EVID_EVQ_OFL_FLT: {
                term_printf("VP_DEV_EVID_EVQ_OFL_FLT\n"); break;
            }
            case VP_DEV_EVID_WDT_FLT: {
                term_printf("VP_DEV_EVID_WDT_FLT\n"); break;
            }
            default: {
                term_printf("0x%x\n", pEvent->eventId); break;
            }
        }
    } else if(pEvent->eventCategory == VP_EVCAT_SIGNALING) {
        switch (pEvent->eventId) {
            case VP_LINE_EVID_HOOK_OFF:   {
                term_printf("VP_LINE_EVID_HOOK_OFF\n"); break;
            }
            case VP_LINE_EVID_HOOK_ON:   {
                term_printf("VP_LINE_EVID_HOOK_ON\n"); break;
            }
            case VP_LINE_EVID_GKEY_DET:{
                term_printf("VP_LINE_EVID_GKEY_DET\n"); break;
            }
            case VP_LINE_EVID_GKEY_REL:   {
                term_printf("VP_LINE_EVID_GKEY_REL\n"); break;
            }
            case VP_LINE_EVID_FLASH:   {
                term_printf("VP_LINE_EVID_FLASH\n"); break;
            }
            case VP_LINE_EVID_STARTPULSE: {
                term_printf("VP_LINE_EVID_STARTPULSE\n"); break;
            }
            case VP_LINE_EVID_DTMF_DIG: {
                term_printf("VP_LINE_EVID_DTMF_DIG\n"); break;
            }
            case VP_LINE_EVID_PULSE_DIG: {
                term_printf("VP_LINE_EVID_PULSE_DIG\n"); break;
            }
            case VP_LINE_EVID_MTONE: {
                term_printf("VP_LINE_EVID_MTONE\n"); break;
            }
            case VP_DEV_EVID_TS_ROLLOVER: {
                term_printf("VP_DEV_EVID_TS_ROLLOVER\n"); break;
            }
            default: {
                term_printf("0x%x\n", pEvent->eventId); break;
            }
        }
    } else if(pEvent->eventCategory == VP_EVCAT_RESPONSE) {
        switch (pEvent->eventId) {
            case VP_DEV_EVID_BOOT_CMP:   {
                term_printf("VP_DEV_EVID_BOOT_CMP\n"); break;
            }
            case VP_LINE_EVID_LLCMD_TX_CMP:   {
                term_printf("VP_LINE_EVID_LLCMD_TX_CMP\n"); break;
            }
            case VP_LINE_EVID_LLCMD_RX_CMP:{
                term_printf("VP_LINE_EVID_LLCMD_RX_CMP\n"); break;
            }
            case VP_DEV_EVID_DNSTR_MBOX:   {
                term_printf("VP_DEV_EVID_DNSTR_MBOX\n"); break;
            }
            case VP_LINE_EVID_RD_OPTION:   {
                term_printf("VP_LINE_EVID_RD_OPTION\n"); break;
            }
            case VP_LINE_EVID_RD_LOOP: {
                term_printf("VP_LINE_EVID_RD_LOOP\n"); break;
            }
            case VP_EVID_CAL_CMP: {
                term_printf("VP_EVID_CAL_CMP\n"); break;
            }
            case VP_EVID_CAL_BUSY: {
                term_printf("VP_EVID_CAL_BUSY\n"); break;
            }
            case VP_LINE_EVID_GAIN_CMP: {
                term_printf("VP_LINE_EVID_GAIN_CMP\n"); break;
            }
            case VP_DEV_EVID_DEV_INIT_CMP: {
                term_printf("VP_DEV_EVID_DEV_INIT_CMP\n"); break;
            }
            case VP_LINE_EVID_LINE_INIT_CMP: {
                term_printf("VP_LINE_EVID_LINE_INIT_CMP\n"); break;
            }
            default: {
                term_printf("0x%x\n", pEvent->eventId); break;
            }
        }
    } else if(pEvent->eventCategory == VP_EVCAT_TEST) {
        switch (pEvent->eventId) {
            case VP_LINE_EVID_TEST_CMP:   {
                term_printf("VP_LINE_EVID_TEST_CMP\n"); break;
            }
            case VP_LINE_EVID_DTONE_DET:   {
                term_printf("VP_LINE_EVID_DTONE_DET\n"); break;
            }
            case VP_LINE_EVID_DTONE_LOSS:{
                term_printf("VP_LINE_EVID_DTONE_LOSS\n"); break;
            }
            case VP_DEV_EVID_STEST_CMP:   {
                term_printf("VP_DEV_EVID_STEST_CMP\n"); break;
            }
            case VP_DEV_EVID_CHKSUM:   {
                term_printf("VP_DEV_EVID_CHKSUM\n"); break;
            }
            default: {
                term_printf("0x%x\n", pEvent->eventId); break;
            }
        }
    } else if(pEvent->eventCategory == VP_EVCAT_PROCESS) {
        switch (pEvent->eventId) {
            case VP_LINE_EVID_MTR_CMP:   {
                term_printf("VP_LINE_EVID_MTR_CMP\n"); break;
            }
            case VP_LINE_EVID_MTR_ABORT:   {
                term_printf("VP_LINE_EVID_MTR_ABORT\n"); break;
            }
            case VP_LINE_EVID_CID_DATA:{
                term_printf("VP_LINE_EVID_CID_DATA\n"); break;
            }
            case VP_LINE_EVID_RING_CAD:   {
                term_printf("VP_LINE_EVID_RING_CAD\n"); break;
            }
            default: {
                term_printf("0x%x\n", pEvent->eventId); break;
            }
        }
    } else {
        term_printf("0x%x\n", pEvent->eventId);
    }
    term_printf("VpEventType.channelId = 0x%x\n", pEvent->channelId);
    term_printf("VpEventType.deviceId = 0x%x\n", pEvent->deviceId);
    term_printf("VpEventType.eventData = 0x%x\n", pEvent->eventData);
    term_printf("VpEventType.parmHandle = 0x%x\n", pEvent->parmHandle);
    term_printf("VpEventType.hasResults = 0x%x\n", pEvent->hasResults);
}

/*
 * PrintGetOptionResults()
 * This function prints the Get option results.
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * None
 */
static VpStatusType
PrintGetOptionResults(
    VpEventType *pEvent)
{
    uint16 optionId;
    VpOptionPulseType pulseTimings;
    VpOptionEventMaskType eventMask;
    VpStatusType status = VP_STATUS_FAILURE;

    optionId = pEvent->eventData;
    if(optionId > VP_OPTION_ID_DTMF_MODE) {
        term_printf("Invalid Event: VP_LINE_EVID_RD_OPTION:%d\n", optionId);
        return VP_STATUS_FAILURE;
    } else {
        term_printf("Received event: VP_LINE_EVID_RD_OPTION:%d\n", optionId);
    }

    switch (optionId) {
        case VP_DEVICE_OPTION_ID_PULSE: {
            status = VpGetResults(pEvent, &pulseTimings);
            if(status == VP_STATUS_SUCCESS) {
                term_printf("Option Read: VpOptionPulseType.breakMin = %u\n",
                                pulseTimings.breakMin);
                term_printf("Option Read: VpOptionPulseType.breakMax = %u\n",
                                pulseTimings.breakMax);
                term_printf("Option Read: VpOptionPulseType.makeMin = %u\n",
                                pulseTimings.makeMin);
                term_printf("Option Read: VpOptionPulseType.makeMax = %u\n",
                                pulseTimings.makeMax);
                term_printf("Option Read: VpOptionPulseType.interDigitMin = "
                                "%u\n", pulseTimings.interDigitMin);
                term_printf("Option Read: VpOptionPulseType.flashMin = %u\n",
                                pulseTimings.flashMin);
                term_printf("Option Read: VpOptionPulseType.flashMax = %u\n",
                                pulseTimings.flashMax);
            }
            break;
        }
        case VP_OPTION_ID_EVENT_MASK: {
            status = VpGetResults(pEvent, &eventMask);
            if(status == VP_STATUS_SUCCESS) {
                term_printf("Option Read: VpOptionEventMaskType.faults = "
                                "0x%.4x\n", eventMask.faults);
                term_printf("Option Read: VpOptionEventMaskType.signaling = "
                                "0x%.4x\n", eventMask.signaling);
                term_printf("Option Read: VpOptionEventMaskType.response = "
                                "0x%.4x\n", eventMask.response);
                term_printf("Option Read: VpOptionEventMaskType.test = "
                                "0x%.4x\n", eventMask.test);
                term_printf("Option Read: VpOptionEventMaskType.process = "
                                "0x%.4x\n", eventMask.process);
                term_printf("Option Read: VpOptionEventMaskType.fxo = "
                                "0x%.4x\n", eventMask.fxo);
                term_printf("Option Read: VpOptionEventMaskType.packet = "
                                "0x%.4x\n", eventMask.packet);
            }
            break;
        }
        default: {
            term_printf("ERROR: Option Read: OptionId = %d Not yet "
                            "implemented \n", optionId);
            break;
        }
    }

    return status;
} /* PrintGetOptionResults() */

/* Print the Get Loop Condition results */
static VpStatusType
PrintGetLoopCondResults(
    VpEventType *pEvent)
{
    VpLoopCondResultsType loopCond;
    char *batName[] = {"VP_BATTERY_UNDEFINED", "VP_BATTERY_1",
                        "VP_BATTERY_2", "VP_BATTERY_3"};
    char *feedRegName[] = {"VP_DF_UNDEFINED", "VP_DF_ANTI_SAT_REG",
                        "VP_DF_CNST_CUR_REG", "VP_DF_RES_FEED_REG"};
    VpStatusType status;

    status = VpGetResults(pEvent, &loopCond);
    if(status == VP_STATUS_SUCCESS) {
        term_printf("LoopCond: VpLoopCondResultsType.rloop = %d\n",
                        loopCond.rloop);
        term_printf("LoopCond: VpLoopCondResultsType.ilg = %d\n",
                        loopCond.ilg);
        term_printf("LoopCond: VpLoopCondResultsType.imt = %d\n",
                        loopCond.imt);
        term_printf("LoopCond: VpLoopCondResultsType.vsab = %d\n",
                        loopCond.vsab);
        term_printf("LoopCond: VpLoopCondResultsType.vbat1 = %d\n",
                        loopCond.vbat1);
        term_printf("LoopCond: VpLoopCondResultsType.vbat2 = %d\n",
                        loopCond.vbat2);
        term_printf("LoopCond: VpLoopCondResultsType.vbat3 = %d\n",
                        loopCond.vbat3);
        term_printf("LoopCond: VpLoopCondResultsType.mspl = %d\n",
                        loopCond.mspl);

        if(loopCond.selectedBat < VP_NUM_BATTERY_TYPES) {
            term_printf("LoopCond: VpLoopCondResultsType.selectedBat = %s\n",
                            batName[loopCond.selectedBat]);
        } else {
            term_printf("LoopCond: VpLoopCondResultsType.selectedBat = %d\n",
                            loopCond.selectedBat);
        }
        if(loopCond.dcFeedReg < VP_NUM_DC_FEED_TYPES) {
            term_printf("LoopCond: VpLoopCondResultsType.dcFeedReg = %s\n",
                            feedRegName[loopCond.dcFeedReg]);
        } else {
            term_printf("LoopCond: VpLoopCondResultsType.dcFeedReg = %d \n",
                            loopCond.dcFeedReg);
        }
    }

    return status;

} /* PrintGetLoopCondResults() */

/*
 * ProcessDtmfDigits()
 * This function Processes detected DTMF digits.
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * None
 */
#define MAX_USER_CMD_SIZE       (25)
#define CMD_END_DIGIT           (VP_DIG_ASTER)
unsigned char command[MAX_LINES_ON_BOARD][MAX_USER_CMD_SIZE];
bool commandInProgress[MAX_LINES_ON_BOARD];
unsigned int commandLength[MAX_LINES_ON_BOARD];

static VpStatusType ProcessDtmfDigits(
    uint8 channelId,
    uint8 digit)
{
    unsigned int commandOpcode;
    int i;
    VpStatusType status = VP_STATUS_SUCCESS;
    uint16 getOptionId = 0;
#ifdef DEMONSTRATE_TEST
    uint16 line;
#endif /* DEMONSTRATE_TEST */
#ifdef DEMONSTRATE_METERING
    uint16 numMeter;
#endif /* DEMONSTRATE_METERING */
#ifdef DEMONSTRATE_HOWLER_TONE
    VpProfileDataType anyTonePrf[] = {0x00, 0x02, 0x00, 12, 0x00, 0x00, 0x00,
                        0x00, 0x11, 0x8A, 0x05, 0x12, 0x16, 0x02, 0x05, 0x12};

    #define HOWLER_TYPE_BYTE_INDEX 5
    #define UK_HOWLER_TONE_BYTE 0x04
    #define AUSTRALIAN_HOWLER_TONE_BYTE 0x08
    VpProfileDataType cadencePrf[] = {0x00, 0x03, 0x00, 0x02, 0x00,
                                                        UK_HOWLER_TONE_BYTE};
#endif /* DEMONSTRATE_HOWLER_TONE */

    if(channelId >= MAX_LINES_ON_BOARD) {
        term_printf("ERROR: Invalid Channel in the event\n.");
    }

    /* Coversion from API digit code to human readable code */
    if(digit == VP_DIG_ZERO) {
        digit = 0;
    }

    if(commandLength[channelId] == 0) {
        if(digit == CMD_END_DIGIT) {
            PrintHelp();
            return status;
        }
    }

    if((commandLength[channelId] == MAX_USER_CMD_SIZE - 1) &&
        (digit != CMD_END_DIGIT)){
        term_printf("Invalid Command length; Ignoring command.");
        commandLength[channelId] = 0;
        PrintHelp();
        return status;
    }

    if(digit != CMD_END_DIGIT){
        command[channelId][commandLength[channelId]++] = digit;
        return status;
    }

    /* Process command */
    commandOpcode = (command[channelId][0]<<4) + command[channelId][1];

    switch (commandOpcode) {
        case 0x00: {
            /* Set the delay in the receive path */
            getOptionId = 0;
            for(i = 2; i<commandLength[channelId]; i++){
                getOptionId = getOptionId*10 + command[channelId][i];
            }

            if(getOptionId >= VP_OPTION_ID_DTMF_MODE) {
                term_printf("Maximum allowed option ID = %d\n",
                                VP_OPTION_ID_DTMF_MODE);
                getOptionId = VP_OPTION_ID_DTMF_MODE;
            }
            term_printf("Reading CH:%d Option = %d\n", channelId, getOptionId);
            status = VpGetOption(&lineCtx[channelId], VP_NULL, getOptionId, 0);
            commandLength[channelId] = 0;
            break;
        }
#ifdef DEMONSTRATE_TEST
        case 0x10: {
            /* Set the delay in the receive path */
            if(testOnGoing == FALSE) {
                line = 0;
                for(i = 2; i<commandLength[channelId]; i++){
                    line = line*10 + command[channelId][i];
                }

                if(line >= MAX_LINES_ON_BOARD) {
                    term_printf("Maximum allowed line ID = %d\n",
                                    MAX_LINES_ON_BOARD - 1);
                    line = MAX_LINES_ON_BOARD - 1;
                }
                testLineId = line;
                term_printf("Setting test Line Id = %d\n", testLineId);
            } else {
                term_printf("ERROR: Test Line Id can't be changed \n");
                term_printf("\tOngoing test on line = %d\n", testLineId);
            }
            commandLength[channelId] = 0;
            break;
        }
        case 0x11: {
            if(testOnGoing == FALSE) {
                term_printf("Performing test on line = %d\n", testLineId);
                status = LineTester(TRUE);
            } else {
                term_printf("ERROR: Test can't be started\n");
                term_printf("\tOngoing test on line = %d\n", testLineId);
            }
            commandLength[channelId] = 0;
            break;
        }
#endif /* DEMONSTRATE_TEST */
#ifdef DEMONSTRATE_METERING
        case 0x20: {
            if(meteringGoingOn[channelId] == FALSE) {
                numMeter = 0;
                for(i = 2; i<commandLength[channelId]; i++){
                    numMeter = numMeter*10 + command[channelId][i];
                }
                commandLength[channelId] = 0; /* Command interpretation ends
                                               * here */


                term_printf("Initialising Metering = %d\n", channelId);
                status = VpInitMeter(&lineCtx[channelId], &METER_PROFILE[0]);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                #define ON_METER    20
                #define OFF_METER   20
                term_printf("Starting metering on line = %d, NumMeter = %d\n",
                                channelId, numMeter);
                status = VpStartMeter(&lineCtx[channelId], ON_METER, OFF_METER,
                                        numMeter);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                if(numMeter > 0) {
                    meteringGoingOn[channelId] = TRUE;
                } else {
                    meteringGoingOn[channelId] = FALSE;
                }
                break;

            } else {
                term_printf("ERROR: Metering can't be started (it is on going)\n");
                commandLength[channelId] = 0;
            }
            break;
        }
#endif /* DEMONSTRATE_METERING */

#ifdef DEMONSTRATE_HOWLER_TONE
        case 0x30: /* Intentional fall through */
        case 0x31: {
            if(commandOpcode == 0x31) {
                term_printf("Playing Australian howler tone...\n");
                cadencePrf[HOWLER_TYPE_BYTE_INDEX] =AUSTRALIAN_HOWLER_TONE_BYTE;
            } else {
                term_printf("Playing U.K howler tone...\n");
            }

            status = VpSetLineTone(&lineCtx[channelId], anyTonePrf,
                                                        cadencePrf, VP_NULL);
            commandLength[channelId] = 0;
            break;
        }
#endif /* DEMONSTRATE_HOWLER_TONE */

        default: {
            term_printf("Command Not Understood CMD = %.2X\n", commandOpcode);
            commandLength[channelId] = 0;
            break;
        }
    }
    return status;
}

/*
 * PrintHelp()
 * This function prints Help.
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * None
 */
static void PrintHelp(void)
{
    term_printf("\n"
                "Command     arguments       End  Comments\n");
    term_printf("<*>                              Print this help\n");
    term_printf("<00>        <Option ID>     <*>  Get an Option\n");

#ifdef DEMONSTRATE_TEST
    term_printf("<10>        <testLineId>    <*>  Set the line for testing\n");
    term_printf("<11>                        <*>  Perform Testing\n");
#endif /* DEMONSTRATE_TEST */

#ifdef DEMONSTRATE_METERING
    term_printf("<20>        <num meter>     <*>  Start metering (Pulses specified)\n");
#endif /* DEMONSTRATE_METERING */

#ifdef DEMONSTRATE_HOWLER_TONE
    term_printf("<30>                        <*>  Play U.K howler tone\n");
    term_printf("<31>                        <*>  Play Australian howler tone\n");
#endif /* DEMONSTRATE_HOWLER_TONE */
}

/*
 * PrintHelp()
 * This function resets command state.
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * None
 */
static void ResetCommandState(void)
{
    int i;
    for(i=0; i<MAX_LINES_ON_BOARD; i++){

        commandInProgress[i] = FALSE;
        commandLength[i] = 0;
    }
}

#ifdef DEMONSTRATE_CID

#undef FIXED_MESSAGE

/* Caller ID Related global variables */
#ifdef FIXED_MESSAGE
uint8 message[] = {0x04, 0x12, 0x30, 0x39, 0x33, 0x30, 0x31, 0x32, 0x32, 0x34,
                    0x36, 0x30, 0x39, 0x35, 0x35, 0x35, 0x31, 0x32, 0x31,
                    0x32, 0x51};
uint8 msgLen = sizeof(message);
#else
uint8 message[100];  /* Message buffer */
uint8 msgLen = 0;
#endif /* FIXED_MESSAGE */
uint8 msgIndex = 0;
uint8 msgSize = 0;
const uint8 BUF_SIZE = 18;
#define VCP_MAX_INIT_OR_SEND_CID_BUF_SIZE 32
#define VCP_MAX_CONTINUE_CID_BUF_SIZE 16

/*
 * PrepareCidMessage()
 * This function prepares a Caller ID message
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * None
 */
static void PrepareCidMessage(void)
{
#ifndef FIXED_MESSAGE
    char *month = "09";
    char *day = "30";
    char *hour = "12";
    char *minute = "24";
    uint16 number = 123;
    char *name = "LEGERITYMINI";

    uint8 msgByteCount;
    uint16 checksum = 0;
    uint8 checkSumByte = 0;

    AppCreateBuffer(month, day, hour, minute, number, name, message);
    msgLen = message[1]+2;  /* Length of Message, plus type and len of message
                             * field */

    for (msgByteCount = 0; msgByteCount < msgLen; msgByteCount++) {
        checksum += message[msgByteCount];
    }
    checkSumByte = (uint8)((~(checksum % 256))+1);
    message[msgLen] = checkSumByte;
    msgLen++;
#endif /* FIXED_MESSAGE */
}

/*
 * GetCallerDataPointer()
 * This function returns Caller ID data buffer pointer.
 *
 * Preconditions:
 * This funciton SHOULD be called with 'startFromBegining' flag == TRUE first
 *
 * Postconditions:
 * None
 */
static uint8* /* Returns a pointer from where Caller ID data has to be read */
GetCallerDataPointer(
    bool startFromBegining,  /* If set indicates start from the beginging
                              * of the message */
    uint8 *messageSize)      /* Pointer to the size of the message */
{
    uint8 *retPtr = VP_NULL;
    int remainingLength;

    if(startFromBegining == TRUE) {
        /* First prepare the message */
        PrepareCidMessage();

        /* Send the first block of the message */
        if(msgLen > VCP_MAX_INIT_OR_SEND_CID_BUF_SIZE) {
            *messageSize = VCP_MAX_INIT_OR_SEND_CID_BUF_SIZE;
            msgIndex = VCP_MAX_INIT_OR_SEND_CID_BUF_SIZE; /* Current
            * location for message data */
        } else {
            *messageSize = msgLen;
            msgIndex = msgLen;
        }
        retPtr = &message[0];

    } else {

        if(msgLen == 0) {
            /* Nothing to send */
            *messageSize = 0;
            return retPtr;
        }

        remainingLength = msgLen - msgIndex;

        if(remainingLength <= 0) {
            /* nothing to send */
            *messageSize = 0;
            retPtr = VP_NULL;
        } else {

            retPtr = &message[msgIndex];
            /* Update the index and the length of the message to be sent
             * this time */
           if(remainingLength >= VCP_MAX_CONTINUE_CID_BUF_SIZE) {
                /* Send the next segment */
                *messageSize = VCP_MAX_CONTINUE_CID_BUF_SIZE;
                msgIndex = msgIndex + VCP_MAX_CONTINUE_CID_BUF_SIZE;

            } else {
                /* Send the last segment */
                *messageSize = remainingLength;
                msgIndex = msgIndex + remainingLength;
            }
        }
    }
    return retPtr;
}
#endif /* DEMONSTRATE_CID */

#ifdef DEMONSTRATE_TEST
/*
 * LineTester()
 * This function sequences various line tests.
 *
 * Preconditions:
 * This function should be called "startFromBeg" flag set to TRUE to start
 * the line tests.
 *
 * Postconditions:
 * Leaves the line in testing state unless all the tests are completed.
 */
static VpStatusType
LineTester(
    bool startFromBeg)
{
    VpStatusType status = VP_STATUS_FAILURE;
    static int testSequencer = 0;
    VpTestConcludeType testConclude;
    VpTestNoiseType noiseMeasurementControl;

    /* Basic error checking */
    if(startFromBeg == TRUE) {
        if(testOnGoing == TRUE) {
            /* new test sequence called while one was going on */
            return status;
        } else {
            testOnGoing = TRUE;
        }
        if(testSequencer != 0){
            /* Previous test sequence not completed */
            return status;
        }
    }

    switch (testSequencer) {
        case 0: {
            /* Prepare for running the tests */
            term_printf("Prepare for Testing\n");
            term_printf("Test ID = VP_TEST_ID_PREPARE\n");
            status = VpTestLine(&lineCtx[testLineId],
                                    VP_TEST_ID_PREPARE, VP_NULL, 0);
            testSequencer++;
            break;
        }
        case 1: {
            /* Perform HOOK test */
            term_printf("Test ID = VP_TEST_ID_HOOK_STATUS\n");
            status = VpTestLine(&lineCtx[testLineId],
                                    VP_TEST_ID_HOOK_STATUS, VP_NULL, 0);
            testSequencer++;
            break;
        }
        case 2: {
            /* Some issue in the firmware */
            /* Perform Noise test */
            term_printf("Test ID = VP_TEST_ID_NOISE\n");
            noiseMeasurementControl.integrateTime = 8000; /* 1 Sec */
            noiseMeasurementControl.filter = VP_FILTER_3KHZ;
            noiseMeasurementControl.amp = 5000;
            noiseMeasurementControl.useBuffer = FALSE;
            status = VpTestLine(&lineCtx[testLineId],
                                    VP_TEST_ID_NOISE, &noiseMeasurementControl,
                                    0);
            testSequencer++;
            break;
        }
        case 3: {
            /* Conclude the test */
            term_printf("Concluding test\n");
            term_printf("TestId = VP_TEST_ID_CONCLUDE\n");
            testConclude.abortTest = TRUE;
            status = VpTestLine(&lineCtx[testLineId],
                                    VP_TEST_ID_CONCLUDE, &testConclude, 0);
            testSequencer = 0;
            break;
        }
        default: {
            /* Error case */
            term_printf("Error in Test sequencing\n");
        }
    }

    return status;
} /* LineTester() */

/*
 * TestEventHandler()
 * This function handles the test events (prints the test results and sequences
 * the various tests.
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * None
 */
static VpStatusType
TestEventHandler(
    VpEventType *pEvent)
{
    VpStatusType status = VP_STATUS_FAILURE;
    uint8 chanId = pEvent->channelId;
    VpTestResultType testResult;
    VpOptionTimeslotType pcmTimeSlot;
    int timeslot;

    if(pEvent->eventCategory == VP_EVCAT_TEST) {
        switch  (pEvent->eventId) {
            case VP_DEV_EVID_STEST_CMP: {
                term_printf("Received event: VP_DEV_EVID_STEST_CMP(%d)\n",
                                chanId);
                term_printf("Self Test %s\n",
                                pEvent->eventData ? "FAILED":"PASSED");
                status = VP_STATUS_SUCCESS;
                /* This was be the last test to be performed; Set a flag
                 * to indicate no more tests are running */
                testOnGoing = FALSE;

                /* Setup the time slots as they were before */
                timeslot = 2 * testLineId;
                pcmTimeSlot.rx = timeslot;
                pcmTimeSlot.tx = timeslot;
                term_printf("Line %d TimeSlot->%d, \n", testLineId, timeslot);
                status = VpSetOption(&lineCtx[testLineId], VP_NULL,
                                      VP_OPTION_ID_TIMESLOT, &pcmTimeSlot);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }

                break;
            }
            case VP_LINE_EVID_TEST_CMP: {
                term_printf("Received event: VP_LINE_EVID_TEST_CMP(%d)\n",
                                chanId);
                /* Get the results and exit test mode */
                status = VpGetResults(pEvent, &testResult);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }

                if(testResult.errorCode == VP_TEST_STATUS_SUCCESS) {
                    if(testResult.testId == VP_TEST_ID_PREPARE) {
                        /* Start the test */
                        term_printf("Test ID = VP_TEST_ID_PREPARE "
                                        "Complete\n");
                        /* Next Test*/
                        status = LineTester(FALSE);
                        break;
                    } else if(testResult.testId == VP_TEST_ID_CONCLUDE) {
                        /* Test has been completed */
                        term_printf("Test ID = VP_TEST_ID_CONCLUDE "
                                        "Complete\n");
                        term_printf("Testing complete\n");

                        /* Perform the Self test */
#ifdef USE_2MHZ_PCLK
                        /* First set up timeslots as required by VCP */
                        /* There are no special limitations on timeslots to be
                         * used when using 2MHz PCLK frequency. */
                        pcmTimeSlot.rx = 0 + 2 * testLineId;
#else
                        /* First set up timeslots as required by VCP */
                        pcmTimeSlot.rx = 48; /* Time slot for testing is
                                              * has to be 48 */
#endif /* USE_2MHZ_PCLK */
                        pcmTimeSlot.tx = 0 + 2 * testLineId;
                        term_printf("Test TimeSlot(Line = %d) Rx ->%d, "
                            "Tx->%d\n", testLineId, pcmTimeSlot.rx,
                            pcmTimeSlot.tx);
                        status = VpSetOption(&lineCtx[testLineId], VP_NULL,
                                          VP_OPTION_ID_TIMESLOT, &pcmTimeSlot);

                        term_printf("Starting Self Test\n");
                        status = VpSelfTest(&lineCtx[testLineId]);

                    } else if(testResult.testId == VP_TEST_ID_HOOK_STATUS) {
                        term_printf("Test VP_TEST_ID_HOOK_STATUS Result = %s\n",
                                        testResult.result.hstatus ?
                                        "OFF HOOK":"ON HOOK");
                        /* Next Test*/
                        status = LineTester(FALSE);
                        break;

                    } else if(testResult.testId == VP_TEST_ID_NOISE) {
                        term_printf("Test VP_TEST_ID_NOISE Result;"
                                    " vrms = %u\n", testResult.result.vrms);
                        /* Next Test*/
                        status = LineTester(FALSE);
                        break;
                    } else {
                        term_printf("Do not understand Test Results; "
                                        "TestId = %d\n", testResult.testId);
                    }
                } else {
                    term_printf("Test Failed; errorCode = %d\n",
                                    testResult.errorCode);
                }
            }
        }
    }
    return status;
} /* TestEventHandler() */
#endif  /* DEMONSTRATE_TEST */

#endif /* USING_QUICK_START_APP */


