/** \file quick_start_example.c
 * quick_start_example.c
 *
 * This file provides a simple quick start application built around
 * VP-API.
 *
 * Copyright (c) 2008, Zarlink Semiconductor, Inc.
 */

/* INCLUDES */
#include "app_des.h"    /* Provides Description of hardware and VCP-firmware
                         * product type*/
#ifdef USING_QUICK_START_APP  /* Conditionally compiled file */
#include "vp_hal.h"
#include "vp_api.h"
#include <stdlib.h>

#include <stdio.h>
#ifndef term_printf
/* Define the following as necessary to redirect the debug output */
#define term_printf printf
#endif /* term_printf */

/*******************************************************************************
 * Conditional Compile time options; Change these based on your software       *
 * architecture and other needs.                                               *
 ******************************************************************************/
#undef EXIT_ON_ERROR       /* Define this to exit on error */
#define TEST_ON_LEG_BOARD   /* This is to Test on Zarlink Semiconductor development board */

#define USE_DUMMY_FRAMER    /* Makes use of dummy framer implemenation. Once
                             * actual framer driver is developed, remove this
                             * conditional compile and use correct files in the
                             * project.     */
#ifdef USE_DUMMY_FRAMER
#include "framer_dummy.h" /* Include this file for bringing in Framer
                           * functionality. Presently the inlcuded file is a
                           * dummy implementation. */
#endif /* USE_DUMMY_FRAMER */

#define BOOTLOADING_VCP /* Define this if you would like QuickStartAppMain() to
                         * perform bootloading of the VCP. */

#define POLL_VCP_EVENTS /* Define this if you would like QuickStartAppMain() to
                         * poll for VCP events and handle them. If VCP events
                         * are handled through a interrupt service routine (in
                         * an interrupt context) you do not need
                         * to define this. */

/*******************************************************************************
 * VCP-Ve790 Product specific quick start demonstration features               *
 ******************************************************************************/
#ifdef VP_CC_VCP_790_SERIES
#define DEMONSTRATE_TEST            /* Define this to demonstrate Test
                                     * functions */
#endif /* VP_CC_VCP_790_SERIES */

/*******************************************************************************
 * VCP-Ve880 Product specific quick start demonstration features               *
 ******************************************************************************/
#ifdef VP_CC_VCP_880_SERIES
#undef DEMONSTRATE_LINE_INIT        /* Define this to demonstrate Line
                                     * initialization function. */
#endif /* VP_CC_VCP_880_SERIES */

/*******************************************************************************
 * Common VCP quick start demonstration features                               *
 ******************************************************************************/
#define DEMONSTRATE_DTMF_DECODE     /* Demonstrate DTMF decoding */
#define DEMONSTRATE_METERING        /* Demonstrate metering */
#define HANDLE_DTMF_CMDS            /* Define this to enable processing
                                     * DTMF commands */
#define DEMONSTRATE_GET_LINE_STATE  /* Demonstrate reading line state */
#define DEMONSTRATE_CID             /* Define this to demonstrate CID
                                     * functions */
#define DEMONSTRATE_GET_LOOP_COND   /* Demonstrate get loop condition command */

#define DEMONSTRATE_HOWLER_TONE     /* Demonstrate howler tone generation */

/*******************************************************************************
 * Feature demonstration specific options                                      *
 ******************************************************************************/
#ifdef DEMONSTRATE_CID
#undef CID_DEMO_USE_SENDCID             /* Define to use VpSendCid()
                                         * (vs. VpInitCid) */
#define RING_CAD_STD RING_CAD_US_CID    /* Macro to select the country selection
                                         * for the CID */
#endif /* DEMONSTRATE_CID */

/* Define a generic API results type to collect results */
#define MAX_MPI_LENGTH (15)
typedef union {
    VpLoopCondResultsType loopCond;
    VpTestResultType test;
    VpChkSumType chkSum;
    uint8 lowLevelCmdBuf[MAX_MPI_LENGTH];
    VpRelGainResultsType relGain;
    VpVersionInfoType version;
    VpPacketStatisticsType packet;

    /* Option Results */
    VpOptionPulseType pulseModeOption;
    VpOptionCriticalFltType critFltOption;
    VpOptionZeroCrossType zeroCrossOption;
    uint16 ramp2StbyOption;
    VpOptionDeviceIoType deviceIoOption;
    VpOptionRingControlType ringControlOption;
    VpOptionTimeslotType timeslotOption;
    VpOptionCodecType codecOption;
    VpOptionPcmHwyType pcmHwyOption;
    VpOptionLoopbackType loopbackOption;
    VpOptionLineStateType lineStateOption;
    VpOptionEventMaskType eventMaskOption;
    VpOptionDtmfModeType dtmfModeOption;
    VpDeviceIoAccessDataType deviceIoAccess;
    VpOptionPcmTxRxCntrlType pcmTxRxCntrl;
    VpOptionUsTranscoderType usTranscoderOption;
    VpOptionDsTranscoderType dsTranscoderOption;
    VpOptionEchoCancelerType echoCancelerOption;
    VpOptionUtdCoeffType utdCoeffTypeOption;
} VpApiResultsType;


#ifdef DEMONSTRATE_TEST
#include "vp_test_lib.h" /* Import VoicePath Test Library */

/* Global variables for running the test */
#define TEST_HANDLE (0xabcd)
VptlTestTopologyType topology = {
    .rhl =        5022.812364,
    .rImt =       3010.0,
    .rImtHigh =  75000.0,
    .rIlg =       6040.0,
    .rIlgHigh = 150000.0,
    .rSa =      200000.0,
    .rSb =      200000.0,
    .cTip =         15.0e-9,
    .cRing =        15.0e-9
};
VptlTestInputType testInput;
VptlCalFactorsType calFactors;
VptlTestResultType testResult;
VptlTestTempType testTemp;
VptlTestCtxType testCtx;

static bool TestInputOutput(bool fillInp, VpTestIdType testId,
    VptlTestInputType *pIestInput, VptlTestResultType *pTestResult);

static VptlTestStatusType ImplementOneTest(VptlEventType *pVptlEvent,
    VpTestIdType testId);

static VpStatusType SequenceTests(VptlEventType *pVptlEvent,
    VpTestIdType testId);

static void PrintRes(const char *inputStr, VptlFloatType res);

static bool IsTestSupported(VpTestIdType testId);

static bool IsATestRunning(void);

static VpStatusType StartPcmSelfTest(void);
static VpStatusType ConcludePcmSelfTest(VpEventType *pEvent);

uint8 testLineId = 1;       /* Line for which test needs to be performed */

/* Spcecial flags to indicate the state of the test and the test sequence */
bool onGoingTestSeq = FALSE;
bool onGoingTest = FALSE;


int curTestIndex = 0;
VptlTestIdType testSequence[] = {
    VPTL_TID_OPEN_DC_VOLTAGE,
    VPTL_TID_OPEN_AC_VOLTAGE,
    VPTL_TID_3ELE_RES,
    VPTL_TID_DC_LOOP_RES,
    VPTL_TID_ROH,
    VPTL_TID_REN,
    VPTL_TID_NOISE,
    VPTL_TID_DC_FEED_SELF_TEST,
    VPTL_TID_ON_OFF_HOOK_SELF_TEST,
    VPTL_TID_RINGING_SELF_TEST,
    VPTL_TID_TRANSMISSION_SELF_TEST,
    VPTL_TID_TRANS_HYBRID_LOSS,
    VPTL_TID_RD_LOOP_BAT_COND_TEST,

};
#define MAX_NUM_TEST (sizeof(testSequence)/sizeof(VptlTestIdType))

char *testIdString[] = {
    /* DevNotes: This needs to be changed if the number/order of the
     * test changes */
    "VPTL_TID_3ELE_RES",
    "VPTL_TID_DC_FEED_SELF_TEST",
    "VPTL_TID_NOISE",
    "VPTL_TID_ON_OFF_HOOK_SELF_TEST",
    "VPTL_TID_OPEN_DC_VOLTAGE",
    "VPTL_TID_OPEN_AC_VOLTAGE",
    "VPTL_TID_RD_LOOP_BAT_COND_TEST",
    "VPTL_TID_REN",
    "VPTL_TID_RINGING_SELF_TEST",
    "VPTL_TID_RINGING_MONITOR_TEST",
    "VPTL_TID_ROH",
    "VPTL_TID_TRANS_HYBRID_LOSS",
    "VPTL_TID_TRANSMISSION_SELF_TEST",
    "VPTL_TID_DC_LOOP_RES",
    "VPTL_TID_3ELE_CAP",
    "VPTL_TID_DIALING_SELF_TEST",
    "VPTL_TID_DTMF_PULSE_MSRMNT",
    "VPTL_TID_FOREIGN_AC_CURRENT",
    "VPTL_TID_HOWLER_TEST",
    "VPTL_TID_METERING_SELF_TEST",
    "VPTL_TID_DISTANCE_TO_OPEN",
    "VPTL_TID_TONE_GEN",
    "VPTL_TID_DRAW_BREAK_DIALTONE",
    "VPTL_TID_5ELE_RES",
    "VPTL_TID_ARB_TONE",
    "VPTL_TID_CALIBRATE",
    "VPTL_TID_FUSE_TEST",
    "VPTL_TID_INWRD_CUR",
    "VPTL_TID_SNR_QNTZ_DIST",
};

char *testReturnString[] = {
    "VPTL_STATUS_RUNNING",
    "VPTL_STATUS_DONE",
    "VPTL_STATUS_ABORTED",
    "VPTL_STATUS_ERROR_INVALID_ARG",
    "VPTL_STATUS_ERROR_UNEXPECTED_EVENT",
    "VPTL_STATUS_ERROR_VP_GENERAL",
    "VPTL_STATUS_ERROR_VP_TEST",
    "VPTL_STATUS_ERROR_UNKNOWN",
    "VPTL_STATUS_ERROR_MATH",
    "VPTL_STATUS_ERROR_RESOURCE_NA",
    "VPTL_STATUS_TEST_NOT_SUPPORTED",
};
#endif /* DEMONSTRATE_TEST */

#ifdef DEMONSTRATE_METERING
bool meteringGoingOn[MAX_LINES_ON_BOARD];
#endif /* DEMONSTRATE_METERING */


/* Create an instance of VCP device context and Device Object
 * Note that memory for these instances can be obtained from memory management
 * functions (like malloc). */
VpVcpDeviceObjectType devObj[MAX_DEVICES_ON_BOARD];
VpDevCtxType devCtx[MAX_DEVICES_ON_BOARD];

/* Create instance of VCP line context and line Object
 * Note that memory for these instances can be obtained from memory management
 * functions (like malloc). */
VpLineCtxType lineCtx[MAX_LINES_ON_BOARD];
VpVcpLineObjectType lineObj[MAX_LINES_ON_BOARD];

/* It is assumed here that boot load image is stored in the global memory.
 * Notes about storing boot image:
 * Every system is different in its mechanisms in storing the data in FLASH
 * Some systems might just make use FLASH as just an extension of their overall
 * memory (either directly or through mapping appropriate page). In such systems
 * obtaining boot image is nothing but obtaining right address (and page) where
 * image is stored.
 * Other systems might make use of FLASH file system. In such systems obtaining
 * boot image is like reading a file. In this case the boot image file would
 * have to be read from the FLASH. Flash file system would provide appprpriate
 * calls to perform such read accesses.
 *
 * Other notes about length of boot image that can be passed to VpBootLoad:
 * VpBootImage function can take the entire length of the boot image or it can
 * be called many times with smaller block sizes. This is described in the
 * VP-API specification.
 */
extern VpImagePtrType bootImage; /* pointer to the VCP firmware image in RAM */
extern unsigned long dspBootStreamLen;  /* size of the VCP firmware image */

/* If using a compressed VCP boot image, then VpBootLoad() requires some
 * temporary scratch memory to do the decompression.  Here we allocate it
 * statically, for simplicity. */
static VpScratchMemType scratchMem;

/* Include the appropriate profiles */
#ifdef VP_CC_VCP_790_SERIES
#include "profiles8_790.h"
#else
#include "profiles8_880.h"
#endif /* VP_CC_VCP_790_SERIES */

/*******************************************************************************
 *                      Low level command debug interface                      *
 ******************************************************************************/
#define LOW_LEVEL_CMD_MAX_LEN   30  /* Maximum length of command to be
                                     * excercised */

/* Example low level commands */
#define LOOP_SUPER_VISION_RD    0xc3
#define LOOP_SUPER_VISION_LEN   8

#define DC_FEED_RD              0xc7
#define DC_FEED_LEN             14

#define READ_VERSION            0x73
#define READ_VERSION_LEN        1


static int presentCommandLength;  /* Indicates the length of the command */

/*******************************************************************************
 *                      Quick Start Application related variables              *
 ******************************************************************************/
typedef enum {
    ONHOOK_NO_RING = 0,
    ONHOOK_RING,
    APP_LINEST_ENUM_SIZE = FORCE_STANDARD_C_ENUM_SIZE /* Portability Req. */
} AppLineStateType;

AppLineStateType lineState[MAX_LINES_ON_BOARD];     /* Application current
                                                     * line state */
uint32 timeStampCounter = 0;
/*******************************************************************************
 *                      Function Declarations                                  *
 ******************************************************************************/
static VpStatusType InitAfterBoot(void);
static void PrintEvent(VpEventType *pEvent);
static VpStatusType PrintGetOptionResults(VpEventType *pEvent,
                                                VpApiResultsType *pApiRslts);
static VpStatusType PrintGetLoopCondResults(VpEventType *pEvent,
                                                VpApiResultsType *pApiRslts);
static VpStatusType ProcessDtmfDigits(uint8 channelId, uint8 digit);
static void PrintHelp(void);
static void ResetCommandState(void);
#ifdef DEMONSTRATE_CID
#include "cli_utilities.h"
static uint8* GetCallerDataPointer(bool startFromBegining, uint8 *messageSize);
#endif /* DEMONSTRATE_CID */

/* strings corresponding to VpStatusType */
char *retstring[] = {
    "VP_STATUS_SUCCESS",
    "VP_STATUS_FAILURE",
    "VP_STATUS_FUNC_NOT_SUPPORTED",
    "VP_STATUS_INVALID_ARG",
    "VP_STATUS_MAILBOX_BUSY",
    "VP_STATUS_ERR_VTD_CODE",
    "VP_STATUS_OPTION_NOT_SUPPORTED",
    "VP_STATUS_ERR_VERIFY",
    "VP_STATUS_DEVICE_BUSY",
    "VP_STATUS_MAILBOX_EMPTY",
    "VP_STATUS_ERR_MAILBOX_DATA",
    "VP_STATUS_ERR_HBI",
    "VP_STATUS_ERR_IMAGE",
    "VP_STATUS_IN_CRTCL_SECTN",
    "VP_STATUS_DEV_NOT_INITIALIZED",
    "VP_STATUS_ERR_PROFILE",
    "VP_STATUS_INVALID_VOICE_STREAM"};

/* strings corresponding to VpLineStateType */
char *statestring[] = {
    "STANDBY",
    "TIP_OPEN",
    "ACTIVE",
    "ACTIVE_POLREV",
    "TALK",
    "TALK_POLREV",
    "OHT",
    "OHT_POLREV",
    "DISCONNECT",
    "RINGING",
    "RINGING_POLREV",
    "FXO_OHT",
    "FXO_LOOP_OPEN",
    "FXO_LOOP_CLOSE",
    "FXO_TALK",
    "FXO_RING_GND"
};

/*
 * QuickStartAppMain()
 * This function is the main function for VP-API quick start application.
 * This function could be called from the "main()" after the necessary board
 * specific initialization (based on the board). If there is no such board set
 * up requied, then this function could be renamed as "main" itself.
 *
 * This function implements a simple demo application using the VP-API. It
 * implements the following high level state machine.
 *
 * Hook State:
 * State 1, Reset on hook:      OHT state
 * State 2, First off hook:     Dial tone
 * State 3, First on hook:      Ringing
 * State 4, Second off hook:    Morse code message "CQ...."
 * State 5, Second on hook:     Go back state 1
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * See introduction.
 */
VpStatusType QuickStartAppMain(void)
{
    VpDeviceIdType deviceId = USE_CHIP_SELECT_ID; /* Chip select ID of the VCP */
    VpStatusType status;
    VpEventType event;
    bool eventPresent = FALSE;
    int i;

    /* term_printf("\n DeviceOject = %d, Line Object = %d\n",
                sizeof(VpDevCtxType) + sizeof(VpVcpDeviceObjectType),
                sizeof(VpLineCtxType) + sizeof(VpVcpLineObjectType)); */

    term_printf("\n\nWelcome to the Zarlink Semiconductor VP-API quick start demo"
                " application.\nUsing VP-API version %x.%x.%x\n\n",
                VP_API_VERSION_MAJOR_NUM, VP_API_VERSION_MINOR_NUM,
                VP_API_VERSION_MINI_NUM);

    /* Initialize application variables */
    for (i = 0; i < MAX_LINES_ON_BOARD; i++)  {
      lineState[i] = ONHOOK_NO_RING;
#ifdef DEMONSTRATE_METERING
      meteringGoingOn[i] = FALSE;
#endif /* DEMONSTRATE_METERING */
    }

    /* Reset the DTMF command decoder state machine */
    ResetCommandState();

#ifdef USE_DUMMY_FRAMER
    /* Initialize and configure the framer; This step may be required if clocks
     * provided to the SLAC are derived from the framer device. */
    FliuInit();
#endif /* USE_DUMMY_FRAMER */

#ifdef BOOTLOADING_VCP
    /* Create and initialize VCP device context */
    term_printf("Creating Device Object...");
    status = VpMakeDeviceObject(VP_DEV_VCP_SERIES, deviceId, &devCtx[0],
                                &devObj[0]);
    if(status != VP_STATUS_SUCCESS){ return status; /* Something is wrong */ }
    term_printf("done\n");

    /* Bootload the VCP */
    term_printf("Downloading VCP boot image...");
    status = VpBootLoad(&devCtx[0], VP_BOOT_STATE_FIRSTLAST, bootImage,
                        dspBootStreamLen, &scratchMem, VP_BOOT_MODE_VERIFY);
    if(status != VP_STATUS_SUCCESS){ return status; /* Something is wrong */ }
    term_printf("done\n");

#else /* BOOTLOADING_VCP */
    /* Bootloading performed else where so go ahead and create the line objects
     * and initialize the device */
    status = InitAfterBoot();
    if(status != VP_STATUS_SUCCESS){ return status; /* Something is wrong */ }
#endif /* BOOTLOADING_VCP */


#ifdef POLL_VCP_EVENTS
    /* An eternal loop for Quick Start Application */
    while(1){
        eventPresent = VpGetEvent(&devCtx[0], &event);
        if(event.status != VP_STATUS_SUCCESS) {
            if(event.status < VP_STATUS_NUM_TYPES) {
                term_printf("ERROR: VpGetEvent = %s.\n\n",
                                retstring[event.status]);
            } else {
                term_printf("ERROR: VpGetEvent = %d.\n\n", event.status);
            }
        } else if(eventPresent == TRUE) {
            status = VtdEventHandler(&event);
            if(status != VP_STATUS_SUCCESS) {
#ifdef EXIT_ON_ERROR
                return status; /* Something is wrong */
#else
                if(event.status < VP_STATUS_NUM_TYPES) {
                    term_printf("ERROR: VtdEventHandler = %s.\n\n",
                                    retstring[status]);
                } else {
                    term_printf("ERROR: VtdEventHandler = %d.\n\n", status);
                }
#endif
            }
        }
    }
#else
    return VP_STATUS_SUCCESS;
#endif /* POLL_VCP_EVENTS */
} /* QuickStartAppMain() */


/*
 * InitAfterBoot()
 * This function creats line contexts and also initiates device initialization
 * functions.
 *
 * Preconditions:
 * Assumes the device is boot loaded and boot complete event has been received.
 *
 * Postconditions:
 * Calls VpDeviceInit().
 */
static VpStatusType InitAfterBoot(void)
{
    VpStatusType status;
    int i;
    VpTermType termType;

    /* Create all the line contexts */
    for (i = 0; i < MAX_LINES_ON_BOARD; i++)  {
        term_printf("creating Line Object %d, TermType = ", i);
#ifdef VP_CC_VCP_790_SERIES
        termType = VP_TERM_FXS_TO_TL;
        term_printf("VP_TERM_FXS_TO_TL...");
#else
        termType = VP_TERM_FXS_GENERIC;
        term_printf("VP_TERM_FXS_GENERIC...");
#endif /* VP_CC_VCP_790_SERIES */

        /* Ignore the termination type for now. This functionality will
         * be supported later */
        status = VpMakeLineObject(termType, i, &lineCtx[i], &lineObj[i],
                                    &devCtx[0]);
        if(status != VP_STATUS_SUCCESS){
            return status; /* Something is wrong */
        }
        term_printf("done\n");
    }

#ifdef DEMONSTRATE_CID
    term_printf("Initializing CID Profile table\n");
    status = VpInitProfile(&devCtx[0], VP_PROFILE_CID, VP_PTABLE_INDEX1,
                            &CLI_TYPE1_US[0]);

    if(status != VP_STATUS_SUCCESS){ return status; /* Something is wrong */ }
#endif /* DEMONSTRATE_CID */

    /* Initialize the device */
    term_printf("Initializing VCP...");
    status = VpInitDevice(&devCtx[0], &DEV_PROFILE[0], &AC_600R_FXS[0],
                            &DC_25MA_CC[0], &RING_20HZ_SINE[0], VP_PTABLE_NULL,
                            VP_PTABLE_NULL);
    if(status != VP_STATUS_SUCCESS){ return status; /* Something is wrong */ }
    term_printf("done\n");
    return status;
}

/*
 * VtdEventHandler()
 * This function handles the events from the VCP and takes necessary action
 * based on those events.
 *
 * Preconditions:
 * This function could be called from either the interrupt context or it could
 * be called from a routine that is polling for events. In either case this
 * function SHOULD be passed a pointer to an event that was filled by the
 * VpGetEvent().
 *
 * Postconditions:
 * This line state/device states are modified by this function.
 * This function returns VP_STATUS_SUCCESS if everything went right
 */
VpStatusType VtdEventHandler(VpEventType *pEvent)
{
    VpStatusType status = VP_STATUS_FAILURE;
    int i;
    uint8 chanId = pEvent->channelId;
    VpOptionTimeslotType pcmTimeSlot;
    VpOptionCodecType temp = VP_OPTION_MLAW;
    VpOptionEventMaskType eventMask;
    VpOptionDtmfModeType dtmfMode;
    uint8 dtmfDigit;
    bool leadingEdge = FALSE;
    int timeslot;
#ifdef DEMONSTRATE_GET_LINE_STATE
    VpLineStateType getLineStateVal;
#endif /* DEMONSTRATE_GET_LINE_STATE */

#ifdef DEMONSTRATE_CID
    uint8 callerIdDataSize, *cidDataPtr;
#ifdef CID_DEMO_USE_SENDCID
    static bool firstBreakInterval = FALSE;
#endif
#endif /* DEMONSTRATE_CID */
    VpApiResultsType apiResults;

    dtmfMode.dtmfControlMode = VP_OPTION_DTMF_DECODE_ON;
    dtmfMode.direction = VP_DIRECTION_US;

    /* Read the results if there is one */
    if(pEvent->hasResults == TRUE) {
        status = VpGetResults(pEvent, &apiResults);
        if(status != VP_STATUS_SUCCESS){
            return status;
        }
    }

#ifdef DEMONSTRATE_TEST
    /* If a test is being performed, pass all the events to the test handler */
    if ((IsATestRunning() == TRUE) && (pEvent->channelId == testLineId)) {
        VptlEventType vptlEvent;
        vptlEvent.pResult = &apiResults;
        vptlEvent.pVpEvent = pEvent;
        status = SequenceTests(&vptlEvent, 0); /* Don't care about the testId */
        return status; /* Comsume events here */
    }
#endif /* DEMONSTRATE_TEST */


    /* Take necessary action based on the event */
    if(pEvent->eventCategory == VP_EVCAT_RESPONSE) {
        switch  (pEvent->eventId) {
            case VP_DEV_EVID_BOOT_CMP: {
                /* Read the results associated with this event */
                term_printf("Received event: VP_EVID_BOOT_CMP\n");
                term_printf("Checksum Data:\n");
                term_printf("VpChkSumType.loadChkSum = 0x%lx\n",
                                apiResults.chkSum.loadChecksum);
                term_printf("VpChkSumType.vInfo.vtdRevCode = 0x%x\n",
                                apiResults.chkSum.vInfo.vtdRevCode);
                term_printf("VpChkSumType.vInfo.swProductId = 0x%x\n",
                                apiResults.chkSum.vInfo.swProductId);
                term_printf("VpChkSumType.vInfo.swVerMajor = 0x%x\n",
                                apiResults.chkSum.vInfo.swVerMajor);
                term_printf("VpChkSumType.vInfo.swVerMinor = 0x%x\n",
                                apiResults.chkSum.vInfo.swVerMinor);

                /* Create Line Objects and Initialize the device */
                status = InitAfterBoot();
                break;
            }
            case VP_DEV_EVID_DEV_INIT_CMP: {
                term_printf("Received event: VP_DEV_EVID_DEV_INIT_CMP(%d)\n",
                                chanId);

                /* Veryify Calibration was successful */
                term_printf("Device initialization complete Event Data: 0x%x\n",
                                pEvent->eventData);

                if(pEvent->eventData == 0) {
                    term_printf("ERROR: Device initialization Failed; "
                                "NO SLAC found; Exiting...\n");

                    status = VP_STATUS_FAILURE;
                    break;
                }

                /* Setup the event mask that is different compared to default
                 * masks; API masks all event (other than non-maskable). So
                 * Unmask all necessary events */
                eventMask.faults = VP_EVCAT_FAULT_UNMASK_ALL;
                eventMask.signaling = VP_EVCAT_SIGNALING_UNMASK_ALL;
                eventMask.response = VP_EVCAT_RESPONSE_UNMASK_ALL;
                eventMask.test = VP_EVCAT_TEST_UNMASK_ALL;
                eventMask.process = VP_EVCAT_PROCESS_UNMASK_ALL;
                eventMask.fxo = VP_EVENT_MASK_ALL; /* Not applicable to VCP */
                eventMask.packet = VP_EVENT_MASK_ALL;  /* Not applicable */

                term_printf("Setting up events masks...\n");
                status = VpSetOption(VP_NULL, pEvent->pDevCtx,
                                      VP_OPTION_ID_EVENT_MASK, &eventMask);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }

#ifdef DEMONSTRATE_LINE_INIT
                /* Initialize a line; Note that this application is not written
                 * to handle event generated as a consequence of this function
                 * call properly. */
                i = 1;
                term_printf("Initialize a line %d", i);
                status = VpInitLine(&lineCtx[i], &AC_600R_FXS[0],
                                    &DC_25MA_CC[0], &RING_20HZ_SINE[0]);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                term_printf("done\n");
#endif  /* DEMONSTRATE_LINE_INIT */

                /* Configure a line differently */
                i = 1;
                term_printf("Configuring line %d for different DC feed, "
                                "Ringing Profile...", i);
                status = VpConfigLine(&lineCtx[i], &AC_600R_FXS[0],
                                    &DC_25MA_CC[0], &RING_20HZ_SINE[0]);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                term_printf("done\n");

                /* Initialize all the lines now ... */
                for(i=0; i< MAX_LINES_ON_BOARD; i++){

                    term_printf("Line %d ", i);

                    /* Initialize Ringing parameters */
                    term_printf("Ring Init, ");
#ifdef DEMONSTRATE_CID
#ifdef CID_DEMO_USE_SENDCID
                    status = VpInitRing(&lineCtx[i], &RING_CAD_NOCID[0],
                                            VP_PTABLE_NULL);
#else /* !CID_DEMO_USE_SENDCID */
                    status = VpInitRing(&lineCtx[i], &RING_CAD_STD[0],
                                            VP_PTABLE_INDEX1);
#endif /* !CID_DEMO_USE_SENDCID */
#else /* !DEMONSTRATE_CID */
                    status = VpInitRing(&lineCtx[i], &RING_CAD_NOCID[0],
                                            VP_PTABLE_NULL);
#endif /* DEMONSTRATE_CID */
                    if(status != VP_STATUS_SUCCESS){
                        break; /* Something is wrong */
                    }

                    if(i >= LINES_FROM_WHICH_LCAS_IS_PRESENT) {
                        /* Place the LCAS in normal state */
                        term_printf("LCAS State->Normal, ");
                        status = VpSetRelayState(&lineCtx[i], VP_RELAY_NORMAL);
                        if(status != VP_STATUS_SUCCESS){
                            break; /* Something is wrong */
                        }
                    }

                    /* Configure the PCM time slot; Presently there is one to
                     * one mapping of channels and time slots used by them. This
                     * could be modified(if needed) as appropriate. */
                    timeslot = 2 * i;   /* Leave the slots in between for
                                         * Testing */
                    pcmTimeSlot.rx = timeslot;
                    pcmTimeSlot.tx = timeslot;
                    term_printf("TimeSlot->%d, ", timeslot);
                    status = VpSetOption(&lineCtx[i], VP_NULL,
                                          VP_OPTION_ID_TIMESLOT, &pcmTimeSlot);
                    if(status != VP_STATUS_SUCCESS){
                        break; /* Something is wrong */
                    }

                    /* Configure CODEC type */
                    term_printf("CODEC->Mu-Law, ");
                    status = VpSetOption(&lineCtx[i], VP_NULL,
                                            VP_OPTION_ID_CODEC, &temp);
                    if(status != VP_STATUS_SUCCESS){
                        break; /* Something is wrong */
                    }

#ifdef DEMONSTRATE_DTMF_DECODE
                    if(i<MAX_DTMF_DETECTION_LINES){
                        /* Enable DTMF detection for certain lines */
                        term_printf("DTMF->ON, ");
                        status = VpSetOption(&lineCtx[i], VP_NULL,
                                        VP_OPTION_ID_DTMF_MODE, &dtmfMode);
                        term_printf("DTMF Status=0x%.8lx, ",
                                        dtmfMode.dtmfDetectionSetting);
                        term_printf("DTMF Remaing Resources=%d, ",
                                        dtmfMode.dtmfResourcesRemaining);
                        if(status != VP_STATUS_SUCCESS){
                            break; /* Something is wrong */
                        }
                    }

                    if(i== 1){
                        /* Enable DTMF detection for certain lines */
                        term_printf("\nDTMF->OFF, ");
                        dtmfMode.dtmfControlMode = VP_OPTION_DTMF_DECODE_OFF;
                        status = VpSetOption(&lineCtx[i], VP_NULL,
                                        VP_OPTION_ID_DTMF_MODE, &dtmfMode);
                        term_printf("DTMF Status=0x%.8lx, ",
                                        dtmfMode.dtmfDetectionSetting);
                        term_printf("DTMF Remaing Resources=%d, \n",
                                        dtmfMode.dtmfResourcesRemaining);
                        if(status != VP_STATUS_SUCCESS){
                            break; /* Something is wrong */
                        }
                        dtmfMode.dtmfControlMode = VP_OPTION_DTMF_DECODE_ON;
                    }

#endif /* DEMONSTRATE_DTMF_DECODE */
#ifdef DEMONSTRATE_GET_LINE_STATE
                    /* Get current line state (should be disconnect) */
                    status = VpGetLineState(&lineCtx[i], &getLineStateVal);
                    if(status != VP_STATUS_SUCCESS){
                        break; /* Something is wrong */
                    }

                    /* Place the Line in OHT state */
                    term_printf("Line State: %s -> OHT\n",
                                    statestring[getLineStateVal]);
                    status = VpSetLineState(&lineCtx[i], VP_LINE_OHT);
#else
                    /* Place the Line in OHT state */
                    term_printf("Line State: OHT\n");
                    status = VpSetLineState(&lineCtx[i], VP_LINE_OHT);
#endif /* DEMONSTRATE_GET_LINE_STATE */
                    if(status != VP_STATUS_SUCCESS){
                        break; /* Something is wrong */
                    }
                }

                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }

#ifdef DEMONSTRATE_GET_LOOP_COND
                /*Issue a low level command */
                term_printf("Issuing Get loop condition command...");
                status = VpGetLoopCond(&lineCtx[0], 0);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                term_printf("done\n");
#endif  /* DEMONSTRATE_GET_LOOP_COND */

                break;
            }
            case VP_LINE_EVID_LINE_INIT_CMP: {
                term_printf("Received event: VP_DEV_EVID_LINE_INIT_CMP(%d)\n",
                                chanId);
                status = VP_STATUS_SUCCESS;
                break;
            }
            case VP_LINE_EVID_LLCMD_RX_CMP: {
                term_printf("Received event: VP_EVID_LLCMD_RX_CMP(%d)\n",
                            chanId);
                term_printf("Data = ");
                for(i = 0; i<presentCommandLength; i++){
                    term_printf("%2.2hX ", apiResults.lowLevelCmdBuf[i]);
                }
                term_printf("\n");
                break;
            }
            case VP_LINE_EVID_LLCMD_TX_CMP: {
                term_printf("Received event: VP_EVID_LLCMD_TX_CMP(%d)\n",
                                chanId);
                status = VP_STATUS_SUCCESS;
                break;
            }
            case VP_DEV_EVID_DNSTR_MBOX:    {
                term_printf("Received event: VP_DEV_EVID_DNSTR_MBOX(%d)\n",
                    pEvent->eventData);
                status = VP_STATUS_SUCCESS;
                break;
            }
            case VP_LINE_EVID_RD_OPTION: {
                /* Print the results */
                status = PrintGetOptionResults(pEvent, &apiResults);
                break;
            }
            case VP_LINE_EVID_RD_LOOP: {
                /* Print the results */
                status = PrintGetLoopCondResults(pEvent, &apiResults);
                break;
            }
        }
    }else if(pEvent->eventCategory == VP_EVCAT_SIGNALING) {
        switch  (pEvent->eventId) {
            case VP_LINE_EVID_HOOK_ON: {
                term_printf("Received event: VP_EVID_HOOK_ON(%d); TS = %u\n",
                            chanId, pEvent->parmHandle);
#ifdef DEMONSTRATE_GET_LINE_STATE
                /* Get current line state */
                status = VpGetLineState(pEvent->pLineCtx, &getLineStateVal);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
#endif /* DEMONSTRATE_GET_LINE_STATE */

                /* Toggle between on-hook ring and on-hook no ring state */
                if(lineState[chanId] == ONHOOK_NO_RING){

#if defined(DEMONSTRATE_CID) && !defined(CID_DEMO_USE_SENDCID)
                    /* Initialize CallerID data */
                    cidDataPtr = GetCallerDataPointer(TRUE, &callerIdDataSize);
                    if((callerIdDataSize > 0) && (cidDataPtr != VP_NULL)) {
                        term_printf("Writing CID data (VpInitCid()) "
                                        "size = %d\n", callerIdDataSize);
                        status = VpInitCid(pEvent->pLineCtx, callerIdDataSize,
                            cidDataPtr);

                        if(status != VP_STATUS_SUCCESS){
                            break; /* Something is wrong */
                        }
                    }
#endif
                    lineState[chanId] = ONHOOK_RING;
#ifdef DEMONSTRATE_GET_LINE_STATE
                    term_printf("Line State: %s -> Ringing\n",
                                    statestring[getLineStateVal]);
#else
                    term_printf("Line State: Ringing\n");
#endif /* DEMONSTRATE_GET_LINE_STATE */
#if defined(DEMONSTRATE_CID) && defined(CID_DEMO_USE_SENDCID)
                    firstBreakInterval = TRUE;
#endif
                    status = VpSetLineState(pEvent->pLineCtx, VP_LINE_RINGING);
                    if(status != VP_STATUS_SUCCESS){
                        break; /* Something is wrong */
                    }

                } else {
                    lineState[chanId] = ONHOOK_NO_RING;
#ifdef DEMONSTRATE_GET_LINE_STATE
                    term_printf("Line State: %s -> OHT\n",
                                    statestring[getLineStateVal]);
#else
                    term_printf("Line State: OHT\n");
#endif /* DEMONSTRATE_GET_LINE_STATE */
                    status = VpSetLineState(pEvent->pLineCtx, VP_LINE_OHT);
                    if(status != VP_STATUS_SUCCESS){
                        break; /* Something is wrong */
                    }
                }
                break;
            }
            case VP_LINE_EVID_HOOK_OFF: {
                term_printf("Received event: VP_EVID_HOOK_OFF(%d); TS = %u\n",
                            chanId, pEvent->parmHandle);

#ifdef DEMONSTRATE_GET_LINE_STATE
                /* Get current line state */
                status = VpGetLineState(pEvent->pLineCtx, &getLineStateVal);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }

                term_printf("Line State: %s -> Talk, ",
                                statestring[getLineStateVal]);
#else
                term_printf("Line State: Talk, ");
#endif /* DEMONSTRATE_GET_LINE_STATE */
                status = VpSetLineState(pEvent->pLineCtx, VP_LINE_TALK);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }

                if(lineState[chanId] == ONHOOK_RING) {
#ifdef VP_CC_VCP_790_SERIES
                    term_printf("Feeding message\n");
                    /* User picks up in response to ringing; Generate a tone
                     * message instead of dial tone */
                    status = VpSetLineTone(pEvent->pLineCtx, &TONE_ONEKHZ_L[0],
                                            &TONE_CAD_SIT[0], VP_NULL);
#else
                    term_printf("Feeding ringback\n");
                    /* User picks up in response to ringing; Generate a tone
                     * message instead of dial tone */
                    status = VpSetLineTone(pEvent->pLineCtx, &TONE_RINGBACK[0],
                                            TONE_CAD_RNGBACK, VP_NULL);
#endif
                } else {
                    term_printf("Feeding Dial Tone\n");
                    status = VpSetLineTone(pEvent->pLineCtx, &TONE_DIAL[0],
                                            VP_PTABLE_NULL, VP_NULL);
                }
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                break;

            }
            case VP_LINE_EVID_DTMF_DIG: {
                /* DTMF digit detected; Print the digit */
                dtmfDigit = pEvent->eventData & 0x0f;
                if((pEvent->eventData & 0x10) == 0) {
                    leadingEdge = FALSE;
                } else {
                    leadingEdge = TRUE;
                }
#ifdef HANDLE_DTMF_CMDS
                term_printf("Line: %d, DTMF Digit: %d, Timestamp %u: "
                            "Edge: %s\n",
                            pEvent->channelId, dtmfDigit,
                            pEvent->parmHandle,
                            leadingEdge ? "Leading":"Trailing");
                if(leadingEdge == TRUE) {
                    status = ProcessDtmfDigits(pEvent->channelId, dtmfDigit);
                } else {
                    status = VP_STATUS_SUCCESS; /* Nothing to do */
                }
#else
                term_printf("Line: %d, DTMF Digit: %d, Timestamp %u: "
                            "Edge: %s\n",
                            pEvent->channelId, dtmfDigit,
                            pEvent->parmHandle,
                            leadingEdge ? "Leading":"Trailing");
                status = VP_STATUS_SUCCESS; /* Event is consumed */
#endif /* HANDLE_DTMF_CMDS */
                break;
            }
            case VP_DEV_EVID_TS_ROLLOVER: {
                timeStampCounter++;
                /* Uncomment if necessary */
                term_printf("Received event: VP_DEV_EVID_TS_ROLLOVER; Count: "
                                "%lu \n", timeStampCounter);
#ifdef DEMONSTRATE_TEST
                /* Initiate testing */
                if(timeStampCounter == 1) {
                    term_printf("\nStarting scheduled tests at TimeStamp "
                        "counter = 1\n");
                    status = StartPcmSelfTest();
                    break;
                } else if(timeStampCounter == 2) {
                    term_printf("\nStarting scheduled tests at TimeStamp "
                        "counter = 2\n");

                    /* Start the test sequence */
                    status = SequenceTests(VP_NULL, VPTL_NUM_TID_CODES);
                    break;

                } else {
                    status = VP_STATUS_SUCCESS; /* Event is consumed */
                    break;
                }
#else
                status = VP_STATUS_SUCCESS; /* Event is consumed */
                break;
#endif /* DEMONSTRATE_TEST */
            }
        }
    } else if(pEvent->eventCategory == VP_EVCAT_PROCESS) {
        switch  (pEvent->eventId) {
            case VP_LINE_EVID_RING_CAD: {
                term_printf("Received event: VP_LINE_EVID_RING_CAD(%d)",
                                chanId);
                status = VP_STATUS_SUCCESS;
                if (pEvent->eventData == VP_RING_CAD_MAKE) {
                    term_printf("make \n");
                } else if (pEvent->eventData == VP_RING_CAD_BREAK) {
                    term_printf("break \n");
#if defined(DEMONSTRATE_CID) && defined(CID_DEMO_USE_SENDCID)
                    /* Demonstrating U.S. Type I caller ID; wait for the first
                     * off interval in the ringing cadence.  Do not send the
                     * data again in subsequent off intervals. */
                    if (!firstBreakInterval) {
                        break;
                    }
                    firstBreakInterval = FALSE;
                    /* Send first chunk of CallerID data */
                    cidDataPtr = GetCallerDataPointer(TRUE, &callerIdDataSize);
                    if((callerIdDataSize > 0) && (cidDataPtr != VP_NULL)) {
                        term_printf("Writing CID data (VpSendCid()) "
                                        "size = %d\n", callerIdDataSize);
                        status = VpSendCid(pEvent->pLineCtx, callerIdDataSize,
                            VP_PTABLE_INDEX1, cidDataPtr);
                    }
#endif
                } else {
                    term_printf("done; ");
                }
                break;
            }
#ifdef DEMONSTRATE_CID
            case VP_LINE_EVID_CID_DATA: {
                term_printf("Received event: VP_LINE_EVID_CID_DATA(%d); - %s\n",
                                chanId,
                                (pEvent->eventData == VP_CID_DATA_TX_DONE) ?
                                 "CID Complete":"Is there any more data ?");

                    /* Get the subsequent CID data blocks */
                    cidDataPtr = GetCallerDataPointer(FALSE, &callerIdDataSize);
                    if((callerIdDataSize > 0) && (cidDataPtr != VP_NULL)) {

                        if(pEvent->eventData == VP_CID_DATA_TX_DONE) {
                            term_printf("Abroting CID due to OFF-Hook\n");
                            status = VP_STATUS_SUCCESS;
                            break;
                        }

                        term_printf("Writing CID data (VpContinueCid()) "
                                        "size=%d\n", callerIdDataSize);
                        status = VpContinueCid(pEvent->pLineCtx,
                                        callerIdDataSize, cidDataPtr);
                        if(status != VP_STATUS_SUCCESS){
                            break; /* Something is wrong */
                        }
                    }

                status = VP_STATUS_SUCCESS;
                break;
            }
#endif /* DEMONSTRATE_CID */
            case VP_LINE_EVID_MTR_CAD: {
                term_printf("Received event: VP_LINE_EVID_MTR_CAD(%d) "
                    "Number of pulses sent = %d\n", chanId, pEvent->eventData);
                status = VP_STATUS_SUCCESS;
                break;
            }
            case VP_LINE_EVID_MTR_CMP: {
                term_printf("Received event: VP_LINE_EVID_MTR_CMP(%d)\n",
                                chanId);
                meteringGoingOn[chanId] = FALSE;
                status = VP_STATUS_SUCCESS;
                break;
            }
            case VP_LINE_EVID_MTR_ABORT: {
                term_printf("Received event: VP_LINE_EVID_MTR_ABORT(%d) "
                    "Number of pulses sent = %d\n", chanId, pEvent->eventData);
                meteringGoingOn[chanId] = FALSE;/* Aborted for whatever reason*/
                status = VP_STATUS_SUCCESS;
                break;
            }
            case VP_LINE_EVID_TONE_CAD: {
                term_printf("Received event: VP_LINE_EVID_TONE_CAD(%d)"
                    "Event Data = %d\n", chanId, pEvent->eventData);
                status = VP_STATUS_SUCCESS;
                break;
            }
        }
    }
#ifdef DEMONSTRATE_TEST
    else if(pEvent->eventCategory == VP_EVCAT_TEST) {
        switch (pEvent->eventId) {
            case VP_DEV_EVID_STEST_CMP: {
                status = ConcludePcmSelfTest(pEvent);
                break;
            }
        }
    }
#endif /* DEMONSTRATE_TEST */

    if(status != VP_STATUS_SUCCESS){

        /* Either the event was not handled or there is error in handling the
         * event. Print the event. */
        term_printf("ERROR: Unknown event or Error while handling the event; "
                    "Printing the event...");
        PrintEvent(pEvent);

        /* If Event has a result and it has not been handled until now,
         * Clear it */
        if(pEvent->hasResults == TRUE) {
            term_printf("WARNING: Unhandled results, Clearing it...\n");
            VpClearResults(pEvent->pDevCtx);
        }
    }

    return status;
} /* VtdEventHandler() */


/*
 * PrintEvent()
 * This function an un-handled event.
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * None
 */
static void PrintEvent(
    VpEventType *pEvent)
{
    char *eventCategory[] = {
                                "VP_EVCAT_FAULT",
                                "VP_EVCAT_SIGNALING",
                                "VP_EVCAT_RESPONSE",
                                "VP_EVCAT_TEST",
                                "VP_EVCAT_PROCESS",
                                "VP_EVCAT_FXO",
                                "VP_EVCAT_PACKET",
    };

    if(pEvent->eventCategory < VP_NUM_EVCATS) {
        term_printf("VpEvent Cat:Id = %s:",
                        eventCategory[pEvent->eventCategory]);
    } else {
        term_printf("VpEvent Cat:Id = 0x%x:",
                        pEvent->eventCategory);
    }

    if(pEvent->eventCategory == VP_EVCAT_FAULT) {
        switch (pEvent->eventId) {
            case VP_DEV_EVID_BAT_FLT:   {
                term_printf("VP_DEV_EVID_BAT_FLT\n"); break;
            }
            case VP_DEV_EVID_CLK_FLT:   {
                term_printf("VP_DEV_EVID_CLK_FLT\n"); break;
            }
            case VP_LINE_EVID_THERM_FLT:{
                term_printf("VP_LINE_EVID_THERM_FLT\n"); break;
            }
            case VP_LINE_EVID_DC_FLT:   {
                term_printf("VP_DEVP_LINE_EVID_DC_FLT\n"); break;
            }
            case VP_LINE_EVID_AC_FLT:   {
                term_printf("VP_LINE_EVID_AC_FLT\n"); break;
            }
            case VP_DEV_EVID_EVQ_OFL_FLT: {
                term_printf("VP_DEV_EVID_EVQ_OFL_FLT\n"); break;
            }
            case VP_DEV_EVID_WDT_FLT: {
                term_printf("VP_DEV_EVID_WDT_FLT\n"); break;
            }
            default: {
                term_printf("0x%x\n", pEvent->eventId); break;
            }
        }
    } else if(pEvent->eventCategory == VP_EVCAT_SIGNALING) {
        switch (pEvent->eventId) {
            case VP_LINE_EVID_HOOK_OFF:   {
                term_printf("VP_LINE_EVID_HOOK_OFF\n"); break;
            }
            case VP_LINE_EVID_HOOK_ON:   {
                term_printf("VP_LINE_EVID_HOOK_ON\n"); break;
            }
            case VP_LINE_EVID_GKEY_DET:{
                term_printf("VP_LINE_EVID_GKEY_DET\n"); break;
            }
            case VP_LINE_EVID_GKEY_REL:   {
                term_printf("VP_LINE_EVID_GKEY_REL\n"); break;
            }
            case VP_LINE_EVID_FLASH:   {
                term_printf("VP_LINE_EVID_FLASH\n"); break;
            }
            case VP_LINE_EVID_STARTPULSE: {
                term_printf("VP_LINE_EVID_STARTPULSE\n"); break;
            }
            case VP_LINE_EVID_DTMF_DIG: {
                term_printf("VP_LINE_EVID_DTMF_DIG\n"); break;
            }
            case VP_LINE_EVID_PULSE_DIG: {
                term_printf("VP_LINE_EVID_PULSE_DIG\n"); break;
            }
            case VP_LINE_EVID_MTONE: {
                term_printf("VP_LINE_EVID_MTONE\n"); break;
            }
            case VP_DEV_EVID_TS_ROLLOVER: {
                term_printf("VP_DEV_EVID_TS_ROLLOVER\n"); break;
            }
            default: {
                term_printf("0x%x\n", pEvent->eventId); break;
            }
        }
    } else if(pEvent->eventCategory == VP_EVCAT_RESPONSE) {
        switch (pEvent->eventId) {
            case VP_DEV_EVID_BOOT_CMP:   {
                term_printf("VP_DEV_EVID_BOOT_CMP\n"); break;
            }
            case VP_LINE_EVID_LLCMD_TX_CMP:   {
                term_printf("VP_LINE_EVID_LLCMD_TX_CMP\n"); break;
            }
            case VP_LINE_EVID_LLCMD_RX_CMP:{
                term_printf("VP_LINE_EVID_LLCMD_RX_CMP\n"); break;
            }
            case VP_DEV_EVID_DNSTR_MBOX:   {
                term_printf("VP_DEV_EVID_DNSTR_MBOX\n"); break;
            }
            case VP_LINE_EVID_RD_OPTION:   {
                term_printf("VP_LINE_EVID_RD_OPTION\n"); break;
            }
            case VP_LINE_EVID_RD_LOOP: {
                term_printf("VP_LINE_EVID_RD_LOOP\n"); break;
            }
            case VP_EVID_CAL_CMP: {
                term_printf("VP_EVID_CAL_CMP\n"); break;
            }
            case VP_EVID_CAL_BUSY: {
                term_printf("VP_EVID_CAL_BUSY\n"); break;
            }
            case VP_LINE_EVID_GAIN_CMP: {
                term_printf("VP_LINE_EVID_GAIN_CMP\n"); break;
            }
            case VP_DEV_EVID_DEV_INIT_CMP: {
                term_printf("VP_DEV_EVID_DEV_INIT_CMP\n"); break;
            }
            case VP_LINE_EVID_LINE_INIT_CMP: {
                term_printf("VP_LINE_EVID_LINE_INIT_CMP\n"); break;
            }
            default: {
                term_printf("0x%x\n", pEvent->eventId); break;
            }
        }
    } else if(pEvent->eventCategory == VP_EVCAT_TEST) {
        switch (pEvent->eventId) {
            case VP_LINE_EVID_TEST_CMP:   {
                term_printf("VP_LINE_EVID_TEST_CMP\n"); break;
            }
            case VP_LINE_EVID_DTONE_DET:   {
                term_printf("VP_LINE_EVID_DTONE_DET\n"); break;
            }
            case VP_LINE_EVID_DTONE_LOSS:{
                term_printf("VP_LINE_EVID_DTONE_LOSS\n"); break;
            }
            case VP_DEV_EVID_STEST_CMP:   {
                term_printf("VP_DEV_EVID_STEST_CMP\n"); break;
            }
            case VP_DEV_EVID_CHKSUM:   {
                term_printf("VP_DEV_EVID_CHKSUM\n"); break;
            }
            default: {
                term_printf("0x%x\n", pEvent->eventId); break;
            }
        }
    } else if(pEvent->eventCategory == VP_EVCAT_PROCESS) {
        switch (pEvent->eventId) {
            case VP_LINE_EVID_MTR_CMP:   {
                term_printf("VP_LINE_EVID_MTR_CMP\n"); break;
            }
            case VP_LINE_EVID_MTR_ABORT:   {
                term_printf("VP_LINE_EVID_MTR_ABORT\n"); break;
            }
            case VP_LINE_EVID_CID_DATA:{
                term_printf("VP_LINE_EVID_CID_DATA\n"); break;
            }
            case VP_LINE_EVID_RING_CAD:   {
                term_printf("VP_LINE_EVID_RING_CAD\n"); break;
            }
            default: {
                term_printf("0x%x\n", pEvent->eventId); break;
            }
        }
    } else {
        term_printf("0x%x\n", pEvent->eventId);
    }
    term_printf("VpEventType.channelId = 0x%x\n", pEvent->channelId);
    term_printf("VpEventType.deviceId = 0x%x\n", pEvent->deviceId);
    term_printf("VpEventType.eventData = 0x%x\n", pEvent->eventData);
    term_printf("VpEventType.parmHandle = 0x%x\n", pEvent->parmHandle);
    term_printf("VpEventType.hasResults = 0x%x\n", pEvent->hasResults);
}

/*
 * PrintGetOptionResults()
 * This function prints the Get option results.
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * None
 */
static VpStatusType
PrintGetOptionResults(
    VpEventType *pEvent,
    VpApiResultsType *pApiRslts)
{
    uint16 optionId;
    VpStatusType status = VP_STATUS_SUCCESS;

    optionId = pEvent->eventData;
    if(optionId > VP_OPTION_ID_DTMF_MODE) {
        term_printf("Invalid Event: VP_LINE_EVID_RD_OPTION:%d\n", optionId);
        return VP_STATUS_FAILURE;
    } else {
        term_printf("Received event: VP_LINE_EVID_RD_OPTION:%d\n", optionId);
    }

    switch (optionId) {
        case VP_DEVICE_OPTION_ID_PULSE: {
            term_printf("Option Read: VpOptionPulseType.breakMin = %u\n",
                            pApiRslts->pulseModeOption.breakMin);
            term_printf("Option Read: VpOptionPulseType.breakMax = %u\n",
                            pApiRslts->pulseModeOption.breakMax);
            term_printf("Option Read: VpOptionPulseType.makeMin = %u\n",
                            pApiRslts->pulseModeOption.makeMin);
            term_printf("Option Read: VpOptionPulseType.makeMax = %u\n",
                            pApiRslts->pulseModeOption.makeMax);
            term_printf("Option Read: VpOptionPulseType.interDigitMin = "
                            "%u\n", pApiRslts->pulseModeOption.interDigitMin);
            term_printf("Option Read: VpOptionPulseType.flashMin = %u\n",
                            pApiRslts->pulseModeOption.flashMin);
            term_printf("Option Read: VpOptionPulseType.flashMax = %u\n",
                            pApiRslts->pulseModeOption.flashMax);
            break;
        }
        case VP_OPTION_ID_EVENT_MASK: {
            term_printf("Option Read: VpOptionEventMaskType.faults = "
                            "0x%.4x\n", pApiRslts->eventMaskOption.faults);
            term_printf("Option Read: VpOptionEventMaskType.signaling = "
                            "0x%.4x\n", pApiRslts->eventMaskOption.signaling);
            term_printf("Option Read: VpOptionEventMaskType.response = "
                            "0x%.4x\n", pApiRslts->eventMaskOption.response);
            term_printf("Option Read: VpOptionEventMaskType.test = "
                            "0x%.4x\n", pApiRslts->eventMaskOption.test);
            term_printf("Option Read: VpOptionEventMaskType.process = "
                            "0x%.4x\n", pApiRslts->eventMaskOption.process);
            term_printf("Option Read: VpOptionEventMaskType.fxo = "
                            "0x%.4x\n", pApiRslts->eventMaskOption.fxo);
            term_printf("Option Read: VpOptionEventMaskType.packet = "
                            "0x%.4x\n", pApiRslts->eventMaskOption.packet);
            break;
        }
        default: {
            term_printf("ERROR: Option Read: OptionId = %d Not yet "
                            "implemented \n", optionId);
            break;
        }
    }

    return status;
} /* PrintGetOptionResults() */

/* Print the Get Loop Condition results */
static VpStatusType
PrintGetLoopCondResults(
    VpEventType *pEvent,
    VpApiResultsType *pApiRslts)
{
    char *batName[] = {"VP_BATTERY_UNDEFINED", "VP_BATTERY_1",
                        "VP_BATTERY_2", "VP_BATTERY_3"};
    char *feedRegName[] = {"VP_DF_UNDEFINED", "VP_DF_ANTI_SAT_REG",
                        "VP_DF_CNST_CUR_REG", "VP_DF_RES_FEED_REG"};

    term_printf("LoopCond: VpLoopCondResultsType.rloop = %d\n",
                    pApiRslts->loopCond.rloop);
    term_printf("LoopCond: VpLoopCondResultsType.ilg = %d\n",
                    pApiRslts->loopCond.ilg);
    term_printf("LoopCond: VpLoopCondResultsType.imt = %d\n",
                    pApiRslts->loopCond.imt);
    term_printf("LoopCond: VpLoopCondResultsType.vsab = %d\n",
                    pApiRslts->loopCond.vsab);
    term_printf("LoopCond: VpLoopCondResultsType.vbat1 = %d\n",
                    pApiRslts->loopCond.vbat1);
    term_printf("LoopCond: VpLoopCondResultsType.vbat2 = %d\n",
                    pApiRslts->loopCond.vbat2);
    term_printf("LoopCond: VpLoopCondResultsType.vbat3 = %d\n",
                    pApiRslts->loopCond.vbat3);
    term_printf("LoopCond: VpLoopCondResultsType.mspl = %d\n",
                    pApiRslts->loopCond.mspl);

    if(pApiRslts->loopCond.selectedBat < VP_NUM_BATTERY_TYPES) {
        term_printf("LoopCond: VpLoopCondResultsType.selectedBat = %s\n",
                        batName[pApiRslts->loopCond.selectedBat]);
    } else {
        term_printf("LoopCond: VpLoopCondResultsType.selectedBat = %d\n",
                        pApiRslts->loopCond.selectedBat);
    }
    if(pApiRslts->loopCond.dcFeedReg < VP_NUM_DC_FEED_TYPES) {
        term_printf("LoopCond: VpLoopCondResultsType.dcFeedReg = %s\n",
                        feedRegName[pApiRslts->loopCond.dcFeedReg]);
    } else {
        term_printf("LoopCond: VpLoopCondResultsType.dcFeedReg = %d \n",
                        pApiRslts->loopCond.dcFeedReg);
    }

    return VP_STATUS_SUCCESS;

} /* PrintGetLoopCondResults() */

/*
 * ProcessDtmfDigits()
 * This function Processes detected DTMF digits.
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * None
 */
#define MAX_USER_CMD_SIZE       (25)
#define CMD_END_DIGIT           (VP_DIG_ASTER)
unsigned char command[MAX_LINES_ON_BOARD][MAX_USER_CMD_SIZE];
bool commandInProgress[MAX_LINES_ON_BOARD];
unsigned int commandLength[MAX_LINES_ON_BOARD];

static VpStatusType ProcessDtmfDigits(
    uint8 channelId,
    uint8 digit)
{
    unsigned int commandOpcode;
    int i;
    VpStatusType status = VP_STATUS_SUCCESS;
    uint16 getOptionId = 0;
#ifdef DEMONSTRATE_TEST
    uint16 line, testId;
#endif /* DEMONSTRATE_TEST */
#ifdef DEMONSTRATE_METERING
    uint16 numMeter;
#endif /* DEMONSTRATE_METERING */
#ifdef DEMONSTRATE_HOWLER_TONE
    VpProfileDataType anyTonePrf[] = {0x00, 0x02, 0x00, 12, 0x00, 0x00, 0x00,
                        0x00, 0x11, 0x8A, 0x05, 0x12, 0x16, 0x02, 0x05, 0x12};

    #define HOWLER_TYPE_BYTE_INDEX 5
    #define UK_HOWLER_TONE_BYTE 0x04
    #define AUSTRALIAN_HOWLER_TONE_BYTE 0x08
    VpProfileDataType cadencePrf[] = {0x00, 0x03, 0x00, 0x02, 0x00,
                                                        UK_HOWLER_TONE_BYTE};
#endif /* DEMONSTRATE_HOWLER_TONE */
   uint16 lowLevelCmdLineId;
   uint16 tempCmd;
   uint8 lowLevelCmd;
   uint8 lowLevelCmdLen;

    if(channelId >= MAX_LINES_ON_BOARD) {
        term_printf("ERROR: Invalid Channel in the event\n.");
    }

    /* Coversion from API digit code to human readable code */
    if(digit == VP_DIG_ZERO) {
        digit = 0;
    }

    if(commandLength[channelId] == 0) {
        if(digit == CMD_END_DIGIT) {
            PrintHelp();
            return status;
        }
    }

    if((commandLength[channelId] == MAX_USER_CMD_SIZE - 1) &&
        (digit != CMD_END_DIGIT)){
        term_printf("Invalid Command length; Ignoring command.");
        commandLength[channelId] = 0;
        PrintHelp();
        return status;
    }

    if(digit != CMD_END_DIGIT){
        command[channelId][commandLength[channelId]++] = digit;
        return status;
    }

    /* Process command */
    commandOpcode = (command[channelId][0]<<4) + command[channelId][1];

    switch (commandOpcode) {
        case 0x00: {
            /* Get an option */
            getOptionId = 0;
            for(i = 2; i<commandLength[channelId]; i++){
                getOptionId = getOptionId*10 + command[channelId][i];
            }

            if(getOptionId >= VP_OPTION_ID_DTMF_MODE) {
                term_printf("Invalid Option ID(%d); Maximum allowed option ID ="
                    " %d\n", getOptionId, VP_OPTION_ID_DTMF_MODE);
            } else {
                term_printf("Reading CH:%d Option = %d\n",
                    channelId, getOptionId);
                status = VpGetOption(&lineCtx[channelId], VP_NULL,
                    getOptionId, 0);
            }
            commandLength[channelId] = 0;
            break;
        }
#ifdef DEMONSTRATE_TEST
        case 0x10: {
            /* Set the line for testing */
            if(IsATestRunning() == FALSE) {

                line = 0;
                for(i = 2; i<commandLength[channelId]; i++){
                    line = line*10 + command[channelId][i];
                }
                if(line >= MAX_LINES_ON_BOARD) {
                    term_printf("Invalid line Id(%d); Maximum allowed line ID ="
                        " %d\n", line, MAX_LINES_ON_BOARD - 1);
                } else {
                    testLineId = line;
                    term_printf("Setting test Line Id = %d\n", testLineId);
                }
            } else {
                term_printf("ERROR: Test Line Id can't be changed \n");
                term_printf("\tOngoing test on line = %d\n", testLineId);
            }
            commandLength[channelId] = 0;
            break;
        }
        case 0x11: {
            /* Perform testing on all lines */
            if(IsATestRunning() == FALSE) {
                term_printf("Performing test on line = %d\n", testLineId);
                if(channelId == testLineId) {
                    term_printf("\nWARNING: You have chosen to run the "
                        "test on the same line that you are using to "
                        "control\n");
                }
                status = SequenceTests(VP_NULL, VPTL_NUM_TID_CODES);
            } else {
                term_printf("ERROR: Test can't be started\n");
                term_printf("\tOngoing test on line = %d\n", testLineId);
            }
            commandLength[channelId] = 0;
            break;
        }
        case 0x12: {
            /* Perform a specified test */
            if(IsATestRunning() == FALSE) {
                term_printf("Performing test on line = %d\n", testLineId);

                /* Extract the TestId */
                testId = 0;
                for(i = 2; i<commandLength[channelId]; i++){
                    testId = testId*10 + command[channelId][i];
                }
                if(IsTestSupported(testId) == TRUE) {

                    if(channelId == testLineId) {
                        term_printf("\nWARNING: You have chosen to run the "
                            "test on the same line that you are using to "
                            "control\n");
                    }

                    status = SequenceTests(VP_NULL, testId);
                } else {
                    term_printf("Invalid Test Id = %d\n", testId);
                }

            } else {
                term_printf("ERROR: Test can't be started\n");
                term_printf("\tOngoing test on line = %d\n", testLineId);
            }
            commandLength[channelId] = 0;
            break;
        }
        case 0x19: {
            /* Perform PCM self-test */
            if(IsATestRunning() == FALSE) {
                term_printf("Performing PCM self test on line = %d\n",
                    testLineId);
                if(channelId == testLineId) {
                    term_printf("\nWARNING: You have chosen to run the "
                        "test on the same line that you are using to "
                        "control\n");
                }
                status = StartPcmSelfTest();
            } else {
                term_printf("ERROR: Test can't be started\n");
                term_printf("\tOngoing test on line = %d\n", testLineId);
            }
            commandLength[channelId] = 0;
            break;
        }
#endif /* DEMONSTRATE_TEST */
#ifdef DEMONSTRATE_METERING
        case 0x20: {
            /* Start metering */
            if(meteringGoingOn[channelId] == FALSE) {
                numMeter = 0;
                for(i = 2; i<commandLength[channelId]; i++){
                    numMeter = numMeter*10 + command[channelId][i];
                }
                commandLength[channelId] = 0; /* Command interpretation ends
                                               * here */


                term_printf("Initialising Metering = %d\n", channelId);
                status = VpInitMeter(&lineCtx[channelId], &METER_PROFILE[0]);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                #define ON_METER    20
                #define OFF_METER   20
                term_printf("Starting metering on line = %d, NumMeter = %d\n",
                                channelId, numMeter);
                status = VpStartMeter(&lineCtx[channelId], ON_METER, OFF_METER,
                                        numMeter);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                if(numMeter > 0) {
                    meteringGoingOn[channelId] = TRUE;
                } else {
                    meteringGoingOn[channelId] = FALSE;
                }
                break;

            } else {
                term_printf("ERROR: Metering can't be started (it is on going)\n");
                commandLength[channelId] = 0;
            }
            break;
        }
#endif /* DEMONSTRATE_METERING */

#ifdef DEMONSTRATE_HOWLER_TONE
        case 0x30: /* Intentional fall through */
        case 0x31: {
            /* Play Howler tone */
            if(commandOpcode == 0x31) {
                term_printf("Playing Australian howler tone...\n");
                cadencePrf[HOWLER_TYPE_BYTE_INDEX] =AUSTRALIAN_HOWLER_TONE_BYTE;
            } else {
                term_printf("Playing U.K howler tone...\n");
            }

            status = VpSetLineTone(&lineCtx[channelId], anyTonePrf,
                                                        cadencePrf, VP_NULL);
            commandLength[channelId] = 0;
            break;
        }
#endif /* DEMONSTRATE_HOWLER_TONE */

        case 0x40: {
            /* Read the user requested low level command */
            /* First parse the command make sure there are no errors in it */
            if(commandLength[channelId] < 4) {
                term_printf("ERROR: Low level command is LineId is two digits\n");
                commandLength[channelId] = 0;
                break;
            } else {
                lowLevelCmdLineId = command[channelId][2] * 10 +
                                                        command[channelId][3];
                if(lowLevelCmdLineId >= MAX_LINES_ON_BOARD) {
                    term_printf("ERROR: Invalid line id = %d\n", lowLevelCmdLineId);
                    commandLength[channelId] = 0;
                    break;
                }
            }
            if(commandLength[channelId] < 7) {
                term_printf("ERROR: Low level command is 3 digits\n");
                commandLength[channelId] = 0;
                break;
            } else {
                tempCmd = command[channelId][4] * 100 +
                                command[channelId][5] * 10 +
                                command[channelId][6];
                if(tempCmd > 0xff) {
                    term_printf("ERROR: Invalid Low Level Command 0x%x\n",
                                                                lowLevelCmd);
                    commandLength[channelId] = 0;
                    break;
                } else if((tempCmd & 0x01) != 0x01) {
                    term_printf("ERROR: Not a Low Level Read command 0x%X\n",
                                    tempCmd);
                    commandLength[channelId] = 0;
                    break;
                }
                lowLevelCmd = tempCmd;
            }
            if(commandLength[channelId] > 9) {
                term_printf("ERROR: Invalid Low Level Command length\n");
                commandLength[channelId] = 0;
                break;
            } else {
                lowLevelCmdLen = 0;
                for(i = 7; i<commandLength[channelId]; i++){
                    lowLevelCmdLen = lowLevelCmdLen*10 + command[channelId][i];
                }
                if(lowLevelCmdLen > MAX_MPI_LENGTH) {
                    term_printf("ERROR: Invalid Low Level Command length %d\n",
                                                                lowLevelCmdLen);
                    commandLength[channelId] = 0;
                    break;
                }
                /* It is OK issue a low level read command */
                presentCommandLength = lowLevelCmdLen;
                term_printf("Issuing low level command Line = %d, Cmd = 0x%X, "
                    "Length = %d\n", lowLevelCmdLineId, lowLevelCmd,
                    lowLevelCmdLen);
                status = VpLowLevelCmd(&lineCtx[lowLevelCmdLineId],
                            &lowLevelCmd, lowLevelCmdLen, 0);
                commandLength[channelId] = 0;
            }

            break;
        }
        default: {
            term_printf("Command Not Understood CMD = %.2X\n", commandOpcode);
            commandLength[channelId] = 0;
            break;
        }
    }
    return status;
}

/*
 * PrintHelp()
 * This function prints Help.
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * None
 */
static void PrintHelp(void)
{
#ifdef DEMONSTRATE_TEST
    int i;
#endif /* DEMONSTRATE_TEST */

    term_printf("\n"
                "Command     arguments       End  Comments\n");
    term_printf("<*>                              Print this help\n");
    term_printf("<00>        <Option ID>     <*>  Get an Option\n");

#ifdef DEMONSTRATE_TEST
    term_printf("<10>        <testLineId>    <*>  Set the line for testing\n");
    term_printf("<11>                        <*>  Perform All line tests\n");
    term_printf("<12>        <testId>        <*>  Perform the specified test\n");
    term_printf("                                  Where <testId> is one of:\n");
    /* Print the available tests in increasing order of testId */
    for(i=0;i<VPTL_NUM_TID_CODES; i++) {
        if(IsTestSupported(i) == TRUE) {
            term_printf("                                   %2d:%s\n", i,
                testIdString[i]);
        }
    }
    term_printf("<19>                        <*>  Perform PCM Self test\n");
#endif /* DEMONSTRATE_TEST */

#ifdef DEMONSTRATE_METERING
    term_printf("<20>        <num meter>     <*>  Start metering (Pulses specified)\n");
#endif /* DEMONSTRATE_METERING */

#ifdef DEMONSTRATE_HOWLER_TONE
    term_printf("<30>                        <*>  Play U.K howler tone\n");
    term_printf("<31>                        <*>  Play Australian howler tone\n");
#endif /* DEMONSTRATE_HOWLER_TONE */

    term_printf("<40>        <line> <Cmd> <len> <*>  Read Low level Command\n");
    term_printf("                                 line => 2 digits line ID\n");
    term_printf("                                 cmd => 3 digit Command (decimal)\n");
    term_printf("                                 len => command length\n");

} /* PrintHelp() */

/*
 * PrintHelp()
 * This function resets command state.
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * None
 */
static void ResetCommandState(void)
{
    int i;
    for(i=0; i<MAX_LINES_ON_BOARD; i++){

        commandInProgress[i] = FALSE;
        commandLength[i] = 0;
    }
}

#ifdef DEMONSTRATE_CID

#undef FIXED_MESSAGE

/* Caller ID Related global variables */
#ifdef FIXED_MESSAGE
uint8 message[] = {0x04, 0x12, 0x30, 0x39, 0x33, 0x30, 0x31, 0x32, 0x32, 0x34,
                    0x36, 0x30, 0x39, 0x35, 0x35, 0x35, 0x31, 0x32, 0x31,
                    0x32, 0x51};
uint8 msgLen = sizeof(message);
#else
uint8 message[100];  /* Message buffer */
uint8 msgLen = 0;
#endif /* FIXED_MESSAGE */
uint8 msgIndex = 0;
uint8 msgSize = 0;
const uint8 BUF_SIZE = 18;
#define VCP_MAX_INIT_OR_SEND_CID_BUF_SIZE 32
#define VCP_MAX_CONTINUE_CID_BUF_SIZE 16

/*
 * PrepareCidMessage()
 * This function prepares a Caller ID message
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * None
 */
static void PrepareCidMessage(void)
{
#ifndef FIXED_MESSAGE
    char *month = "09";
    char *day = "30";
    char *hour = "12";
    char *minute = "24";
    uint16 number = 123;
    char *name = "LEGERITYMINI";

    uint8 msgByteCount;
    uint16 checksum = 0;
    uint8 checkSumByte = 0;

    AppCreateBuffer(month, day, hour, minute, number, name, message);
    msgLen = message[1]+2;  /* Length of Message, plus type and len of message
                             * field */

    for (msgByteCount = 0; msgByteCount < msgLen; msgByteCount++) {
        checksum += message[msgByteCount];
    }
    checkSumByte = (uint8)((~(checksum % 256))+1);
    message[msgLen] = checkSumByte;
    msgLen++;
#endif /* FIXED_MESSAGE */
}

/*
 * GetCallerDataPointer()
 * This function returns Caller ID data buffer pointer.
 *
 * Preconditions:
 * This funciton SHOULD be called with 'startFromBegining' flag == TRUE first
 *
 * Postconditions:
 * None
 */
static uint8* /* Returns a pointer from where Caller ID data has to be read */
GetCallerDataPointer(
    bool startFromBegining,  /* If set indicates start from the beginging
                              * of the message */
    uint8 *messageSize)      /* Pointer to the size of the message */
{
    uint8 *retPtr = VP_NULL;
    int remainingLength;

    if(startFromBegining == TRUE) {
        /* First prepare the message */
        PrepareCidMessage();

        /* Send the first block of the message */
        if(msgLen > VCP_MAX_INIT_OR_SEND_CID_BUF_SIZE) {
            *messageSize = VCP_MAX_INIT_OR_SEND_CID_BUF_SIZE;
            msgIndex = VCP_MAX_INIT_OR_SEND_CID_BUF_SIZE; /* Current
            * location for message data */
        } else {
            *messageSize = msgLen;
            msgIndex = msgLen;
        }
        retPtr = &message[0];

    } else {

        if(msgLen == 0) {
            /* Nothing to send */
            *messageSize = 0;
            return retPtr;
        }

        remainingLength = msgLen - msgIndex;

        if(remainingLength <= 0) {
            /* nothing to send */
            *messageSize = 0;
            retPtr = VP_NULL;
        } else {

            retPtr = &message[msgIndex];
            /* Update the index and the length of the message to be sent
             * this time */
           if(remainingLength >= VCP_MAX_CONTINUE_CID_BUF_SIZE) {
                /* Send the next segment */
                *messageSize = VCP_MAX_CONTINUE_CID_BUF_SIZE;
                msgIndex = msgIndex + VCP_MAX_CONTINUE_CID_BUF_SIZE;

            } else {
                /* Send the last segment */
                *messageSize = remainingLength;
                msgIndex = msgIndex + remainingLength;
            }
        }
    }
    return retPtr;
}
#endif /* DEMONSTRATE_CID */

#ifdef DEMONSTRATE_TEST
/*
 * TestInputOutput()
 * This function either fills in the default test inputs or prints the test
 * results from the given test results structure.
 *
 */
static bool                             /* TRUE = Function successful */
TestInputOutput(
    bool fillInp,                       /* If TRUE, this function fills the
                                         * default inputs for the indicated
                                         * test. Else it prints the test
                                         * results.*/
    VpTestIdType testId,                /* Test ID */
    VptlTestInputType *pTestInput,      /* Test inputs; Filled when inputs
                                         * are asked; Else it could still be
                                         * used while printing the results */
    VptlTestResultType *pTestResult)    /* Test results are printed if indicated
                                         * to do so */
{
    bool retVal = TRUE;

    /* Check for error conditions in the input */
    if(((fillInp == TRUE) && (pTestInput == VP_NULL)) ||
       ((fillInp == FALSE) &&
        ((pTestInput == VP_NULL) || (pTestResult == VP_NULL)))) {
        return FALSE;
    }

    if(fillInp == TRUE) {
        term_printf("------------------------------------------------------\n");
    }

    /* Fill the test inputs or print the test results */
    switch (testId) {
        case VPTL_TID_OPEN_DC_VOLTAGE: {
            if(fillInp == TRUE) {
                /* Prepare for running the tests */
                pTestInput->openDcV.testType = VPTL_OPEN_DC_WO_AUTO_DISCONNECT;
                pTestInput->openDcV.tolerance = VPTL_OPEN_DC_TYP_TOL;
                pTestInput->openDcV.relayState = VP_RELAY_NORMAL;
                term_printf("Using default inputs for Foreign DC Voltage test\n");
            } else {
                term_printf("\nForeign DC Voltage Test Results\n");
                term_printf("\tTest Passed=%s\n",
                    pTestResult->result.openDcV.testPassed ? "Yes":"No");
                if(pTestResult->result.openDcV.testPassed == TRUE) {
                    term_printf("\tTip lead Voltage = %f Volts\n",
                        pTestResult->result.openDcV.vTip);
                    term_printf("\tRing lead Voltage = %f Volts\n",
                        pTestResult->result.openDcV.vRing);
                    term_printf("\tDifferential Voltage = %f Volts\n",
                        pTestResult->result.openDcV.vDiff);
                }
            }
            break;
        }
        case VPTL_TID_NOISE: {
            if(fillInp == TRUE) {
                /* Prepare for running the tests */
                pTestInput->noise.outputSignalPower = VPTL_NULL_POWER_dBm0;
                pTestInput->noise.pAcProfileForTest = VP_PTABLE_NULL;
                pTestInput->noise.pAcProfileAfterTest = VP_PTABLE_NULL;
                pTestInput->noise.filter = VP_FILTER_NONE;
                term_printf("Using default inputs for Noise test\n");
            } else {
                term_printf("\nNoise Test results\n");
                term_printf("\tnoise=%f dBm0\n",
                    pTestResult->result.noise.noise);
            }
            break;
        }
        case VPTL_TID_OPEN_AC_VOLTAGE: {
            if(fillInp == TRUE) {
                /* Prepare for running the tests */
                pTestInput->openAcV.relayState = VP_RELAY_NORMAL;
                term_printf("Using default inputs for Foreign AC Voltage test\n");
            } else {
                term_printf("\nForeign AC Voltage Test Results\n");
                term_printf("\tvtip=%f Vrms\n\tvring=%f Vrms\n\tvdiff=%f Vrms\n",
                    pTestResult->result.openAcV.vTip,
                    pTestResult->result.openAcV.vRing,
                    pTestResult->result.openAcV.vDiff);

            }
            break;
        }
        case VPTL_TID_DC_LOOP_RES: {
            if(fillInp == TRUE) {
                /* Prepare for running the tests */
                #define I_TEST_HIGH 0.012
                pTestInput->dcLoopRes.iTestHigh = I_TEST_HIGH;
                pTestInput->dcLoopRes.testType = VPTL_DC_LOOP_RES_OFF_COMP;
                term_printf("Using default inputs for DC loop resistance test\n");
            } else {
                term_printf("\nDC loop resistance Test Results\n");
                term_printf("\tRLOOP=%f Ohms\n\t",
                    pTestResult->result.dcLoopRes.rLoop);
                term_printf("ILG1=%f A\n\t",
                    pTestResult->result.dcLoopRes.iLg);
                term_printf("ILG2=%f A\n\t",
                    pTestResult->result.dcLoopRes.iLg2);
                term_printf("Voltage Saturation=%s\n\t",
                    (pTestResult->result.dcLoopRes.voltageSaturation == TRUE)?
                    "Yes":"No");
                term_printf("CMC Degradation=%s\n",
                    (pTestResult->result.dcLoopRes.degradedByCmc == TRUE)?
                    "Yes":"No");
            }
            break;
        }
        case VPTL_TID_ROH: {
            if(fillInp == TRUE) {
                /* No test inputs */
            } else {
                term_printf("\nReceiver Off-Hook Test Results\n");
                term_printf("\tOff-Hook=%s\n",
                    (pTestResult->result.roh.receiverOffHook == TRUE)?
                    "Yes":"No");
                term_printf("\tRloop out of range=%s\n",
                    (pTestResult->result.roh.rLoopOutOfRange == TRUE)?
                    "Yes":"No");
            }
            break;
        }
        case VPTL_TID_REN: {
            if(fillInp == TRUE) {
                /* Prepare for running the tests */
                term_printf("Using default inputs for REN test\n");
                pTestInput->ren.renTestType = VPTL_REN_REGULAR_PHNE_TEST;
                #define V_REN_TEST  12  /* Signal for REN test in Vrms */
                pTestInput->ren.vRenTest = V_REN_TEST;
            } else {
                term_printf("\nRinger Equivalence Numer Test Results\n");
                term_printf("\tREN=%f \n", pTestResult->result.ren.ren);
                if(pTestInput->ren.renTestType == VPTL_REN_REGULAR_PHNE_TEST) {
                    term_printf("\tMeasured Zload=%f Ohms\n",
                        pTestResult->result.ren.measuredZload);
                    term_printf("\tCMC Degradation=%s\n",
                        (pTestResult->result.ren.degradedByCmc == TRUE)?
                        "yes":"no");
                } else {
                    term_printf("\tMeasured Cload=%f nF\n",
                        1.0e9 * pTestResult->result.ren.measuredCap);
                }
            }
            break;
        }
        case VPTL_TID_TRANSMISSION_SELF_TEST: {
            if(fillInp == TRUE) {
                /* Prepare for running the tests */
                term_printf("Using default inputs for Transmission self-test\n");
                pTestInput->transmission.tone.freq =
                    TRANSMISSION_TEST_SIGNAL_FREQ;
                pTestInput->transmission.pAcProfileForTest =
                    pTestInput->transmission.pAcProfileAfterTest =
                    VP_PTABLE_NULL;
                pTestInput->transmission.tone.level =
                    TRANSMISSION_TEST_SIGNAL_AMP;
                pTestInput->transmission.zSourceReal =
                    TRANSMISSION_TEST_Z_SOURCE_R;
                pTestInput->transmission.zSourceImaginary =
                    TRANSMISSION_TEST_Z_SOURCE_I;
                pTestInput->transmission.tolerance =
                    TRANSMISSION_TEST_TOLERANCE;
            } else {
                term_printf("\nTransmission Self Test Results\n");
                term_printf("\tTest Passed=%s\n",
                    pTestResult->result.transmission.testPassed ? "Yes":"No");
                term_printf("\tTest Aborted due to off-hook=%s\n",
                    pTestResult->result.transmission.testAbortedDueToOffHook ?
                    "Yes":"No");
                term_printf("\tTrans-Hybrid Loss (Z Source = Open Circuit)=%f\n",
                    pTestResult->result.transmission.thlOpen);
                term_printf("\tTrans-Hybrid Loss (Z Source = Test load)=%f\n",
                    pTestResult->result.transmission.thlRload);
                term_printf("\tExpected Trans-Hybrid Loss Lower Limit=%f\n",
                    pTestResult->result.transmission.thlExpectedLL);
                if(pTestResult->result.transmission.thlExpectedUL ==
                    VERY_HIGH_THL) {
                    term_printf("\tExpected Trans-Hybrid Loss Upper Limit="
                                    "VERY_HIGH_THL\n");
                } else {
                    term_printf("\tExpected Trans-Hybrid Loss Upper Limit=%f\n",
                        pTestResult->result.transmission.thlExpectedUL);
                }
            }
            break;
        }
        case VPTL_TID_DC_FEED_SELF_TEST: {
            if(fillInp == TRUE) {
                /* Nothing to change */
            } else {
                term_printf("\nDC Feed Self-Test Results\n");
                term_printf("\tTest Passed=%s\n",
                    pTestResult->result.dcFeed.testPassed ? "Yes":"No");
                term_printf("\tMeasured Test load =%f Ohms\n",
                    pTestResult->result.dcFeed.rloadMsrd);
                term_printf("\tMeasured High Bat =%f Volts\n",
                    pTestResult->result.dcFeed.vBat);
                term_printf("\tMeasured VTR =%f Volts\n",
                    pTestResult->result.dcFeed.vTr);

                term_printf("\tCurrent through test load (Bat=High, Pol=Normal)"
                    " =%f Amps\n",
                    pTestResult->result.dcFeed.imtHighBatNormalFeed);
                term_printf("\tCurrent through test load (Bat=High, Pol=Revers"
                    "e) =%f Amps\n",
                    pTestResult->result.dcFeed.imtHighBatReverseFeed);

                term_printf("\tCurrent through test load (Bat=Low, Pol=Normal)"
                    " =%f Amps\n",
                    pTestResult->result.dcFeed.imtLowBatNormalFeed);

                term_printf("\tCurrent through test load (Bat=Boost, Pol=Norma"
                    "l) =%f Amps\n",
                    pTestResult->result.dcFeed.imtBoostBatNormalFeed);
            }
            break;
        }
        case VPTL_TID_ON_OFF_HOOK_SELF_TEST: {
            if(fillInp == TRUE) {
                /* Prepare for running the tests */
                term_printf("Using default inputs for ON-OFF Hook Self-Test\n");
                pTestInput->onOffHook.overrideOffHook = FALSE;

            } else {
                term_printf("\nON-OFF Hook Self-Test Results\n");
                switch (pTestResult->result.onOffHook.status) {
                    case VPTL_HOOK_SELF_TEST_UNKNOWN:
                        term_printf("UNKNOWN\n");  break;
                    case VPTL_HOOK_SELF_TEST_ABORTED:
                        term_printf("ABORTED\n");  break;
                    case VPTL_HOOK_SELF_TEST_HW_FAULT:
                        term_printf("HW_FAULT\n"); break;
                    case VPTL_HOOK_SELF_TEST_PASSED:
                        term_printf("PASSED\n");   break;
                    default:
                        term_printf("SW_BUG\n");   break;

                    }
            }
            break;
        }
        case VPTL_TID_RINGING_SELF_TEST: {
            if(fillInp == TRUE) {
                /* Prepare for running the tests */
                term_printf("Using default inputs for Ringing Self-Test test\n");

                pTestInput->ringing.useInternalRinging = TRUE;
                pTestInput->ringing.vRinging = RINGING_SELF_TEST_VOLTAGE;
                pTestInput->ringing.pRingProfileAfterTest = VP_PTABLE_NULL;
                pTestInput->ringing.pRingProfileForTest = VP_PTABLE_NULL;

            } else {
                term_printf("\nRinging Self-Test Results\n");
                term_printf("\tTest Passed=%s\n",
                    pTestResult->result.ringing.testPassed ? "Yes":"No");
                term_printf("\tON hook to OFF hook transition detected=%s\n",
                    pTestResult->result.ringing.on2OffHookChangeDetected ?
                    "Yes":"No");
                if(pTestResult->result.ringing.measuredRloop == VPTL_OPEN_CIRCUIT) {
                    term_printf("\tRLOOP=Open\n");
                } else {
                    term_printf("\tRLOOP=%f Ohms\n",
                    pTestResult->result.ringing.measuredRloop);
                }

            }
            break;
        }
        case VPTL_TID_3ELE_RES: {
            bool foreignVI = TRUE;
            if(fillInp == TRUE) {
                /* Prepare for running the tests */
                term_printf("Using default inputs for Three-Element Resistance Test\n");
                pTestInput->ter.highImpedanceTest = FALSE;
                pTestInput->ter.measureForeignVI = TRUE;
            } else {
                term_printf("\nThree-Element Resistance Results\n");

                if((pTestResult->result.ter.rtg == VPTL_HARD_FAULT) ||
                   (pTestResult->result.ter.rrg == VPTL_HARD_FAULT)){

                    /* Hard faults are present; Print the results */
                    if(pTestResult->result.ter.rtg == VPTL_HARD_FAULT) {
                        term_printf("\tRTG=");
                        term_printf("FAULT\n");
                    }

                    if(pTestResult->result.ter.rrg == VPTL_HARD_FAULT) {
                        term_printf("\tRRG=");
                        term_printf("FAULT\n");
                    }

                    /* No need to print other results */
                    foreignVI = FALSE; /* Since this test could not be run */

                } else {

                    if(pTestResult->result.ter.rtr == VPTL_HARD_FAULT) {

                        term_printf("\tRTR=");
                        term_printf("FAULT\n");

                        PrintRes("RGND", pTestResult->result.ter.rGnd);

                        /* Other results are not applicable */

                    } else {
                        /* No hard faults */
                        PrintRes("RTG", pTestResult->result.ter.rtg);
                        PrintRes("RRG", pTestResult->result.ter.rrg);
                        PrintRes("RTR", pTestResult->result.ter.rtr);
                    }
                }

                if(foreignVI == TRUE) {
                    term_printf("\tvTip=");
                    term_printf("%f V\n", pTestResult->result.ter.vTip);

                    term_printf("\tvRing=");
                    term_printf("%f V\n", pTestResult->result.ter.vRing);

                    term_printf("\tiTip=");
                    term_printf("%f A\n", pTestResult->result.ter.iTip);

                    term_printf("\tiRing=");
                    term_printf("%f A\n", pTestResult->result.ter.iRing);
                }
            }
            break;
        }
        case VPTL_TID_RD_LOOP_BAT_COND_TEST: {
            if(fillInp == TRUE) {
                /* Nothing to be done */

            } else {
                term_printf("\nRead Loop and Battery Condition Test Results\n");

                term_printf("Read Loop and Battery Condition Test\n");
                if(pTestResult->result.loopBatCond.rloop ==
                                                VPTL_IMPEDANCE_NOT_MEASURED){
                    term_printf("\tRLOOP=not valid\n");
                } else {
                    term_printf("\tRLOOP=%f Ohms\n",
                        pTestResult->result.loopBatCond.rloop);
                }

                if(pTestResult->result.loopBatCond.ilg ==
                                                    VPTL_CURRENT_NOT_MEASURED) {
                    term_printf("\tILG=not valid\n");
                } else {
                    term_printf("\tILG=%f Amps\n",
                        pTestResult->result.loopBatCond.ilg);
                }
                if(pTestResult->result.loopBatCond.imt ==
                                                    VPTL_CURRENT_NOT_MEASURED) {
                    term_printf("\tIMT=not valid\n");
                } else {
                    term_printf("\tIMT=%f Amps\n",
                        pTestResult->result.loopBatCond.imt);
                }

                if(pTestResult->result.loopBatCond.vsab ==
                                                    VPTL_VOLTAGE_NOT_MEASURED) {
                    term_printf("\tVSAB=not valid\n");
                } else {
                    term_printf("\tVSAB=%f Volts\n",
                        pTestResult->result.loopBatCond.vsab);
                }
                term_printf("\tVBL=%f Volts\n",
                    pTestResult->result.loopBatCond.vbl);
                term_printf("\tVBH=%f Volts\n",
                    pTestResult->result.loopBatCond.vbh);
                term_printf("\tVBP=%f Volts\n",
                    pTestResult->result.loopBatCond.vbp);
                term_printf("\tMSPL=%f Volts\n",
                    pTestResult->result.loopBatCond.mspl);
            }
            break;
        }
        case VPTL_TID_TRANS_HYBRID_LOSS: {
            if(fillInp == TRUE) {
                /* Prepare for running the tests */
                term_printf("Using default inputs for Transhybrid Loss Test\n");

                #define THL_TEST_FREQ (1010.0) /* Hz */
                #define THL_TEST_LEVEL (-10.0) /* dBm0 */

                pTestInput->transHybridLoss.tone.freq = THL_TEST_FREQ;
                pTestInput->transHybridLoss.tone.level = THL_TEST_LEVEL;
                pTestInput->transHybridLoss.pAcProfileForTest =
                    pTestInput->transHybridLoss.pAcProfileAfterTest =
                    VP_PTABLE_NULL;

            } else {
                term_printf("\n Transhybrid Loss Results\n");
                term_printf("\techo=%f dBm0\n\tTHL=%f dB\n",
                    pTestResult->result.transHybridLoss.echo,
                    pTestResult->result.transHybridLoss.loss);
            }
            break;
        }
        default: {
            /* Error case */
            term_printf("Error in Test sequencing\n");
            retVal = FALSE;
            break;
        }
    }

    /* DevNotes:
     * If any changes are required for the test inputs then a function could
     * be called from here to change the default inputs. The inputs to that
     * function could be communicated through global variables.
     */

    if(fillInp == FALSE) {
        term_printf("------------------------------------------------------\n");
    }

    return retVal;
} /* TestInputOutput() */

/*
 * ImplementOneTest()
 * This function carries out a given test.
 */
static VptlTestStatusType /* Indicates the state of the test */
ImplementOneTest(
    VptlEventType *pVptlEvent, /* If NULL, start the given test from the
                                * beginning. Else an intermediate state of the
                                * test. Event is being passed. */
    VpTestIdType testId)       /* This argument is used only if the first
                                * argument is NULL. */
{
    VptlTestStatusType testStatus;

    if(pVptlEvent == VP_NULL) {
        /* Start a new test */
        /* Fill the default test inputs */
        if((testId < 0) || (testId > VPTL_NUM_TID_CODES)) {
            /* Not checking for if a given test is supported because it is
             * checked else where. */
            term_printf("\nStarting a new test; Test ID = %d\n", testId);
        } else {
            term_printf("\nStarting a new test; Test ID = %s\n",
                testIdString[testId]);
        }

        if(TestInputOutput(TRUE, testId, &testInput, VP_NULL) == FALSE) {
            term_printf("ERROR: Failure while filling test inputs\n");
            return VPTL_STATUS_ERROR_UNKNOWN;
        }

        /* Start the test */
        testStatus = VptlRunTest(&lineCtx[testLineId], testId, TEST_HANDLE,
            &topology, &testInput, VP_NULL, &testResult, &testTemp, &testCtx);
    } else {
        /* Continue the present test */
        testStatus =  VptlEventHandler(&testCtx, pVptlEvent);
    }

    if(testStatus == VPTL_STATUS_DONE) {
        /* Print the test results */
        if(TestInputOutput(FALSE, testResult.testId, &testInput, &testResult)
                                                                    == FALSE) {
            term_printf("ERROR: Failure while printing test results\n");
            testStatus = VPTL_STATUS_ERROR_UNKNOWN;
        }

    } else if(testStatus == VPTL_STATUS_RUNNING){
        /* Nothing to do... wait for more test stages to complete */
    } else {
        /* Print the error */
        if(testStatus <= VPTL_STATUS_TEST_NOT_SUPPORTED) {
            term_printf("ERROR in test library = %s\n",
                                                testReturnString[testStatus]);
        } else {
            term_printf("ERROR in test library = %d\n", testStatus);
        }
    }

    return testStatus;

} /* ImplementOneTest() */


/*
 * SequenceTests()
 * This function sequences various tests.
 */
static VpStatusType
SequenceTests(
    VptlEventType *pVptlEvent,  /* If NULL, this argument indicates test
                                 * needs to start from the beginning, Else
                                 * a continuing test */
    VpTestIdType testId)        /* Used only if the first argument is NULL.
                                 * Has a special meaning to run all the tests if
                                 * it has a value equal of "VPTL_NUM_TID_CODES".
                                 */
{
    VptlTestStatusType testStatus;
    bool singleStepTests = FALSE;
    VpTestIdType testSeqTestId = testId;
    int i;

    /* Check for errors when asked to run a new test */
    if((pVptlEvent == VP_NULL) && (onGoingTest == TRUE)) {
        term_printf("ERROR: Can't start a new test. Already a test running\n");
        return VP_STATUS_FAILURE;
    }
    if((pVptlEvent == VP_NULL) && (onGoingTest == FALSE) &&
        (onGoingTestSeq == TRUE)){
        term_printf("ERROR: Can't start a new test. Already a test sequence "
                    "is running\n");
        return VP_STATUS_FAILURE;
    }

    /* Start a new test */
    if(pVptlEvent == VP_NULL) {
        if(testId == VPTL_NUM_TID_CODES) {
            /* Check to make sure all the tests in the test sequence are
             * supported */
            for(i=0; i<MAX_NUM_TEST; i++) {
                if(IsTestSupported(testSequence[i]) != TRUE) {
                    term_printf("ERROR: Can't start the test sequence. Test"
                        " sequence has a test that is not supported (%s)\n",
                        testIdString[testSequence[i]]);
                    return VP_STATUS_FAILURE;
                }
            }

            /* A sequence of tests are being requested */
            onGoingTestSeq = TRUE;
            onGoingTest = TRUE;

            /* Extract the testIds */
            testSeqTestId = testSequence[curTestIndex++];

        } else {
            /* Only one test is being requested */
            onGoingTest = TRUE;
            testSeqTestId = testId;
        }
    }

    /* Carry out the test. Sequence through other tests if the test is of
     * single step */
    do {
        testStatus = ImplementOneTest(pVptlEvent, testSeqTestId);

        /* Exit this loop if necessary */
        if(testStatus == VPTL_STATUS_DONE) {
            if(onGoingTestSeq == FALSE) {
                /* Single test */
                onGoingTest = FALSE;
                singleStepTests = FALSE;
            } else {

                /* A sequence of tests. One of them is complete */
                if(curTestIndex < MAX_NUM_TEST) {
                    /* Start the next one */
                    /* Change the arguments for running the next test */
                    testSeqTestId = testSequence[curTestIndex++];
                    pVptlEvent = VP_NULL;

                    singleStepTests = TRUE;
                } else {
                    /* No more tests to run */
                    singleStepTests = FALSE;

                    /* The last test of the test sequence is complete */
                    onGoingTest = FALSE;
                    onGoingTestSeq = FALSE;
                    curTestIndex = 0;
                }
            }
        } else {
            singleStepTests = FALSE;
        }

    } while(singleStepTests == TRUE);

    if((testStatus == VPTL_STATUS_DONE) || (testStatus == VPTL_STATUS_RUNNING)){

        return VP_STATUS_SUCCESS;

    } else {
        term_printf("ERROR in test sequencing\n");
        return VP_STATUS_FAILURE;
    }

} /* SequenceTests() */

/* This function is used to print a resistance value */
static void
PrintRes(
    const char *inputStr,
    VptlFloatType res)
{
    term_printf("\t%s=", inputStr);

    if(res == VPTL_OPEN_CIRCUIT) {

        term_printf("OPEN\n");

    } else if(res == VPTL_HARD_FAULT) {

        term_printf("FAULT\n");

    } else if(res == VPTL_IMPEDANCE_NOT_MEASURED) {

        term_printf("NOT MEASURED\n");

    } else {

        term_printf("%f Ohms\n", res);
    }
} /* PrintRes() */

/*
 * This functions returns TRUE if a given test is supported. Else it returns
 * FALSE. The decision is based on the test package.
 */
static bool
IsTestSupported(
    VpTestIdType testId)
{
    if((testId < 0) || (testId >= VPTL_NUM_TID_CODES) ||
       (testId == VPTL_TID_CALIBRATE) ||
       (testId == VPTL_TID_5ELE_RES) ||
       (testId == VPTL_TID_3ELE_CAP) ||
       (testId == VPTL_TID_DISTANCE_TO_OPEN) ||
       (testId == VPTL_TID_FOREIGN_AC_CURRENT) ||
       (testId == VPTL_TID_DTMF_PULSE_MSRMNT) ||
       (testId == VPTL_TID_SNR_QNTZ_DIST) ||
       (testId == VPTL_TID_ARB_TONE) ||
       (testId == VPTL_TID_TONE_GEN) ||
       (testId == VPTL_TID_DRAW_BREAK_DIALTONE) ||
       (testId == VPTL_TID_INWRD_CUR) ||
       (testId == VPTL_TID_METERING_SELF_TEST) ||
       (testId == VPTL_TID_DIALING_SELF_TEST) ||
       (testId == VPTL_TID_HOWLER_TEST) ||
       (testId == VPTL_TID_FUSE_TEST)) {

        return FALSE;
    } else {
        return TRUE;
    }
}

/*
 * This function is used to check if a test is running.
 */
static bool         /* Returns true if a test is running; Else it returns
                     * false */
IsATestRunning(
    void)
{
    if((onGoingTestSeq == TRUE) || (onGoingTest == TRUE)) {
        return TRUE;
    } else {
        return FALSE;
    }
} /* IsATestRunning() */

/*
 * This function starts the PCM self test function on the selected line for
 * testing.
 */
static VpStatusType
StartPcmSelfTest(
    void)
{
    VpOptionTimeslotType pcmTimeSlot;
    VpStatusType status;

    if(IsATestRunning() == TRUE) {
        term_printf("\nSkipping PCM Self Test because some other tests are "
            "being carried out\n");
        return VP_STATUS_SUCCESS;
    }

    /* Perform the Self test */
#ifdef USE_2MHZ_PCLK
    /* First set up timeslots as required by VCP */
    /* There are no special limitations on timeslots to be used when using
     * 2MHz PCLK frequency. */
    pcmTimeSlot.rx = 0 + 2 * testLineId;
#else
    /* First set up timeslots as required by VCP */
    pcmTimeSlot.rx = 48; /* Time slot for testing is has to be 48 */
#endif /* USE_2MHZ_PCLK */
    pcmTimeSlot.tx = 0 + 2 * testLineId;
    term_printf("PCM Self-Test TimeSlot(Line = %d) Rx ->%d, Tx->%d\n",
        testLineId, pcmTimeSlot.rx, pcmTimeSlot.tx);
    status = VpSetOption(&lineCtx[testLineId], VP_NULL, VP_OPTION_ID_TIMESLOT,
                &pcmTimeSlot);
    if(status != VP_STATUS_SUCCESS){
        return status; /* Something is wrong */
    }

    term_printf("Starting PCM Self-Test\n");
    status = VpSelfTest(&lineCtx[testLineId]);

    /* DevNotes: This test is not setting a flag for other tests to know
     * the PCM self test is going on. */

    return status;

} /* StartPcmSelfTest() */

static VpStatusType
ConcludePcmSelfTest(
    VpEventType *pEvent)
{
    int timeslot;
    VpOptionTimeslotType pcmTimeSlot;
    VpStatusType status;

    term_printf("Received event: VP_DEV_EVID_STEST_CMP(%d)\n",
        pEvent->channelId);
    term_printf("PCM Self-Test %s\n", pEvent->eventData ? "FAILED":"PASSED");

    /* Setup the time slots as they were before */
    timeslot = 2 * testLineId;
    pcmTimeSlot.rx = timeslot;
    pcmTimeSlot.tx = timeslot;
    term_printf("Restoring regular time slot Line %d TimeSlot->%d, \n",
        testLineId, timeslot);
    status = VpSetOption(&lineCtx[testLineId], VP_NULL, VP_OPTION_ID_TIMESLOT,
        &pcmTimeSlot);

    return status;

} /* ConcludePcmSelfTest() */

#endif  /* DEMONSTRATE_TEST */

#endif /* USING_QUICK_START_APP */


