/** \file qs_le71hr0030_dat.c
 * qs_le71hr0030_dat.c
 *
 *  This file is a Quick-Start application for the LE71HR030 Liine module.  It
 * accompanies a profiles8.c and profiles8.h file set to configure the devices
 * and lines on the -0030 LM.  The purpose of this applicaiton is to
 * demonstrate the capabilities of the Vp790 chipset via the API w.r.t Tone
 * Generation, DTMF Generation, and Dial Pulse Detection.
 *
 * Copyright (c) 2008, Zarlink Semiconductor, Inc.
 */

/*
 * These headers are for UVB/Linux purposes only.  The user will not find them
 * in the API
 */
#include <sys/time.h>
#include <signal.h>
#include "uvb_xs.h"

/*
 * These files are required by all applications.  "vp_api.h" includes all of
 * the API libraries required by the application.  "profiles8.h" includes the
 * profiles used by the application.  At a minimum, there must be a Device
 * Profile for the device to initialize.
 */
#include "vp_api.h"
#include "profiles8.h"

#include <stdio.h>

/* UVB specific values for device ID depending on the MPI Mode */
#define MPI_3WIRE 0x30
#define MPI_4WIRE 0x10

/* Current configuration of Line Module */
#define MPI_MODE MPI_3WIRE

/* Application specific #define that improve code readability */
#define NUM_LINES     4
#define NUM_DEVICES   1

/*
 * Application memory space for the API required Device/Line contexts and
 * objects.  This application uses the simplest memory scheme -- global,
 * but the API will work with any memory (management) scheme.
 */
VpDevCtxType pDevCtx[NUM_DEVICES];
VpLineCtxType pLineCtx[NUM_LINES];
Vp790DeviceObjectType pDevObj[NUM_DEVICES];
Vp790LineObjectType pLineObj[NUM_LINES];

/*
 * Variable that will toggle for each Hook Event to alternate between ringing
 * and non-ringing times.
 */
bool ringingTime[NUM_LINES] = {FALSE, FALSE, FALSE, FALSE};

/**
 * Function:  InitQuickStartSettings()
 *
 * Description:  Function used to initialize the device and lines of the device
 * to settings not provided by default in the API.  This function should be the
 * first function called after a device init complete event occurs.
 */
void
InitQuickStartSetttings(
    VpEventType *pEvent)
{
    uint8 lineNum;
    VpDevCtxType *pDevCtx = pEvent->pDevCtx;
    uint16 handle = 0;

    /*
     * For this demo, we want the API to detect the dial pulses on the FXS
     * line, then initialize the ringing cadence.
     */
    VpOptionPulseModeType pulseMode = VP_OPTION_PULSE_DECODE_ON;
    VpSetOption(VP_NULL, pDevCtx, VP_OPTION_ID_PULSE_MODE, &pulseMode);

    /*
     * This loop is ok because all lines are FXS lines.  If this function
     * is called for an FXO line, no harm -- it just returns a not supported
     * error
     */

    for (lineNum = 0; lineNum < NUM_LINES; lineNum++) {
        VpInitRing(&pLineCtx[lineNum], RING_CAD_STD, VP_PTABLE_NULL);
        VpSetLineState(&pLineCtx[lineNum], VP_LINE_STANDBY);
    }

    /*
     * Read the event mask contents so we can unmask only the FXS signaling
     * events.
     */
    VpGetOption(VP_NULL, pDevCtx, VP_OPTION_ID_EVENT_MASK, handle);

    return;
}

/**
 * Function:  InitReadOptions()
 *
 * Description:  Function used to initialize the device event masks to only
 * allow those used by this application.
 */
void
InitReadOptions(
    VpEventType *pEvent)
{
    VpDevCtxType *pDevCtx = pEvent->pDevCtx;
    VpOptionEventMaskType eventMask;

    VpGetResults(pEvent, &eventMask);

    eventMask.signaling &= ~(VP_LINE_EVID_HOOK_OFF | VP_LINE_EVID_HOOK_ON);
    eventMask.signaling &= ~(VP_LINE_EVID_FLASH | VP_LINE_EVID_STARTPULSE);
    eventMask.signaling &= ~(VP_LINE_EVID_PULSE_DIG);

    VpSetOption(VP_NULL, pDevCtx, VP_OPTION_ID_EVENT_MASK, &eventMask);
}


/**
 * Function:  ProcessSignalingEvent()
 *
 * Description:  Function used to perform some action on the FXS lines when a
 * signaling event occurs.
 */
void
ProcessSignalingEvent(
    VpEventType *pEvent)
{
    uint8 chanId = pEvent->channelId;
    VpLineCtxType *pLineCtx = pEvent->pLineCtx;
    uint16 evDat = pEvent->eventData;

    /*
     * The tone profiles (pointers) in the following list are created from the
     * Profile Wizard
     */
    VpProfilePtrType tonePtrs[] = {
        TONE_DIAL, TONE_DIAL, TONE_BUSY, TONE_RINGBACK
    };

    /*
     * The cadence profiles (pointers) in the following list are created from
     * the Profile Wizard
     */
    VpProfilePtrType cadPtrs[] = {
        TONE_CAD_DIAL0, TONE_CAD_DIAL1, TONE_CAD_BUSY0, TONE_CAD_RNGBACK
    };

    switch (pEvent->eventId) {
        case VP_LINE_EVID_HOOK_OFF:
            VpSetLineState(pLineCtx, VP_LINE_ACTIVE);
            VpSetLineTone(pLineCtx, TONE_DIAL, TONE_CAD_DIAL0, VP_NULL);
            break;

        case VP_LINE_EVID_HOOK_ON:
            VpSetLineTone(pLineCtx, VP_PTABLE_NULL, VP_PTABLE_NULL, VP_NULL);
            if(ringingTime[chanId] == FALSE) {
                VpSetLineState(pLineCtx, VP_LINE_STANDBY);
                ringingTime[chanId] = TRUE;
            } else {
                VpSetLineState(pLineCtx, VP_LINE_RINGING);
                ringingTime[chanId] = FALSE;
            }
            break;

        case VP_LINE_EVID_PULSE_DIG:
            if (evDat <= 4) {
                VpSetLineTone(pLineCtx, tonePtrs[evDat-1], cadPtrs[evDat-1],
                    VP_NULL);
            }
            break;

        case VP_LINE_EVID_STARTPULSE:
            VpSetLineTone(pLineCtx, VP_PTABLE_NULL, VP_PTABLE_NULL, VP_NULL);
            break;

        case VP_LINE_EVID_FLASH:
            VpSendSignal(pLineCtx, VP_DIGIT_GENERATION_DIAL_HOOK_FLASH, &evDat);
            break;

        default:
            break;
    }
    return;
}

/**
 * Function:  alarmHandle()
 *
 * Description: This function is called at a periodic rate to perform all
 * required API operations.  It implements the functional requirements of
 * the application mentioned in the header.
 */
void
alarmHandle(
    int val)
{
    int deviceNum;
    bool deviceEventStatus = FALSE;
    VpEventType pEvent;
    VpDevCtxType *pDevCtxLocal;

    /*
     * This loop will query the FXS device for events, and when an event is
     * found (deviceEventStatus = TRUE), will parse the event and perform
     * further operations.
     */
    for (deviceNum = 0; deviceNum < NUM_DEVICES; deviceNum++) {
        pDevCtxLocal = &pDevCtx[deviceNum];
        VpApiTick(pDevCtxLocal, &deviceEventStatus);
        if(deviceEventStatus == TRUE) {
            while(VpGetEvent(pDevCtxLocal, &pEvent)) {
                /*
                 * Process the Signaling Events and initialize the lines when
                 * a Device Init Complete event occurs.
                 */
                if(pEvent.eventCategory == VP_EVCAT_SIGNALING) {
                    ProcessSignalingEvent(&pEvent);
                } else if (pEvent.eventCategory == VP_EVCAT_RESPONSE) {
                    if (pEvent.eventId == VP_DEV_EVID_DEV_INIT_CMP) {
                        InitQuickStartSetttings(&pEvent);
                    } else if (pEvent.eventId == VP_LINE_EVID_RD_OPTION) {
                        InitReadOptions(&pEvent);
                    }
                }
            }
        }
    }
    return;
}

/**
 * Function:  UVB_Brd_Init()
 *
 * Description: This function sets the PCLK frequencies of the UVB and connects
 * the I/O data to the small and DIN conectors.  The user will only need this
 * function if this application is running on the UVB.  Otherwise it should be
 * replaced with an equivalent hardware setup.
 */
int
UVB_Brd_Init(void)
{
    int reg_offs_freq = 0;
    int reg_offs_connect = 6;
    int mask = 0xffff;
    int value_freq = 0x0660;
    int value_connect = 0xC;

    if(CPLD_Write( reg_offs_connect, mask, value_connect ) < 0) {
        return -1;
    }
    if(CPLD_Write( reg_offs_freq, mask, value_freq ) < 0) {
        return -1;
    }

    return 0;
}

/**
 * Function:  main()
 *
 * Description: This function initializes the API and Devices/Lines used on
 * the -0030 LM.  It also starts the "alarmHandle" function to be called at
 * a periodic rate.
 */
int
main(void)
{
    struct itimerval timeStruct, timeOldVal;
    int deviceNum, lineNum;
    VpStatusType status;
    VpDevCtxType *pDevCtxLocal = VP_NULL;
    VpLineCtxType *pLineCtxLocal = VP_NULL;

    /*
     * The MPI layer of the API accepts an application specific deviceId not
     * used by the API. For the UVB, the deviceId corresponds to a CPLD value
     * that connects the Data I/O signals to either the small connector or the
     * DIN connector, and in 3-wire (MPI) or 4-wire (SPI) mode.
     */
    uint8 hrdCfg = MPI_MODE;

    /*
     * The following structure is specific to Linux.  It sets up a timed call
     * to a user specified function.  For the purpose of this demo, the timing
     * is 5mS and the function that will be called is "alarmHandle".  The
     * timer and function call does not start until "setitimer()" and
     * "signal()" are called (below).
     */
    timeStruct.it_interval.tv_sec = 0;
    timeStruct.it_interval.tv_usec = 5000;
    timeStruct.it_value.tv_sec = 0;
    timeStruct.it_value.tv_usec = 5000;

    timeOldVal.it_interval.tv_sec = 0;
    timeOldVal.it_interval.tv_usec = 0;
    timeOldVal.it_value.tv_sec = 0;
    timeOldVal.it_value.tv_usec = 0;

    /*
     * UVB_Brd_Int() is an initialization specific to the UVB board that sets
     * the UVB PCLK frequencies (to 8.192MHz) and connects the digital signals
     * to both the small and DIN connectors.
     */
    if(UVB_Brd_Init() < 0) {
        return -1;
    };

    /*
     * Simple loop to create the device/line contexts and objects for the
     * API
     */
    for (deviceNum = 0; deviceNum < NUM_DEVICES; deviceNum++) {
        pDevCtxLocal = &pDevCtx[deviceNum];
        VpMakeDeviceObject(VP_DEV_790_SERIES, ((deviceNum << 3) | hrdCfg),
            pDevCtxLocal, &pDevObj[deviceNum]);

        /* Make the Line Objects */
        for (lineNum = 0; lineNum < NUM_LINES; lineNum++) {
            pLineCtxLocal = &pLineCtx[lineNum];
            VpMakeLineObject(VP_TERM_FXS_GENERIC, lineNum, pLineCtxLocal,
                &pLineObj[lineNum], pDevCtxLocal);
        }
    }

    /*
     * Initialize the device and lines associated with this device.  Then start
     * the API tick and look for the Device Init Complete Event.  The device
     * and lines are not ready for any other function until the device init
     * event occurs.
     */

    pDevCtxLocal = &pDevCtx[0];
    status = VpInitDevice(pDevCtxLocal, DEV_PROFILE, AC_600R_FXS, DC_25MA_CC,
        RING_20HZ_SINE, VP_PTABLE_NULL, VP_PTABLE_NULL);

    if(status != VP_STATUS_SUCCESS) {
        printf("\n\rDevice Did Not Initialize Properly\t%d\n\r", status);
        fflush(NULL);
        return -1;
    }

    /*
     * The following two function calls were described above as Linux specific.
     * They are used to call a user defined function at set increments.  The
     * user may have several other methods to perform the same.
     */
    signal(SIGALRM, alarmHandle);
    setitimer(ITIMER_REAL, &timeStruct, &timeOldVal);

    /*
     * Once the "tick" is setup, the main program can enter a "while(1)" loop
     * and the user specified function will take over the application (with
     * the help of the Linux OS)
     */
    while(1);
}


