/** \file vtd_mem_test.c
 * vtd_mem_test.c
 *
 * This file contains the platform independent code used to test the read/write
 * access to the VCP2 class of devices over the HBI bus using the
 * Hardware Abstraction Layer (HAL).
 */

#include <stdio.h>
#include <stdlib.h>
#include "vp_hal.h"
#include "brd_init.h"

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

/* Conditional Compile Flags */
#undef EXIT_ON_HW_VERSION_REG_FAIL		/* define this condition compile flag to
                                         * exit upon failure to read
                                         * Hardware Version Register. */

#undef HBI_HAL_DEBUG

#ifdef HBI_HAL_DEBUG
#define NEED_MORE_DEBUG
#endif /* HBI_HAL_DEBUG */

#define NEED_MORE_DEBUG

#define TEST_SMALL_BLOCKS

#ifdef TEST_SMALL_BLOCKS
#define NEED_MORE_DEBUG
#endif /* TEST_SMALL_BLOCKS */

/* Macro for extracting the offset value from an AMBA address */
#define GET_AMBA_OFFSET_ADD(AMBA)	((AMBA[1] & 0xFF) / 2)

#define TEST_MULTI_WORD_RD_WR   /* Enable this conditional compile flag to
                                 * verify more than one word at a time */

#define TEST_BOOT_LOAD_FUNCTION   /* Define this to enable to test the
                                   * VpHalHbiBootWr() function
                                   */


#ifdef TEST_MULTI_WORD_RD_WR
#define VERIFY_BLOCK_SIZE (16)  /* Specify a verify block size */
#if ((128 % VERIFY_BLOCK_SIZE) != 0) || (VERIFY_BLOCK_SIZE > 32)
#error Please specify an appropriate verify block size
#endif /* VERIFY_BLOCK_SIZE */
#endif /* TEST_MULTI_WORD_RD_WR */


/* The following structure is used to define memroy blocks in VCP2 */
typedef struct {
    uint32 start;              	/* AMBA Start Address; AMBA addresses are 32
                                 * bits wide */
    uint16 length;			   	/* Length of memory block being tested */
    const char* text;			/* Memory block descriptor */
} MemDescriptorType;


unsigned int chipSelect = USE_CHIP_SELECT_ID;

/*
 * This function tests the VCP2 Memory blocks. It verifies all DM, PM but
 * it does not verify the CM. It performs this test through Page 255. Before it
 * does the memory testing it first checks to see if it can identify the
 * hardware silicon version by reading the hardware silicon version register.
 *
 * This function returns TRUE if it recognizes silicon revision version and
 * memory test passes. If Memory test fails, it returns FALSE.
 */
int DoVtdMemTest(void)
{
    uint16 ambaAdd[2];
    uint8 offset;
#ifdef NEED_MORE_DEBUG
    uint16 rdBase[2];
#endif /* NEED_MORE_DEBUG */
    uint16 hbiCmd;
    uint16 wrData, rdData;
    uint16 count;
    int i;
    uint8 hardwareVersion;

#ifdef TEST_MULTI_WORD_RD_WR
    uint16 wrDataMw[VERIFY_BLOCK_SIZE], rdDataMw[VERIFY_BLOCK_SIZE];
#endif /* TEST_MULTI_WORD_RD_WR */

#ifdef TEST_REVERSE_ORDER_WR_RD
    uint16 wrDataRo[REVERSE_ORDER_SIZE], rdDataRo;
#endif /* TEST_REVERSE_ORDER_WR_RD */

#ifdef  TEST_BOOT_LOAD_FUNCTION
    #define BOOT_WRITE_TEST_NUM_WORDS (3)
    int numWords;
    uint8 bootFuncWriteData[2 * BOOT_WRITE_TEST_NUM_WORDS + 10];
    uint16 bootDataRead[BOOT_WRITE_TEST_NUM_WORDS];
    uint16 inputData;
#endif  /* TEST_BOOT_LOAD_FUNCTION */


    /* Call the HAL layer function to initialize CPLD(if any) and the HBI
     * configuration register */
    printf("Initializing CPLD and PINCONFIG register...\n");
    if(VpHalHbiInit(chipSelect) == FALSE){
        printf("Failed in VpHalHbiInit() HAL function");
        return FALSE;
    }

    /* Select the Page 255 */
    hbiCmd = 0xfe00 + 255;
    printf("Selecting Page 255\n");
    if(VpHalHbiCmd(chipSelect, hbiCmd) == FALSE) {
        printf("Failed in VpHalHbiCmd() HAL function");
        return FALSE;
    }

    /***************************************************************************
     * First Read VCP2 version register
     **************************************************************************/
    printf("Verifying HWREVCODE register...\n");
    ambaAdd[0] = 0x7fff;		/* AMBA address of the Hardware Revision Code */
    ambaAdd[1] = 0xfc42;        /* Register */
    offset = GET_AMBA_OFFSET_ADD(ambaAdd);
    
    /* Set the Page 255 base address */
    hbiCmd = 0x8681;
    printf("Setting Page 255 base register: 0x%04x%04x\n",
        ambaAdd[0], ambaAdd[1]);
    if(VpHalHbiWrite(chipSelect, hbiCmd, 1, ambaAdd)  == FALSE) {
        printf("Failed in VpHalHbiWrite() HAL function");
        return FALSE;
    }
#ifdef NEED_MORE_DEBUG
    VpHalHbiRead(chipSelect, 0x8601, 1, rdBase);
    printf("Read Page 255 base register: 0x%04x%04x\n",
        rdBase[0], rdBase[1]);
#endif /* NEED_MORE_DEBUG */
    /* Read and extract hardware revision code */
    hbiCmd = 0x0000 + (offset<<8);
    printf("Offset %04x   hbiCmd %04x\n",offset,hbiCmd);
    if(VpHalHbiRead(chipSelect, hbiCmd, 0, &rdData) == FALSE) {
        printf("Failed in VpHalHbiRead() HAL function");
        return FALSE;
    }
    printf("HBI-Access:Add:0x%04x%04x, Read: 0x%04x\n",
        ambaAdd[0], ambaAdd[1], rdData);
    hardwareVersion = (rdData & 0x00f8)>>3;
    printf("\n***********************************************\n");
    if(hardwareVersion == 1) {
        printf("Emulator found.\n");
    } else if(hardwareVersion == 2) {
        printf("Revision AAA VCP2 silicon found.\n");
    } else if(hardwareVersion == 3) {
        printf("Revision ABA VCP2 silicon found.\n");
    } else {
        /* NOTE: Future versions of Silicon revision need modification here..*/
        printf("ERROR: Unrecognized silicon revision code = %d; ",
                    hardwareVersion);
#ifdef EXIT_ON_HW_VERSION_REG_FAIL
        printf("Exiting memory test...\n");
        return FALSE;
#else
        printf("But contniuing with memory test...\n");
#endif /*  EXIT_ON_HW_VERSION_REG_FAIL */
    }
    printf("\n***********************************************\n");

    /* Read the PCLKB register */
    #define PCLKSEL_B_OFF 8
    #define PCLKSEL_A_OFF 9
    hbiCmd = 0x8000 + (PCLKSEL_B_OFF<<8);
    if(VpHalHbiRead(chipSelect, hbiCmd, 0, &rdData) == FALSE) {
        printf("Failed in VpHalHbiRead() HAL function");
        return FALSE;
    }

    printf("PCLKSEL Register = 0x%04x", rdData);

    hbiCmd = 0x8000 + (PCLKSEL_A_OFF<<8);
    if(VpHalHbiRead(chipSelect, hbiCmd, 0, &rdData) == FALSE) {
        printf("Failed in VpHalHbiRead() HAL function");
        return FALSE;
    }

    printf("%04x\n", rdData);
    printf("Please make sure this register setting is correct based on the\n"
           "clock signal provided to the VCP2 device.\n"
           "NOTE: By default this register is configured to autodetect\n"
           "PCLK freq given FS.\n\n");

    /***************************************************************************
     * Now perform memory read write test
     **************************************************************************/
    printf("Starting Memory Test...\n");

    ambaAdd[0] = 0x000A;
    ambaAdd[1] = 0x0000;
    
    srand(0);

#ifdef TEST_SMALL_BLOCKS
    for(count = 0; count < 0x5; count++) {
#else
    for(count = 0; count < MemTable[index].length; count++) {
#endif /* TEST_SMALL_BLOCKS */

        offset = count & 0x7F;

        /* Set the Page 255 base address for every 128 words */
        if(offset == 0) {
            /* Set the Page 255 base address */
            hbiCmd = 0x8681;

#ifdef NEED_MORE_DEBUG
            printf("Setting Page 255 base register: 0x%04x%04x\n",
                ambaAdd[0], ambaAdd[1]);
#endif /* NEED_MORE_DEBUG */
            VpHalHbiWrite(chipSelect, hbiCmd, 1, ambaAdd);
#ifdef NEED_MORE_DEBUG
            VpHalHbiRead(chipSelect, 0x8601, 1, rdBase);
            printf("Read Page 255 base register: 0x%04x%04x\n",
                rdBase[0], rdBase[1]);
#endif /* NEED_MORE_DEBUG */
#ifdef HBI_HAL_DEBUG
            printf("-");
#endif /* HBI_HAL_DEBUG */
        }

        /* Now perform a paged write and read access */
        /* Write Paged offset */
        hbiCmd = 0x0080 + (offset<<8);
        wrData = rand();
        VpHalHbiWrite(chipSelect, hbiCmd, 0, &wrData);
        /* Read Paged offset */
        hbiCmd = 0x0000 + (offset<<8);
        VpHalHbiRead(chipSelect, hbiCmd, 0, &rdData);

#ifdef NEED_MORE_DEBUG
        printf("HBI-Access:Add:0x%04x%04x, Write:0x%.4x, Read: 0x%04x\n",
                ambaAdd[0], ambaAdd[1], wrData, rdData);
#endif /* NEED_MORE_DEBUG */
        if(rdData != wrData) {
            return FALSE;
        }
        ambaAdd[1] += 1;
        if (ambaAdd[1] == 0) {
            ambaAdd[0] += 1;
        }
    }
#ifdef HBI_HAL_DEBUG
    printf("\n<<< Testing %s Done\n", MemTable[index].text);
#endif /* HBI_HAL_DEBUG */

    /***************************************************************************
     * Now perform memory multi word read write test
     **************************************************************************/
#ifdef TEST_MULTI_WORD_RD_WR
    printf("Starting Memory multi word (Access size = %d) HBI access "
                "verification...\n", VERIFY_BLOCK_SIZE);

    ambaAdd[0] = 0x000A;
    ambaAdd[1] = 0x0100;
    
    srand(100);

#ifdef TEST_SMALL_BLOCKS
    for(count = 0; count < 0x5; count = count + VERIFY_BLOCK_SIZE) {
#else
    for(count = 0; count < MemTable[index].length; count = count + VERIFY_BLOCK_SIZE) {
#endif /* TEST_SMALL_BLOCKS */

        offset = count & 0x7F;

        /* Set the Page 255 base address for every 128 words */
        if(offset == 0) {
            /* Set the Page 255 base address */
            hbiCmd = 0x8681;

#ifdef NEED_MORE_DEBUG
            printf("Setting Page 255 base register: 0x%04x%04x\n",
                ambaAdd[0], ambaAdd[1]);
#endif /* NEED_MORE_DEBUG */
            VpHalHbiWrite(chipSelect, hbiCmd, 1, ambaAdd);
#ifdef NEED_MORE_DEBUG
            VpHalHbiRead(chipSelect, 0x8601, 1, rdBase);
            printf("Read Page 255 base register: 0x%04x%04x\n",
                rdBase[0], rdBase[1]);
#endif /* NEED_MORE_DEBUG */
#ifdef HBI_HAL_DEBUG
            printf("-");
#endif /* HBI_HAL_DEBUG */
        }

        /* Now perform a paged write and read access */
        /* Write Paged offset */
        hbiCmd = 0x0080 + (offset<<8) + (VERIFY_BLOCK_SIZE - 1);
        for(i=0; i<VERIFY_BLOCK_SIZE; i++){
    		wrDataMw[i] = rand();
        }
        VpHalHbiWrite(chipSelect, hbiCmd, VERIFY_BLOCK_SIZE - 1, wrDataMw);
        /* Read Paged offset */
        hbiCmd = 0x0000 + (offset<<8) + (VERIFY_BLOCK_SIZE - 1);
        VpHalHbiRead(chipSelect, hbiCmd, VERIFY_BLOCK_SIZE - 1, rdDataMw);

        for(i=0; i<VERIFY_BLOCK_SIZE; i++) {
#ifdef NEED_MORE_DEBUG
            printf("HBI-Access:Add:0x%04x%04x, Write:0x%.4x, Read: 0x%04x\n",
                    ambaAdd[0], ambaAdd[1], wrDataMw[i], rdDataMw[i]);
#endif /* NEED_MORE_DEBUG */
            if(rdDataMw[i] != wrDataMw[i]) {
                return FALSE;
            }
            ambaAdd[1] += 1;
            if (ambaAdd[1] == 0) {
                ambaAdd[0] += 1;
            }
        }
    }
#ifdef HBI_HAL_DEBUG
    printf("\n<<< Testing %s Done\n", MemTable[index].text);
#endif /* HBI_HAL_DEBUG */
#endif /* TEST_MULTI_WORD_RD_WR */


#ifdef  TEST_BOOT_LOAD_FUNCTION
    /* Test the VpHalHbiBootWr() function by writing known data (containing)
     * HBI commands + data and the by reading the data that was written */

    /* Select page 255 */
    i = 0;
    bootFuncWriteData[i++] = 0xfe;
    bootFuncWriteData[i++] = 0xff;
    /* Write page 255 base address */
    bootFuncWriteData[i++] = 0x86;
    bootFuncWriteData[i++] = 0x81;
    bootFuncWriteData[i++] = 0x00;
    bootFuncWriteData[i++] = 0x0A;
    bootFuncWriteData[i++] = 0x02;
    bootFuncWriteData[i++] = 0x00;

    /* Write to the selected page (write three words) */
    bootFuncWriteData[i++] = 0x00;
    bootFuncWriteData[i++] = 0x82;

    bootFuncWriteData[i++] = 0xac;
    bootFuncWriteData[i++] = 0xbc;

    bootFuncWriteData[i++] = 0xde;
    bootFuncWriteData[i++] = 0xad;

    bootFuncWriteData[i++] = 0xbe;
    bootFuncWriteData[i++] = 0xef;

    printf("\nTesting the VpHalHbiBootWr() function\n");

    /* Write the data */
    numWords = (i>>1)-1;
    VpHalHbiBootWr(chipSelect, numWords, &bootFuncWriteData[0]);

    /* Read the data (read three words) */
    hbiCmd = 0x0002;
    VpHalHbiRead(chipSelect, hbiCmd, 2, &bootDataRead[0]);

    for(count = 0; count <= 2; count++) {

        inputData = bootFuncWriteData[2 * count + 10];
        inputData = inputData<<8;
        inputData |= bootFuncWriteData[2 * count + 10 + 1];

        if(bootDataRead[count] != inputData) {
            printf("Data does not match: Wrote 0x%x, Read: 0x%04x\n",
                inputData, bootDataRead[count]);
            return FALSE;
        }
    }
#endif  /* TEST_BOOT_LOAD_FUNCTION */

    return TRUE;
} /* DoVtdMemTest */

/*
 * Main function. Takes chipselect as an input argument.
 */
int main(int argc, char *argv[])
{

#ifdef LEGERITY_UVB_DEMO_BOARD
    /*
     * Zarlink UVB provides two connectors where a HBI devices could be
     * mounted.
     */

    if(argc >= 2) {
        /* User has specified a chip select other than default */
        if(sscanf(argv[1], "%x", &chipSelect) != 1) {
            printf("\n ***ERROR*** Expecting a chip select value; "
                   "Ex: 0x%x or 0x%x\n", DIN_CON_CS1_HBI8, SMALL_CON_CS1_HBI8);
            return 0;
        }

        if((chipSelect != DIN_CON_CS1_HBI8) &&
           (chipSelect != SMALL_CON_CS1_HBI8)) {

            printf("\n ***ERROR*** Expecting a chip select value; "
                   "Ex: 0x%x or 0x%x\n", DIN_CON_CS1_HBI8, SMALL_CON_CS1_HBI8);
            return 0;
        }
    }
#endif /* LEGERITY_UVB_DEMO_BOARD */

    /* First initialize the board */
    InitializeBoard();

    if(DoVtdMemTest()==TRUE){
        printf("\nSUCCESS:: Memory test passed\n");
    } else {
        printf("\nERROR:: Memory test Failed\n");
    }

    return 0;
}

