/** \file mpi_test_vcp2.c
 * mpi_test_vcp2.c
 *
 * Copyright (c) 2008, Zarlink Semiconductor, Inc.
 */

/* INCLUDES */
#include "app_des.h"    /* Provides Description of hardware and VCP-firmware
                         * product type*/
#include "vp_api.h"
#include "vp_hal.h"
#include <stdlib.h>
#include <stdio.h>
#ifndef term_printf
/* Define the following as necessary to redirect the debug output */
#define term_printf printf
#endif /* term_printf */

/*******************************************************************************
 * Conditional Compile time options; Change these based on your software       *
 * architecture and other needs.                                               *
 ******************************************************************************/
#undef EXIT_ON_ERROR       /* Define this to exit on error */
#define POLL_VCP_EVENTS /* Define this if you would like MpiTestAppMain() to
                         * poll for VCP events and handle them. If VCP events
                         * are handled through a interrupt service routine (in
                         * an interrupt context) you do not need
                         * to define this. */

/* Create an instance of VCP device context and Device Object
 * Note that memory for these instances can be obtained from memory management
 * functions (like malloc). */
VpVcp2DeviceObjectType devObj[MAX_DEVICES_ON_BOARD];
VpDevCtxType devCtx[MAX_DEVICES_ON_BOARD];

/* Create instance of VCP line context and line Object
 * Note that memory for these instances can be obtained from memory management
 * functions (like malloc). */
VpLineCtxType lineCtx[MAX_LINES_ON_BOARD];
VpVcp2LineObjectType lineObj[MAX_LINES_ON_BOARD];

/* It is assumed here that boot load image is stored in the global memory.
 * Notes about storing boot image:
 * Every system is different in its mechanisms in storing the data in FLASH
 * Some systems might just make use FLASH as just an extension of their overall
 * memory (either directly or through mapping appropriate page). In such systems
 * obtaining boot image is nothing but obtaining right address (and page) where
 * image is stored.
 * Other systems might make use of FLASH file system. In such systems obtaining
 * boot image is like reading a file. In this case the boot image file would
 * have to be read from the FLASH. Flash file system would provide appprpriate
 * calls to perform such read accesses.
 *
 * Other notes about length of boot image that can be passed to VpBootLoad:
 * VpBootImage function can take the entire length of the boot image or it can
 * be called many times with smaller block sizes. This is described in the
 * VP-API specification.
 */
extern VpImagePtrType bootImage; /* pointer to the VCP firmware image in RAM */
extern unsigned long mpiTestBootImageLen;  /* size of the VCP firmware image */

/* If using a compressed VCP2 boot image, then VpBootLoad() requires some
 * temporary scratch memory to do the decompression.  Here we allocate it
 * statically, for simplicity. */
static VpScratchMemType scratchMem;

VpStatusType VtdEventHandler(VpEventType *pEvent);

/* Include the appropriate profiles */
#include "profiles8_790.h"

uint32 timeStampCounter = 0;
/*******************************************************************************
 *                      Function Declarations                                  *
 ******************************************************************************/
static VpStatusType InitAfterBoot(void);
static void PrintEvent(VpEventType *pEvent);

/* strings corresponding to VpStatusType */
char *retstring[] = {
    "VP_STATUS_SUCCESS",
    "VP_STATUS_FAILURE",
    "VP_STATUS_FUNC_NOT_SUPPORTED",
    "VP_STATUS_INVALID_ARG",
    "VP_STATUS_MAILBOX_BUSY",
    "VP_STATUS_ERR_VTD_CODE",
    "VP_STATUS_OPTION_NOT_SUPPORTED",
    "VP_STATUS_ERR_VERIFY",
    "VP_STATUS_DEVICE_BUSY",
    "VP_STATUS_MAILBOX_EMPTY",
    "VP_STATUS_ERR_MAILBOX_DATA",
    "VP_STATUS_ERR_HBI",
    "VP_STATUS_ERR_IMAGE",
    "VP_STATUS_IN_CRTCL_SECTN",
    "VP_STATUS_DEV_NOT_INITIALIZED",
    "VP_STATUS_ERR_PROFILE",
    "VP_STATUS_INVALID_VOICE_STREAM"};

/* strings corresponding to VpLineStateType */
char *statestring[] = {
    "STANDBY",
    "TIP_OPEN",
    "ACTIVE",
    "ACTIVE_POLREV",
    "TALK",
    "TALK_POLREV",
    "OHT",
    "OHT_POLREV",
    "DISCONNECT",
    "RINGING",
    "RINGING_POLREV",
    "FXO_OHT",
    "FXO_LOOP_OPEN",
    "FXO_LOOP_CLOSE",
    "FXO_TALK",
    "FXO_RING_GND"
};

/*
 * Displays the least-significant 'bits' bits of 'val', with a space between
 * every four bits.
 */
void
PrintBinary(
    int8 bits,
    uint32 val)
{
    int8 bit;

    for (bit = bits - 1; bit >= 0; bit--) {
        /* Print a space between nybbles. */
        if (!((bit + 1) % 4) && ((bit + 1) != bits)) {
            term_printf(" ");
        }
        term_printf("%c", '0' + (uint8)((val >> bit) & 1));
    }
} /* PrintBinary() */

/*
 * MpiTestAppMain()
 * This function is the main function for VP-API MPI Test application.
 * This function could be called from the "main()" after the necessary board
 * specific initialization (based on the board). If there is no such board set
 * up requied, then this function could be renamed as "main" itself.
 */
VpStatusType
MpiTestAppMain(
    void)
{
    VpDeviceIdType deviceId = USE_CHIP_SELECT_ID; /* Chip select ID of the VCP */
    VpStatusType status;
    VpEventType event;
    bool eventPresent = FALSE;

    term_printf("\n\nWelcome to the Zarlink Semiconductor VP-API MPI Test application.\n\n");

    /* Create and initialize VCP device context */
    term_printf("Creating Device Object...");
    status = VpMakeDeviceObject(VP_DEV_VCP2_SERIES, deviceId, &devCtx[0],
                                &devObj[0]);
    if(status != VP_STATUS_SUCCESS) {
        return status; /* Something is wrong */
    }
    term_printf("done\n");

    /* Bootload the VCP2 */
    term_printf("Downloading VCP2 boot image...");
    status = VpBootLoad(&devCtx[0], VP_BOOT_STATE_FIRSTLAST, bootImage,
                        mpiTestBootImageLen, &scratchMem, VP_BOOT_MODE_VERIFY);
    if(status != VP_STATUS_SUCCESS) {
        return status; /* Something is wrong */
    }
    term_printf("done\n");

    if(status != VP_STATUS_SUCCESS) {
        return status; /* Something is wrong */
    }
    term_printf("done\n");

#ifdef POLL_VCP_EVENTS
    /* An eternal loop for Quick Start Application */
    while(1){
        eventPresent = VpGetEvent(&devCtx[0], &event);
        if(event.status != VP_STATUS_SUCCESS) {
            if(event.status < VP_STATUS_NUM_TYPES) {
                term_printf("ERROR: VpGetEvent = %s.\n\n",
                                retstring[event.status]);
            } else {
                term_printf("ERROR: VpGetEvent = %d.\n\n", event.status);
            }
        } else if(eventPresent == TRUE) {
            status = VtdEventHandler(&event);
            if(status != VP_STATUS_SUCCESS) {
#ifdef EXIT_ON_ERROR
                return status; /* Something is wrong */
#else
                if(event.status < VP_STATUS_NUM_TYPES) {
                    term_printf("ERROR: VtdEventHandler = %s.\n\n",
                                    retstring[status]);
                } else {
                    term_printf("ERROR: VtdEventHandler = %d.\n\n", status);
                }
#endif
            }
        }
    }
#else
    return VP_STATUS_SUCCESS;
#endif /* POLL_VCP_EVENTS */
} /* QuickStartAppMain() */


/*
 * InitAfterBoot()
 * This function creats line contexts and also initiates device initialization
 * functions.
 *
 * Preconditions:
 * Assumes the device is boot loaded and boot complete event has been received.
 *
 * Postconditions:
 * Calls VpInitDevice().
 */
static VpStatusType
InitAfterBoot(
    void)
{
    VpStatusType status;
    int i;
    VpTermType termType;

    /* Create all the line contexts */
    for (i = 0; i < MAX_LINES_ON_BOARD; i++)  {
        term_printf("creating Line Object %d, TermType = ", i);

        termType = VP_TERM_FXS_TO_TL;
        term_printf("VP_TERM_FXS_TO_TL...");

        /* Ignore the termination type for now. This functionality will
         * be supported later */
        status = VpMakeLineObject(termType, i, &lineCtx[i], &lineObj[i],
                                    &devCtx[0]);
        if(status != VP_STATUS_SUCCESS){
            return status; /* Something is wrong */
        }
        term_printf("done\n");
    }

    /* Initialize the device */
    term_printf("Initializing VCP...");
    /*status = VpInitDevice(&devCtx[0], &DEV_PROFILE[0], &AC_600R_FXS[0],
                            &DC_25MA_CC[0], &RING_20HZ_SINE[0], VP_PTABLE_NULL,
                            VP_PTABLE_NULL);
    */
    status = VpInitDevice(&devCtx[0], &DEV_PROFILE[0], VP_PTABLE_NULL,
                            VP_PTABLE_NULL, VP_PTABLE_NULL, VP_PTABLE_NULL,
                            VP_PTABLE_NULL);

    if(status != VP_STATUS_SUCCESS){
        return status; /* Something is wrong */
    }
    term_printf("done\n");
    return status;
}

/*
 * VtdEventHandler()
 * This function handles the events from the VCP and takes necessary action
 * based on those events.
 *
 * Preconditions:
 * This function could be called from either the interrupt context or it could
 * be called from a routine that is polling for events. In either case this
 * function SHOULD be passed a pointer to an event that was filled by the
 * VpGetEvent().
 *
 * Postconditions:
 * This line state/device states are modified by this function.
 * This function returns VP_STATUS_SUCCESS if everything went right
 */
VpStatusType
VtdEventHandler(
    VpEventType *pEvent)
{
    VpStatusType status = VP_STATUS_FAILURE;
    VpChkSumType chkSumInfo;
    uint8 chanId = pEvent->channelId;
    VpOptionEventMaskType eventMask;

    /* Take necessary action based on the event */
    if(pEvent->eventCategory == VP_EVCAT_RESPONSE) {
        switch  (pEvent->eventId) {
            case VP_DEV_EVID_BOOT_CMP: {
                /* Read the results associated with this event */
                term_printf("Received event: VP_EVID_BOOT_CMP\n");
                term_printf("Reading Checksum...");
#define  TEST_POLL_RESPONSE_EVENT
#ifdef  TEST_POLL_RESPONSE_EVENT
                pEvent->eventId = 0;
                status = VpGetResults(pEvent, VP_NULL);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                PrintEvent(pEvent);
#endif /* TEST_POLL_RESPONSE_EVENT */

                status = VpGetResults(pEvent, &chkSumInfo);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }
                term_printf("done. Checksum Data:\n");
                term_printf("VpChkSumType.loadChkSum = 0x%lx\n",
                                chkSumInfo.loadChecksum);
                term_printf("VpChkSumType.vInfo.vtdRevCode = 0x%x\n",
                                chkSumInfo.vInfo.vtdRevCode);
                term_printf("VpChkSumType.vInfo.swProductId = 0x%x\n",
                                chkSumInfo.vInfo.swProductId);
                term_printf("VpChkSumType.vInfo.swVerMajor = 0x%x\n",
                                chkSumInfo.vInfo.swVerMajor);
                term_printf("VpChkSumType.vInfo.swVerMinor = 0x%x\n",
                                chkSumInfo.vInfo.swVerMinor);

                /* Create Line Objects and Initialize the device */
                status = InitAfterBoot();
                break;
            }
            case VP_DEV_EVID_DEV_INIT_CMP: {
                term_printf("Received event: VP_DEV_EVID_DEV_INIT_CMP(%d)\n",
                                chanId);

                /* Veryify Calibration was successful */
                term_printf("Device initialization complete.  Detected SLACs:");
                PrintBinary(16, (int32) pEvent->eventData);
                term_printf("\n");

                if(pEvent->eventData == 0) {
                    term_printf("ERROR: Device initialization Failed; "
                                "No SLAC found; MPI Test requires at least one SLAC; Exiting...\n");

                    status = VP_STATUS_FAILURE;
                    break;
                }

                eventMask.faults = VP_EVCAT_FAULT_UNMASK_ALL;
                eventMask.signaling = VP_DEV_EVID_TS_ROLLOVER;
                eventMask.response = VP_DEV_EVID_DNSTR_MBOX;
                eventMask.test = VP_EVCAT_TEST_UNMASK_ALL;
                eventMask.process = VP_EVCAT_PROCESS_UNMASK_ALL;
                eventMask.fxo = VP_EVENT_MASK_ALL; /* Not applicable to VCP */
                eventMask.packet = VP_EVENT_MASK_ALL;  /* Not applicable */

                term_printf("Setting up events masks...\n");
                status = VpSetOption(VP_NULL, pEvent->pDevCtx,
                                      VP_OPTION_ID_EVENT_MASK, &eventMask);
                if(status != VP_STATUS_SUCCESS){
                    break; /* Something is wrong */
                }

                term_printf("done\n");
                break;
            }
            case VP_LINE_EVID_LINE_INIT_CMP: {
                term_printf("Received event: VP_DEV_EVID_LINE_INIT_CMP(%d)\n",
                                chanId);
                status = VP_STATUS_SUCCESS;
                break;
            }
            case VP_DEV_EVID_DNSTR_MBOX:    {
                term_printf("Received event: VP_DEV_EVID_DNSTR_MBOX(%d)\n",
                    pEvent->eventData);
                status = VP_STATUS_SUCCESS;
                break;
            }
        }
    } else if(pEvent->eventCategory == VP_EVCAT_SIGNALING) {
        switch  (pEvent->eventId) {
            case VP_LINE_EVID_HOOK_ON: {
                term_printf("Received event: VP_EVID_HOOK_ON(%d); TS = %u\n",
                            chanId, pEvent->parmHandle);
                break;
            }
            case VP_LINE_EVID_HOOK_OFF: {
                term_printf("Received event: VP_EVID_HOOK_OFF(%d); TS = %u\n",
                            chanId, pEvent->parmHandle);
                break;
            }
            case VP_DEV_EVID_TS_ROLLOVER: {
                timeStampCounter++;
                /* Uncomment if necessary */
                term_printf("Received event: VP_DEV_EVID_TS_ROLLOVER; Count: "
                                "%lu \n", timeStampCounter);
                status = VP_STATUS_SUCCESS; /* Event is consumed */
                break;
            }
        }
    } else if (pEvent->eventCategory == VP_EVCAT_TEST) {
        /*
         * If all is well, you should get "0" for the line field and "success"
         * (0) for the param field. If there is a failure, you will get the SLAC
         * number that the failure was on for the line field. In the param field
         * you will get the expected data to be read back in the high byte and
         * the actual data read back for that failure in the low byte.
         */
        switch ( pEvent->eventId) {
            case VP_DEV_EVID_STEST_CMP: {
                if (pEvent->eventData != 0)
                {
                    term_printf("MPI problem detected on SLAC #%d: expected to read back 0x%2.2X, got 0x%2.2X\n",
                                pEvent->channelId,
                                (pEvent->eventData >> (int16)8),
                                (pEvent->eventData & 0x00FF));
                    status = VP_STATUS_ERR_VERIFY;
                } else {
                    term_printf("No mpi errors...\n");
                    status = VP_STATUS_SUCCESS;
                }
                break;
            }
        }
    }

    if(status != VP_STATUS_SUCCESS){

        /* Either the event was not handled or there is error in handling the
         * event. Print the event. */
        term_printf("ERROR: Unknown event or Error while handling the event; "
                    "Printing the event...");
        PrintEvent(pEvent);

        /* If Event has a result and it has not been handled until now,
         * Clear it */
        if(pEvent->hasResults == TRUE) {
            term_printf("WARNING: Unhandled results, Clearing it...\n");
            VpClearResults(pEvent->pDevCtx);
        }
    }

    return status;
} /* VtdEventHandler() */


/*
 * PrintEvent()
 * This function an un-handled event.
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * None
 */
static void
PrintEvent(
    VpEventType *pEvent)
{
    char *eventCategory[] = {
        "VP_EVCAT_FAULT",
        "VP_EVCAT_SIGNALING",
        "VP_EVCAT_RESPONSE",
        "VP_EVCAT_TEST",
        "VP_EVCAT_PROCESS",
        "VP_EVCAT_FXO",
        "VP_EVCAT_PACKET",
    };

    if(pEvent->eventCategory < VP_NUM_EVCATS) {
        term_printf("VpEvent Cat:Id = %s:",
                        eventCategory[pEvent->eventCategory]);
    } else {
        term_printf("VpEvent Cat:Id = 0x%x:",
                        pEvent->eventCategory);
    }

    if(pEvent->eventCategory == VP_EVCAT_FAULT) {
        switch (pEvent->eventId) {
            case VP_DEV_EVID_BAT_FLT:   {
                term_printf("VP_DEV_EVID_BAT_FLT\n"); break;
            }
            case VP_DEV_EVID_CLK_FLT:   {
                term_printf("VP_DEV_EVID_CLK_FLT\n"); break;
            }
            case VP_LINE_EVID_THERM_FLT:{
                term_printf("VP_LINE_EVID_THERM_FLT\n"); break;
            }
            case VP_LINE_EVID_DC_FLT:   {
                term_printf("VP_DEVP_LINE_EVID_DC_FLT\n"); break;
            }
            case VP_LINE_EVID_AC_FLT:   {
                term_printf("VP_LINE_EVID_AC_FLT\n"); break;
            }
            case VP_DEV_EVID_EVQ_OFL_FLT: {
                term_printf("VP_DEV_EVID_EVQ_OFL_FLT\n"); break;
            }
            case VP_DEV_EVID_WDT_FLT: {
                term_printf("VP_DEV_EVID_WDT_FLT\n"); break;
            }
            default: {
                term_printf("0x%x\n", pEvent->eventId); break;
            }
        }
    } else if(pEvent->eventCategory == VP_EVCAT_SIGNALING) {
        switch (pEvent->eventId) {
            case VP_LINE_EVID_HOOK_OFF:   {
                term_printf("VP_LINE_EVID_HOOK_OFF\n"); break;
            }
            case VP_LINE_EVID_HOOK_ON:   {
                term_printf("VP_LINE_EVID_HOOK_ON\n"); break;
            }
            case VP_LINE_EVID_GKEY_DET:{
                term_printf("VP_LINE_EVID_GKEY_DET\n"); break;
            }
            case VP_LINE_EVID_GKEY_REL:   {
                term_printf("VP_LINE_EVID_GKEY_REL\n"); break;
            }
            case VP_LINE_EVID_FLASH:   {
                term_printf("VP_LINE_EVID_FLASH\n"); break;
            }
            case VP_LINE_EVID_STARTPULSE: {
                term_printf("VP_LINE_EVID_STARTPULSE\n"); break;
            }
            case VP_LINE_EVID_DTMF_DIG: {
                term_printf("VP_LINE_EVID_DTMF_DIG\n"); break;
            }
            case VP_LINE_EVID_PULSE_DIG: {
                term_printf("VP_LINE_EVID_PULSE_DIG\n"); break;
            }
            case VP_LINE_EVID_MTONE: {
                term_printf("VP_LINE_EVID_MTONE\n"); break;
            }
            case VP_DEV_EVID_TS_ROLLOVER: {
                term_printf("VP_DEV_EVID_TS_ROLLOVER\n"); break;
            }
            default: {
                term_printf("0x%x\n", pEvent->eventId); break;
            }
        }
    } else if(pEvent->eventCategory == VP_EVCAT_RESPONSE) {
        switch (pEvent->eventId) {
            case VP_DEV_EVID_BOOT_CMP:   {
                term_printf("VP_DEV_EVID_BOOT_CMP\n"); break;
            }
            case VP_LINE_EVID_LLCMD_TX_CMP:   {
                term_printf("VP_LINE_EVID_LLCMD_TX_CMP\n"); break;
            }
            case VP_LINE_EVID_LLCMD_RX_CMP:{
                term_printf("VP_LINE_EVID_LLCMD_RX_CMP\n"); break;
            }
            case VP_DEV_EVID_DNSTR_MBOX:   {
                term_printf("VP_DEV_EVID_DNSTR_MBOX\n"); break;
            }
            case VP_LINE_EVID_RD_OPTION:   {
                term_printf("VP_LINE_EVID_RD_OPTION\n"); break;
            }
            case VP_LINE_EVID_RD_LOOP: {
                term_printf("VP_LINE_EVID_RD_LOOP\n"); break;
            }
            case VP_EVID_CAL_CMP: {
                term_printf("VP_EVID_CAL_CMP\n"); break;
            }
            case VP_EVID_CAL_BUSY: {
                term_printf("VP_EVID_CAL_BUSY\n"); break;
            }
            case VP_LINE_EVID_GAIN_CMP: {
                term_printf("VP_LINE_EVID_GAIN_CMP\n"); break;
            }
            case VP_DEV_EVID_DEV_INIT_CMP: {
                term_printf("VP_DEV_EVID_DEV_INIT_CMP\n"); break;
            }
            case VP_LINE_EVID_LINE_INIT_CMP: {
                term_printf("VP_LINE_EVID_LINE_INIT_CMP\n"); break;
            }
            default: {
                term_printf("0x%x\n", pEvent->eventId); break;
            }
        }
    } else if(pEvent->eventCategory == VP_EVCAT_TEST) {
        switch (pEvent->eventId) {
            case VP_LINE_EVID_TEST_CMP:   {
                term_printf("VP_LINE_EVID_TEST_CMP\n"); break;
            }
            case VP_LINE_EVID_DTONE_DET:   {
                term_printf("VP_LINE_EVID_DTONE_DET\n"); break;
            }
            case VP_LINE_EVID_DTONE_LOSS:{
                term_printf("VP_LINE_EVID_DTONE_LOSS\n"); break;
            }
            case VP_DEV_EVID_STEST_CMP:   {
                term_printf("VP_DEV_EVID_STEST_CMP\n"); break;
            }
            case VP_DEV_EVID_CHKSUM:   {
                term_printf("VP_DEV_EVID_CHKSUM\n"); break;
            }
            default: {
                term_printf("0x%x\n", pEvent->eventId); break;
            }
        }
    } else if(pEvent->eventCategory == VP_EVCAT_PROCESS) {
        switch (pEvent->eventId) {
            case VP_LINE_EVID_MTR_CMP:   {
                term_printf("VP_LINE_EVID_MTR_CMP\n"); break;
            }
            case VP_LINE_EVID_MTR_ABORT:   {
                term_printf("VP_LINE_EVID_MTR_ABORT\n"); break;
            }
            case VP_LINE_EVID_CID_DATA:{
                term_printf("VP_LINE_EVID_CID_DATA\n"); break;
            }
            case VP_LINE_EVID_RING_CAD:   {
                term_printf("VP_LINE_EVID_RING_CAD\n"); break;
            }
            default: {
                term_printf("0x%x\n", pEvent->eventId); break;
            }
        }
    } else {
        term_printf("0x%x\n", pEvent->eventId);
    }
    term_printf("VpEventType.channelId = 0x%x\n", pEvent->channelId);
    term_printf("VpEventType.deviceId = 0x%x\n", pEvent->deviceId);
    term_printf("VpEventType.eventData = 0x%x\n", pEvent->eventData);
    term_printf("VpEventType.parmHandle = 0x%x\n", pEvent->parmHandle);
    term_printf("VpEventType.hasResults = 0x%x\n", pEvent->hasResults);
}


