/** \file dvp_api.c
 * dvp_api.c
 *
 *  This file contains the implementation of VCP part of the VP-API
 */

/* INCLUDES */
#include "vp_api.h"

#if defined (VP_CC_VCP_SERIES)  /* Compile Further only if required */

#include "dvp_api.h"
#include "vp_api_int.h"
#include "boot_common.h"
#include "dvp_hbi.h"


/******************************************************************************
 *                        VCP SPECIFIC DEFINES                                *
 ******************************************************************************/
/* VCP specific termination types */
typedef enum
{
    VCP_TERM_GENERIC_FXS,       /* Generic SLIC, No LCAS */
    VCP_FXS_79232_75282,        /* FXS termination with SLIC = Le79232,
                                 * LCAS = Le75282 */
    VCP_FXS_79232_75181,        /* FXS termination with SLIC = Le79232,
                                 * LCAS = Le75181 */
    VCP_FXS_79252_TO_TL,        /* FXS termination with SLIC = Le79252
                                 * With test load and testout relay */
    VCP_FXS_792x2_TITO_TL_R,    /* FXS termination with SLIC = Le79232 or
                                 * Le79252; with single control for test-in/
                                 * test-out relays; with test load; with
                                 * optional external ringing */
    VCP_FXS_792x2_CO_TL,        /* FXS termination with SLIC = Le79232
                                 * Le79242 or Le79252 and cutover relay and
                                 * test load resistor */
    VCP_FXS_79232_RR,           /* FXS termination with SLIC = Le79232
                                 * and relay for external ringing relay
                                 * and reset relay. Reset relay is not
                                 * controlled by the VCP. It is controlled
                                 * by the host to realize the reset function
                                 * by disconnecting the ringing generator from
                                 * the ringing bus. */
    /* DevNotes: The following termination type is not supported in the VCP
     * firmware */
    VCP_FXS_792x2_RDT,          /* FXS termination with SLIC = Le79232
                                 * Le79242 or Le79252 and multipurpose relay
                                 * that is used for ringing, disconnecting the
                                 * loop and also for providing access to testout
                                 * bus. */

    VCP_TERM_ENUM_SIZE = FORCE_STANDARD_C_ENUM_SIZE /* Portability Req. */
} VcpTermType;

#define VCP_RING_CAD_PROF_COUNT  4 /* Software simulated Ringing cadence profile
                             * table size;
                             * Note that VCP provides only one type of
                             * cadence table; Software breaks it in to
                             * two types of tables: one, tone cadence
                             * and two, ringing cadence profile;
                             * Ringing cadence is stored first in the VCP
                             * cadence profile table followed by tone cadence
                             */

#define VCP_TONECAD_PROF_TYPE  VP_PROFILE_RINGCAD /* Since in VCP there is only
                                                   * one type of cadence table*/

#define VCP_TS_RESL_2_API_TS_RESL  (2) /* VCP time stamp resolution (125micro
                                        * seconds) to API time stamp resolution
                                        * (0.5 milli seconds) scaling factor */

#define VCP_SW_TS_CNTR_RESL  (16 - VCP_TS_RESL_2_API_TS_RESL) /* VCP software
                                        * time stamp counter resolution. */

#define VCP_MAX_CID_DATA_LENGTH (32)  /* Maximum CID data length (in bytes) */

/*
 * Events
 *   Here we define an enumeration for all possible VCP S/W events. Note that
 *   this enumeration is broken up (with comments) into the 5 interrupt types,
 *   with invalid event codes specified as VCP_EVID_Tx_RSVDn.
 *   Also, we set up some #define constants to indicate the number of valid
 *   events within each event type. These are used by VcpGetEvent() to provide
 *   extra verification on the events reported by the VCP.
 */
/* Event types returned by the VCP */
typedef enum
{
    VCP_EVID_NO_EVENT        = 0x0000,
    VCP_EVID_INVALID         = 0x0001,
    /* Type 1: Fault Events */
    VCP_EVID_BAT_FLT         = 0x0010,
    VCP_EVID_CLK_FLT         = 0x0011,
    VCP_EVID_THERM_FLT       = 0x0012,
    VCP_EVID_DC_FLT          = 0x0013,
    VCP_EVID_AC_FLT          = 0x0014,
    VCP_EVID_T1_RSVD5        = 0x0015,
    VCP_EVID_T1_RSVD6        = 0x0016,
    VCP_EVID_T1_RSVD7        = 0x0017,
    VCP_EVID_T1_RSVD8        = 0x0018,
    VCP_EVID_T1_RSVD9        = 0x0019,
    VCP_EVID_T1_RSVDA        = 0x001A,
    VCP_EVID_T1_RSVDB        = 0x001B,
    VCP_EVID_T1_RSVDC        = 0x001C,
    VCP_EVID_T1_RSVDD        = 0x001D,
    VCP_EVID_T1_RSVDE        = 0x001E,
    VCP_EVID_T1_RSVDF        = 0x001F,
    /* Type 2: Signaling Events */
    VCP_EVID_HOOK_OFF        = 0x0020,
    VCP_EVID_HOOK_ON         = 0x0021,
    VCP_EVID_GKEY_ON         = 0x0022,
    VCP_EVID_GKEY_OFF        = 0x0023,
    VCP_EVID_FLASH           = 0x0024,
    VCP_EVID_STARTPULSE      = 0x0025,
    VCP_EVID_DTMF_DIG        = 0x0026,
    VCP_EVID_PULSE_DIG       = 0x0027,
    VCP_EVID_MTONE           = 0x0028,
    VCP_EVID_TS_ROLLOVER     = 0x0029,
    VCP_EVID_T2_RSVDA        = 0x002A,
    VCP_EVID_T2_RSVDB        = 0x002B,
    VCP_EVID_T2_RSVDC        = 0x002C,
    VCP_EVID_T2_RSVDD        = 0x002D,
    VCP_EVID_T2_RSVDE        = 0x002E,
    VCP_EVID_T2_RSVDF        = 0x002F,
    /* Type 3: Response Events */
    VCP_EVID_BOOT_CMP        = 0x0030,
    VCP_EVID_MPI_TX_CMP      = 0x0031,
    VCP_EVID_MPI_RX_CMP      = 0x0032,
    VCP_EVID_DNSTR_MBOX      = 0x0033,
    VCP_EVID_RD_OPTION       = 0x0034,
    VCP_EVID_RD_LOOP         = 0x0035,
    VCP_EVID_CAL_CMP         = 0x0036,
    VCP_EVID_CAL_BUSY        = 0x0037,
    VCP_EVID_GAIN_CMP        = 0x0038,
    VCP_EVID_T3_RSVD9        = 0x0039,
    VCP_EVID_T3_RSVDA        = 0x003A,
    VCP_EVID_LINE_INIT_CMP   = 0x003B,
    VCP_EVID_IO_ACCESS_CMP   = 0x003C,
    VCP_EVID_T3_RSVDD        = 0x003D,
    VCP_EVID_T3_RSVDE        = 0x003E,
    VCP_EVID_T3_RSVDF        = 0x003F,
    /* Type 4: Test Events */
    VCP_EVID_TEST_CMP        = 0x0040,
    VCP_EVID_T4_RSVD1        = 0x0041, /* Used to be VCP_EVID_TEST_STOP event */
    VCP_EVID_DTONE_DET       = 0x0042,
    VCP_EVID_DTONE_LOSS      = 0x0043,
    VCP_EVID_STEST_CMP       = 0x0044,
    VCP_EVID_CHKSUM          = 0x0045,
    VCP_EVID_T4_RSVD6        = 0x0046,
    VCP_EVID_T4_RSVD7        = 0x0047,
    VCP_EVID_T4_RSVD8        = 0x0048,
    VCP_EVID_T4_RSVD9        = 0x0049,
    VCP_EVID_T4_RSVDA        = 0x004A,
    VCP_EVID_T4_RSVDB        = 0x004B,
    VCP_EVID_T4_RSVDC        = 0x004C,
    VCP_EVID_T4_RSVDD        = 0x004D,
    VCP_EVID_T4_RSVDE        = 0x004E,
    VCP_EVID_T4_RSVDF        = 0x004F,
    /* Type 5: Call Processing Events */
    VCP_EVID_MTR_CMP         = 0x0050,
    VCP_EVID_MTR_ABORT       = 0x0051,
    VCP_EVID_CID_DATA        = 0x0052,
    VCP_EVID_RING_CAD        = 0x0053,
    VCP_EVID_SIGNAL_CMP      = 0x0054,
    VCP_EVID_MTR_CAD         = 0x0055,
    VCP_EVID_TONE_CAD        = 0x0056,
    VCP_EVID_T5_RSVD7        = 0x0057,
    VCP_EVID_T5_RSVD8        = 0x0058,
    VCP_EVID_T5_RSVD9        = 0x0059,
    VCP_EVID_T5_RSVDA        = 0x005A,
    VCP_EVID_T5_RSVDB        = 0x005B,
    VCP_EVID_T5_RSVDC        = 0x005C,
    VCP_EVID_T5_RSVDD        = 0x005D,
    VCP_EVID_T5_RSVDE        = 0x005E,
    VCP_EVID_T5_RSVDF        = 0x005F,
    VCP_EVID_ENUM_SIZE = FORCE_STANDARD_C_ENUM_SIZE /* Portability Req. */
} VcpEventIdType;

#define VCP_EVID_NUM_T1      (VCP_EVID_T1_RSVD5 - 0x0010)
#define VCP_EVID_NUM_T2      (VCP_EVID_T2_RSVDA - 0x0020)
#define VCP_EVID_NUM_T3      (VCP_EVID_T3_RSVDD - 0x0030)
#define VCP_EVID_NUM_T4      (VCP_EVID_T4_RSVD6 - 0x0040)
#define VCP_EVID_NUM_T5      (VCP_EVID_T5_RSVD7 - 0x0050)
#define VCP_EVID_NUM_TYPES   5

typedef enum
{
    /* VoiceEdge API line states */
    VCP_LINE_STANDBY,            /* Normal on-hook, low power standby */
    VCP_LINE_TIP_OPEN,           /* Ground start signaling state */
    VCP_LINE_ACTIVE,             /* Normal off-hook active state */
    VCP_LINE_ACTIVE_POLREV,      /* Normal active w/reverse polarity */
    VCP_LINE_OHT,                /* On hook tx */
    VCP_LINE_OHT_POLREV,         /* On hook tx w/reverse polarity */

    VCP_LINE_NOCURRENT,          /* Current denial state */
    VCP_LINE_NOCURRENT_POLREV,   /* Current denial w/reverse polarity */

    VCP_LINE_DISCONNECT,         /* Line out of service state */
    VCP_LINE_RINGING,            /* Place ringing on the line */
    VCP_LINE_RINGING_POLREV,     /* Place reverse polarity ringing on the line*/
    VCP_NUM_LINE_STATES,

    VCP_LINE_VP_MASK             = 0x80,

    /* VoicePath-2 API line states */
    VCP_VP_LINE_STANDBY          = VCP_LINE_VP_MASK | VP_LINE_STANDBY,
    VCP_VP_LINE_TIP_OPEN         = VCP_LINE_VP_MASK | VP_LINE_TIP_OPEN,
    VCP_VP_LINE_ACTIVE           = VCP_LINE_VP_MASK | VP_LINE_ACTIVE,
    VCP_VP_LINE_ACTIVE_POLREV    = VCP_LINE_VP_MASK | VP_LINE_ACTIVE_POLREV,
    VCP_VP_LINE_TALK             = VCP_LINE_VP_MASK | VP_LINE_TALK,
    VCP_VP_LINE_TALK_POLREV      = VCP_LINE_VP_MASK | VP_LINE_TALK_POLREV,
    VCP_VP_LINE_OHT              = VCP_LINE_VP_MASK | VP_LINE_OHT,
    VCP_VP_LINE_OHT_POLREV       = VCP_LINE_VP_MASK | VP_LINE_OHT_POLREV,
    VCP_VP_LINE_DISCONNECT       = VCP_LINE_VP_MASK | VP_LINE_DISCONNECT,
    VCP_VP_LINE_RINGING          = VCP_LINE_VP_MASK | VP_LINE_RINGING,
    VCP_VP_LINE_RINGING_POLREV   = VCP_LINE_VP_MASK | VP_LINE_RINGING_POLREV,
    VCP_LINESTATE_ENUM_SIZE = FORCE_STANDARD_C_ENUM_SIZE /* Portability Req. */
} VcpLineStateType;

typedef enum
{
    VP_TS_ON,       /* Turn on TX & RX PCM, no change in TS */
    VP_TS_OFF,      /* Turn off TX/RX PCM, no change in TS */
    VP_TS_SET,      /* Change Time Slot assignments */
    VP_TS_ON_SET,   /* Change Time Slot assign, and turn ON PCM */
    VP_TS_ENUM_SIZE = FORCE_STANDARD_C_ENUM_SIZE /* Portability Req. */
} VcpTimeslotCtrlType;

typedef enum
{
    VCP_INPUT_HOOK       = 0,
    VCP_INPUT_RAW_HOOK   = 1,
    VCP_INPUT_GKEY       = 2,
    VCP_INPUT_THERM_FLT  = 3,
    VCP_INPUT_CLK_FLT    = 4,
    VCP_INPUT_AC_FLT     = 5,
    VCP_INPUT_DC_FLT     = 6,
    VCP_INPUT_PBAT_FLT   = 7,
    VCP_INPUT_HBAT_FLT   = 8,
    VCP_INPUT_LBAT_FLT   = 9,
    VCP_NUM_INPUT_TYPES  = 10,
    VCP_INPUT_ENUM_SIZE = FORCE_STANDARD_C_ENUM_SIZE /* Portability Req. */
} VcpInputType;

typedef union
{
    uint16                      rampTime;
    VpOptionPulseType           pulse;
    VpOptionPulseModeType       pulseMode;
    VpOptionTimeslotType        timeslot;
    VpOptionCodecType           codec;
    VpOptionPcmHwyType          pcmHwy;
    VpOptionLoopbackType        loopback;
    VpOptionLineStateType       lineState;
    VpOptionCriticalFltType     criticalFlt;
    VpOptionZeroCrossType       zeroCross;
    VpOptionEventMaskType       eventMask;
    VpOptionRingControlType     ringControl;
    VpOptionDtmfModeType        dtmfControl;
    VpOptionDeviceIoType        deviceIo;
    VpOptionPcmTxRxCntrlType    pcmTxRxCntrl;
} VcpOptionDataType;

typedef union
{
    VpTestConcludeType        tconclude;
    VpTestSetSenseGainType    senseGain;
    VpTestOpenVType           openV;
    VpTestDiffVacType         diffVac;
    VpTestDcRLoopType         dcRLoop;
    VpTestAcRLoopType         acRLoop;
    VpTestNullIGndType        nullIGnd;
    VpTestResType             res;
    VpTestCapType             cap;
    VpTestIdcType             idc;
    VpTestIacType             iac;
    VpTestKeypadType          keypad;
    VpTestNoiseType           noise;
    VpTestSnrQDistorType      snrqd;
    VpTestArb1ToneType        arb1Tone;
    VpTestToneGenType         toneGen;
    VpTestDialToneType        dialTone;
    VpTestTransHybridLossType hybridLoss;
    VpTestRenType             ren;
} VcpTestArgType;


/* VCP specific API function prototypes */
static VpStatusType VcpMakeLineObject(VpTermType termType, uint8 channelId,
    VpLineCtxType *pLineCtx, void *pLineObj, VpDevCtxType *pDevCtx);

/* Initialization functions */
static VpStatusType VcpInitDevice(
    VpDevCtxType *pDevCtx, VpProfilePtrType pDevProfile,
    VpProfilePtrType pAcProfile, VpProfilePtrType pDcProfile,
    VpProfilePtrType pRingProfile, VpProfilePtrType pFxoAcProfile,
    VpProfilePtrType pFxoCfgProfile);

static VpStatusType VcpInitLine(
    VpLineCtxType *pLineCtx, VpProfilePtrType pAcProfile,
    VpProfilePtrType pDcProfile, VpProfilePtrType pRingProfile);

static VpStatusType VcpConfigLine(
    VpLineCtxType *pLineCtx, VpProfilePtrType pAcProfile,
    VpProfilePtrType pDcProfile, VpProfilePtrType pRingProfile);

static VpStatusType VcpSetBatteries(
    VpLineCtxType *pLineCtx,
    VpBatteryModeType battMode,
    VpBatteryValuesType *pBatt);

static VpStatusType VcpSetBFilter(
    VpLineCtxType *pLineCtx,
    VpBFilterModeType bFiltMode,
    VpProfilePtrType pAcProfile);

static VpStatusType VcpCalCodec(
    VpLineCtxType *pLineCtx, VpDeviceCalType mode);

static VpStatusType VcpCalLine(
    VpLineCtxType *pLineCtx);

static VpStatusType VcpInitRing(VpLineCtxType *pLineCtx,
    VpProfilePtrType pCadProfile, VpProfilePtrType pCidProfile);

static VpStatusType VcpInitCid(
    VpLineCtxType *pLineCtx, uint8 length, uint8p pCidData);

static VpStatusType VcpInitMeter(
    VpLineCtxType *pLineCtx, VpProfilePtrType pMeterProfile);

static VpStatusType VcpInitProfile(
    VpDevCtxType *pDevCtx, VpProfileType type,
    VpProfilePtrType pProfileIndex, VpProfilePtrType pProfile);

static VpStatusType VcpSoftReset(
    VpDevCtxType *pDevCtx);

/* Control functions */
static VpStatusType VcpSetLineState(
    VpLineCtxType *pLineCtx, VpLineStateType state);

static VpStatusType VcpSetLineTone(
    VpLineCtxType *pLineCtx, VpProfilePtrType pToneProfile,
    VpProfilePtrType pCadProfile, VpDtmfToneGenType *pDtmfControl);

static VpStatusType VcpSetRelayState(
    VpLineCtxType *pLineCtx, VpRelayControlType rState);

static VpStatusType VcpSetRelGain(
    VpLineCtxType *pLineCtx, uint16 txLevel,
    uint16 rxLevel, uint16 handle);

static VpStatusType VcpSendSignal(
    VpLineCtxType *pLineCtx,
    VpSendSignalType type,
    void *pStruct);

static VpStatusType VcpSendCid(
    VpLineCtxType *pLineCtx, uint8 length,
    VpProfilePtrType pCidProfile, uint8p pCidData);

static VpStatusType VcpContinueCid(
    VpLineCtxType *pLineCtx, uint8 length, uint8p pCidData);

static VpStatusType VcpStartMeter(
    VpLineCtxType *pLineCtx, uint16 onTime,
    uint16 offTime, uint16 numMeters);

static VpStatusType VcpSetOption(
    VpLineCtxType *pLineCtx, VpDevCtxType *pDevCtx,
    VpOptionIdType option, void *pValue);

static VpStatusType VcpDeviceIoAccess(
    VpDevCtxType *pDevCtx,
    VpDeviceIoAccessDataType *pDeviceIoData);

static VpStatusType VcpLowLevelCmd(
    VpLineCtxType *pLineCtx, uint8p cmd, uint8 len, uint16 handle);

/* Status and query functions */
static bool VcpGetEvent(
    VpDevCtxType *pDevCtx,  VpEventType *pEvent);

static VpStatusType VcpGetLineStatus(
    VpLineCtxType *pLineCtx, VpInputType input, bool *pStatus);

static VpStatusType VcpGetDeviceStatus(
    VpDevCtxType *pDevCtx, VpInputType input, uint32 *pDeviceStatus);

static VpStatusType VcpGetLoopCond(
    VpLineCtxType *pLineCtx, uint16 handle);

static VpStatusType VcpGetOption(
    VpLineCtxType *pLineCtx, VpDevCtxType *pDevCtx,
    VpOptionIdType option, uint16 handle);

static VpStatusType VcpGetLineState(
    VpLineCtxType *pLineCtx, VpLineStateType *pCurrentState);

static VpStatusType VcpFlushEvents(
    VpDevCtxType *pDevCtx);

static VpStatusType VcpGetResults(
    VpEventType *pEvent, void *pResults);

static VpStatusType VcpClearResults(
    VpDevCtxType *pDevCtx);

/* Testing functions */
static VpStatusType VcpTestLine(
    VpLineCtxType *pLineCtx, VpTestIdType test,
    const void *pArgsUntyped, uint16 handle);

static VpStatusType VcpCodeCheckSum(
    VpDevCtxType *pDevCtx, uint16 handle);

static VpStatusType VcpSelfTest(
    VpLineCtxType *pLineCtx);

static VpStatusType VcpFillTestBuf(
    VpLineCtxType *pLineCtx, uint16 length,
    VpVectorPtrType pData);

/* VCP API specific helper functions */
static void InitDeviceObjMembers(
    VpVcpDeviceObjectType *pDevObj);

static VpStatusType InitAndConfigLine(
    VpLineCtxType *pLineCtx, VpProfilePtrType pAcProfile,
    VpProfilePtrType pDcProfile, VpProfilePtrType pRingProfile,
    VcpMbFuncIdType mailBoxCmd);

static VpStatusType GetCheckSumMessage(
    VpDeviceIdType deviceId, VpChkSumType *pResults);

static VpStatusType GetMpiRxMessage(
    VpDeviceIdType deviceId, uint8p pResults);

static VpStatusType GetRdLoopCondMessage(
    VpDevCtxType *pDevCtx, VpLoopCondResultsType *pResults);

static VpStatusType GetGainCmpMessage(
    VpDeviceIdType deviceId, VpRelGainResultsType *pResults);

static VpStatusType GetRdOptionMessage(
    VpDevCtxType *pDevCtx, VcpOptionDataType *pResults);

static VpStatusType GetIoAccessResponse(
    VpDeviceIdType deviceId, VpDeviceIoAccessDataType *pResults);

static VpStatusType GetTestCmpMessage(
    VpDeviceIdType deviceId, VpTestResultType *pResults);

static VcpEventIdType ConvertApiEventId2VcpEventId(
    VpEventType *pEvent);

static bool ConvertVcpEventId2ApiEventId(
    uint16 eventWord, VpEventCategoryType *eventCat, uint16 *eventId);

/* VCP Access Helper functions Prototypes */
static VpStatusType WriteMailBoxCmd(
    VpDeviceIdType deviceId, uint8 length, uint16 *pCmdBuf);

static VpStatusType MailboxAcquire(
    VpDeviceIdType deviceId);

static void MailboxRelease(
    VpDeviceIdType deviceId);

static void MailboxAddCommand(
    VpDeviceIdType deviceId, VcpMbFuncIdType funcId, uint16 handle);

static uint8 MailboxAddArg(
    VpDeviceIdType deviceId, uint16 arg, uint8 offset);

static uint8 MailboxAddProfile(
    VpDeviceIdType deviceId, VpProfileType type,
    VpProfilePtrType pProfile, uint8 offset);

static uint8 WriteProfile(
    VpDeviceIdType deviceId, uint8 page, uint8 offset,
    uint8 length, VpProfilePtrType pProfile, bool compact);

static bool FftSizeIsValid(
    uint16 fftSize);

static uint8 GetProfileTableMaxCount(
    VpProfileType profileType);

static VpStatusType SetVcpDefaultOptions(
    VpDevCtxType *pDevCtx, VpLineCtxType *pLineCtx);

static uint8 GetTotalDtmfResources(uint32 linewiseStatus,
    uint8 remainingResources);

static bool UpdateProfileTableStatus(VpVcpDeviceObjectType *pDevObj,
                VpProfileType type, int profileIndex,VpProfilePtrType pProfile);

static bool IsItAValidProfilePtr(VpVcpDeviceObjectType *pDevObj,
                VpProfileType type, VpProfilePtrType pProfile);

static bool IsThereAValidProfileTableEntry(VpVcpDeviceObjectType *pDevObj,
                VpProfileType type, VpProfilePtrType pProfileIndex);

static uint16 * GetProfileTableStatusPtr(VpVcpDeviceObjectType *pDevObj,
                VpProfileType type);

/******************************************************************************
 *                   VCP Specific System Configuration Functions              *
 ******************************************************************************/
/**
 * VpMakeVcpDeviceObject()
 * This function initializes the device context and device object to handle
 * VCP functionality.
 *
 * Preconditions:
 * The passed device object and device context pointers must be valid. The
 * device object pointer type should be for VCP device object type. Passed
 * context/objects are assumed to be un-initialized.
 *
 * Postconditions:
 * Initializes device context and device object based on input parameters and to
 * be able to handle VCP functionality.
 */
VpStatusType
VpMakeVcpDeviceObject(
    VpDevCtxType *pDevCtx,          /**< Device Context to be initialized by
                                     * other input */
    VpVcpDeviceObjectType *pDevObj) /**< Device Object to be pointed to by
                                     * device context*/
{

    /* First reset all members of the device object to their default values */
    InitDeviceObjMembers(pDevObj);

    if (pDevCtx != VP_NULL) {
        return VpMakeVcpDeviceCtx(pDevCtx, pDevObj);
    } else {
        return VP_STATUS_SUCCESS;
    }
} /* VpMakeVcpDeviceObject() */

/**
 * VpMakeVcpDeviceCtx()
 * This function initializes the device context to handle VCP functionality.
 *
 * Preconditions:
 * This function should be called after initializing the device object. This
 * function can be called more than once without modifying the contents of the
 * device object.
 *
 * Postconditions:
 * Initializes device context to be able to handle VCP functionality.
 */
VpStatusType
VpMakeVcpDeviceCtx(
    VpDevCtxType *pDevCtx,          /**< Device Context to be initialized by
                                     * other input */
    VpVcpDeviceObjectType *pDevObj) /**< Device Object to be pointed to by
                                     * device context*/
{
    if((pDevCtx == VP_NULL) || (pDevObj == VP_NULL)) {
        return VP_STATUS_INVALID_ARG;
    }
    /* Initialize the members of device context */
    pDevCtx->pDevObj = pDevObj;
    pDevCtx->deviceType = VP_DEV_VCP_SERIES;

    /* Initialize the function pointers (for those functions that VCP
     * supports) */
    pDevCtx->funPtrsToApiFuncs.MakeLineObject = VcpMakeLineObject;

    pDevCtx->funPtrsToApiFuncs.BootLoad = VpVcpVppBootLoad;
    pDevCtx->funPtrsToApiFuncs.InitDevice = VcpInitDevice;
    pDevCtx->funPtrsToApiFuncs.InitLine = VcpInitLine;
    pDevCtx->funPtrsToApiFuncs.ConfigLine = VcpConfigLine;
    pDevCtx->funPtrsToApiFuncs.InitRing = VcpInitRing;
    pDevCtx->funPtrsToApiFuncs.InitCid = VcpInitCid;
    pDevCtx->funPtrsToApiFuncs.InitMeter = VcpInitMeter;
    pDevCtx->funPtrsToApiFuncs.InitProfile = VcpInitProfile;
    pDevCtx->funPtrsToApiFuncs.SoftReset = VcpSoftReset;
    pDevCtx->funPtrsToApiFuncs.SetBFilter = VcpSetBFilter;
    pDevCtx->funPtrsToApiFuncs.SetBatteries = VcpSetBatteries;

    pDevCtx->funPtrsToApiFuncs.SetLineState = VcpSetLineState;
    pDevCtx->funPtrsToApiFuncs.SetLineTone = VcpSetLineTone;
    pDevCtx->funPtrsToApiFuncs.SetRelGain = VcpSetRelGain;
    pDevCtx->funPtrsToApiFuncs.SendCid = VcpSendCid;
    pDevCtx->funPtrsToApiFuncs.ContinueCid = VcpContinueCid;
    pDevCtx->funPtrsToApiFuncs.StartMeter = VcpStartMeter;
    pDevCtx->funPtrsToApiFuncs.SetOption = VcpSetOption;
    pDevCtx->funPtrsToApiFuncs.DeviceIoAccess = VcpDeviceIoAccess;
    pDevCtx->funPtrsToApiFuncs.GetLineState = VcpGetLineState;
    pDevCtx->funPtrsToApiFuncs.LowLevelCmd = VcpLowLevelCmd;

    pDevCtx->funPtrsToApiFuncs.GetEvent = VcpGetEvent;
    pDevCtx->funPtrsToApiFuncs.GetLineStatus = VcpGetLineStatus;
    pDevCtx->funPtrsToApiFuncs.GetDeviceStatus = VcpGetDeviceStatus;
    pDevCtx->funPtrsToApiFuncs.GetLoopCond = VcpGetLoopCond;
    pDevCtx->funPtrsToApiFuncs.GetOption = VcpGetOption;
    pDevCtx->funPtrsToApiFuncs.FlushEvents = VcpFlushEvents;
    pDevCtx->funPtrsToApiFuncs.GetResults = VcpGetResults;
    pDevCtx->funPtrsToApiFuncs.ClearResults = VcpClearResults;

    pDevCtx->funPtrsToApiFuncs.CodeCheckSum = VcpCodeCheckSum;
    pDevCtx->funPtrsToApiFuncs.SelfTest = VcpSelfTest;

    /* Initialize Product specific function support */
    if(pDevObj->productType == VP_VCP_PRODUCT_VE_880) {
        pDevCtx->funPtrsToApiFuncs.CalLine = VP_NULL;
        pDevCtx->funPtrsToApiFuncs.CalCodec = VP_NULL;
        pDevCtx->funPtrsToApiFuncs.SetRelayState = VP_NULL;
        pDevCtx->funPtrsToApiFuncs.SendSignal = VcpSendSignal;
        pDevCtx->funPtrsToApiFuncs.TestLine = VP_NULL;
        pDevCtx->funPtrsToApiFuncs.FillTestBuf = VP_NULL;

    } else {

        /* Must be VP_VCP_PRODUCT_VE_790 */
        pDevCtx->funPtrsToApiFuncs.CalLine = VcpCalLine;
        pDevCtx->funPtrsToApiFuncs.CalCodec = VcpCalCodec;
        pDevCtx->funPtrsToApiFuncs.SetRelayState = VcpSetRelayState;
        pDevCtx->funPtrsToApiFuncs.SendSignal = VP_NULL;
        pDevCtx->funPtrsToApiFuncs.TestLine = VcpTestLine;
        pDevCtx->funPtrsToApiFuncs.FillTestBuf = VcpFillTestBuf;
    }

    return VP_STATUS_SUCCESS;
} /* VpMakeVcpDeviceCtx() */

/*
 * VpMakeVcpLineObject()
 * This function initializes a line context using the information that is
 * passed. This function is like a C++ constructor. It initializes the passed
 * line context and line object based on the paramters provided. The passed line
 * object type should match with the type of device object type. It also
 * establishes a link between the device context and the line context. This
 * function SHOULD NEVER be called outside the VP-API. This is an VP-API
 * internal function.
 *
 * Preconditions:
 * This function assumes device context has already been created and
 * initialized. This function should only be called after downloading the boot
 * image to the device when applicable.
 *
 * Postconditions:
 * This function initializes the line context/line object. Line related VP-API
 * functions can be called after calling this function.
 */
static VpStatusType
VcpMakeLineObject(
    VpTermType termType,
    uint8 channelId,
    VpLineCtxType *pLineCtx,
    void *pVoidLineObj,
    VpDevCtxType *pDevCtx)
{
    VpVcpLineObjectType *pLineObj = pVoidLineObj;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 dvpTerm;
    VpStatusType retVal = VP_STATUS_SUCCESS;

    /* Basic error checking */
    if (channelId >= pDevObj->maxChannels) {
        return VP_STATUS_INVALID_ARG;
    }

    /* Convert API Line termination type to VCP specific termination type */
    switch (termType) {
        case VP_TERM_FXS_GENERIC:{
            dvpTerm = VCP_TERM_GENERIC_FXS;
            break;
        }
        case VP_TERM_FXS_75181: {
            dvpTerm = VCP_FXS_79232_75181;
            break;
        }
        case VP_TERM_FXS_TO_TL:
        {
            dvpTerm = VCP_FXS_79252_TO_TL;
            break;
        }
        case VP_TERM_FXS_75282: {
            dvpTerm = VCP_FXS_79232_75282;
            break;
        }
        case VP_TERM_FXS_TITO_TL_R: {
            dvpTerm = (uint16)VCP_FXS_792x2_TITO_TL_R;
            break;
        }
        case VP_TERM_FXS_CO_TL: {
            dvpTerm = VCP_FXS_792x2_CO_TL;
            break;
        }
        case VP_TERM_FXS_RR: {
            dvpTerm = VCP_FXS_79232_RR;
            break;
        }
        default: {
            return  VP_STATUS_ERR_VTD_CODE;
        }
    }

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    pLineObj->termType = termType;

    pLineCtx->pLineObj = pLineObj;
    pLineCtx->pDevCtx = pDevCtx;

    /* Establish the link between device context to line context */
    pDevCtx->pLineCtx[channelId] = pLineCtx;

    /* Initialize line object */
    pLineObj->channelId = channelId;
    pLineObj->termType = termType;
    pLineObj->calCompleteEventMask = FALSE;
    /* Note that we are not initializing the 'lineId' member because we do not
     * know what it should be initialized (that information is not provided
     * through the VpMakeLineObj() function) and the basic type of this member
     * is defined by the customer and hence this variable could not have one
     * default value. */

    /* Write the Line termination type to the VCP */
    if(!VpHbiWr(deviceId, VP_REG_LINE_TERM_TYPE(channelId), &dvpTerm)){
        retVal = VP_STATUS_ERR_HBI;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retVal;
}  /* VpMakeVcpLineObject() */

/******************************************************************************
 *                   VCP Specific Initialization Functions                    *
 ******************************************************************************/
/*
 * VcpInitDevice()
 * This function calibrates the device and initializes all lines on the device
 * with the AC, DC, and Ringing parameters passed. See VP-API reference guide
 * for more information.
 *
 * Preconditions:
 * This function should be called only after creating and initializing the
 * device context.
 *
 * Postconditions:
 * Device is calibrated and all lines associated with this device are
 * initialized with the AC, DC, and Ringing Paramaters passed (DC and Ringing
 * apply to FXS type lines only).
 */
static VpStatusType
VcpInitDevice(
    VpDevCtxType *pDevCtx,          /**< Pointer to device context for the
                                     * device that will be initialized */
    VpProfilePtrType pDevProfile,   /**< Pointer to Device Profile */
    VpProfilePtrType pAcProfile,    /**< Pointer to AC Profile that is applied
                                     * to all FXS lines on this device */
    VpProfilePtrType pDcProfile,    /**< Pointer to DC Profile that is applied
                                     * to all FXS lines on this device */
    VpProfilePtrType pRingProfile,  /**< Pointer to Ringing Profile that is
                                     * applied to all FXS lines on this device*/
    VpProfilePtrType pFxoAcProfile, /**< Pointer to AC Profile that is applied
                                     * to all FXO lines on this device */
    VpProfilePtrType pFxoCfgProfile)/**< Pointer to Config Profile that is
                                     * applied to all FXO lines on this device*/
{
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint8 argIndex = VP_MB_ARGLIST_INDEX;
    VpStatusType retval;

    /* Basic error checking */
    if(pDevProfile == VP_PTABLE_NULL) {
        /* Nothing can be done without the device profile */
        return VP_STATUS_INVALID_ARG;
    }

    /* Profile pointer error checking */
    if((IsItAValidProfilePtr(pDevObj, VP_PROFILE_DEVICE, pDevProfile)
            == FALSE) ||
       (IsItAValidProfilePtr(pDevObj, VP_PROFILE_AC, pAcProfile) == FALSE) ||
       (IsItAValidProfilePtr(pDevObj, VP_PROFILE_DC, pDcProfile) == FALSE) ||
       (IsItAValidProfilePtr(pDevObj, VP_PROFILE_RING, pRingProfile) == FALSE)){

        return VP_STATUS_ERR_PROFILE;
    }

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
        retval = VP_STATUS_MAILBOX_BUSY;

    } else {
        retval = VP_STATUS_SUCCESS;

        /* First mark that the device is not initialized */
        pDevObj->deviceInitialized = FALSE;

        /* Add the InitSystem command to the mailbox */
        MailboxAddCommand(deviceId, VCP_MBFUNC_INITDEVICE, VCP_VP_LINE_DISCONNECT);

        /* Add the profile arguments to the mailbox */
        argIndex = MailboxAddProfile(deviceId, VP_PROFILE_DEVICE, pDevProfile,
                                        argIndex);
        if (argIndex == 0) retval = VP_STATUS_INVALID_ARG;

        argIndex = MailboxAddProfile(deviceId, VP_PROFILE_AC, pAcProfile,
                                        argIndex);
        if (argIndex == 0) retval = VP_STATUS_INVALID_ARG;

        argIndex = MailboxAddProfile(deviceId, VP_PROFILE_DC, pDcProfile,
                                        argIndex);
        if (argIndex == 0) retval = VP_STATUS_INVALID_ARG;

        argIndex = MailboxAddProfile(deviceId, VP_PROFILE_RING, pRingProfile,
                                        argIndex);
        if (argIndex == 0) retval = VP_STATUS_INVALID_ARG;

        /* VCP Presently supports only FXS, so,  ignore FXO profiles */

        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);
    }
    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpInitDevice() */

/**
 * VcpInitLine()
 * Initializes a line of the  VCP by loading the given profiles, places the line
 * to known state, loads default options and generates an event.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpInitLine(
    VpLineCtxType *pLineCtx,
    VpProfilePtrType pAcProfile,        /* AC Profile Selector     */
    VpProfilePtrType pDcProfile,        /* DC Profile Selector     */
    VpProfilePtrType pRingProfile       /* Ring Profile Selector   */
    )
{
    return InitAndConfigLine(pLineCtx, pAcProfile, pDcProfile,
                                pRingProfile, VCP_MBFUNC_RESETLINE);
} /* VcpInitLine() */

/**
 * VcpConfigLine()
 * Configures a line of the  VCP by loading the given profiles.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpConfigLine(
    VpLineCtxType *pLineCtx,
    VpProfilePtrType pAcProfile,        /* AC Profile Selector     */
    VpProfilePtrType pDcProfile,        /* DC Profile Selector     */
    VpProfilePtrType pRingProfile       /* Ring Profile Selector   */
    )
{
    return InitAndConfigLine(pLineCtx, pAcProfile, pDcProfile,
                                pRingProfile, VCP_MBFUNC_INITLINE);
} /* VcpConfigLine() */

/**
 * VcpSetBatteries()
 *  This function sets the battery values in the device, which for some devices
 * may result in improved feed performance.
 *
 * Preconditions:
 *  The device must be created and initialized before calling this function.
 *
 * Postconditions:
 *  The device is programmed to use the battery values provided, or programmed
 * to use measured battery voltages.
 */
static VpStatusType
VcpSetBatteries(
    VpLineCtxType *pLineCtx,
    VpBatteryModeType battMode,
    VpBatteryValuesType *pBatt)
{
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;

#define VCP_BATT_EN_LEN     (11)
    uint8 batteryEnable[VCP_BATT_EN_LEN] = {
        VCP_VP790_DCR2_WRT, VCP_VP790_DRC2_BATOVER,
        VCP_VP790_LOW_BATT_WRT, 0x00, 0x00,
        VCP_VP790_HIGH_BATT_WRT, 0x00, 0x00,
        VCP_VP790_POS_BATT_WRT, 0x00, 0x00
    };

#define VCP_BATT_DIS_LEN     (2)
    uint8 batteryDisable[VCP_BATT_DIS_LEN] = {
        VCP_VP790_DCR2_WRT, 0x00
    };

    if(pDevObj->productType == VP_VCP_PRODUCT_VE_880) {
        return VP_STATUS_FUNC_NOT_SUPPORTED;
    }

    if (battMode == VP_BATT_MODE_EN) {
        batteryEnable[3] = (uint8)((pBatt->batt1 >> 8) & 0xFF);
        batteryEnable[4] = (uint8)(pBatt->batt1 & 0xFF);

        batteryEnable[6] = (uint8)((pBatt->batt2 >> 8) & 0xFF);
        batteryEnable[7] = (uint8)(pBatt->batt2 & 0xFF);

        batteryEnable[9] = (uint8)((pBatt->batt3 >> 8) & 0xFF);
        batteryEnable[10] = (uint8)(pBatt->batt3 & 0xFF);

        return VcpLowLevelCmd(pLineCtx, batteryEnable, VCP_BATT_EN_LEN, 0);
    } else {
        return VcpLowLevelCmd(pLineCtx, batteryDisable, VCP_BATT_DIS_LEN, 0);
    }
}

/**
 * VcpSetBFilter()
 *  This function enables the B-Filter and programs it with the B-Filter values
 * provided in the ac profile, or disables the B-Filter.
 *
 * Preconditions:
 *  The line must be created and initialized before calling this function.
 *
 * Postconditions:
 *  The B-Filter is programmed to either disabled values or to the values passed
 * in the AC-Profile.
 */
static VpStatusType
VcpSetBFilter(
    VpLineCtxType *pLineCtx,
    VpBFilterModeType bFiltMode,
    VpProfilePtrType pAcProfile)
{
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;

    if (bFiltMode == VP_BFILT_EN) {
        /* use the incoming ac profile pointed to by pAcProfile */
        return InitAndConfigLine(pLineCtx, pAcProfile, VP_NULL, \
                VP_NULL, VCP_MBFUNC_INITLINE);
    } else if (pDevObj->productType == VP_VCP_PRODUCT_VE_880) {
        /*
         * create an acProfile on the fly that only disables the bfilter
         * coeffs for an 880 device
         */
         VpProfileDataType acBfiltDisable[] = {
              /* AC Profile with only BFilters disabled*/
             0x00, 0x00, 0x00, 0x14, 0x00, 0x12, 
             0x86, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
             0x00, 0x00, 0x00, 0x00,
             0x96, 0x00, 0x00
         };
         return InitAndConfigLine(pLineCtx, acBfiltDisable, VP_NULL, \
                VP_NULL, VCP_MBFUNC_INITLINE);
    } else {
        /*
         * create an acProfile on the fly that only disables the bfilter
         * coeffs for an 790 device
         */
         VpProfileDataType acBfiltDisable[] = {
              /* AC Profile with only BFilters disabled*/
             0x00, 0x00, 0x00, 0x18, 0x00, 0x16, 
             0x86, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
             0x00, 0x00, 0x00, 0x00,
             0x96, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
         };
         return InitAndConfigLine(pLineCtx, acBfiltDisable, VP_NULL, \
                VP_NULL, VCP_MBFUNC_INITLINE);
    }
}

/**
 * VcpCalCodec()
 * Starts a SLAC (CODEC) calibration for one of the SLAC devices controlled by
 * the VCP. The SLAC that needs to be calibrated is determined by identifying
 * the SLAC with which line context is associated with.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpCalCodec(
    VpLineCtxType *pLineCtx,   /* Selects the target line number */
    VpDeviceCalType mode       /* Specifies the method for calibrating a SLAC */
    )
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint8 argIndex = VP_MB_ARGLIST_INDEX;
    VpStatusType retval;
    uint8 line = pLineObj->channelId;

    /* Basic error checking */
    if (mode >= VP_NUM_DEV_CAL_TYPES) {
        return VP_STATUS_INVALID_ARG;
    }
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
        retval = VP_STATUS_MAILBOX_BUSY;

    } else {
        /* Add the Calibrate command to the mailbox */
        MailboxAddCommand(deviceId, VCP_MBFUNC_CALSLAC, 0x0000);

        /* Add the line number and halt mode arguments to the mailbox */
        argIndex = MailboxAddArg(deviceId, (uint16)line, argIndex);
        argIndex = MailboxAddArg(deviceId, (uint16)mode, argIndex);

        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);

        retval = VP_STATUS_SUCCESS;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpCalCodec() */

/**
 * VcpCalLine()
 *  Starts a SLIC calibration for one of the SLIC devices controlled by the
 *   VCP.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpCalLine(
    VpLineCtxType *pLineCtx
    )
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint8 argIndex = VP_MB_ARGLIST_INDEX;
    VpStatusType retval;
    uint8 line = pLineObj->channelId;

    /* Basic error checking */
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
        retval = VP_STATUS_MAILBOX_BUSY;

    } else {
        /* Add the Calibrate command to the mailbox */
        MailboxAddCommand(deviceId, VCP_MBFUNC_CALSLIC, 0x0000);

        /* Add the line number and halt mode arguments to the mailbox */
        argIndex = MailboxAddArg(deviceId, (uint16)line, argIndex);

        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);

        retval = VP_STATUS_SUCCESS;
    }
    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpCalLine() */

/**
 * VcpInitRing()
 *  Initializes the VCP Ring State for the ring cadence and
 *  caller ID behavior.
 *
 * Refer to the VP API User's Guide for details about this function.
 */
static VpStatusType
VcpInitRing(
    VpLineCtxType *pLineCtx,            /* Pointer to line context  */
    VpProfilePtrType pCadProfile,       /* Cadence profile selector */
    VpProfilePtrType pCidProfile        /* Caller ID profile selector */
    )
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint8 argIndex = VP_MB_ARGLIST_INDEX;
    VpStatusType retval;
    uint8 line = pLineObj->channelId;

    /* Basic error checking */
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Profile pointer error checking */
    if((IsItAValidProfilePtr(pDevObj, VP_PROFILE_RINGCAD, pCadProfile)
            == FALSE) ||
       (IsItAValidProfilePtr(pDevObj, VP_PROFILE_CID, pCidProfile) == FALSE)){

        return VP_STATUS_ERR_PROFILE;
    }

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
        retval = VP_STATUS_MAILBOX_BUSY;

    } else {
        retval = VP_STATUS_SUCCESS;

        /* Add the InitRing command to the mailbox */
        MailboxAddCommand(deviceId, VCP_MBFUNC_INITRING, 0x0000);

        /* Add the line argument to the mailbox */
        argIndex = MailboxAddArg(deviceId, (uint16)line,   argIndex);

        /* Add the profile arguments to the mailbox */
        argIndex = MailboxAddProfile(deviceId, VP_PROFILE_RINGCAD, pCadProfile,
                                        argIndex);
        if (argIndex == 0) retval = VP_STATUS_INVALID_ARG;

        argIndex = MailboxAddProfile(deviceId, VP_PROFILE_CID, pCidProfile,
                                        argIndex);
        if (argIndex == 0) retval = VP_STATUS_INVALID_ARG;

        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpInitRing() */


/**
 * VcpInitCid()
 *  Initializes the caller ID buffer for the next  VCP Ring State.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpInitCid(
    VpLineCtxType *pLineCtx,
    uint8 length,
    uint8p pCidData
    )
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint8 line = pLineObj->channelId;

    /* Maximum length of CID command buffer */
    #define MAX_INIT_CID_CMD_LEN (VCP_MAX_CID_DATA_LENGTH + 4)
    uint16 cidCmdBuf[MAX_INIT_CID_CMD_LEN];
    uint8 cmdLen = 0;
    int i;

    /* Basic error checking */
    if((length > VCP_MAX_CID_DATA_LENGTH) || (length == 0)){
        return VP_STATUS_INVALID_ARG;
    }
    if(pCidData == VP_NULL) {
        return VP_STATUS_INVALID_ARG;
    }
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Fill the command */
    cidCmdBuf[cmdLen++] = VCP_MBFUNC_INIT_CID;
    cidCmdBuf[cmdLen++] = 0; /* This field is reserved */
    cidCmdBuf[cmdLen++] = line;
    cidCmdBuf[cmdLen++] = length;
    for(i = 0; i<length; i++){
        cidCmdBuf[cmdLen++] = pCidData[i];
    }

    return WriteMailBoxCmd(deviceId, cmdLen, cidCmdBuf);
} /* VcpInitCid() */

/**
 * VcpInitMeter()
 *  Initializes 'set and forget' parameters for metering.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpInitMeter(
    VpLineCtxType *pLineCtx,
    VpProfilePtrType pMeterProfile
    )
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint8 argIndex = VP_MB_ARGLIST_INDEX;
    VpStatusType retval;
    uint8 line = pLineObj->channelId;

    /* Basic error checking */
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Profile pointer error checking */
    if(IsItAValidProfilePtr(pDevObj, VP_PROFILE_METER, pMeterProfile) == FALSE){
        return VP_STATUS_ERR_PROFILE;
    }

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
        retval = VP_STATUS_MAILBOX_BUSY;

    } else {
        retval = VP_STATUS_SUCCESS;

        /* Add the InitRing command to the mailbox */
        MailboxAddCommand(deviceId, VCP_MBFUNC_INITMETER, 0x0000);

        /* Add the line argument to the mailbox */
        argIndex = MailboxAddArg(deviceId, (uint16)line,   argIndex);

        /* Add the profile arguments to the mailbox */
        argIndex = MailboxAddProfile(deviceId, VP_PROFILE_METER, pMeterProfile,
                                        argIndex);
        if (argIndex == 0) retval = VP_STATUS_INVALID_ARG;

        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpInitMeter() */


/**
 * VcpInitProfile()
 *  Initializes the indexed VCP profile slot with a profile.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpInitProfile(
    VpDevCtxType *pDevCtx,              /* Device context to be used */
    VpProfileType type,                 /* Type of profile to load */
    VpProfilePtrType pProfileIndex,     /* Profile index selector */
    VpProfilePtrType pProfile           /* Pointer to the profile data */
    )
{
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint8 page;
    uint8 offset;
    uint16 lenBytes;
    bool compact;
    uint8 selector;
    int index;
    VpStatusType status = VP_STATUS_SUCCESS;

    /* Basic error checking */
    if(pProfileIndex == VP_PTABLE_NULL){
        return VP_STATUS_INVALID_ARG;
    }
    if (type >= VP_PROFILE_FXO_CONFIG) {    /* Profiles beyond this are not
                                             * supported in VCP */
        return VP_STATUS_INVALID_ARG;
    }
    /* Get the index user wants to store the profile */
    index = VpGetProfileIndex(pProfileIndex);
    if(index < 0) {
        /* Error in profile index */
        return VP_STATUS_INVALID_ARG;
    }
    if((index + 1) > GetProfileTableMaxCount(type)) { /* index is 0 based */
        return VP_STATUS_INVALID_ARG;
    }
    /* Update profile table entry status; Also detect some more errors */
    if(UpdateProfileTableStatus(pDevObj, type, index, pProfile) == FALSE) {
        return VP_STATUS_ERR_PROFILE;
    }

    if(pProfile == VP_PTABLE_NULL) {
        /* We just erased a profile table slot and nothing else to do */
        return VP_STATUS_SUCCESS;
    }

    selector = index;

	/* DevNotes:  All code for handling the (compact = FALSE) case can be
	 * removed if and when we converge the 790 HBI memory layout to match
	 * the 880 layout. */

    lenBytes = pProfile[VP_PROFILE_LENGTH] + 4;
    compact = TRUE;
    switch (type) {
        case VP_PROFILE_DEVICE:
            page = VP_REG_DEV_PROF_PAGE(selector);
            offset = VP_REG_DEV_PROF_OFFSET(selector);
            break;
        case VP_PROFILE_AC:
            page = VP_REG_AC_PROF_PAGE(selector);
            offset = VP_REG_AC_PROF_OFFSET(selector);
            break;
        case VP_PROFILE_DC:
            page = VP_REG_DC_PROF_PAGE(selector);
            offset = VP_REG_DC_PROF_OFFSET(selector);
            break;
        case VP_PROFILE_RING:
            page = VP_REG_RING_PROF_PAGE(selector);
            offset = VP_REG_RING_PROF_OFFSET(selector);
            break;
        case VP_PROFILE_TONE:
            page = VP_REG_TONE_PROF_PAGE(selector);
            offset = VP_REG_TONE_PROF_OFFSET(selector);
            break;
        case VP_PROFILE_METER:
            page = VP_REG_MTR_PROF_PAGE(selector);
            offset = VP_REG_MTR_PROF_OFFSET(selector);
            break;
        case VP_PROFILE_TONECAD:
            selector = selector + VCP_RING_CAD_PROF_COUNT; /* VCP provides only
             * one type of cadence table; API simulates two types of
             * cadence tables. */
             /* Intentional code fall through */
        case VP_PROFILE_RINGCAD:
            page = VP_REG_CAD_PROF_PAGE(selector);
            offset = VP_REG_CAD_PROF_OFFSET(selector);
            break;
        case VP_PROFILE_CID:
            page = VP_REG_CID_PROF_PAGE(selector);
            offset = VP_REG_CID_PROF_OFFSET(selector);
            break;
        default:
            /* 'type' is bounds-checked above, So this is not expected
            * to happen; This code is present to get rid of the warning */
            status = VP_STATUS_INVALID_ARG;
    }

	VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

	/* Write profile data to HBI. */
    if(status == VP_STATUS_SUCCESS) {
        if(WriteProfile(deviceId, page, offset, (uint8)lenBytes, pProfile,
                        compact) == 0) {
            status = VP_STATUS_ERR_HBI;
        }
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return status;
} /* VcpInitProfile() */

/**
 * VcpSoftReset()
 *  Resets the VCP without requiring a boot load.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpSoftReset(
    VpDevCtxType *pDevCtx
    )
{
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint16 retval;

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* DevNotes: The wrapper function for this function must make sure nothing
     * else is going on when this function was called. */
    /* First reset all members of the device object to their default values */
    InitDeviceObjMembers(pDevObj);

    if (
        /* Call VpHbiVcpReset() to halt the VCP and rewrite the HBI pin
         * configuration. */
        (VpHbiVcpVppReset(deviceId) == FALSE) ||

        /* Set Page 255 base register to point to CM 0. */
        (VpHbiVcpVppSetBase255(deviceId, 0x1C000) == FALSE) ||

        /* Write CM 0.  This causes the VCP to start execution.
         * Note, this assumes that the first instruction in VCP code
         * memory is NOP.  There's no way for the host to read the
         * VCP's code memory. */
        (VpHbiWr(deviceId, 255, 0, 1, 0) == FALSE)
    ) {
        retval = VP_STATUS_ERR_HBI;
    } else {
        retval = VP_STATUS_SUCCESS;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpSoftReset() */


/******************************************************************************
 *                   VCP Specific Control Functions                           *
 ******************************************************************************/
/**
 * VcpSetLineState()
 *  Sets a line controlled by the VCP into the given state.
 *
 * Refer to the VP API User's Guide for details about this function.
 */
static VpStatusType
VcpSetLineState(
    VpLineCtxType *pLineCtx,
    VpLineStateType state
    )
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint8 argIndex = VP_MB_ARGLIST_INDEX;
    VpStatusType retval;
    uint8 line = pLineObj->channelId;

    /* Basic error checking */
    if (state >= VP_LINE_FXO_OHT) { /* Presently VCP handles only
                                     * FXS termination */
        return VP_STATUS_INVALID_ARG;
    }
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    state = state | VCP_LINE_VP_MASK;

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
        retval = VP_STATUS_MAILBOX_BUSY;

    } else {
        /* Add the Line State command to the mailbox */
        MailboxAddCommand(deviceId, VCP_MBFUNC_LINESTATE, 0x0000);

        /* Add the line number and state arguments to the mailbox */
        argIndex = MailboxAddArg(deviceId, (uint16)line,  argIndex);
        argIndex = MailboxAddArg(deviceId, (uint16)state, argIndex);

        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);

        retval = VP_STATUS_SUCCESS;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpSetLineState() */


/**
 * VcpSetLineTone()
 *  Places a cadenced tone on the indicated line.
 *
 * Refer to the VP-API User's Guide for details about this function.
 */
static VpStatusType
VcpSetLineTone(
    VpLineCtxType *pLineCtx,
    VpProfilePtrType pToneProfile,
    VpProfilePtrType pCadProfile,
    VpDtmfToneGenType *pDtmfControl
    )
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint8 argIndex = VP_MB_ARGLIST_INDEX;
    VpStatusType retval;
    uint8 line = pLineObj->channelId;

    /* Basic error checking */
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Profile pointer error checking */
    if((IsItAValidProfilePtr(pDevObj, VP_PROFILE_TONE, pToneProfile)
            == FALSE) ||
       (IsItAValidProfilePtr(pDevObj, VP_PROFILE_TONECAD, pCadProfile)
            == FALSE)){

        return VP_STATUS_ERR_PROFILE;
    }

    if((pDevObj->productType == VP_VCP_PRODUCT_VE_790) &&
       (pDtmfControl != VP_NULL)) {
        /* VCP-790 does not support DTMF tone generation yet. */
        return VP_STATUS_INVALID_ARG;
    }

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
        retval = VP_STATUS_MAILBOX_BUSY;

    } else {
        retval = VP_STATUS_SUCCESS;

        /* Add the Line Tone command to the mailbox */
        MailboxAddCommand(deviceId, VCP_MBFUNC_LINETONE, 0x0000);

        /* Add the line number argument to the mailbox */
        argIndex = MailboxAddArg(deviceId, (uint16)line, argIndex);

        /* Add the profile arguments to the mailbox */
        argIndex = MailboxAddProfile(deviceId, VP_PROFILE_TONE, pToneProfile,
                                        argIndex);
        if (argIndex == 0) retval = VP_STATUS_INVALID_ARG;

        argIndex = MailboxAddProfile(deviceId, VP_PROFILE_TONECAD,
                                        pCadProfile, argIndex);
        if (argIndex == 0) retval = VP_STATUS_INVALID_ARG;

        /* Add DTMF tone generation info to the mailbox */
        if(pDtmfControl == VP_NULL){
            /* No digit requested. */
            MailboxAddArg(deviceId, VP_DIG_NONE, argIndex);
        } else {
            argIndex = MailboxAddArg(deviceId, (uint16)pDtmfControl->toneId, argIndex);
            MailboxAddArg(deviceId, (uint16)pDtmfControl->dir, argIndex);
        }

        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpSetLineTone() */


/**
 * VcpSetRelayState()
 *  Configures the VCP-controlled relays.
 *
 * Refer to the VP API User's Guide for details about this function.
 */
static VpStatusType
VcpSetRelayState(
    VpLineCtxType *pLineCtx,
    VpRelayControlType rState
    )
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint8 argIndex = VP_MB_ARGLIST_INDEX;
    VpStatusType retval;
    uint8 line = pLineObj->channelId;

    /* Basic error checking */
    if((rState > VP_RELAY_RINGING_TEST) || (rState < 0)) {
        return VP_STATUS_INVALID_ARG;
    }
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* DevNotes: It would be more efficient here if VpRelayState() constants
     *           were bitmasks.
     */

    /* No need to check for VP_TERM_FXS_75282 because all relay states are
     * supported (at least the ones that till this point) */
    if (pLineObj->termType == VP_TERM_FXS_75181) {
        /* This line termination type supports only certain types of
         * relay states */
        if ((rState != VP_RELAY_NORMAL) &&
            (rState != VP_RELAY_TALK) &&
            (rState != VP_RELAY_RINGING)) {
            return VP_STATUS_INVALID_ARG;
        }
    } else if ((pLineObj->termType == VP_TERM_FXS_TO_TL) ||
        (pLineObj->termType == VP_TERM_FXS_CO_TL)) {
        /* This line termination type supports only certain types of
         * relay states */
        if ((rState != VP_RELAY_NORMAL) &&
            (rState != VP_RELAY_TALK) && /* Same as normal */
            (rState != VP_RELAY_TESTOUT) &&
            (rState != VP_RELAY_DISCONNECT) && /* Same configuration as
                                                * testout */
            (rState != VP_RELAY_BRIDGED_TEST) &&
            (rState != VP_RELAY_SPLIT_TEST)) {
            return VP_STATUS_INVALID_ARG;
        }
    } else if (pLineObj->termType == VP_TERM_FXS_GENERIC) {
        /* This line termination type supports only certain types of
         * relay states */
        if ((rState != VP_RELAY_NORMAL) &&
            (rState != VP_RELAY_TALK) && /* Same as normal */
            (rState != VP_RELAY_BRIDGED_TEST)) {
            return VP_STATUS_INVALID_ARG;
        }
    } else if (pLineObj->termType == VP_TERM_FXS_TITO_TL_R) {
        /* This line termination type supports only certain types of
         * relay states */
        if ((rState != VP_RELAY_NORMAL) &&
            (rState != VP_RELAY_TALK) &&
            (rState != VP_RELAY_RINGING) &&
            (rState != VP_RELAY_TESTOUT) &&
            (rState != VP_RELAY_DISCONNECT) && /* Same configuration as
                                                * testout */
            (rState != VP_RELAY_BRIDGED_TEST) &&
            (rState != VP_RELAY_SPLIT_TEST) &&
            (rState != VP_RELAY_RINGING_TEST)) {
            return VP_STATUS_INVALID_ARG;
        }
    } else if (pLineObj->termType == VP_TERM_FXS_RDT) {
        /* This line termination type supports only certain types of
         * relay states */
        if ((rState != VP_RELAY_NORMAL) &&
            (rState != VP_RELAY_TALK) &&
            (rState != VP_RELAY_RINGING) &&
            (rState != VP_RELAY_TESTOUT) &&
            (rState != VP_RELAY_DISCONNECT) && /* Same configuration as
                                                * testout */
            (rState != VP_RELAY_BRIDGED_TEST) &&
            (rState != VP_RELAY_SPLIT_TEST)) {
            return VP_STATUS_INVALID_ARG;
        }
    } else if (pLineObj->termType == VP_TERM_FXS_RR) {
        /* This line termination type supports only certain types of
         * relay states */
        if ((rState != VP_RELAY_NORMAL) &&
            (rState != VP_RELAY_TALK) &&
            (rState != VP_RELAY_RINGING) &&
            (rState != VP_RELAY_BRIDGED_TEST) &&
            (rState != VP_RELAY_RESET)) {
            return VP_STATUS_INVALID_ARG;
        }

    } /* Else, no checking to support relay states or line termination types
       * for which API has still not been updated yet... */

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
        retval = VP_STATUS_MAILBOX_BUSY;

    } else {
        /* Add the Line State command to the mailbox */
        MailboxAddCommand(deviceId, VCP_MBFUNC_RELAYSTATE, 0x0000);

        /* Add the line number and state arguments to the mailbox */
        argIndex = MailboxAddArg(deviceId, (uint16)line,   argIndex);
        argIndex = MailboxAddArg(deviceId, (uint16)rState, argIndex);

        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);

        retval = VP_STATUS_SUCCESS;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpSetRelayState() */

/**
 * VcpSetRelGain()
 *  Adjusts the relative gain for the transmit and receive paths for the
 *  specified line.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpSetRelGain(
    VpLineCtxType *pLineCtx,
    uint16 txLevel,
    uint16 rxLevel,
    uint16 handle
    )
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint8 argIndex = VP_MB_ARGLIST_INDEX;
    VpStatusType retval;
    uint8 line = pLineObj->channelId;

    /* Basic error checking */
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
        retval = VP_STATUS_MAILBOX_BUSY;

    } else {
        /* Add the Line State command to the mailbox */
        MailboxAddCommand(deviceId, VCP_MBFUNC_RELGAIN, handle);

        /* Add the line number and level arguments to the mailbox */
        argIndex = MailboxAddArg(deviceId, (uint16)line, argIndex);
        argIndex = MailboxAddArg(deviceId, txLevel, argIndex);
        argIndex = MailboxAddArg(deviceId, rxLevel, argIndex);

        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);

        retval = VP_STATUS_SUCCESS;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpSetRelGain() */

/**
 * VcpSendSignal()
 * This function is used to send a specified signal on the line.
 *
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType VcpSendSignal(
    VpLineCtxType *pLineCtx,
    VpSendSignalType type,
    void *pStruct)
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint8 line = pLineObj->channelId;
    VpSendMsgWaitType *pSendMsg;
    #define MAX_SEND_SIGNAL_CMD_LEN (8)
    uint16 cidCmdBuf[MAX_SEND_SIGNAL_CMD_LEN];
    uint8 cmdLen = 0;

    /* Basic error testing */
    if(type != VP_SENDSIG_MSG_WAIT_PULSE) {
        return VP_STATUS_INVALID_ARG;
    }

    cidCmdBuf[cmdLen++] = VCP_MBFUNC_SEND_SIGNAL;
    cidCmdBuf[cmdLen++] = 0; /* This field is reserved */
    cidCmdBuf[cmdLen++] = line;
    cidCmdBuf[cmdLen++] = (uint16)type;

    if(pStruct == VP_NULL) {

        /* User wants to abort the signal */
        cidCmdBuf[cmdLen++] = 0;
        cidCmdBuf[cmdLen++] = 0;
        cidCmdBuf[cmdLen++] = 0;
        cidCmdBuf[cmdLen++] = 0;

    } else {

        /* Pass on the user defined signal */
        pSendMsg = pStruct;
        cidCmdBuf[cmdLen++] = pSendMsg->voltage;
        cidCmdBuf[cmdLen++] = pSendMsg->onTime;
        cidCmdBuf[cmdLen++] = pSendMsg->offTime;
        cidCmdBuf[cmdLen++] = pSendMsg->cycles;

    }

    return WriteMailBoxCmd(deviceId, cmdLen, cidCmdBuf);
} /* VcpSendSignal() */

/**
 * VcpSendCid()
 *  Sends on-demand Caller ID.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpSendCid(
    VpLineCtxType *pLineCtx,
    uint8 length,
    VpProfilePtrType pCidProfile,
    uint8p pCidData
    )
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint8 line = pLineObj->channelId;

    /* Maximum length of CID command buffer */
    #define MAX_SEND_CID_CMD_LEN (VCP_MAX_CID_DATA_LENGTH + 5)
    uint16 cidCmdBuf[MAX_SEND_CID_CMD_LEN];
    uint8 cmdLen = 0;
    int profileIndex;
    int i;

    /* Basic error checking */
    if((length > VCP_MAX_CID_DATA_LENGTH) || (length == 0)){
        return VP_STATUS_INVALID_ARG;
    }
    if(pCidData == VP_NULL) {
        return VP_STATUS_INVALID_ARG;
    }
    /* Basic error checking */
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Get the index of the profile */
    /* DevNotes: Presently VCP permits only CallerID profile indexes not an
     * actual profile; Once an actual profile is supported check for the profile
     * type as well */
    profileIndex = VpGetProfileIndex(pCidProfile);
    if(profileIndex < 0) {
        /* Error in profile index or profile pointer */
        return VP_STATUS_INVALID_ARG;
    }
    if((profileIndex + 1) > GetProfileTableMaxCount(VP_PROFILE_CID)) {
        /* index is 0 based */
        return VP_STATUS_INVALID_ARG;
    }

    /* Fill the command */
    cidCmdBuf[cmdLen++] = VCP_MBFUNC_SENDCID;
    cidCmdBuf[cmdLen++] = 0; /* This field is reserved */
    cidCmdBuf[cmdLen++] = line;
    cidCmdBuf[cmdLen++] = (profileIndex + 1) | 0x8000; /* To indicate
                                                        * profile index */
    cidCmdBuf[cmdLen++] = length;
    for(i = 0; i<length; i++){
        cidCmdBuf[cmdLen++] = pCidData[i];
    }

    return WriteMailBoxCmd(deviceId, cmdLen, cidCmdBuf);
} /* VcpSendCid() */

/**
 * VcpContinueCid()
 *  Continues a Caller ID sequence by updating the CID message buffer.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpContinueCid(
    VpLineCtxType *pLineCtx,
    uint8 length,
    uint8p pCidData
    )
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint8 line = pLineObj->channelId;
    uint8 cmdLen = 0;
    int i;

    /* Maximum length of CID command buffer */
    #define VCP_MAX_CID_CONTINUE_DATA_LENGTH (16)
    #define MAX_CONTINUE_CID_CMD_LEN (VCP_MAX_CID_CONTINUE_DATA_LENGTH + 4)
    uint16 cidCmdBuf[MAX_CONTINUE_CID_CMD_LEN];

    /* Basic error checking */
    if ((length > VCP_MAX_CID_CONTINUE_DATA_LENGTH) || (length == 0) ||
        (pCidData == VP_NULL)) {
        return VP_STATUS_INVALID_ARG;
    }

    /* Basic error checking */
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Fill the command */
    cidCmdBuf[cmdLen++] = VCP_MBFUNC_CONT_CID;
    cidCmdBuf[cmdLen++] = 0; /* This field is reserved */
    cidCmdBuf[cmdLen++] = line;
    cidCmdBuf[cmdLen++] = length;
    for(i = 0; i<length; i++){
        cidCmdBuf[cmdLen++] = pCidData[i];
    }

    return WriteMailBoxCmd(deviceId, cmdLen, cidCmdBuf);
} /* VcpContinueCid() */

/**
 * VcpStartMeter()
 *  Places a cadenced meter tone on to a VoicePath VCP line.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpStartMeter(
    VpLineCtxType *pLineCtx,    /* Selects the desired line */
    uint16 onTime,          /* Pulse on time, in units of 10ms  */
    uint16 offTime,         /* Pulse off time, in units of 10ms */
    uint16 numMeters        /* Number of meter pulse to automatically perform */
    )
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint8 argIndex = VP_MB_ARGLIST_INDEX;
    VpStatusType retval;
    uint8 line = pLineObj->channelId;

    /* Basic error checking */
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
        retval = VP_STATUS_MAILBOX_BUSY;

    } else {
        retval = VP_STATUS_SUCCESS;

        /* Add the Start Meter command to the mailbox */
        MailboxAddCommand(deviceId, VCP_MBFUNC_METER, 0x0000);

        /* Add the line number argument to the mailbox */
        argIndex = MailboxAddArg(deviceId, (uint16)line, argIndex);

        /* Add the cadence arguments to the mailbox */
        argIndex = MailboxAddArg(deviceId, onTime, argIndex);
        argIndex = MailboxAddArg(deviceId, offTime, argIndex);

        /* Add the numMeters argument to the mailbox */
        argIndex = MailboxAddArg(deviceId, numMeters, argIndex);

        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpStartMeter() */

/**
 * VcpSetOption()
 * This function is used to set various options that are supported by VCP.
 * For a detailed description of which options are available for VCP please
 * see VP-API user guide.
 *
 * Preconditions:
 * Device/Line context should be created and initialized. For applicable devices
 * bootload should be performed before calling the function.
 *
 * Postconditions:
 * This function sets the requested option.
 */
#define MAX_SET_OPT_CMD_LEN 20

static VpStatusType
VcpSetOption(
    VpLineCtxType *pLineCtx,
    VpDevCtxType *pDevCtx,
    VpOptionIdType option,
    void *value)
{
    VpDevCtxType *pDevCtxLocal;
    VpVcpLineObjectType *pLineObj = VP_NULL;
    uint16 msgLen;
    uint16 msgBuf[MAX_SET_OPT_CMD_LEN];
    VpStatusType retval;
    uint8 line = 0;
    VpDeviceIdType deviceId;
    VpVcpDeviceObjectType *pDevObj;
    VpOptionCodecType *codec;
    VpOptionPulseModeType *pulseMode;
    VpOptionPulseType *pulse;
    VpOptionPcmHwyType *pcmHwy;
    VpOptionLoopbackType *loopback;
    VpOptionLineStateType *lineState;
    VpOptionZeroCrossType *zeroCross;
    VpOptionEventMaskType *eventMask;
    VpOptionCriticalFltType *critical;
    VpOptionRingControlType *ringControl;
    VpOptionDtmfModeType *pDtmfControl = VP_NULL;
    VpOptionDeviceIoType *pDeviceIoControl;
    VpOptionPcmTxRxCntrlType *pPcmTxRxCntrl;
    bool readDtmfStatus = FALSE;
    bool issueSetOptionCmd = TRUE;
    uint16 dtmfStatus[2];
    uint32 dtmfStatusLong;
    uint16 remainingDtmfResources;
    VpOptionDtmfModeControlType dtmfControl = VP_NUM_OPTION_DTMF_IDS;
    uint16 eventMaskData;
    int i;
    bool calCmpEventMask;

    if((option < 0) ||
       (option > VP_OPTION_ID_PCM_TXRX_CNTRL) ||
       (option == VP_OPTION_ID_RESERVED_1) ||
       (option == VP_OPTION_ID_RESERVED_2)) {
        /* These options are not understood by VCP */
        return VP_STATUS_INVALID_ARG;
    }

    if(pLineCtx != VP_NULL) {
        pDevCtxLocal = pLineCtx->pDevCtx;
        pDevObj = pDevCtxLocal->pDevObj;
        deviceId = pDevObj->deviceId;
        pLineObj = pLineCtx->pLineObj;
        line = pLineObj->channelId;
        if (line >= pDevObj->maxChannels) {
            return VP_STATUS_INVALID_ARG;
        }

        /* Make sure a device specific option is not being set */
        if((option == VP_DEVICE_OPTION_ID_PULSE) ||
           (option == VP_DEVICE_OPTION_ID_CRITICAL_FLT) ||
           (option == VP_DEVICE_OPTION_ID_RAMP2STBY) ||
           (option == VP_DEVICE_OPTION_ID_DEVICE_IO)) {
            return VP_STATUS_INVALID_ARG;
        }

    } else {        /* Device context pointer must be valid */

        pDevObj = pDevCtx->pDevObj;
        deviceId = pDevObj->deviceId;

        /* For device options and for line specific options (when device context
         * is passed) set the requested option to the device and to all lines.*/
        line = VP_ALL_LINES;
    }
    /* Basic error checking */
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED; /* Because options are going to
        * be overwritten upon calling initialization */
    }

    retval = VP_STATUS_SUCCESS;
    msgLen = 0;

    /* Add the Set Option command to the mailbox. */
    msgBuf[msgLen++] = (uint16)VCP_MBFUNC_SETOPTION;
    msgBuf[msgLen++] = 0x0000;                          /* dummy context info */

    /* Add the line number and option arguments to the mailbox. */
    msgBuf[msgLen++] = (uint16)line;
    if(pDevObj->productType == VP_VCP_PRODUCT_VE_880) {
        if(option == VP_OPTION_ID_EVENT_MASK) {
            /* VCP now has a better event mask command; The previous one did not
             * mask the device event masks if the line were to any line other
             * than zero. The following option does not have such restrictions*/
            option = VP_OPTION_ID_RESERVED_EVENT_MASK_VCP;
        }
    }

    msgBuf[msgLen++] = (uint16)option;

    /* Add the option arguments to the mailbox .*/
    switch (option) {
        case VP_OPTION_ID_TIMESLOT:
            msgBuf[msgLen++] = VP_TS_SET; /* Time slots need not have to be
                                           * turned on here */
            msgBuf[msgLen++] = (uint16)((VpOptionTimeslotType *)value)->tx;
            msgBuf[msgLen++] = (uint16)((VpOptionTimeslotType *)value)->rx;
            break;

        case VP_OPTION_ID_CODEC:
            codec = value;
            if ((*codec == VP_OPTION_ALAW) ||
                (*codec == VP_OPTION_MLAW) ||
                (*codec == VP_OPTION_LINEAR)) {

                msgBuf[msgLen++] = (uint16)*codec;
            } else {
                retval = VP_STATUS_INVALID_ARG;
            }
            break;

        case VP_OPTION_ID_PULSE_MODE:
            pulseMode = value;
            if ((*pulseMode == VP_OPTION_PULSE_DECODE_OFF) ||
                (*pulseMode == VP_OPTION_PULSE_DECODE_ON)) {

                msgBuf[msgLen++] = (uint16)*pulseMode;
            } else {
                retval = VP_STATUS_INVALID_ARG;
            }
            break;

        case VP_DEVICE_OPTION_ID_PULSE:
            pulse =  value;
            if (pulse->makeMin > pulse->makeMax) {
                retval = VP_STATUS_INVALID_ARG;
            }

            if (pulse->breakMin > pulse->breakMax) {
                retval = VP_STATUS_INVALID_ARG;
            }

            if (pulse->flashMin > pulse->flashMax) {
                retval = VP_STATUS_INVALID_ARG;
            }

            if (retval == VP_STATUS_SUCCESS) {
                /* Only issue command if args are valid. */
                msgBuf[msgLen++] = pulse->makeMin;
                msgBuf[msgLen++] = pulse->makeMax;
                msgBuf[msgLen++] = pulse->breakMin;
                msgBuf[msgLen++] = pulse->breakMax;
                msgBuf[msgLen++] = pulse->interDigitMin;
                msgBuf[msgLen++] = pulse->flashMin;
                msgBuf[msgLen++] = pulse->flashMax;
                msgBuf[msgLen++] = pulse->flashMax + 1; /* API does not have
                * "onHookMin" field anymore */
            }
            break;

        case VP_OPTION_ID_PCM_HWY:
            pcmHwy = value;
            if ((*pcmHwy == VP_OPTION_HWY_A) ||
                (*pcmHwy == VP_OPTION_HWY_B)) {

                msgBuf[msgLen++] = (uint16)*pcmHwy;
            } else {
                retval = VP_STATUS_INVALID_ARG;
            }
            break;

        case VP_OPTION_ID_LOOPBACK:
            loopback = value;
            if ((*loopback == VP_OPTION_LB_OFF) ||
                (*loopback == VP_OPTION_LB_TIMESLOT) ||
                (*loopback == VP_OPTION_LB_DIGITAL)) {

                msgBuf[msgLen++] = (uint16)*loopback;
            } else {
                retval = VP_STATUS_INVALID_ARG;
            }
            break;

        case VP_OPTION_ID_LINE_STATE:
            lineState = value;
            if (lineState->battRev) {
                msgBuf[msgLen++] = 0x0001;
            } else {
                msgBuf[msgLen++] = 0x0000;
            }
            if (lineState->bat < VP_NUM_OPTION_BAT_IDS) {
                msgBuf[msgLen++] = (uint16)lineState->bat;
            } else {
                retval = VP_STATUS_INVALID_ARG;
            }
            break;

        case VP_OPTION_ID_RESERVED_EVENT_MASK_VCP: /* New VCP option */
        case VP_OPTION_ID_EVENT_MASK:
            eventMask = value;
            /* Mask off undefined events; Unmask non-maskable events */
            eventMaskData = (eventMask->faults |
                                VP_EVCAT_FAULT_UNDEFINED |
                                VCP_EVCAT_FAULT_NOT_NEEDED) &
                                (VCP_EVCAT_FAULT_NEEDED &
                                (~VP_API_NONMASK_FAULT_EVENTS));
            msgBuf[msgLen++] = eventMaskData;
#ifdef DEBUG_VCP_API_EVENT
            term_printf("EventMask->faults = %x\n", eventMaskData);
#endif /* DEBUG_VCP_API_EVENT */
            eventMaskData = (eventMask->signaling |
                                VP_EVCAT_SIGNALING_UNDEFINED |
                                VCP_EVCAT_SIGNALLING_NOT_NEEDED) &
                                (VCP_EVCAT_SIGNALLING_NEEDED &
                                (~VP_API_NONMASK_SIGNALING_EVENTS));
            /* DevNotes: Note that this section of the code should
             * protected, but it is not, Such protection would not
             * be needed if one is implemented at the top level. This is on the
             * TODO list */
            if((eventMaskData & VP_DEV_EVID_TS_ROLLOVER) != 0) {
                /* Unmask so that API could keep track of time */
                eventMaskData = eventMaskData & (~VP_DEV_EVID_TS_ROLLOVER);
                /* Set a flag so that this event can be masked for
                 * the application */
                pDevObj->timeStampEventMask = TRUE;
            } else {
                pDevObj->timeStampEventMask = FALSE;
            }
            msgBuf[msgLen++] = eventMaskData;
#ifdef DEBUG_VCP_API_EVENT
            term_printf("EventMask->signaling = %x\n", eventMaskData);
#endif /* DEBUG_VCP_API_EVENT */
            eventMaskData = (eventMask->response |
                                VP_EVCAT_RESPONSE_UNDEFINED |
                                VCP_EVCAT_RESPONSE_NOT_NEEDED ) &
                                (VCP_EVCAT_RESPONSE_NEEDED &
                                (~VP_API_NONMASK_RESPONSE_EVENTS));
            /* Calibration complete event is handled in software */
            if((eventMaskData & VP_EVID_CAL_CMP) != 0) {
                calCmpEventMask = TRUE;
                eventMaskData = eventMaskData & (~VP_EVID_CAL_CMP);
            } else {
                calCmpEventMask = FALSE;
            }
            /* Generate software mask information */
            if(line != VP_ALL_LINES) {
                /* A line context was passed */
                pLineObj->calCompleteEventMask = calCmpEventMask;
            } else {
                for(i = 0; i < pDevObj->maxChannels; i++) {
                    if(pDevCtx->pLineCtx[i] != VP_NULL) {
                        pLineObj = pDevCtx->pLineCtx[i]->pLineObj;
                        pLineObj->calCompleteEventMask = calCmpEventMask;
                    }
                }
            }
            /* VCP does not understand these events and these are non maskable
             * events */
            eventMaskData |= VP_DEV_EVID_DEV_INIT_CMP;
            msgBuf[msgLen++] = eventMaskData;
#ifdef DEBUG_VCP_API_EVENT
            term_printf("EventMask->response = %x\n", eventMaskData);
#endif /* DEBUG_VCP_API_EVENT */
            eventMaskData = (eventMask->test |
                                VP_EVCAT_TEST_UNDEFINED) &
                                (~VP_API_NONMASK_TEST_EVENTS);
            msgBuf[msgLen++] = eventMaskData;
#ifdef DEBUG_VCP_API_EVENT
            term_printf("EventMask->test = %x\n", eventMaskData);
#endif /* DEBUG_VCP_API_EVENT */
            eventMaskData = (eventMask->process |
                                VP_EVCAT_PROCESS_UNDEFINED) &
                                (~ VP_API_NONMASK_PROCESS_EVENTS);
            if(pDevObj->productType == VP_VCP_PRODUCT_VE_880) {
                /* VCP-880 does not support this event. So, mask it */
                eventMaskData = eventMaskData | VP_LINE_EVID_MTR_CAD;
            }

            msgBuf[msgLen++] = eventMaskData;
#ifdef DEBUG_VCP_API_EVENT
            term_printf("EventMask->process = %x\n", eventMaskData);
#endif /* DEBUG_VCP_API_EVENT */
            break;

        case VP_OPTION_ID_ZERO_CROSS:
            zeroCross = value;
            if(*zeroCross < 0){
                retval = VP_STATUS_INVALID_ARG;
                break;
            }
            if(pDevObj->productType == VP_VCP_PRODUCT_VE_880) {
                if(*zeroCross > VP_OPTION_ZC_NONE) {
                    retval = VP_STATUS_INVALID_ARG;
                    break;
                }
            } else {
                /* Must be VP_VCP_PRODUCT_VE_790 */
                if(*zeroCross >= VP_OPTION_ZC_NONE) {
                    retval = VP_STATUS_INVALID_ARG;
                    break;
                }
            }
            msgBuf[msgLen++] = *zeroCross;
            break;
        case VP_DEVICE_OPTION_ID_RAMP2STBY:
            msgBuf[msgLen++] = *((uint16 *)value);
            break;
        case VP_DEVICE_OPTION_ID_CRITICAL_FLT:
            critical = value;
            if(pDevObj->productType == VP_VCP_PRODUCT_VE_880) {
                /* VCP-880 does not support state transition to disconnect
                 * upon fault detection. */
                if((critical->acFltDiscEn == TRUE) ||
                   (critical->dcFltDiscEn == TRUE)) {
                    retval = VP_STATUS_INVALID_ARG;
                    break;
                }
            }
            msgBuf[msgLen++] = critical->acFltDiscEn |
                               (critical->dcFltDiscEn<<1) |
                               (critical->thermFltDiscEn<<3);
            break;
        case VP_OPTION_ID_RING_CNTRL:
            ringControl = value;
            if(ringControl->zeroCross < 0){
                retval = VP_STATUS_INVALID_ARG;
                break;
            }
            if(pDevObj->productType == VP_VCP_PRODUCT_VE_880) {
                if(ringControl->zeroCross > VP_OPTION_ZC_NONE) {
                    retval = VP_STATUS_INVALID_ARG;
                    break;
                }
            } else {
                /* Must be VP_VCP_PRODUCT_VE_790 */
                if(ringControl->zeroCross >= VP_OPTION_ZC_NONE) {
                    retval = VP_STATUS_INVALID_ARG;
                    break;
                }
            }
            msgBuf[msgLen++] = ringControl->zeroCross;
            msgBuf[msgLen++] = ringControl->ringExitDbncDur;
            msgBuf[msgLen++] = ringControl->ringTripExitSt | VCP_LINE_VP_MASK;
            break;
        case VP_OPTION_ID_DTMF_MODE:
            pDtmfControl = value;
            dtmfControl = pDtmfControl->dtmfControlMode;

            if(dtmfControl < VP_NUM_OPTION_DTMF_IDS){

                if(dtmfControl == VP_OPTION_DTMF_DECODE_ON) {
                   /* VCP supports DTMF detection only in Upstream direction */
                   if(pDtmfControl->direction == VP_DIRECTION_US) {
                        msgBuf[msgLen++] = dtmfControl;
                        readDtmfStatus = TRUE;
                    } else {
                        retval = VP_STATUS_INVALID_ARG;
                    }
                } else if(dtmfControl == VP_OPTION_DTMF_DECODE_OFF) {
                    /* Direction does not matter when one is turning OFF DTMF
                     * detection */
                    msgBuf[msgLen++] = dtmfControl;
                    readDtmfStatus = TRUE;
                } else {
                    /* User just wants to get the status of DTMF detectors */
                    readDtmfStatus = TRUE;
                    issueSetOptionCmd = FALSE;
                }
            } else {
                retval = VP_STATUS_INVALID_ARG;
            }
            break;
        case VP_DEVICE_OPTION_ID_DEVICE_IO:
            pDeviceIoControl = value;
            msgBuf[msgLen++] = pDeviceIoControl->directionPins_31_0 >> 16;
            msgBuf[msgLen++] = pDeviceIoControl->directionPins_31_0 & 0xffff;
            msgBuf[msgLen++] = pDeviceIoControl->directionPins_63_32 >> 16;
            msgBuf[msgLen++] = pDeviceIoControl->directionPins_63_32 & 0xffff;
            msgBuf[msgLen++] = pDeviceIoControl->outputTypePins_31_0 >> 16;
            msgBuf[msgLen++] = pDeviceIoControl->outputTypePins_31_0 & 0xffff;
            msgBuf[msgLen++] = pDeviceIoControl->outputTypePins_63_32 >> 16;
            msgBuf[msgLen++] = pDeviceIoControl->outputTypePins_63_32 & 0xffff;
            break;
        case VP_OPTION_ID_PCM_TXRX_CNTRL:
            pPcmTxRxCntrl = value;
            if((*pPcmTxRxCntrl < VP_OPTION_PCM_BOTH) ||
               (*pPcmTxRxCntrl > VP_OPTION_PCM_TX_ONLY)) {
                retval = VP_STATUS_INVALID_ARG;
            } else {
                msgBuf[msgLen++] = *pPcmTxRxCntrl;
            }
            break;
        default:
            retval = VP_STATUS_INVALID_ARG;
            break;
    }

    /* Check for buffer overflow. */
    if (msgLen > MAX_SET_OPT_CMD_LEN) {
        /* Indicates a software bug. Increase the buffer size. */
        retval = VP_STATUS_ERR_HBI;
    }

    /*
     * If we made it through all of that without detecting an error condition,
     * go ahead and write the command to the mailbox.
     */
    if (retval == VP_STATUS_SUCCESS) {
        VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);/* CRITICAL
                                                           * SECTION => */

        if(readDtmfStatus == TRUE) {
            /* Read the status of DTMF detection resources from the
             * direct page */
            VpHbiRd(deviceId, VP_REG_DTMF_REMAINING_RSRC,
                        &remainingDtmfResources);
            pDtmfControl->dtmfResourcesRemaining = remainingDtmfResources;

            VpHbiRd(deviceId, VP_REG_DTMF_ENABLE_STATUS, dtmfStatus);
            dtmfStatusLong = ((uint32)dtmfStatus[0]<<16) + dtmfStatus[1];
            pDtmfControl->dtmfDetectionSetting = dtmfStatusLong;

            if(dtmfControl != VP_OPTION_DTMF_GET_STATUS) {
                if(dtmfControl == VP_OPTION_DTMF_DECODE_ON) {
                    if((remainingDtmfResources == 0) || (line == VP_ALL_LINES)){

                        retval = VP_STATUS_DEVICE_BUSY;
                        issueSetOptionCmd = FALSE; /* No more resources
                        * available; So no point in issuing the command */
                    } else {
                        /* Compute the end result that would be present in
                         * the direct page registers after the command was
                         * processed */
                        if((dtmfStatusLong & ((uint32)1<<line)) == 0) {
                            /* Line already does not have DTMF enabled */
                            pDtmfControl->dtmfDetectionSetting =
                                dtmfStatusLong | ((uint32)1<<line);
                            pDtmfControl->dtmfResourcesRemaining =
                                    remainingDtmfResources - 1;
                        } else {
                            /* DTMF is already enabled for this line; No
                             * need to issue a command once again */
                            issueSetOptionCmd = FALSE;
                        }
                    }
                } else {
                    /* user must be turning off DTMF detection; Update the
                     * Data that this option returns */
                    if(line == VP_ALL_LINES){
                        pDtmfControl->dtmfDetectionSetting = 0;
                        pDtmfControl->dtmfResourcesRemaining =
                            GetTotalDtmfResources(dtmfStatusLong,
                                            remainingDtmfResources);
                    } else {
                        if((dtmfStatusLong & ((uint32)1<<line)) != 0) {
                            /* Line has a DTMF channel enabled that is being
                             * disabled */
                            pDtmfControl->dtmfDetectionSetting =
                                dtmfStatusLong & (~((uint32)1<<line));
                            pDtmfControl->dtmfResourcesRemaining =
                                    remainingDtmfResources + 1;
                        } else {
                            /* DTMF is not enabled on this line; So no need
                             * to issue the command */
                            issueSetOptionCmd = FALSE;
                            retval = VP_STATUS_SUCCESS;
                        }
                    }
                }
            }
        }

        if(issueSetOptionCmd == TRUE) {
            /* Ensure that the Mailbox is available for the host */
            if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
                retval = VP_STATUS_MAILBOX_BUSY;
            } else {
                /* Write the message to the mailbox. */
                VpHbiWr(deviceId, VP_REG_MB_DOWN_BUF(0, msgLen), msgBuf);
                /* Turn the mailbox over to the VCP (raise the flag) */
                MailboxRelease(deviceId);
            }
        }

        VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);/* CRITICAL
                                                          * SECTION <= */
    }

    return retval;
} /* VcpSetOption() */

/**
 * VcpDeviceIoAccess()
 * This function is used to set GPIO pins provided in the VTD. Please see
 * VP-API user guide for more details.
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * None
 */
static VpStatusType
VcpDeviceIoAccess(
    VpDevCtxType *pDevCtx,
    VpDeviceIoAccessDataType *pDeviceIoData)
{
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    #define MAX_DEVICE_IO_CMD_LEN (3 + 9)
    uint16 cidCmdBuf[MAX_SEND_CID_CMD_LEN];
    uint8 cmdLen = 0;
    VpDeviceIoAccessType accessType = pDeviceIoData->accessType;

    /* Basic error checking */
    if((pDevCtx == VP_NULL) || (pDeviceIoData == VP_NULL)){
        return VP_STATUS_INVALID_ARG;
    }
    if((accessType != VP_DEVICE_IO_WRITE) &&
       (accessType != VP_DEVICE_IO_READ)){
        return VP_STATUS_INVALID_ARG;
    }

    cidCmdBuf[cmdLen++] = VCP_MBFUNC_DEVICE_IO;
    cidCmdBuf[cmdLen++] = 0; /* This field is reserved */
    cidCmdBuf[cmdLen++] = 0; /* Line fied is not used */

    /* Fill rest of the arguments */
    cidCmdBuf[cmdLen++] = accessType;
    cidCmdBuf[cmdLen++] = pDeviceIoData->accessMask_31_0 >> 16;
    cidCmdBuf[cmdLen++] = pDeviceIoData->accessMask_31_0 & 0xffff;
    cidCmdBuf[cmdLen++] = pDeviceIoData->accessMask_63_32 >> 16;
    cidCmdBuf[cmdLen++] = pDeviceIoData->accessMask_63_32 & 0xffff;
    cidCmdBuf[cmdLen++] = pDeviceIoData->deviceIOData_31_0 >> 16;
    cidCmdBuf[cmdLen++] = pDeviceIoData->deviceIOData_31_0 & 0xffff;
    cidCmdBuf[cmdLen++] = pDeviceIoData->deviceIOData_63_32 >> 16;
    cidCmdBuf[cmdLen++] = pDeviceIoData->deviceIOData_63_32 & 0xffff;

    return WriteMailBoxCmd(deviceId, cmdLen, cidCmdBuf);

} /* VcpDeviceIoAccess() */


/**
 * VcpLowLevelCmd()
 *  Execute a direct SLAC MPI command.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpLowLevelCmd(
    VpLineCtxType *pLineCtx,
    uint8p cmd,
    uint8 len,
    uint16 handle
    )
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint8 argIndex = VP_MB_ARGLIST_INDEX;
    bool isReadCmd = *cmd & 0x01;
    uint8 ctr;
    VpStatusType retval;
    uint8 line = pLineObj->channelId;

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL
                                                         * SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
        retval = VP_STATUS_MAILBOX_BUSY;

    } else {
        /* Add the MPI command to the mailbox */
        MailboxAddCommand(deviceId, VCP_MBFUNC_MPICMD, handle);

        /* Add the line number, cmd, and length arguments to the mailbox */
        argIndex = MailboxAddArg(deviceId, (uint16)line,   argIndex);
        argIndex = MailboxAddArg(deviceId, (uint16)*cmd++, argIndex);
        argIndex = MailboxAddArg(deviceId, (uint16)len,    argIndex);
        argIndex = MailboxAddArg(deviceId, (uint16)handle, argIndex);

        /* For write commands, add the cmd data to the mailbox. */
        if (!isReadCmd) {
            for (ctr = 0; ctr < len; ctr++) {
                /*
                 * NOTE: MPI data bytes are currently NOT packed into words for
                 * the HBI transaction. That is, only one byte of each word sent
                 * through the HBI is meaningful.
                 */
                argIndex = MailboxAddArg(deviceId, (uint16)*cmd++, argIndex);
            }
        }
        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);

        retval = VP_STATUS_SUCCESS;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpLowLevelCmd() */




/******************************************************************************
 *                   VCP Specific Status and Query Functions                  *
 ******************************************************************************/
/**
 * VcpGetEvent()
 *  Gets an event (if available) from the VCP.
 *
 * Refer to the VP-API User's Guide for events that can be obtained from this
 * function.
 */
static bool
VcpGetEvent(
    VpDevCtxType *pDevCtx,
    VpEventType *pEvent)
{
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint8 channelId;
    uint16 eventWord;
    uint16 paramWord;
    uint16 eventId;
    VpStatusType retval;
    VpEventCategoryType eventCat;
    uint16 mbFlag;
    uint16 mbEventId;
    bool responsePresent = FALSE;
    bool readOptionResponse = FALSE;
    uint16 readOptionResponseOptId = 0;
    VpVcpLineObjectType *plineObj;
    uint16 productId;

    /* First initialize the event data */
    pEvent->status = VP_STATUS_SUCCESS;
    pEvent->channelId = 0;
    pEvent->eventId = 0; /* No event for VP-API */
    pEvent->pLineCtx = VP_NULL;
    pEvent->deviceId = deviceId;
    pEvent->pDevCtx = pDevCtx;
    pEvent->eventCategory = VP_EVCAT_FAULT;
    pEvent->parmHandle = 0;
    pEvent->eventData = 0;
    pEvent->hasResults = FALSE;
    /* Note that we are not initializing the 'lineId' member because we do not
     * know what it should be initialized because the basic type of this member
     * is defined by the customer and hence this variable could not have one
     * default value. */

    /* NOTE: Critical section protection is not required if this function is
     * called from the interrupt context (as is recommended for VCP). If however
     * there are implementations where a task would be launched upon hardware
     * interrupt, and if that task were to call this function, there is a
     * possibility that task could be interrupted to invoke another HBI
     * transaction. Hence the critical section protection. Another reason why
     * this may be required is: The process that is receiving downstream
     * (signaling information from CO; like T1/E1 framer) may have higher
     * priority compared VCP interrupts.
     */
    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Read the event register */
    VpHbiRd(deviceId, HW_Reg_INTIND, &eventWord);
    VpHbiRd(deviceId, HW_Reg_INTPARAM, &paramWord);

    /* See if a mailbox response is present */
    VpHbiRd(deviceId, HW_Reg_MBOXFLAG, &mbFlag);
    if((mbFlag & VP_BIT_MBOXFLAG_RSP) == 0) {
        responsePresent = TRUE;

        /* Read the event corresponding to mailbox response */
        VpHbiRd(deviceId, VP_REG_MB_UP_BUF(0, 1), &mbEventId);

        /* If the response mailbox contains an option that is being read back
         * read the option ID that is being read back */
        if(mbEventId == VCP_EVID_RD_OPTION) {
            readOptionResponse = TRUE;

            /* Read the option Id corrsponding to the option that is being
             * read */
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(3, 1), &readOptionResponseOptId);

        } else if(mbEventId == VCP_EVID_BOOT_CMP) {

            /* Received a boot complete event. Capture the VCP product related
             * information for future use. */

            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(6, 1), &productId);
            #define VE_880_PRODUCT_MASK 0x80
            if((productId & VE_880_PRODUCT_MASK) == VE_880_PRODUCT_MASK) {
                pDevObj->productType = VP_VCP_PRODUCT_VE_880;
                #ifdef DEBUG_VCP_API_EVENT
                term_printf("VCP Product VP_VCP_PRODUCT_VE_880 found\n");
                #endif /* DEBUG_VCP_API_EVENT */
            } else {
                pDevObj->productType = VP_VCP_PRODUCT_VE_790;
                #ifdef DEBUG_VCP_API_EVENT
                term_printf("VCP Product VP_VCP_PRODUCT_VE_790 found\n");
                #endif /* DEBUG_VCP_API_EVENT */
            }

            /* Reassign the function pointers as per the VCP product type */
            if(VpMakeVcpDeviceCtx(pDevCtx, pDevObj) != VP_STATUS_SUCCESS) {
                #ifdef DEBUG_VCP_API
                term_printf("Error during reinitializing the funciton pointers"
                            " after booting.\n");
                #endif /* DEBUG_VCP_API */
            }
        }
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    if(ConvertVcpEventId2ApiEventId(eventWord, &eventCat, &eventId) == FALSE) {
        if(eventWord == VCP_EVID_NO_EVENT) {
            /* No event present, hence no need to return false; Host could
             * be polling for events */
            pEvent->status = VP_STATUS_SUCCESS;
        } else {
            /* Event not understood; Some kind of VCP/API understanding
             * problem */
            #ifdef DEBUG_VCP_API
            term_printf("VCP Event not understood = 0x%.4x\n", eventWord);

            #endif /* DEBUG_VCP_API */
            pEvent->status = VP_STATUS_FAILURE;
        }
        return FALSE;
    }

    /* Finish up the API's return event structure */
    pEvent->eventCategory = eventCat;
    pEvent->eventId = eventId;
    pEvent->channelId = channelId = (uint8)eventWord & 0x1F;
    if (channelId < pDevObj->maxChannels) {
        pEvent->pLineCtx = pDevCtx->pLineCtx[channelId];

        if(pEvent->pLineCtx != VP_NULL) {
            plineObj = pEvent->pLineCtx->pLineObj;
            pEvent->lineId  = plineObj->lineId;
        }
    }
    pEvent->parmHandle = paramWord;

    /* Event pre-generation processing */
    if(pDevObj->deviceInitialized == FALSE){

        if((eventCat == VP_EVCAT_RESPONSE) &&
           (eventId == VP_EVID_CAL_CMP)) {
            /* This event must be due to VpInitDevice() function call */

            /* Now device initialization is complete; Set the flag so that we
             * do not do this again and also so that options can be set */
            pDevObj->deviceInitialized = TRUE;

            /* Device has been initialized; Set up default options */
            retval = SetVcpDefaultOptions(pDevCtx, VP_NULL);
            if(retval != VP_STATUS_SUCCESS){
                /* No Point in continuing... Exit */
                pEvent->status = retval;
                return FALSE; /* No device init complete event */
            }

            /* This event contains information about SLACs found;
             * May need it in future, store it */
            pDevObj->slacsFoundDuringCal = paramWord;

            /* Swap the information regarding SLACs found */
            pEvent->eventData = paramWord;
            pEvent->parmHandle = 0;

            /* API-II requires the event in response to be device init complete
             * and not cal complete; So change the event */
            pEvent->eventId = eventId = VP_DEV_EVID_DEV_INIT_CMP;
        }
    }

    /* Event post processing */
    if(eventCat == VP_EVCAT_SIGNALING) {
        if(eventId == VP_LINE_EVID_DTMF_DIG) {
            #define VCP_EVENT_DTMF_DIGIT_TIME_STAMP_MASK  (0xffe0)
            #define VCP_EVENT_DTMF_DIGIT_MASK (0x1f)
            pEvent->eventData = paramWord & VCP_EVENT_DTMF_DIGIT_MASK;
            pEvent->parmHandle =
                ((VCP_EVENT_DTMF_DIGIT_TIME_STAMP_MASK & paramWord)>>
                    VCP_TS_RESL_2_API_TS_RESL) +
                (pDevObj->timeStampCntrHB<<VCP_SW_TS_CNTR_RESL);

        } else if(eventId == VP_LINE_EVID_PULSE_DIG) {
            #define VCP_EVENT_PULSE_DIGIT_MASK (0xff) /* Invalid pulse
                * digit is communicated as 0xff. Note that DTMF does not
                * equivalent invalid digit reporting mechanism. */
            pEvent->eventData = paramWord & VCP_EVENT_PULSE_DIGIT_MASK;
            pEvent->parmHandle = 0; /* No time stamp for pulse digit */
        } else if((eventId == VP_LINE_EVID_HOOK_OFF) ||
                  (eventId == VP_LINE_EVID_HOOK_ON) ||
                  (eventId == VP_LINE_EVID_GKEY_DET) ||
                  (eventId == VP_LINE_EVID_GKEY_REL)) {
            /* These events have timestamp information; Correct the scale */
            pEvent->parmHandle =
                (paramWord >> VCP_TS_RESL_2_API_TS_RESL) +
                (pDevObj->timeStampCntrHB<<VCP_SW_TS_CNTR_RESL);
        } else if(eventId == VP_DEV_EVID_TS_ROLLOVER) {
            pDevObj->timeStampCntrHB++;
            if(pDevObj->timeStampCntrHB == (1<<VCP_TS_RESL_2_API_TS_RESL)){
                pDevObj->timeStampCntrHB = 0;
                if(pDevObj->timeStampEventMask == TRUE) {
                    /* User has masked this event; So do not generate an event*/
                    return FALSE;
                }
            } else {
                /* Since software is keeping track remaining time stamp counter
                 * events */
                return FALSE;
            }
        }
    } else if(eventCat == VP_EVCAT_FAULT) {
        /* All the events in this category return the data in the same way */
        pEvent->eventData = pEvent->parmHandle;
        pEvent->parmHandle = 0;
    } else if(eventCat == VP_EVCAT_PROCESS) {
        if((eventId == VP_LINE_EVID_RING_CAD) ||
           (eventId == VP_LINE_EVID_CID_DATA) ||
           (eventId == VP_LINE_EVID_SIGNAL_CMP) ||
           (eventId == VP_LINE_EVID_MTR_ABORT) ||
           (eventId == VP_LINE_EVID_MTR_CAD) ||
           (eventId == VP_LINE_EVID_TONE_CAD)) {
            pEvent->eventData = pEvent->parmHandle;
            pEvent->parmHandle = 0;
        }
    } else if(eventCat == VP_EVCAT_TEST) {
        if((eventId == VP_LINE_EVID_DTONE_DET) ||
           (eventId == VP_LINE_EVID_DTONE_LOSS)) {
            /* These events do no contain time stamp; They contain sample
             * counters from the start of the test */
            pEvent->eventData = pEvent->parmHandle;
            pEvent->parmHandle = 0;
        } else if(eventId == VP_DEV_EVID_STEST_CMP) {
            /* Contains test pass/fail information */
            pEvent->eventData = pEvent->parmHandle;
            pEvent->parmHandle = 0;
        }
    } else if(eventCat == VP_EVCAT_RESPONSE) {
        if(eventId == VP_EVID_CAL_CMP){
            /* This event is not handled in the hardware, It is handled in
             * software; Check to see if this event is enabled */
            if(pDevCtx->pLineCtx[channelId] == VP_NULL){ /* NOTE: Calibration
                 * complete is a line specific event */
                /* Since there is no line object associated with the event
                 * toss the event */

                /* DevNotes: VCP returns the line number of the first line
                 * in a SLAC when SLAC calibration is done. It is possible that
                 * there are no line contexts corresponding to that line. In
                 * such a case the calibration complete event would be lost.
                 */
                return FALSE;
            } else {
                plineObj = pDevCtx->pLineCtx[channelId]->pLineObj;
                if(plineObj->calCompleteEventMask == TRUE){
                    /* Event is masked */
                    return FALSE;
                }
            }
        } else if(eventId == VP_LINE_EVID_LINE_INIT_CMP) {
            /* We need to set the line state to disconnect after initLine to
             * patch a bug in the vcp firmware 1.98
             */

            /* Set the line state to Disconnect */
            retval = VcpSetLineState(pDevCtx->pLineCtx[channelId], VP_LINE_DISCONNECT);

            if(retval != VP_STATUS_SUCCESS){
                /* No Point in continuing... Exit */
                pEvent->status = retval;
                return FALSE; /* No device init complete event */
            }

            /* Setup default options for this line */
            retval = SetVcpDefaultOptions(VP_NULL,
                                            pDevCtx->pLineCtx[channelId]);
            if(retval != VP_STATUS_SUCCESS){
                /* No Point in continuing... Exit */
                pEvent->status = retval;
                return FALSE; /* No device init complete event */
            }
        } else if(eventId == VP_DEV_EVID_IO_ACCESS_CMP) {
            /* Indicates the type of access (Read/Write) */
            pEvent->eventData = pEvent->parmHandle;
            pEvent->parmHandle = 0;
        }
    }

    /* Actions if there is a response in the response mail box */
    if(responsePresent == TRUE){
        if(mbEventId == (eventWord >> 8)) {
            /* There is a response corresponding to the event */
            pEvent->hasResults = TRUE;

            /* Set the event Id data member to a value corrsponding to option
             * Id if the response mail box contains option read back */
            if(readOptionResponse == TRUE) {
                if(readOptionResponseOptId ==
                    VP_OPTION_ID_RESERVED_EVENT_MASK_VCP) {
                    /* Outside world does not know this option */
                    readOptionResponseOptId = VP_OPTION_ID_EVENT_MASK;
                }
                pEvent->eventData = readOptionResponseOptId;
            }
        }
    }

    if ((pEvent->eventCategory == VP_EVCAT_SIGNALING) &&
         ((pEvent->eventId == VP_LINE_EVID_HOOK_OFF) ||
          (pEvent->eventId == VP_LINE_EVID_HOOK_ON) ||
          (pEvent->eventId == VP_LINE_EVID_STARTPULSE) ||
          (pEvent->eventId == VP_LINE_EVID_FLASH) ||
          (pEvent->eventId == VP_LINE_EVID_PULSE_DIG))) {

        VpLineStateType currentState;

        if (VcpGetLineState(pEvent->pLineCtx, &currentState)
            == VP_STATUS_SUCCESS) {
            if (currentState == VP_LINE_TIP_OPEN) {
                return FALSE;
            }
        }
    }

    return TRUE;
} /* VcpGetEvent() */

/**
 * VcpGetLineStatus()
 *  Returns the status for any given input for the given line.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpGetLineStatus(
    VpLineCtxType *pLineCtx,
    VpInputType input,
    bool *pStatus
    )
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;

    uint32 deviceStatus;
    uint8 line = pLineObj->channelId;
    VpStatusType funcStatus;

    funcStatus = VcpGetDeviceStatus(pLineCtx->pDevCtx, input, &deviceStatus);
    if(funcStatus != VP_STATUS_SUCCESS) {
        return funcStatus;
    }

    if(deviceStatus & ((uint32)1<<line)){
        *pStatus = TRUE;
    } else {
        *pStatus = FALSE;
    }

    return VP_STATUS_SUCCESS;
} /* VcpGetLineStatus() */

/**
 * VcpGetDeviceStatus()
 *  Returns the status for all lines, multiplexed into a 32 bit return value.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpGetDeviceStatus(
    VpDevCtxType *pDevCtx,
    VpInputType input,
    uint32 *pDeviceStatus
    )
{
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint16 data[2];

    if (input >= VCP_NUM_INPUT_TYPES) {
        /* 'input' out of range, just return 0. */
        *pDeviceStatus = 0;
        return VP_STATUS_INVALID_ARG;
    }
    /* Basic error checking */
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    /* Map VP-API input type to that of VCP  */
    if(input == VP_INPUT_BAT1_FLT) {
        input = VCP_INPUT_LBAT_FLT;
    } else if(input == VP_INPUT_BAT3_FLT){
        input = VCP_INPUT_PBAT_FLT;
    }

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    if(VpHbiRd(deviceId, VP_REG_STAT_INPUTS(input), data) == FALSE){
        data[0] = data[1] = 0;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    *pDeviceStatus = (((uint32)data[0] << 16) | data[1]);

    return VP_STATUS_SUCCESS;
} /* VcpGetDeviceStatus() */

/**
 * VcpGetLoopCond()
 *  Returns the current loop conditions for the specified line.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpGetLoopCond(
    VpLineCtxType *pLineCtx,
    uint16 handle
    )
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint8 argIndex = VP_MB_ARGLIST_INDEX;
    VpStatusType retval;
    uint8 line = pLineObj->channelId;

    /* Basic error checking */
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
        retval = VP_STATUS_MAILBOX_BUSY;

    } else {
        /* Add the "get loop conditions" command to the mailbox */
        MailboxAddCommand(deviceId, VCP_MBFUNC_GETLOOPCOND, handle);

        /* Add the line number to the mailbox */
        argIndex = MailboxAddArg(deviceId, (uint16)line, argIndex);

        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);

        retval = VP_STATUS_SUCCESS;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpGetLoopCond() */


/**
 * VcpGetOption()
 *  Initiates a VCP Option read command.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpGetOption(
    VpLineCtxType *pLineCtx,
    VpDevCtxType *pDevCtx,
    VpOptionIdType option,
    uint16 handle
    )
{
    VpVcpLineObjectType *pLineObj;
    VpDevCtxType *pDevCtxLocal;
    VpVcpDeviceObjectType *pDevObj;

    uint8 argIndex = VP_MB_ARGLIST_INDEX;
    VpStatusType retval;
    uint8 line = 0;
    VpDeviceIdType deviceId;
    VpOptionIdType optionMax;

    /* Basic error checking */
    optionMax = VP_OPTION_ID_PCM_TXRX_CNTRL;
    if((option < 0) ||
       (option > optionMax) ||
       (option == VP_OPTION_ID_RESERVED_1) ||
       (option == VP_OPTION_ID_RESERVED_2)) {
        return VP_STATUS_INVALID_ARG;
    }

    /* Get the basic pointers initialized */
    if(pLineCtx != VP_NULL) {
        pLineObj = pLineCtx->pLineObj;
        line = pLineObj->channelId;
        pDevCtxLocal = pLineCtx->pDevCtx;
        pDevObj = pDevCtxLocal->pDevObj;
        deviceId = pDevObj->deviceId;
    } else {
        /* Device context pointer must be valid; because of the error checking
         * in VpGetOption() */
        pDevObj = pDevCtx->pDevObj;
        deviceId = pDevObj->deviceId;

        /* Make sure a line specific option is not being requested */
        if((option != VP_DEVICE_OPTION_ID_PULSE) &&
           (option != VP_DEVICE_OPTION_ID_CRITICAL_FLT) &&
           (option != VP_DEVICE_OPTION_ID_RAMP2STBY) &&
           (option != VP_DEVICE_OPTION_ID_DEVICE_IO)) {
            /* Must be one of the line specific options */
            return VP_STATUS_INVALID_ARG;
        }
    }

    if(pDevObj->productType == VP_VCP_PRODUCT_VE_880) {
        if(option == VP_OPTION_ID_EVENT_MASK) {
            /* VCP now has a better event mask command; The previous one did not
             * mask the device event masks if the line were to any line other
             * than zero. The following option does not have such restrictions*/
            option = VP_OPTION_ID_RESERVED_EVENT_MASK_VCP;
        }
    }

    /* DevNotes: Check if a given option is device specific or line specific
     * and a corresponding pointer is passed */
    /* Basic error checking */
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED; /* Because options are going to
        * be overwritten upon calling initialization and hence any read would
        * not be of any meaning */
    }

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
        retval = VP_STATUS_MAILBOX_BUSY;

    } else {
        /* Add the Get Option command to the mailbox, with pValue as
         * the context */
        MailboxAddCommand(deviceId, VCP_MBFUNC_GETOPTION, handle);

        /* Add the line number and option arguments to the mailbox */
        argIndex = MailboxAddArg(deviceId, (uint16)line,   argIndex);
        argIndex = MailboxAddArg(deviceId, (uint16)option, argIndex);

        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);

        retval = VP_STATUS_SUCCESS;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpGetOption() */

/**
 * VcpGetLineState()
 * This function is used to retrieve the line state of a given line.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpGetLineState(
    VpLineCtxType *pLineCtx,
    VpLineStateType *pCurrentState)
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint8 line = pLineObj->channelId;
    uint16 lineState;

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */
    VpHbiRd(deviceId, VP_REG_LINE_STATE(line), &lineState);
    /* One register contains line state for two lines */
    lineState = ((line & 1) == 0) ? (lineState >> 8) : (lineState & 0xff);
    /* Convert from VCP state to API line state */
    *pCurrentState = lineState & (~VCP_LINE_VP_MASK);
    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return VP_STATUS_SUCCESS;
} /* VcpGetLineState() */


/**
 * VcpFlushEvents()
 *  Flushes all events in the VCP's event queue.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpFlushEvents(
    VpDevCtxType *pDevCtx)
{
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    VpStatusType retval;

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
        retval = VP_STATUS_MAILBOX_BUSY;

    } else {
        /* Add the InitRing command to the mailbox */
        MailboxAddCommand(deviceId, VCP_MBFUNC_FLUSH, 0x0000);

        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);

        retval = VP_STATUS_SUCCESS;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpFlushEvents() */

/**
 * VcpGetResults()
 *  Copies the data corresponding to the specified event from the response
 *  mailbox into the structure pointed to by 'pResults' and frees the mailbox.
 *
 * Refer to the VP API User's Guide for details about this function.
 */
static VpStatusType
VcpGetResults(
    VpEventType *pEvent,
    void *pResults)
{
    VpDevCtxType *pDevCtx = pEvent->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint16 flag;
    VpStatusType retval = VP_STATUS_SUCCESS;
    uint16 mbEventId;
    uint16 chnlId;
    uint16 responseOptionId;

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the mailbox contains valid data */
    VpHbiRd(deviceId, HW_Reg_MBOXFLAG, &flag);
    if ( (flag & VP_BIT_MBOXFLAG_RSP) > 0x00) {     /* If VCP 'owns' the mbox */
        retval = VP_STATUS_MAILBOX_EMPTY;
    } else {
        /* Host owns the response mailbox; Read the corresponding event id */
        VpHbiRd(deviceId, VP_REG_MB_UP_BUF(0, 1), &mbEventId);
    }

    if(pResults == VP_NULL){
        /* User wants to get an event corresponding the message on the response
         * mailbox. Provide that information */

        if(retval != VP_STATUS_SUCCESS) {
            /* No response present */
            pEvent->eventId = 0;
        } else {

            /* A response is present; Translate it to correponding event */

            if(ConvertVcpEventId2ApiEventId(mbEventId<<8,
                &(pEvent->eventCategory), &(pEvent->eventId)) == TRUE){

                /* Read the line number and fill the other elements of the
                 * event */
                VpHbiRd(deviceId, VP_REG_MB_UP_BUF(1, 1), &chnlId);
                if(chnlId < pDevObj->maxChannels) {
                    pEvent->channelId = chnlId;
                    pEvent->pLineCtx = pDevCtx->pLineCtx[chnlId];
                } else {
                    pEvent->channelId = 0;
                    pEvent->pLineCtx = VP_NULL;
                }

                /* Fill the other elements of the event */
                pEvent->status = VP_STATUS_SUCCESS;
                pEvent->deviceId = deviceId;
                if(mbEventId == VCP_EVID_RD_OPTION) {
                    /* If the response corresponds to read option, then we need
                     * to the option ID corresponding to the response */
                    VpHbiRd(deviceId, VP_REG_MB_UP_BUF(3, 1),
                                &responseOptionId);
                    if(responseOptionId ==VP_OPTION_ID_RESERVED_EVENT_MASK_VCP){
                        /* Outside world does not know this option */
                        responseOptionId = VP_OPTION_ID_EVENT_MASK;
                    }

                    pEvent->eventData = responseOptionId;
                } else {
                    pEvent->eventData = 0;
                }
                pEvent->parmHandle = 0;
                pEvent->hasResults = TRUE;
            } else {
                /* Failed to map the event */
                pEvent->eventId = 0;
                retval = VP_STATUS_FAILURE;
            }
        }

    } else if(retval == VP_STATUS_SUCCESS) { /* Reading the response */

        /* Verify that the message in the mailbox matches the event type */
        if((VcpEventIdType)mbEventId != ConvertApiEventId2VcpEventId(pEvent) ) {
            retval = VP_STATUS_ERR_MAILBOX_DATA;
        }

        if (retval == VP_STATUS_SUCCESS) {

            retval = VP_STATUS_INVALID_ARG; /* This will be changed if the
                                             * event is handled */

            if(pEvent->eventCategory == VP_EVCAT_RESPONSE) {
                switch (pEvent->eventId) {
                    case VP_DEV_EVID_BOOT_CMP:
                        retval = GetCheckSumMessage(deviceId,
                                    (VpChkSumType*)pResults);
                        break;
                    case VP_LINE_EVID_LLCMD_RX_CMP:
                        retval = GetMpiRxMessage(deviceId, (uint8p)pResults);
                        break;
                    case VP_LINE_EVID_RD_LOOP:
                        retval = GetRdLoopCondMessage(pDevCtx,
                                    (VpLoopCondResultsType *)pResults);
                        break;
                    case VP_LINE_EVID_GAIN_CMP:
                        retval = GetGainCmpMessage(deviceId,
                                    (VpRelGainResultsType *)pResults);
                        break;
                    case VP_LINE_EVID_RD_OPTION:
                        retval = GetRdOptionMessage(pDevCtx,
                                    (VcpOptionDataType *)pResults);
                        break;
                    case VP_DEV_EVID_IO_ACCESS_CMP:
                        retval = GetIoAccessResponse(deviceId, pResults);
                        break;
                    default:
                        break;
                }
            } else if(pEvent->eventCategory == VP_EVCAT_TEST) {
                switch (pEvent->eventId) {
                    case VP_LINE_EVID_TEST_CMP:
                        retval = GetTestCmpMessage(deviceId,
                                    (VpTestResultType *) pResults);
                        break;
                    case VP_DEV_EVID_CHKSUM:
                        retval = GetCheckSumMessage(deviceId,
                                    (VpChkSumType *)pResults);
                        break;
                    default:
                        break;
                }
            }
        }

        if (retval == VP_STATUS_SUCCESS) {
            /* Return response-mailbox ownership back to the VCP. */
            flag = VP_BIT_MBOXFLAG_RSP;     /* Write '1' to set */
            VpHbiWr(deviceId, HW_Reg_MBOXFLAG, &flag);
        }
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpGetResults() */

/**
 * VcpClearResults()
 *  Clears outstanding results associated with an event (only one result
 * is discarded).
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpClearResults(
    VpDevCtxType *pDevCtx)
{
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 flag;

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the mailbox contains valid data */
    VpHbiRd(deviceId, HW_Reg_MBOXFLAG, &flag);

    if ((flag & VP_BIT_MBOXFLAG_RSP) == 0) {
        /* Host owns the mailox; Release the mailbox */

        flag = VP_BIT_MBOXFLAG_RSP;     /* Write '1' to set */
        VpHbiWr(deviceId, HW_Reg_MBOXFLAG, &flag);

    } /* else nothing to do */

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return VP_STATUS_SUCCESS;
} /* VcpClearResults() */

/**
 * VcpTestLine()
 *  Initiates a line test.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
/*
 * NOTE: The following constant is used to allocate space on the stack for a
 * local copy of the outgoing test command message. It should be tuned to equal
 * the length of the longest test mailbox message.
 */
#define MAX_TEST_CMD_LEN 20

static VpStatusType
VcpTestLine(
    VpLineCtxType *pLineCtx,
    VpTestIdType test,
    const void *pArgsUntyped,
    uint16 handle
    )
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint16 msgLen;
    uint16 msgBuf[MAX_TEST_CMD_LEN];
    VpStatusType retval;
    const VcpTestArgType *pArgs = pArgsUntyped; /* to comply with generic interface
                                           * without any compile warnings */
    uint8 line = pLineObj->channelId;

    /* Basic error checking */
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED; /* Because the device profile is
                                               * not specified. */
    }

    retval = VP_STATUS_SUCCESS;
    msgLen = 0;

    /* Add the Test Line command to the mailbox */
    msgBuf[msgLen++] = (uint16)VCP_MBFUNC_TESTLINE;
    msgBuf[msgLen++] = handle;

    /* Add the line number and test arguments to the mailbox */
    msgBuf[msgLen++] = (uint16)line;
    msgBuf[msgLen++] = (uint16)test;

    /* Add the test arguments to the mailbox */
    switch (test) {
        case VP_TEST_ID_PREPARE:
            /* No parameters for this test type. */
            break;

        case VP_TEST_ID_CONCLUDE:
            /* Error check test parameters. */
            if ((pArgs->tconclude.abortTest != TRUE) &&
                (pArgs->tconclude.abortTest != FALSE)) {

                retval = VP_STATUS_INVALID_ARG;
            }

            if (retval == VP_STATUS_SUCCESS) {
                msgBuf[msgLen++] = (uint16)pArgs->tconclude.abortTest;
            }
            break;

        case VP_TEST_ID_TONE_GEN:
            /* No error checking of these parameters required. */
            msgBuf[msgLen++] = pArgs->toneGen.duration;
            msgBuf[msgLen++] = pArgs->toneGen.freq1;
            msgBuf[msgLen++] = (uint16)pArgs->toneGen.amp1;
            msgBuf[msgLen++] = pArgs->toneGen.freq2;
            msgBuf[msgLen++] = (uint16)pArgs->toneGen.amp2;
            msgBuf[msgLen++] = pArgs->toneGen.freq3;
            msgBuf[msgLen++] = (uint16)pArgs->toneGen.amp3;
            msgBuf[msgLen++] = pArgs->toneGen.freq4;
            msgBuf[msgLen++] = (uint16)pArgs->toneGen.amp4;
            break;

        case VP_TEST_ID_ARB_1TONE:
            /* Error check test parameters. */
            if (FftSizeIsValid(pArgs->arb1Tone.fftSize) == FALSE) {
                retval = VP_STATUS_INVALID_ARG;
            }

            if ((pArgs->arb1Tone.useBuffer != TRUE) &&
                (pArgs->arb1Tone.useBuffer != FALSE)) {

                retval = VP_STATUS_INVALID_ARG;
            }

            if (retval == VP_STATUS_SUCCESS) {
                /* Only issue command if args are valid. */
                msgBuf[msgLen++] = pArgs->arb1Tone.timeout;
                msgBuf[msgLen++] = pArgs->arb1Tone.fftSize;
                msgBuf[msgLen++] = pArgs->arb1Tone.threshold;
                msgBuf[msgLen++] = (uint16)pArgs->arb1Tone.useBuffer;
            }
            break;

        case VP_TEST_ID_KEYPAD:
            /* Error check test parameters. */
            if (pArgs->keypad.kTest >= VP_NUM_KEY_TEST_TYPES) {
                retval = VP_STATUS_INVALID_ARG;
            }

            if (FftSizeIsValid(pArgs->keypad.fftSize) == FALSE) {
                retval = VP_STATUS_INVALID_ARG;
            }
            if ( (pArgs->keypad.kTest == VP_KEY_TEST_PULSE) ||
                 (pArgs->keypad.kTest == VP_KEY_TEST_BOTH) ) {
                if (pArgs->keypad.pulseOpt.makeMin >
                    pArgs->keypad.pulseOpt.makeMax) {

                    retval = VP_STATUS_INVALID_ARG;
                }

                if (pArgs->keypad.pulseOpt.breakMin >
                    pArgs->keypad.pulseOpt.breakMax) {

                    retval = VP_STATUS_INVALID_ARG;
                }

                if (pArgs->keypad.pulseOpt.flashMin >
                    pArgs->keypad.pulseOpt.flashMax) {

                    retval = VP_STATUS_INVALID_ARG;
                }
            }

            if ((pArgs->keypad.useBuffer != TRUE) &&
                (pArgs->keypad.useBuffer != FALSE)) {

                retval = VP_STATUS_INVALID_ARG;
            }

            if (retval == VP_STATUS_SUCCESS) {
                /* Only issue command if args are valid. */
                msgBuf[msgLen++] = (uint16)pArgs->keypad.kTest;
                msgBuf[msgLen++] = pArgs->keypad.timeout;
                msgBuf[msgLen++] = pArgs->keypad.freq;
                msgBuf[msgLen++] = (uint16)pArgs->keypad.amp;
                msgBuf[msgLen++] = pArgs->keypad.fftSize;
                msgBuf[msgLen++] = pArgs->keypad.threshold;
                msgBuf[msgLen++] = pArgs->keypad.pulseOpt.makeMin;
                msgBuf[msgLen++] = pArgs->keypad.pulseOpt.makeMax;
                msgBuf[msgLen++] = pArgs->keypad.pulseOpt.breakMin;
                msgBuf[msgLen++] = pArgs->keypad.pulseOpt.breakMax;
                msgBuf[msgLen++] = pArgs->keypad.pulseOpt.interDigitMin;
                msgBuf[msgLen++] = pArgs->keypad.pulseOpt.flashMin;
                msgBuf[msgLen++] = pArgs->keypad.pulseOpt.flashMax;
                msgBuf[msgLen++] = pArgs->keypad.pulseOpt.flashMax + 1; /* API
                * does not have "onHookMin" field anymore */
                msgBuf[msgLen++] = (uint16)pArgs->keypad.useBuffer;
            }
            break;

        case VP_TEST_ID_OPEN_VDC:
        case VP_TEST_ID_OPEN_VAC:
            /* Error check test parameters. */
            if ((pArgs->openV.tip != VP_TEST_TIP) &&
                (pArgs->openV.tip != VP_TEST_RING)) {

                retval = VP_STATUS_INVALID_ARG;
            }

            if (retval == VP_STATUS_SUCCESS) {
                /*
                 * NOTE: Calibration mode is ignored for open V AC test.
                 * However, we pass 'calMode' to VCP for both AC and DC because
                 * VCP assumes same message format for both test types.
                 */
                msgBuf[msgLen++] = (uint16)pArgs->openV.calMode;
                msgBuf[msgLen++] = pArgs->openV.integrateTime;
                msgBuf[msgLen++] = pArgs->openV.settlingTime;
                msgBuf[msgLen++] = (uint16)pArgs->openV.tip;
            }
            break;

        case VP_TEST_ID_DIFF_VAC:
            /* No error checking of these parameters required. */
            msgBuf[msgLen++] = pArgs->diffVac.integrateTime;
            msgBuf[msgLen++] = (uint16)pArgs->diffVac.gvsaCal;
            msgBuf[msgLen++] = (uint16)pArgs->diffVac.gvsbCal;
            msgBuf[msgLen++] = pArgs->diffVac.settlingTime;
            break;

        case VP_TEST_ID_NOISE:
            /* Error check test parameters. */
            if ((pArgs->noise.filter < 0) ||
                (pArgs->noise.filter >= VP_NUM_TEST_FILTER_TYPES)) {

                retval = VP_STATUS_INVALID_ARG;
            }

            if ((pArgs->noise.useBuffer != TRUE) &&
                (pArgs->noise.useBuffer != FALSE)) {

                retval = VP_STATUS_INVALID_ARG;
            }

            if (retval == VP_STATUS_SUCCESS) {
                msgBuf[msgLen++] = pArgs->noise.integrateTime;
                msgBuf[msgLen++] = (uint16)pArgs->noise.filter;
                msgBuf[msgLen++] = (uint16)pArgs->noise.amp;
                msgBuf[msgLen++] = (uint16)pArgs->noise.useBuffer;
            }
            break;

        case VP_TEST_ID_DIALTONE:
            /* Error check test parameters. */
            if ((pArgs->dialTone.inward != TRUE) &&
                (pArgs->dialTone.inward != FALSE)) {

                retval = VP_STATUS_INVALID_ARG;
            }

            if (FftSizeIsValid(pArgs->dialTone.fftSize) == FALSE) {
                retval = VP_STATUS_INVALID_ARG;
            }

            if ((pArgs->dialTone.useBuffer != TRUE) &&
                (pArgs->dialTone.useBuffer != FALSE)) {

                retval = VP_STATUS_INVALID_ARG;
            }

            if (retval == VP_STATUS_SUCCESS) {
                msgBuf[msgLen++] = (uint16)pArgs->dialTone.inward;
                msgBuf[msgLen++] = pArgs->dialTone.timeout;
                msgBuf[msgLen++] = pArgs->dialTone.detectGran;
                msgBuf[msgLen++] = pArgs->dialTone.threshold;
                msgBuf[msgLen++] = pArgs->dialTone.fftSize;
                msgBuf[msgLen++] = pArgs->dialTone.duration;
                msgBuf[msgLen++] = (uint16)pArgs->dialTone.useBuffer;
            }
            break;

        case VP_TEST_ID_SNR_QDIST:
            /* Error check test parameters. */
            if ((pArgs->snrqd.useBuffer != TRUE) &&
                (pArgs->snrqd.useBuffer != FALSE)) {

                retval = VP_STATUS_INVALID_ARG;
            }

            if (retval == VP_STATUS_SUCCESS) {
                msgBuf[msgLen++] = pArgs->snrqd.integrateTime;
                msgBuf[msgLen++] = (uint16)pArgs->snrqd.amp;
                msgBuf[msgLen++] = (uint16)pArgs->snrqd.useBuffer;
            }
            break;

        case VP_TEST_ID_HYBRID_LOSS:
            /* Error check test parameters. */
            if ((pArgs->hybridLoss.useBuffer != TRUE) &&
                (pArgs->hybridLoss.useBuffer != FALSE)) {

                retval = VP_STATUS_INVALID_ARG;
            }

            if (retval == VP_STATUS_SUCCESS) {
                msgBuf[msgLen++] = pArgs->hybridLoss.integrateTime;
                msgBuf[msgLen++] = pArgs->hybridLoss.freq;
                msgBuf[msgLen++] = (uint16)pArgs->hybridLoss.amp;
                msgBuf[msgLen++] = (uint16)pArgs->hybridLoss.useBuffer;
            }
            break;

        case VP_TEST_ID_3ELE_RES:
            /* Error check test parameters. */
            if ((pArgs->res.calMode != TRUE) &&
                (pArgs->res.calMode != FALSE)) {

                retval = VP_STATUS_INVALID_ARG;
            }

            if (retval == VP_STATUS_SUCCESS) {
                msgBuf[msgLen++] = (uint16)pArgs->res.calMode;
                msgBuf[msgLen++] = (uint16)pArgs->res.vCm;
                msgBuf[msgLen++] = (uint16)pArgs->res.vDiff;
                msgBuf[msgLen++] = pArgs->res.integrateTime;
                msgBuf[msgLen++] = pArgs->res.settlingTime;
            }
            break;

        case VP_TEST_ID_3ELE_CAP:
            /* Error check test parameters. */
            if ((pArgs->cap.calMode != TRUE) &&
                (pArgs->cap.calMode != FALSE)) {

                retval = VP_STATUS_INVALID_ARG;
            }
            if (pArgs->cap.freq > 4000) {
                retval = VP_STATUS_INVALID_ARG;
            }

            if (retval == VP_STATUS_SUCCESS) {
                msgBuf[msgLen++] = (uint16)pArgs->cap.calMode;
                msgBuf[msgLen++] = (uint16)pArgs->cap.vCm;
                msgBuf[msgLen++] = (uint16)pArgs->cap.vDiff;
                msgBuf[msgLen++] = pArgs->cap.freq;
                msgBuf[msgLen++] = (uint16)pArgs->cap.gimtCal;
                msgBuf[msgLen++] = (uint16)pArgs->cap.gilgCal;
                msgBuf[msgLen++] = pArgs->cap.integrateTime;
                msgBuf[msgLen++] = pArgs->cap.settlingTime;
            }
            break;

        case VP_TEST_ID_SET_SENSE_GAIN:
            /* Error check test parameters. */
            if (
                ((pArgs->senseGain.imt != VP_SENSE_GAIN_HIGH) &&
                 (pArgs->senseGain.imt != VP_SENSE_GAIN_LOW)) ||
                ((pArgs->senseGain.ilg != VP_SENSE_GAIN_HIGH) &&
                 (pArgs->senseGain.ilg != VP_SENSE_GAIN_LOW))
            ) {
                retval = VP_STATUS_INVALID_ARG;
            }

            if (retval == VP_STATUS_SUCCESS) {
                msgBuf[msgLen++] = (uint16)pArgs->senseGain.imt;
                msgBuf[msgLen++] = (uint16)pArgs->senseGain.ilg;
            }
            break;

        case VP_TEST_ID_DC_RLOOP:
            /* Error check test parameters. */
            if ((pArgs->dcRLoop.calMode != TRUE) &&
                (pArgs->dcRLoop.calMode != FALSE)) {
                retval = VP_STATUS_INVALID_ARG;
            }

            if (retval == VP_STATUS_SUCCESS) {
                msgBuf[msgLen++] = (uint16)pArgs->dcRLoop.calMode;
                msgBuf[msgLen++] = (uint16)pArgs->dcRLoop.iTestLevel;
                msgBuf[msgLen++] = pArgs->dcRLoop.integrateTime;
                msgBuf[msgLen++] = pArgs->dcRLoop.settlingTime;
            }
            break;

        case VP_TEST_ID_AC_RLOOP:
        case VP_TEST_ID_AC_RLOOP_PHASE:
            /* Error check test parameters. */
            if ((pArgs->acRLoop.calMode != TRUE) &&
                (pArgs->acRLoop.calMode != FALSE)) {
                retval = VP_STATUS_INVALID_ARG;
            }
            if (pArgs->acRLoop.freq > 4000) {
                retval = VP_STATUS_INVALID_ARG;
            }

            if (retval == VP_STATUS_SUCCESS) {
                msgBuf[msgLen++] = (uint16)pArgs->acRLoop.calMode;
                msgBuf[msgLen++] = pArgs->acRLoop.freq;
                msgBuf[msgLen++] = (uint16)pArgs->acRLoop.vTestLevel;
                msgBuf[msgLen++] = (uint16)pArgs->acRLoop.vBias;
                msgBuf[msgLen++] = pArgs->acRLoop.integrateTime;
                msgBuf[msgLen++] = pArgs->acRLoop.settlingTime;
            }
            break;

        case VP_TEST_ID_NULL_IGND:
            /* Error check test parameters. */
            if ((pArgs->nullIGnd.tip != VP_TEST_TIP) &&
                (pArgs->nullIGnd.tip != VP_TEST_RING)) {
                retval = VP_STATUS_INVALID_ARG;
            }

            if (retval == VP_STATUS_SUCCESS) {
                msgBuf[msgLen++] = (uint16)pArgs->nullIGnd.vTestLevel;
                msgBuf[msgLen++] = (uint16)pArgs->nullIGnd.iTarget;
                msgBuf[msgLen++] = (uint16)pArgs->nullIGnd.tip;
                msgBuf[msgLen++] = pArgs->nullIGnd.settlingTime;
            }
            break;

        case VP_TEST_ID_IDC:
            /* Error check test parameters. */
            if ((pArgs->idc.calMode != TRUE) &&
                (pArgs->idc.calMode != FALSE)) {
                retval = VP_STATUS_INVALID_ARG;
            }

            if (retval == VP_STATUS_SUCCESS) {
                msgBuf[msgLen++] = (uint16)pArgs->idc.calMode;
                msgBuf[msgLen++] = (uint16)pArgs->idc.vCm;
                msgBuf[msgLen++] = pArgs->idc.integrateTime;
                msgBuf[msgLen++] = pArgs->idc.settlingTime;
            }
            break;

        case VP_TEST_ID_IAC:
            /* No error checking of these parameters required. */
            msgBuf[msgLen++] = (uint16)pArgs->iac.gimtCal;
            msgBuf[msgLen++] = (uint16)pArgs->iac.gilgCal;
            msgBuf[msgLen++] = (uint16)pArgs->iac.vCm;
            msgBuf[msgLen++] = pArgs->iac.integrateTime;
            msgBuf[msgLen++] = pArgs->iac.settlingTime;
            break;

        case VP_TEST_ID_ALT_REN:
            /* No error checking of these parameters required. */
            msgBuf[msgLen++] = pArgs->ren.integrateTime;
            msgBuf[msgLen++] = (uint16)pArgs->ren.vDiff;
            break;

        case VP_TEST_ID_HOOK_STATUS:
            /* No parameters for this test type. */
            break;

        default:
            if ((test >= VP_TEST_ID_RSVD_MIN) &&
                (test <= VP_TEST_ID_RSVD_MAX)) {

                /* Handle "reserved" test IDs in a generic fashion:
                 * Copy words from '(uint16p)pArgs' into downstream
                 * mailbox.  Number of arguments is contained in the
                 * first element of the '(uint16p)pArgs' array. */
                uint8 i, numArgs = ((uint16p)pArgs)[0];
                for (i = 0; i < numArgs; i++) {
                    msgBuf[msgLen++] = ((uint16p)pArgs)[i + 1];
                }
            } else {
                return VP_STATUS_INVALID_ARG;
            }
    }

    /* Check for buffer overflow. */
    if (msgLen > MAX_TEST_CMD_LEN) {
        /* Indicates a software bug. Increase the buffer size. */
        retval = VP_STATUS_ERR_HBI;
    }

    /*
     * If we made it through all of that without detecting an error condition,
     * go ahead and write the command to the mailbox.
     */
    if (retval == VP_STATUS_SUCCESS) {
        VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL
                                                            * SECTION => */

        /* Ensure that the Mailbox is available for the host */
        if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
            retval = VP_STATUS_MAILBOX_BUSY;
        } else {
            /* Write the message to the mailbox. */
            VpHbiWr(deviceId, VP_REG_MB_DOWN_BUF(0, msgLen), msgBuf);

            /* Turn the mailbox over to the VCP (raise the flag) */
            MailboxRelease(deviceId);
        }

        VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL
                                                           * SECTION <= */
    }

    return retval;
} /* VcpTestLine() */

/**
 * VcpCodeCheckSum()
 *  Initiates a checksum calculation over the VCP's code space.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpCodeCheckSum(
    VpDevCtxType *pDevCtx,
    uint16 handle
    )
{
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    VpStatusType retval;

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
        retval = VP_STATUS_MAILBOX_BUSY;

    } else {
        /* Add the Check Sum command to the mailbox */
        MailboxAddCommand(deviceId, VCP_MBFUNC_CHECKSUM, handle);

        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);

        retval = VP_STATUS_SUCCESS;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpCodeCheckSum() */

/**
 * VcpSelfTest()
 *  Initiates a self test.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpSelfTest(
    VpLineCtxType *pLineCtx)
{
    VpVcpDeviceObjectType *pDevObj;
    VpVcpLineObjectType *pLineObj;
    VpDeviceIdType deviceId;
    VpDevCtxType *pDevCtxLocal;
    uint8 argIndex = VP_MB_ARGLIST_INDEX;
    VpStatusType retval;
    uint8 line;

    pDevCtxLocal = pLineCtx->pDevCtx;
    pDevObj = pDevCtxLocal->pDevObj;
    deviceId = pDevObj->deviceId;
    pLineObj = pLineCtx->pLineObj;
    line = pLineObj->channelId;

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
        retval = VP_STATUS_MAILBOX_BUSY;

    } else {
        /* Add the Self Test command to the mailbox */
        MailboxAddCommand(deviceId, VCP_MBFUNC_SELFTEST, 0x0000);

        /* Add the line number to the mailbox */
        argIndex = MailboxAddArg(deviceId, (uint16)line, argIndex);

        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);

        retval = VP_STATUS_SUCCESS;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpSelfTest() */

/**
 * VcpFillTestBuf()
 *  Loads test vector data into VCP's internal buffers for algorithm testing.
 *
 * Refer to the VoicePath API User's Guide for details about this function.
 */
static VpStatusType
VcpFillTestBuf(
    VpLineCtxType *pLineCtx,
    uint16 length,
    VpVectorPtrType pData
    )
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint16 testLine;
    uint8 testBuf;
    VpStatusType retval = VP_STATUS_INVALID_ARG;
    uint8 line = pLineObj->channelId;

    /* Bounds-checking on arguments: */
    if (length > VP_TEST_BUF_MAX_SAMPLES) {
        length = VP_TEST_BUF_MAX_SAMPLES;
    }

    if (length == 0) {
        /* Nothing to do. */
        return VP_STATUS_SUCCESS;
    }

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    for (testBuf = 0; testBuf < VP_NUM_TEST_BUFS; testBuf++) {

        /* See which line the test buffer is allocated to, if any. */
        if(VpHbiRd(deviceId, VP_REG_TESTBUF_LINE(testBuf), &testLine) == FALSE){
            retval = VP_STATUS_ERR_HBI;
            break;
        }

        if (testLine == line) {
            /* TEST_PREPARE has been previously executed on the specified
             * line, and the current test buffer was allocated. */

            /* Reset the test cursor to 0. */
            if (VpHbiWr(deviceId, VP_REG_TESTBUF_CUR(testBuf), 0) == FALSE) {
                retval = VP_STATUS_ERR_HBI;
                break;
            }

            if (length > 0) {
                /* Write the data at pData into the test buffer starting at
                 * the appropriate AMBA address. */
                if (VpHbiAmbaWrite(deviceId, VP_TEST_BUF_AMBA(testBuf),
                    length - 1, pData) == FALSE) {
                    retval = VP_STATUS_ERR_HBI;
                    break;
                }
            }

            if (length < VP_TEST_BUF_MAX_SAMPLES) {
                /* Fill the remainder of the specified test buffer with
                 * zeroes. */
                if (VpHbiAmbaWrite(deviceId, VP_TEST_BUF_AMBA(testBuf) + length,
                    (VP_TEST_BUF_MAX_SAMPLES - length) - 1, 0) == FALSE) {
                    retval = VP_STATUS_ERR_HBI;
                    break;
                }
            }

            retval = VP_STATUS_SUCCESS;
            break;
        }

        /* If for loop completes, then retval defaults to
         * VP_STATUS_INVALID_ARG. */
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* VcpFillTestBuf() */

/******************************************************************************
 *                   VCP Specific API Support Functions                       *
 ******************************************************************************/
/* This function initializes device object members */
static void InitDeviceObjMembers(
    VpVcpDeviceObjectType *pDevObj)
{
    /* Initialize device Object */
    if (VP_MAX_LINES_PER_DEVICE > VP_VCP_MAX_NUM_LINES) {
        pDevObj->maxChannels = VP_VCP_MAX_NUM_LINES;
    } else {
        pDevObj->maxChannels = VP_MAX_LINES_PER_DEVICE;
    }

    pDevObj->slacsFoundDuringCal = 0;
    pDevObj->deviceInitialized = FALSE;

    pDevObj->devicePrfTblSts = 0;
    pDevObj->acPrfTblSts = 0;
    pDevObj->dcPrfTblSts = 0;
    pDevObj->ringPrfTblSts = 0;
    pDevObj->ringCadPrfTblSts = 0;
    pDevObj->tonePrfTblSts = 0;
    pDevObj->toneCadPrfTblSts = 0;
    pDevObj->meterPrfTblSts = 0;
    pDevObj->cidPrfTblSts = 0;

    pDevObj->timeStampCntrHB = 0;

    pDevObj->timeStampEventMask = FALSE;

    pDevObj->recCnt = 0;

#ifdef VP_CC_MULTI_THREAD_PROTECTION
    /* When called through MakeDeviceObject()
     * Note that resetting this flag is not protected within critical sections.
     * This is because there could nothing else be going on when device is being
     * initialized.
     * When called through SoftReset(), the wrapper function would make sure
     * nothing else is going on.
     */
    pDevObj->lineApiBusy = 0;
#endif /* VP_CC_MULTI_THREAD_PROTECTION */

    /* Initialize to one type; Since there will only be one device object
     * in all programming models (irrespective of the number of device contexts
     * created), this field will be overwritten once the boot load function
     * is completed. It is also expected the only programming model supported
     * is creating more device contexts after bootloading the part. This
     * requirememt arises because the function pointers are present in the
     * device context and not in device object. API-II allows more than one
     * device context. */
    pDevObj->productType = VP_VCP_PRODUCT_VE_790;

} /* InitDeviceObjMembers() */

/**
 * InitAndConfigLine()
 * Performs two functions 1)Calls Line initialization or 2)Configures line.
 * In both cases it loads new profiles.
 */
static VpStatusType
InitAndConfigLine(
    VpLineCtxType *pLineCtx,
    VpProfilePtrType pAcProfile,        /* Specifies AC Profile   */
    VpProfilePtrType pDcProfile,        /* Specifies DC Profile    */
    VpProfilePtrType pRingProfile,      /* Ring Profile Selector   */
    VcpMbFuncIdType mailBoxCmd)         /* VCP mailbox command */
{
    VpVcpLineObjectType *pLineObj = pLineCtx->pLineObj;
    VpDevCtxType *pDevCtx = pLineCtx->pDevCtx;
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    uint8 argIndex = VP_MB_ARGLIST_INDEX;
    VpStatusType retval;
    uint8 line = pLineObj->channelId;

    /* Profile pointer error checking */
    if((IsItAValidProfilePtr(pDevObj, VP_PROFILE_AC, pAcProfile) == FALSE) ||
       (IsItAValidProfilePtr(pDevObj, VP_PROFILE_DC, pDcProfile) == FALSE) ||
       (IsItAValidProfilePtr(pDevObj, VP_PROFILE_RING, pRingProfile) == FALSE)){

        return VP_STATUS_ERR_PROFILE;
    }
    /* Basic error checking */
    if(pDevObj->deviceInitialized == FALSE) {
        return VP_STATUS_DEV_NOT_INITIALIZED;
    }

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if (MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {
        retval = VP_STATUS_MAILBOX_BUSY;

    } else {
        retval = VP_STATUS_SUCCESS;

        /* Add the InitLine command to the mailbox */
        MailboxAddCommand(deviceId, mailBoxCmd, 0x0000);

        /* Add the line number argument to the mailbox */
        argIndex = MailboxAddArg(deviceId, (uint16)line, argIndex);

        /* Add the profile arguments to the mailbox */
        argIndex = MailboxAddProfile(deviceId, VP_PROFILE_AC, pAcProfile,
                                        argIndex);
        if (argIndex == 0) retval = VP_STATUS_INVALID_ARG;

        argIndex = MailboxAddProfile(deviceId, VP_PROFILE_DC, pDcProfile,
                                        argIndex);
        if (argIndex == 0) retval = VP_STATUS_INVALID_ARG;

        argIndex = MailboxAddProfile(deviceId, VP_PROFILE_RING, pRingProfile,
                                        argIndex);
        if (argIndex == 0) retval = VP_STATUS_INVALID_ARG;

        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* InitAndConfigLine() */

/**
 * GetChkSumMessage()
 *  This function copies data from the VCP's upstream mailbox into the structure
 *  pointed to by 'pResults'.
 *
 * Preconditions
 *  Assumes that data in the upstream mailbox is response to a BOOT_CMP or a
 *  CHKSUM_CMP event.
 *  Assumes buffer pointed to by 'pResults' is large enough to hold entire
 *  checksum structure.
 *
 * Postconditions
 *  Checksum and version info is copied from the upstream mailbox to caller's
 *  buffer.
 *  Returns VP_INVALID if an HBI read error occurs.
 *  Returns VP_STATUS_SUCCESS otherwise.
 */
static VpStatusType
GetCheckSumMessage(
    VpDeviceIdType deviceId,
    VpChkSumType* pResults)
{
    uint16 data;

    /* Get the 32bit code load checksum */
    VpHbiRd(deviceId, VP_REG_MB_UP_BUF(3, 1), &data);
    pResults->loadChecksum = (uint32)data << 16;
    VpHbiRd(deviceId, VP_REG_MB_UP_BUF(4, 1), &data);
    pResults->loadChecksum |= (uint32)data;

    /* Get the version info structure data */
    VpHbiRd(deviceId, VP_REG_MB_UP_BUF(5, 1), &pResults->vInfo.vtdRevCode);
    VpHbiRd(deviceId, VP_REG_MB_UP_BUF(6, 1), &data);
    pResults->vInfo.swProductId = (uint8)data;
    VpHbiRd(deviceId, VP_REG_MB_UP_BUF(7, 1), &data);
    pResults->vInfo.swVerMajor = (uint8)(data >> 8);
    pResults->vInfo.swVerMinor = (uint8)(data & 0x00FF);

    return VP_STATUS_SUCCESS;
} /* GetCheckSumMessage() */



/**
 * GetMpiRxMessage()
 *  This function copies data from the VCP's upstream mailbox into the structure
 *  pointed to by 'pResults'.
 * Preconditions
 *  Assumes that data in the upstream mailbox is response to a MPI read command.
 *  Assumes buffer pointed to by 'pResults' is large enough to hold entire MPI
 *  data string.
 * Postconditions
 *  MPI read command data is copied from upstream mailbox to caller's buffer.
 *  Returns VP_INVALID if an HBI read error occurs.
 *  Returns VP_STATUS_SUCCESS otherwise, even if length of MPI data is zero.
 */
static VpStatusType
GetMpiRxMessage(
    VpDeviceIdType deviceId,
    uint8p pResults
    )
{
    uint8 ctr;
    uint16 len;

    /* Get the length of the RX mailbox data. */
    VpHbiRd(deviceId, VP_REG_MB_UP_BUF(2, 1), &len);

    if (len != 0) {
        /*
         * Copy the data from the upstream mailbox into the client's
         * response buffer.
         */
        for (ctr = 0; ctr < len; ctr++) {
            uint16 data16;

            if (!VpHbiRd(deviceId, VP_REG_MB_UP_BUF(ctr+3, 1), &data16)) {
                return VP_STATUS_INVALID_ARG;
            }

            /*
             * WARNING: Only one byte the of 16-bit word read from HBI is
             * useful. This code assumes that the low-order byte contains the
             * MPI data.
             */
            *pResults++ = (uint8)data16;
        }
    }

    return VP_STATUS_SUCCESS;
} /* GetMpiRxMessage() */

/**
 * GetRdLoopCondMessage()
 *  This function copies data from the VCP's upstream mailbox into the structure
 *  pointed to by 'pResults'.
 * Preconditions
 *  Assumes that data in the upstream mailbox is response to a RD_LOOP event.
 *  Assumes buffer pointed to by 'pResults' is large enough to hold entire
 *  loop conditions structure.
 * Postconditions
 *  Loop conditions info is copied from the upstream mailbox to caller's
 *  buffer.
 *  Returns VP_INVALID if an HBI read error occurs.
 *  Returns VP_STATUS_SUCCESS otherwise.
 */
static VpStatusType
GetRdLoopCondMessage(
    VpDevCtxType *pDevCtx,
    VpLoopCondResultsType *pResults
    )
{
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;

    #define RD_LOOP_COND_MSG_LEN 10
    uint16 data[RD_LOOP_COND_MSG_LEN];
    int i, length = RD_LOOP_COND_MSG_LEN;

    if(pDevObj->productType != VP_VCP_PRODUCT_VE_790) {
        length = length - 2;
    }

    /* Get the loopconditions */
    VpHbiRd(deviceId, VP_REG_MB_UP_BUF(3, length), &data[0]);

    /* Fill the results */
    i = 0;
    pResults->rloop = data[i++];
    pResults->ilg   = data[i++];
    pResults->imt   = data[i++];
    pResults->vsab  = data[i++];
    pResults->vbat1 = data[i++];                        /* 5th element */

    pResults->vbat2 = data[i++];
    pResults->vbat3 = data[i++];
    pResults->mspl  = data[i++];
    if(pDevObj->productType == VP_VCP_PRODUCT_VE_790) {
        pResults->selectedBat  = (VpBatteryType)data[i++];
        pResults->dcFeedReg  = (VpDcFeedRegionType)data[i]; /* 10th element */
    } else {
        /* These fields are not supported for VE-880 */
        pResults->selectedBat  = VP_BATTERY_UNDEFINED;
        pResults->dcFeedReg  = VP_DF_UNDEFINED; /* 10th element */
    }

    return VP_STATUS_SUCCESS;
} /* GetRdLoopCondMessage() */

/**
 * GetGainCmpMessage()
 *  This function copies data from the VCP's upstream mailbox into the structure
 *  pointed to by 'pResults'.
 * Preconditions
 *  Assumes that data in the upstream mailbox is response to a GAIN_CMP event.
 *  Assumes buffer pointed to by 'pResults' is large enough to hold entire
 *  relative gain results structure.
 * Postconditions
 *  Gain results info is copied from the upstream mailbox to caller's buffer.
 *  Returns VP_INVALID if an HBI read error occurs.
 *  Returns VP_STATUS_SUCCESS otherwise.
 */
static VpStatusType
GetGainCmpMessage(
    VpDeviceIdType deviceId,
    VpRelGainResultsType *pResults
    )
{
    /* Get the gain results */
    uint16 dataRd;
    VpHbiRd(deviceId, VP_REG_MB_UP_BUF(3, 1), &dataRd);
    pResults->gResult = (VpGainResultType)dataRd;
    VpHbiRd(deviceId, VP_REG_MB_UP_BUF(4, 1), &pResults->gxValue);
    VpHbiRd(deviceId, VP_REG_MB_UP_BUF(5, 1), &pResults->grValue);

    return VP_STATUS_SUCCESS;
} /* GetGainCmpMessage() */

/**
 * GetRdOptionMessage()
 *  This function copies data from the VCP's upstream mailbox into the structure
 *  pointed to by 'pResults'.
 * Preconditions
 *  Assumes buffer pointed to by 'pResults' is large enough to hold entire an
 *  option result.
 * Postconditions
 *  Option data is copied from upstream mailbox to caller's buffer.
 *  Returns VP_INVALID if an HBI read error occurs.
 *  Returns VP_STATUS_ERR_MAILBOX_DATA is mailbox does not contain options
 *  results.
 *  Returns VP_STATUS_SUCCESS otherwise, even if length of MPI data is zero.
 */
static VpStatusType
GetRdOptionMessage(
    VpDevCtxType *pDevCtx,
    VcpOptionDataType *pResults
    )
{
    VpVcpDeviceObjectType *pDevObj = pDevCtx->pDevObj;
    VpDeviceIdType deviceId = pDevObj->deviceId;
    uint16 optionId;
    uint16 data;
    uint16 dtmfStatus[2];
    #define DEVICE_OPTION_DATA_SIZE     (8)
    uint16 readData[DEVICE_OPTION_DATA_SIZE];
    uint8 line;
    VpVcpLineObjectType *pLineObj;

    /* Get the option results, based on the option type */
    VpHbiRd(deviceId, VP_REG_MB_UP_BUF(3, 1), &optionId);
    switch (optionId) {
        case VP_DEVICE_OPTION_ID_PULSE:
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(4, 1), &pResults->pulse.makeMin);
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(5, 1), &pResults->pulse.makeMax);
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(6, 1),
                        &pResults->pulse.breakMin);
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(7, 1), &pResults->pulse.breakMax);
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(8, 1),
                        &pResults->pulse.interDigitMin);
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(9, 1),
                        &pResults->pulse.flashMin);
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(10, 1),
                        &pResults->pulse.flashMax);
            /* No need to read the "On-Hook-Min-Time" since API does not support
             * it anymore  */
            break;
        case VP_DEVICE_OPTION_ID_CRITICAL_FLT:
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(4,  1), &data);
            pResults->criticalFlt.acFltDiscEn = data & 0x01;
            pResults->criticalFlt.dcFltDiscEn = (data & 0x02)>>1;
            pResults->criticalFlt.thermFltDiscEn = (data & 0x08)>>3;
            break;
        case VP_OPTION_ID_PULSE_MODE:
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(4,  1), &data);
            pResults->pulseMode = (VpOptionPulseModeType)data;
            break;
        case VP_OPTION_ID_ZERO_CROSS:
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(4,  1), &data);
            pResults->zeroCross = (VpOptionZeroCrossType)data;
            break;
        case VP_DEVICE_OPTION_ID_RAMP2STBY:
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(4,  1), &pResults->rampTime);
            break;
        case VP_OPTION_ID_TIMESLOT:
            /* No need to read the control information (first word), since it is
             * always time slot setting and is not used by the API */
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(5,  1), &data);
            pResults->timeslot.tx = (uint8)data;
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(6,  1), &data);
            pResults->timeslot.rx = (uint8)data;
            break;
        case VP_OPTION_ID_CODEC:
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(4,  1), &data);
            pResults->codec = (VpOptionCodecType)data;
            break;
        case VP_OPTION_ID_PCM_HWY:
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(4,  1), &data);
            pResults->pcmHwy = (VpOptionPcmHwyType)data;
            break;
        case VP_OPTION_ID_LOOPBACK:
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(4,  1), &data);
            pResults->loopback = (VpOptionLoopbackType)data;
            break;
        case VP_OPTION_ID_LINE_STATE:
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(4,  1), &data);
            pResults->lineState.battRev = (bool)data;
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(5,  1), &data);
            pResults->lineState.bat = (VpOptionBatType)data;
            break;
        case VP_OPTION_ID_RESERVED_EVENT_MASK_VCP:
        case VP_OPTION_ID_EVENT_MASK:
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(1,  1), &data);
            line = data;
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(4,  1), &data);
            pResults->eventMask.faults = data;
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(5,  1), &data);
            if(pDevObj->timeStampEventMask == TRUE) {
                /* This event is handled only in the software */
                data = data | VP_DEV_EVID_TS_ROLLOVER;
            }
            pResults->eventMask.signaling = data;
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(6,  1), &data);
            if (line < pDevObj->maxChannels) {
                if(pDevCtx->pLineCtx[line] != VP_NULL) {
                    pLineObj = pDevCtx->pLineCtx[line]->pLineObj;
                    if(pLineObj->calCompleteEventMask == TRUE) {
                        /* This event is handled only in the software */
                        data = data |  VP_EVID_CAL_CMP;
                    }
                }
            }
            /* VCP does not understand device init complete event */
            data = data & (~VP_DEV_EVID_DEV_INIT_CMP);

            pResults->eventMask.response = data;
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(7,  1), &data);
            pResults->eventMask.test = data;
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(8,  1), &data);
            pResults->eventMask.process = data;
            pResults->eventMask.fxo = VP_EVENT_MASK_ALL;
            pResults->eventMask.packet = VP_EVENT_MASK_ALL;
            break;
        case VP_OPTION_ID_RING_CNTRL:
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(4,  1), &data);
            pResults->ringControl.zeroCross = data;
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(5,  1), &data);
            pResults->ringControl.ringExitDbncDur = data;
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(6,  1), &data);
            pResults->ringControl.ringTripExitSt = data & (~VCP_LINE_VP_MASK);
            break;
        case VP_OPTION_ID_DTMF_MODE:
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(4,  1), &data);
            pResults->dtmfControl.dtmfControlMode = data;
            /* Fill the other remaining elements from the direct page */
            VpHbiRd(deviceId, VP_REG_DTMF_REMAINING_RSRC, &data);
            pResults->dtmfControl.dtmfResourcesRemaining = data;
            VpHbiRd(deviceId, VP_REG_DTMF_ENABLE_STATUS, dtmfStatus);
            pResults->dtmfControl.dtmfDetectionSetting =
                ((uint32)dtmfStatus[0]<<16) + dtmfStatus[1];
            /* VCP supports only tone detection in upstream */
            pResults->dtmfControl.direction = VP_DIRECTION_US;
            break;
        case VP_DEVICE_OPTION_ID_DEVICE_IO:
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(4, DEVICE_OPTION_DATA_SIZE),
                            &readData[0]);
            pResults->deviceIo.directionPins_31_0 =
                                        ((uint32)readData[0]<<16) + readData[1];
            pResults->deviceIo.directionPins_63_32 =
                                        ((uint32)readData[2]<<16) + readData[3];
            pResults->deviceIo.outputTypePins_31_0 =
                                        ((uint32)readData[4]<<16) + readData[5];
            pResults->deviceIo.outputTypePins_63_32 =
                                        ((uint32)readData[6]<<16) + readData[7];
            break;
        case VP_OPTION_ID_PCM_TXRX_CNTRL:
            VpHbiRd(deviceId, VP_REG_MB_UP_BUF(4, 1), &data);
            pResults->pcmTxRxCntrl = data;
            break;
        default:
            break;
    } /* switch(optionId) */

    return VP_STATUS_SUCCESS;
} /* GetRdOptionMessage() */

/**
 * GetIoAccessResponse()
 *  This function copies data from the VCP's upstream mailbox into the structure
 *  pointed to by 'pResults'.
 * Preconditions
 *  Assumes that data in the upstream mailbox is response to a IO read access
  * event.
 *  Assumes buffer pointed to by 'pResults' is large enough to hold entire
 *  IO access data structure.
 * Postconditions
 *  Copies the response to the passed reference pointer
 */
static VpStatusType
GetIoAccessResponse(
    VpDeviceIdType deviceId,
    VpDeviceIoAccessDataType *pResults)
{
    #define IO_READ_RESPONSE_SIZE 9

    uint16 data[IO_READ_RESPONSE_SIZE];
    /* Get the IO access information */
    VpHbiRd(deviceId, VP_REG_MB_UP_BUF(3, IO_READ_RESPONSE_SIZE), &data[0]);

    pResults->accessType = data[0];
    pResults->accessMask_31_0 = ((uint32)data[1]<<16) + data[2];
    pResults->accessMask_63_32 = ((uint32)data[3]<<16) + data[4];
    pResults->deviceIOData_31_0 = ((uint32)data[5]<<16) + data[6];
    pResults->deviceIOData_63_32 = ((uint32)data[7]<<16) + data[8];

    return VP_STATUS_SUCCESS;
} /* GetIoAccessResponse() */

/**
 * GetTestCmpMessage()
 *  This function copies data from the VCP's upstream mailbox into the structure
 *  pointed to by 'pResults'.
 * Preconditions
 *  Assumes that data in the upstream mailbox is response to a line test
 *  command.
 *  Assumes 'pResults' points to an instance of a VpTestResultType.
 * Postconditions
 *  Test result data is copied from upstream mailbox to caller's buffer.
 *  Returns VP_INVALID if an HBI read error occurs.
 *  Returns VP_STATUS_SUCCESS otherwise.
 */
/*
 * NOTE: The following constant is used to allocate space on the stack for a
 * local copy of the incoming test result message. GetTestCmpMessage() returns a
 * fail code if the received message is longer than this limit. If this happens,
 * then this constant should be increased.
 */
#define MAX_TEST_RSP_LEN 20

static VpStatusType
GetTestCmpMessage(
    VpDeviceIdType deviceId,
    VpTestResultType *pResults
    )
{
    uint16 msgLen;
    uint16 msgBuf[MAX_TEST_RSP_LEN];
    VpTestResultKeyType *pKpadResult;

    /* Third word in message is message length (excluding length word itself).*/
    VpHbiRd(deviceId, VP_REG_MB_UP_BUF(2, 1), &msgLen);

    if ((msgLen < 2) || (msgLen > MAX_TEST_RSP_LEN)) {
        /* Message length out of bounds, return error. */
        return VP_STATUS_ERR_HBI;
    }

    /* Copy 'msgLen' words from upstream mailbox into local buffer. */
    VpHbiRd(deviceId, VP_REG_MB_UP_BUF(3, msgLen), msgBuf);

    /* Next word is always test type for this test result. */
    pResults->testId = (VpTestIdType)msgBuf[0];

    /* Next word is always error code for this test result. */
    pResults->errorCode = (VpTestStatusType)msgBuf[1];

    /*
     * Fill caller's result buffer regardless of error code. Some test results
     * return useful information even when the test fails, as in the case of
     * overflow or underflow in the open V AC test.
     */

    /* Parse remaining message according to test type. */
    switch (pResults->testId) {
        case VP_TEST_ID_PREPARE:
        case VP_TEST_ID_CONCLUDE:
            /* No result for these two "tests." */
            break;

        case VP_TEST_ID_TONE_GEN:
            /* No other data for this test, just continue. */
            break;

        case VP_TEST_ID_ARB_1TONE:
            /*
             * Only high freq. and amp. are valid for this test.
             * Others are included for debug/testing only.
             */

            /* Intentional case fall-through here... */
        case VP_TEST_ID_DIALTONE:
            pResults->result.tone.rmsPwr = msgBuf[2];
            pResults->result.tone.highFreq = msgBuf[3];
            pResults->result.tone.highAmp = msgBuf[4];
            pResults->result.tone.secFreq = msgBuf[5];
            pResults->result.tone.secAmp = msgBuf[6];
            pResults->result.tone.thdFreq = msgBuf[7];
            pResults->result.tone.thdAmp = msgBuf[8];
            break;

        case VP_TEST_ID_KEYPAD:
            {
                pKpadResult = &pResults->result.keypad;

                pKpadResult->kTestResult = (VpKeyTestResultType)msgBuf[2];
                pKpadResult->digit = (VpDigitType)msgBuf[3];
                pKpadResult->lowFreq = msgBuf[4];
                pKpadResult->lowAmp = msgBuf[5];
                pKpadResult->highFreq = msgBuf[6];
                pKpadResult->highAmp = msgBuf[7];
                pKpadResult->fftSize = msgBuf[8];
                pKpadResult->minBreak = msgBuf[9];
                pKpadResult->maxBreak = msgBuf[10];
                pKpadResult->aveBreak = msgBuf[11];
                pKpadResult->pulsePerSec = msgBuf[12];
                pKpadResult->hookFlashDet = (bool)msgBuf[13];
                pKpadResult->hookFlashDuration = msgBuf[14];
                pKpadResult->disconnectDet = (bool)msgBuf[15];
            }
            break;

        case VP_TEST_ID_OPEN_VDC:
            /* This result is signed. */
            pResults->result.sVout = (int16)msgBuf[2];
            break;

        case VP_TEST_ID_OPEN_VAC:
        case VP_TEST_ID_DIFF_VAC:
            /* This result is unsigned. */
            pResults->result.uVout = msgBuf[2];
            break;

        case VP_TEST_ID_NOISE:
        case VP_TEST_ID_HYBRID_LOSS:
            pResults->result.vrms = msgBuf[2];
            break;

        case VP_TEST_ID_SNR_QDIST:
            pResults->result.snrqd.vrmsNoise = msgBuf[2];
            pResults->result.snrqd.vrmsSignal = msgBuf[3];
            break;

        case VP_TEST_ID_SET_SENSE_GAIN:
            /* No other data for this test, just continue. */
            break;

        case VP_TEST_ID_DC_RLOOP:
            pResults->result.dcvab.vab = (int16)msgBuf[2];
            pResults->result.dcvab.ilg = (int16)msgBuf[3];
            break;

        case VP_TEST_ID_AC_RLOOP:
            pResults->result.acimt.imt = (int16)msgBuf[2];
            pResults->result.acimt.ilg = (int16)msgBuf[3];
            break;

        case VP_TEST_ID_AC_RLOOP_PHASE:
            pResults->result.acimtp.imt = (int16)msgBuf[2];
            pResults->result.acimtp.ilg = (int16)msgBuf[3];
            pResults->result.acimtp.phase = (int16)msgBuf[3];
            break;

        case VP_TEST_ID_NULL_IGND:
            pResults->result.vcm = (int16)msgBuf[2];
            break;

        case VP_TEST_ID_3ELE_CAP:
            pResults->result.cap.iahl = msgBuf[2];
            pResults->result.cap.iahh = msgBuf[3];
            pResults->result.cap.ibhh = msgBuf[4];
            pResults->result.cap.iall = msgBuf[5];
            pResults->result.cap.ibll = msgBuf[6];
            pResults->result.cap.vabhl = msgBuf[7];
            break;

        case VP_TEST_ID_3ELE_RES:
            pResults->result.res.iahl = (int16)msgBuf[2];
            pResults->result.res.ibhl = (int16)msgBuf[3];
            pResults->result.res.iahh = (int16)msgBuf[4];
            pResults->result.res.ibhh = (int16)msgBuf[5];
            pResults->result.res.iall = (int16)msgBuf[6];
            pResults->result.res.ibll = (int16)msgBuf[7];
            pResults->result.res.vabhl = (int16)msgBuf[8];
            break;

        case VP_TEST_ID_IDC:   /* intentional fall-through */
        case VP_TEST_ID_IAC:
            pResults->result.vix.ia = (int16)msgBuf[2];
            pResults->result.vix.ib = (int16)msgBuf[3];
            break;

        case VP_TEST_ID_ALT_REN:
            pResults->result.ren.imt = (int16)msgBuf[2];
            pResults->result.ren.imtDc = (int16)msgBuf[3];
            break;

        case VP_TEST_ID_HOOK_STATUS:
            pResults->result.hstatus = (bool)msgBuf[2];
            break;

        default:
            if ((pResults->testId >= VP_TEST_ID_RSVD_MIN) &&
                (pResults->testId <= VP_TEST_ID_RSVD_MAX)) {

                /* Handle "reserved" test IDs in a generic fashion:
                 * Copy words from upstream mailbox into 'pResults->result'
                 * buffer.  Number of words to copy is the next value in
                 * the mailbox. */
                uint16 numResults = msgBuf[2];
                VpHbiRd(deviceId, VP_REG_MB_UP_BUF(6, numResults),
                    (uint16p)&pResults->result);
            } else {
                return VP_STATUS_ERR_HBI;
            }
    }

    return VP_STATUS_SUCCESS;
} /* GetTestCmpMessage() */

/**
 * ConvertApiEventId2VcpEventId()
 *  This function converts API event category and eventId type to VCP event id.
 *
 * Preconditions
 *
 * Postconditions
 */
static VcpEventIdType
ConvertApiEventId2VcpEventId(
    VpEventType *pEvent)
{
    int i;
    uint16 eventId, mask;
    VcpEventIdType retVal;

    eventId = pEvent->eventId;

    if((pEvent->eventCategory >= VP_EVCAT_FXO) || (eventId == 0)){
        retVal = VCP_EVID_INVALID;
    } else {

        /* Compute EventId */
        for(i=0; i<sizeof(uint16)*8; i++){
            mask = 1<<i;
            if((mask & eventId) != 0) {
                break;
            }
        }

        retVal = i + (pEvent->eventCategory + 1)*0x10;
    }

    return retVal;
} /* ConvertApiEventId2VcpEventId() */

/**
 * ConvertVcpEventId2ApiEventId()
 *  This function converts a VCP eventId to API-2 event Id. This function
 * returns true if there is a valid API understood event else it returns false.
 *
 * Preconditions
 *
 * Postconditions
 */
static bool
ConvertVcpEventId2ApiEventId(
    uint16 eventWord, VpEventCategoryType *eventCat, uint16 *eventId)
{
    uint8 type; uint16 tempEventId;
    /* The following array maps an event type to the number of valid events
     * within that type */
    uint8 validEvTypeMax[VCP_EVID_NUM_TYPES] = {
        VCP_EVID_NUM_T1, VCP_EVID_NUM_T2, VCP_EVID_NUM_T3,
        VCP_EVID_NUM_T4, VCP_EVID_NUM_T5};

    if (eventWord & 0x8000) {
        /* DevNotes: There is a system event present in the event queue at
         * the begining. This is a bug in the VCP h/w. Because of VCP bug,
         * we won't handle system interrupts until the next VCP revision.
         *
         * Further Notes: Once the first event is removed, we still have many
         * events that are meaningful. Add code to handle such events.
         */

        *eventCat = *eventId = 0;
        return FALSE;
    } else {

        /* Handle the VCP's SW interrupts (API Events) */
        type = (uint8)(eventWord >> 12) & 0x0F; /* Isolate the type */
        tempEventId = eventWord >> 8;           /* Get the eventId  */

        /* Verify that the event is valid */
        if ((type == 0) || (type > 5) ||            /* Verify Type Range    */
            ((tempEventId & 0x0F)                   /* Isolate the subType */
                >= validEvTypeMax[type-1]) ||       /* Verify subType Range */
            (tempEventId == VCP_EVID_T4_RSVD1) ||   /* Removed event */
            (tempEventId == VCP_EVID_T3_RSVD9) ||   /* Removed event */
            (tempEventId == VCP_EVID_T3_RSVDA)){    /* Empty slot */

            /*
             * Invalid interrupt indication
             *   The only reason that this case should be possible, would be
             *   if the VCP returned an invalid event type.
             */
            *eventCat = *eventId = 0;
            return FALSE;
        }

        /* Extract the VP-API specific eventId and category */
        *eventCat = type - 1;
        *eventId = 1<<(tempEventId - type * 0x10);
    }

    return TRUE;
} /* ConvertVcpEventId2ApiEventId() */

/******************************************************************************
 *                              Helper Functions                              *
 ******************************************************************************/
/**
 * WriteMailBoxCmd()
 *  This function writes a complete command to the VCP mail box. Before writing
 * the command it acquires the mail box and after writing the command it
 * releases the maibox.
 *
 * Preconditions
 * None
 *
 * Postconditions
 *  None
 */
static VpStatusType
WriteMailBoxCmd(
    VpDeviceIdType deviceId,
    uint8 length,       /* Exact length (not one less) */
    uint16 *pCmdBuf)
{
    VpStatusType retval;

    /* Basic error checking */
    if(length > VP_REG_MB_DOWN_BUF_SIZE) {
        return VP_STATUS_FAILURE;
    }

    VpSysEnterCritical(deviceId, VP_HBI_CRITICAL_SEC); /* CRITICAL SECTION => */

    /* Ensure that the Mailbox is available for the host */
    if(MailboxAcquire(deviceId) == VP_STATUS_MAILBOX_BUSY) {

        retval = VP_STATUS_MAILBOX_BUSY;

    } else {

        VpHbiWr(deviceId, VP_REG_MB_DOWN_BUF(VP_MB_FUNC_INDEX, length),
                pCmdBuf);

        /* Turn the mailbox over to the VCP (raise the flag) */
        MailboxRelease(deviceId);

        retval = VP_STATUS_SUCCESS;
    }

    VpSysExitCritical(deviceId, VP_HBI_CRITICAL_SEC);  /* CRITICAL SECTION <= */

    return retval;
} /* WriteMailBoxCmd() */

/**
 * MailboxAcquire()
 *  This function ensures that the host has the right to write to the mailbox.
 *  It returns VP_STATUS_MAILBOX_BUSY, if the VCP has the mailbox locked (flag
 *  raised).
 *
 * Preconditions
 * None
 *
 * Postconditions
 *  None
 */
static VpStatusType
MailboxAcquire(VpDeviceIdType deviceId)
{
    uint16 data;

#ifdef WAIT_TO_ACQUIRE_VCP_MB
    uint32 i;
    /* Read the mailbox DownStream Flag */
    /* In most cases, VCP mailbox flag should readily be available. Under rare
     * circumstances when VCP is working on a previous command and has not yet
     * completed processing it and there is another command that needs to be
     * processed it may be OK to wait unitl the VCP returns the mailbox
     * ownership.
     * This approach may be desired if application does not want to handle the
     * (less)unpredictable mail box busy error (and try to re-issue the
     * command). Users of API may want to modify the number of interations in
     * the following loop based on architecture specific needs.
     */
    for(i=1; i<ITERATIONS_TO_WAIT_FOR_VCP_MB; i++) {
        VpHbiRd(deviceId, HW_Reg_MBOXFLAG, &data);
        if ( (data & VP_BIT_MBOXFLAG_CMD) > 0) {   /* If host 'owns' mbox */
        #ifdef DEBUG_VCP_API_MB
            if (!(i % 5000)) {
                term_printf("WARNING: Waiting for VCP mailbox\n");
            }
        #endif /* DEBUG_VCP_API_MB */
        } else {
            return VP_STATUS_SUCCESS;
        }
    }
    return VP_STATUS_MAILBOX_BUSY;
#else
    /* Read the command mailbox flag */
    VpHbiRd(deviceId, HW_Reg_MBOXFLAG, &data);

    /* Determine if mailbox is busy */
    if ( (data & VP_BIT_MBOXFLAG_CMD) > 0) {   /* If host 'owns' mbox */
        return VP_STATUS_MAILBOX_BUSY;
    } else {
        return VP_STATUS_SUCCESS;
    }
#endif /* WAIT_TO_ACQUIRE_VCP_MB */
} /* MailboxAcquire() */

/**
 * MailboxRelease()
 *  This function releases the mailbox to the VCP. This is done by raising the
 *  mailbox flag.
 * Preconditions
 *
 * Postconditions
 *  The mailbox flag is raised.
 */
static void
MailboxRelease(VpDeviceIdType deviceId)
{
    uint16 data;

    /* Give the VCP ownership of the command mailbox */
    data = VP_BIT_MBOXFLAG_CMD;     /* write 1 to set bit */
    VpHbiWr(deviceId, HW_Reg_MBOXFLAG, &data);

} /* MailboxRelease() */
/**
 * MailboxAddCommand()
 *  This function adds a command id and a context handle to the mailbox.
 * Preconditions
 *
 * Postconditions
 *  Command ID and context handle are added to the mailbox.
 */

static void
MailboxAddCommand(
    VpDeviceIdType deviceId,
    VcpMbFuncIdType funcId,      /* Function ID of Mailbox command to issue */
    uint16 contextId            /* Data for the context handle */
    )
{
    /* Copy 'funcId' into a 16-bit variable for VpHbiWr(). */
    uint16 funcId16 = (uint16)funcId;

    VpHbiWr(deviceId, VP_REG_MB_DOWN_BUF(VP_MB_FUNC_INDEX, 1), &funcId16);
    VpHbiWr(deviceId, VP_REG_MB_DOWN_BUF(VP_MB_CONTEXT_INDEX, 1), &contextId);
} /* MailboxAddCommand() */

/**
 * MailboxAddArg()
 *  This function adds an argument to the VCP Downstream mailbox.
 * Preconditions
 *  None
 * Postconditions
 *  'arg' is added to the mailbox 'offset' words from beginning of the mailbox.
 *  The offset of the next free mailbox location is returned.
 */
static uint8
MailboxAddArg(
    VpDeviceIdType deviceId,
    uint16 arg,             /* Argument value to add to the mailbox */
    uint8  offset           /* Offset in maibox at which to write arg */
    )
{
    VpHbiWr(deviceId, VP_REG_MB_DOWN_BUF(offset, 1), &arg);

    return offset + 1;
} /* MailboxAddArg() */

/**
 * MailboxAddProfile()
 *  Adds a profile (either a selector, or the entire profile) to the VCP
 *  message mailbox.
 * Preconditions
 *  None
 * Postconditions
 *  The VCP's mailbox has either the profile selector or the entire profile
 *  loaded into it.
 *  The offset of the next free mailbox location is returned.
 */
static uint8
MailboxAddProfile(
    VpDeviceIdType deviceId,
    VpProfileType type,
    VpProfilePtrType pProfile,
    uint8 offset
    )
{
    uint16 lenBytes;
    bool compact;
    compact = FALSE;

    /* Return error in the unlikely event that profile ID is not valid. */
    if (type >= VP_PROFILE_FXO_CONFIG) return 0; /* Profiles above this
                                                  * are not supported */

    if (pProfile <= VP_PTABLE_MAX_INDEX) {
        /*
         * "Pointer to profile" is actually a profile selector. Add only the
         * selector to the mailbox.
         */
        uint16 selector = (uint16)((unsigned)pProfile);

        /*
         * Check profile-specific bounds on selector. Note that selectors are
         * numbered 1 through 15 with 0 ("VP_NULL") being valid for all.
         */
        if (selector > GetProfileTableMaxCount(type)) return 0;

        if(type == VP_PROFILE_TONECAD) {
            /* Cadence table provided by VCP is split in to two cadence
             * tables by software. */
            if(selector != 0) {
                /* '0' Is a special number that has to be treated specially
                 * because it does not correspond to any table entry but it
                 * means (mostly) means continuous on cadence or do not
                 * change.  */
                selector = selector + VCP_RING_CAD_PROF_COUNT;
            }
        }

        /* Set MSB flag indicating that this is indeed a profile selector. */
        selector |= 0x8000;

        VpHbiWr(deviceId, VP_REG_MB_DOWN_BUF(offset, 1), &selector);
        offset++;

    } else {
        /*
         * 'pProfile' is actually a pointer to a profile array. Handle different
         * profile formats here.
         */

         /* The length field is always sent before the profile itself.  The VCP
            knows whether this field is in bytes or words, based on the profile
            type. */
         #define API2_PROFILE_LENGTH_INDEX 3 /* DevNotes: - replace this with
                                              * the appropriate API2 constant */
         lenBytes = pProfile[API2_PROFILE_LENGTH_INDEX] + 4;
         VpHbiWr(deviceId, VP_REG_MB_DOWN_BUF(offset, 1), &lenBytes);
         offset++;

         /* Write profile data to downstream mailbox. */
         offset += WriteProfile(deviceId,
             VP_REG_MB_DOWN_BUF_PAGE(offset),
             VP_REG_MB_DOWN_BUF_OFFSET(offset),
             lenBytes, pProfile, TRUE);

    }

    return offset;
} /* MailboxAddProfile() */


/**
 * WriteProfile()
 *  This internal helper function is called to write a profile (or any byte
 *  string that may need to be "packed") into a block of HBI memory.
 * Preconditions
 *  'page', 'offset' specify the start HBI address of the destination buffer.
 *  'length' is the length of the source data in bytes (if compact == TRUE) or
 *     words (if compact == FALSE).
 *  'pProfile' is a pointer to the start of the source data.
 *  'compact' determines whether source bytes are to be packed into 16-bit
 *     HBI data words.
 * Postconditions
 *  The source data is copied into the specified HBI buffer.
 *  Returns count of words written to HBI.
 */
static uint8
WriteProfile(
    VpDeviceIdType deviceId,
    uint8 page,
    uint8 offset,
    uint8 length,
    VpProfilePtrType pProfile,
    bool compact
    )
{
    /* Reserve space on the stack for the longest possible profile. */
    uint16 wordBuf[VP_REG_MAX_PROF_SIZE];
    uint8 i;

    if (compact == TRUE) {
        /* Convert length in bytes to length in words. */
        length = (length + 1) >> 1;
    } else {
        /* Length in bytes is same as length in words; no conversion reqd. */
    }

    for (i = 0; i < length; i++) {
        wordBuf[i] = (uint16)*pProfile++;

        if (compact == TRUE) {
            wordBuf[i] |= (uint16)*pProfile++ << 8;
        }
    }

    if (VpHbiWr(deviceId, page, offset, length - 1, wordBuf)) {

        return length;
    }
    return 0;
} /* WriteProfile() */

static bool
FftSizeIsValid(
    uint16 fftSize
    )
{
    if ((fftSize == 0x0000) || (fftSize == 0x0080) || (fftSize == 0x0100) ||
        (fftSize == 0x0200) || (fftSize == 0x0400) || (fftSize == 0x0800) ||
        (fftSize == 0x1000) || (fftSize == 0x2000)) {
        return TRUE;
    }

    return FALSE;
}

/**
 * GetProfileTableMaxCount()
 * This function returns the maximum index permitted for a given type of
 * profile table.
 *
 * Preconditions
 * None.
 *
 * Postconditions
 * None.
 */
static uint8
GetProfileTableMaxCount(
    VpProfileType profileType)
{
    switch (profileType) {
        case VP_PROFILE_DEVICE:     {return VP_REG_DEV_PROF_COUNT; }
        case VP_PROFILE_AC:         {return VP_REG_AC_PROF_COUNT; }
        case VP_PROFILE_DC:         {return VP_REG_DC_PROF_COUNT; }
        case VP_PROFILE_RING:       {return VP_REG_RING_PROF_COUNT; }
        case VP_PROFILE_RINGCAD:    {return VCP_RING_CAD_PROF_COUNT; }
        case VP_PROFILE_TONE:       {return VP_REG_TONE_PROF_COUNT; }
        case VP_PROFILE_METER:      {return VP_REG_MTR_PROF_COUNT; }
        case VP_PROFILE_CID:        {return VP_REG_CID_PROF_COUNT; }
        case VP_PROFILE_TONECAD:    {return VP_REG_CAD_PROF_COUNT
                                            - VCP_RING_CAD_PROF_COUNT; }
        /* Other profile types are not supported by VCP */
        default: return 0;
    }
}

/**
 * SetVcpDefaultOptions()
 * This function executes the options to set the device/lines to API-II
 * standard default settings.  It may be passed a valid device context, or a
 * valid line context.  This is a convenient function for the API itself to use
 * when a device or line is initialized.
 *
 * NOTE: Presently it sets only device default options.
 *
 * Preconditions:
 * Only valid line context or valid device context should be passed. Not both.
 *
 * Postconditions:
 * The device and line associated with this device is initialized with default
 * values.
 */
static VpStatusType
SetVcpDefaultOptions(
    VpDevCtxType *pDevCtx,
    VpLineCtxType *pLineCtx)
{
    VpOptionPulseType pulseOption;
    VpOptionCriticalFltType criticalFlt;
    VpOptionZeroCrossType zeroCross;
    uint16 ramp2Stby;
    VpOptionPulseModeType pulseMode;
    VpOptionCodecType codec;
    VpOptionPcmHwyType pcmHwy;
    VpOptionLoopbackType loopBack;
    VpOptionLineStateType lineState;
    VpOptionEventMaskType eventMask;
    VpOptionRingControlType ringControl;
    VpOptionDtmfModeType dtmfMode;
    VpStatusType status = VP_STATUS_SUCCESS;
    bool singleLineOpt;
    VpOptionPcmTxRxCntrlType pcmCntrl;
    VpDevCtxType *pDevCtxLocal;
    VpVcpDeviceObjectType *pDevObj;

    /* Basic error checking */
    if((pDevCtx == VP_NULL) && (pLineCtx == VP_NULL)) {
        return VP_STATUS_FAILURE;
    }
    if((pDevCtx != VP_NULL) && (pLineCtx != VP_NULL)) {
        return VP_STATUS_FAILURE;
    }

    if(pDevCtx != VP_NULL) {
        /* Setting options for all lines */
        singleLineOpt = FALSE;

        pDevObj = pDevCtx->pDevObj;

    } else {
        /* Setting options for one line */
        singleLineOpt = TRUE;

        pDevCtxLocal = pLineCtx->pDevCtx;
        pDevObj = pDevCtxLocal->pDevObj;
    }

#ifdef DEBUG_VCP_API_DEFLT_OPTN
     term_printf("Setting default options for VCP(%s)...\n",
                    singleLineOpt ? "One Line":"All Lines");
#endif /* DEBUG_VCP_API_DEFLT_OPTN */

    if(singleLineOpt == FALSE) {

        /* Set default pulse option */
        pulseOption.breakMin = VP_OPTION_DEFAULT_DP_BREAK_MIN;
        pulseOption.breakMax = VP_OPTION_DEFAULT_DP_BREAK_MAX;
        pulseOption.makeMin = VP_OPTION_DEFAULT_DP_MAKE_MIN;
        pulseOption.makeMax = VP_OPTION_DEFAULT_DP_MAKE_MAX;
        pulseOption.interDigitMin = VP_OPTION_DEFAULT_DP_INTER_DIG_MIN;
        pulseOption.flashMin = VP_OPTION_DEFAULT_DP_FLASH_MIN;
        pulseOption.flashMax = VP_OPTION_DEFAULT_DP_FLASH_MAX;
        status = VcpSetOption(pLineCtx, pDevCtx, VP_DEVICE_OPTION_ID_PULSE,
                                &pulseOption);
        if(status != VP_STATUS_SUCCESS) {
#ifdef DEBUG_VCP_API_DEFLT_OPTN
            term_printf("ERROR: Failed to set 'Pulse' Default Option %d\n", status);
#endif /* DEBUG_VCP_API_DEFLT_OPTN */
            return status;
        }

        /* Set default Critical fault option */
        if(pDevObj->productType == VP_VCP_PRODUCT_VE_880) {
            /* Automatic transition to disconnect upon fault detection is
             * not supported in VCP-880. */
            criticalFlt.acFltDiscEn = VP_OPTION_DEFAULT_CF_AC_DIS_EN_880;
            criticalFlt.dcFltDiscEn = VP_OPTION_DEFAULT_CF_DC_DIS_EN_880;
        } else {
            /* Must be VP_VCP_PRODUCT_VE_790 */
            criticalFlt.acFltDiscEn = VP_OPTION_DEFAULT_CF_AC_DIS_EN;
            criticalFlt.dcFltDiscEn = VP_OPTION_DEFAULT_CF_DC_DIS_EN;
        }
        criticalFlt.thermFltDiscEn = VP_OPTION_DEFAULT_CF_THERMAL_DIS_EN;
        status = VcpSetOption(pLineCtx, pDevCtx, VP_DEVICE_OPTION_ID_CRITICAL_FLT,
                                &criticalFlt);
        if(status != VP_STATUS_SUCCESS) {
#ifdef DEBUG_VCP_API_DEFLT_OPTN
            term_printf("ERROR: Failed to set 'Critical Fault' Default Option %d\n",
                            status);
#endif /* DEBUG_VCP_API_DEFLT_OPTN */
            return status;
        }
    }

    /* Set default Zero Cross option */
    zeroCross = VP_OPTION_DEFAULT_ZERO_CROSS;
    status = VcpSetOption(pLineCtx, pDevCtx, VP_OPTION_ID_ZERO_CROSS,
                            &zeroCross);
    if(status != VP_STATUS_SUCCESS) {
#ifdef DEBUG_VCP_API_DEFLT_OPTN
        term_printf("ERROR: Failed to set 'Zero Cross' Default Option %d\n",
                        status);
#endif /* DEBUG_VCP_API_DEFLT_OPTN */
        return status;
    }

    if(singleLineOpt == FALSE) {
        /* Set default Ramp to standby option */
        ramp2Stby = VP_OPTION_DEFAULT_RAMP2STBY;
        status = VcpSetOption(pLineCtx, pDevCtx, VP_DEVICE_OPTION_ID_RAMP2STBY,
                                &ramp2Stby);
        if(status != VP_STATUS_SUCCESS) {
#ifdef DEBUG_VCP_API_DEFLT_OPTN
            term_printf("ERROR: Failed to set 'Ramp2Stby' Default Option %d\n",
                            status);
#endif /* DEBUG_VCP_API_DEFLT_OPTN */
            return status;
        }
    }

    /* Set default Pulse mode option */
    pulseMode = VP_OPTION_DEFAULT_PULSE_MODE;
    status = VcpSetOption(pLineCtx, pDevCtx, VP_OPTION_ID_PULSE_MODE,
                            &pulseMode);
    if(status != VP_STATUS_SUCCESS) {
#ifdef DEBUG_VCP_API_DEFLT_OPTN
        term_printf("ERROR: Failed to set 'Pulse Mode' Default Option %d\n",
                        status);
#endif /* DEBUG_VCP_API_DEFLT_OPTN */
        return status;
    }

    /* Set default CODEC option */
    codec = VP_OPTION_DEFAULT_CODEC_MODE;
    status = VcpSetOption(pLineCtx, pDevCtx, VP_OPTION_ID_CODEC, &codec);
    if(status != VP_STATUS_SUCCESS) {
#ifdef DEBUG_VCP_API_DEFLT_OPTN
        term_printf("ERROR: Failed to set 'CODEC' Default Option %d\n",
                        status);
#endif /* DEBUG_VCP_API_DEFLT_OPTN */
        return status;
    }

    /* Set default PCM Highway option */
    pcmHwy = VP_OPTION_DEFAULT_PCM_HWY;
    status = VcpSetOption(pLineCtx, pDevCtx, VP_OPTION_ID_PCM_HWY, &pcmHwy);
    if(status != VP_STATUS_SUCCESS) {
#ifdef DEBUG_VCP_API_DEFLT_OPTN
        term_printf("ERROR: Failed to set 'PCM Highway' Default Option %d\n",
                        status);
#endif /* DEBUG_VCP_API_DEFLT_OPTN */
        return status;
    }

    /* Set default Loopback option */
    loopBack = VP_OPTION_DEFAULT_LOOP_BACK;
    status = VcpSetOption(pLineCtx, pDevCtx, VP_OPTION_ID_LOOPBACK, &loopBack);
    if(status != VP_STATUS_SUCCESS) {
#ifdef DEBUG_VCP_API_DEFLT_OPTN
        term_printf("ERROR: Failed to set 'Loopback' Default Option %d\n",
                        status);
#endif /* DEBUG_VCP_API_DEFLT_OPTN */
        return status;
    }

    /* Set default Line State option */
    lineState.battRev = VP_OPTION_DEFAULT_LS_BAT_REV;
    lineState.bat = VP_OPTION_DEFAULT_LS_BAT;
    status = VcpSetOption(pLineCtx, pDevCtx, VP_OPTION_ID_LINE_STATE,
                            &lineState);
    if(status != VP_STATUS_SUCCESS) {
#ifdef DEBUG_VCP_API_DEFLT_OPTN
        term_printf("ERROR: Failed to set 'Line State' Default Option %d\n",
                        status);
#endif /* DEBUG_VCP_API_DEFLT_OPTN */
        return status;
    }

    /* Set Default event mask */
    if(singleLineOpt == FALSE) {

        /* Even though event mask is line specific there are device specific
         * event masks as well. Hence when a line initialization is done,
         * if we wrote the event mask we could be overwriting the device event
         * masks that were set for other lines as well. So don't do it.
         */
        eventMask.faults = VCP_OPTION_DEFAULT_FAULT_EVENT_MASK;
        eventMask.signaling = VCP_OPTION_DEFAULT_SIGNALING_EVENT_MASK;
        eventMask.response = VCP_OPTION_DEFAULT_RESPONSE_EVENT_MASK;
        eventMask.test = VP_OPTION_DEFAULT_TEST_EVENT_MASK;
        eventMask.process = VP_OPTION_DEFAULT_PROCESS_EVENT_MASK;
        eventMask.fxo = VCP_OPTION_DEFAULT_FXO_EVENT_MASK;
        eventMask.packet = VCP_OPTION_DEFAULT_PACKET_EVENT_MASK;
        status = VcpSetOption(pLineCtx, pDevCtx, VP_OPTION_ID_EVENT_MASK,
                                &eventMask);
        if(status != VP_STATUS_SUCCESS) {
#ifdef DEBUG_VCP_API_DEFLT_OPTN
            term_printf("ERROR: Failed to set 'Event' Default Option %d\n",
                            status);
#endif /* DEBUG_VCP_API_DEFLT_OPTN */
            return status;
        }
    }

    /* Set default Ring Control option */
    ringControl.zeroCross = VP_OPTION_DEFAULT_RC_ZERO_CROSS;
    ringControl.ringExitDbncDur = VP_OPTION_DEFAULT_RC_RING_EXIT_DBNC_VAL;
    ringControl.ringTripExitSt = VP_OPTION_DEFAULT_RC_RING_EXIT_STATE;
    status = VcpSetOption(pLineCtx, pDevCtx, VP_OPTION_ID_RING_CNTRL,
                            &ringControl);
    if(status != VP_STATUS_SUCCESS) {
#ifdef DEBUG_VCP_API_DEFLT_OPTN
        term_printf("ERROR: Failed to set 'Ring Control' Default Option %d\n",
                        status);
#endif /* DEBUG_VCP_API_DEFLT_OPTN */
        return status;
    }

    /* Set default DTMF mode option */
    dtmfMode.dtmfControlMode = VP_OPTION_DEFAULT_DM_DTMF_CONTROL_MODE;
    dtmfMode.direction = VP_OPTION_DEFAULT_DM_DIRECTION;
    status = VcpSetOption(pLineCtx, pDevCtx, VP_OPTION_ID_DTMF_MODE, &dtmfMode);
    if(status != VP_STATUS_SUCCESS) {
#ifdef DEBUG_VCP_API_DEFLT_OPTN
        term_printf("ERROR: Failed to set 'DTMF Mode' Default Option %d\n",
                        status);
#endif /* DEBUG_VCP_API_DEFLT_OPTN */
        return status;
    }

    /* Set default PCM transmit/receive control option */
    pcmCntrl = VP_OPTION_DEFAULT_PCM_TXRX_CNTRL;
    status = VcpSetOption(pLineCtx, pDevCtx, VP_OPTION_ID_PCM_TXRX_CNTRL,
                            &pcmCntrl);
    if(status != VP_STATUS_SUCCESS) {
#ifdef DEBUG_VCP_API_DEFLT_OPTN
        term_printf("ERROR: Failed to set 'PCM Tx/Rx Cntrl' Default Option"
                    " %d\n", status);
#endif /* DEBUG_VCP_API_DEFLT_OPTN */
        return status;
    }

    return status;
} /* SetVcpDefaultOptions() */

/**
 * GetTotalDtmfResources()
 * This function returns total number of DTMF resources available.
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * None
 */
static uint8
GetTotalDtmfResources(
    uint32 linewiseStatus,
    uint8 remainingResources)
{
    int i;
    uint8 totalResources = 0;

    for(i = 0; i<32; i++){
        if((linewiseStatus & 0x01) == 0x01) {
            totalResources++;
            linewiseStatus = linewiseStatus>>1;
        }
    }
    return (totalResources + remainingResources);
} /* GetTotalDtmfResources() */

/**
 * UpdateProfileTableStatus()
 * This function is used to update the status of the profile tables.
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * This function updates the status of profile table status variables.
 * Returns TRUE if it successfully accomplishes its task else it returns
 * FALSE.
 */
static bool
UpdateProfileTableStatus(
    VpVcpDeviceObjectType *pDevObj,
    VpProfileType type,
    int profileIndex,
    VpProfilePtrType pProfile)
{
    uint16 *pData = VP_NULL;
    bool retVal = TRUE;
    int tempIndex;
    uint16 mask = 0;

    /* Obtain profile table status variable pointer */
    pData = GetProfileTableStatusPtr(pDevObj, type);
    if(pData == VP_NULL) {
        /* Profile type not understood */
        return FALSE;
    }

    /* Verify application */
    if(pProfile == VP_PTABLE_NULL) {
        /* A profile table entry is being erased */

        mask = 1<<profileIndex;
        if((*pData & mask) == mask) {
            /* Profie present in the profile slot; Erasing it */
            *pData = *pData & (~mask);
        } /* Else, It should be OK to erase something that does not exit
           * at the first place (to start from clean slate). */
    } else {

        tempIndex = VpGetProfileIndex(pProfile);
        if(tempIndex < 0) {
            /* Perform basic profile error checking */

            if(VpVerifyProfileType(type, pProfile) == FALSE) {
                retVal = FALSE;
            } else {
                /* Valid profile Pointer (Not a index) */
                mask = 1<<profileIndex;
                *pData = *pData | mask;
            }
        } else {
            /* Valid profile index; Host is passing a profile table index
             * instead of a valid profile pointer */
            retVal = FALSE;
        }
    }

    return retVal;
}

/**
 * IsItAValidProfilePtr()
 * This function is used to enquire if a given profile pointer points to a
 * valid profile.
 *
 * Preconditions:
 * None.
 *
 * Postconditions:
 * This function returns a boolean value as defined below.
 * TRUE <== Given Profile pointer points to VP_PTABLE_NULL
 * TRUE <== Given Profile pointer points to a VALID profile table index
 * TRUE <== Given Profile pointer points to a VALID profile pointer
 * FALSE <== Given profile pointer points to invalid profile index
 * FALSE <== Given profile pointer points to a profile that has basic errors
 */
static bool
IsItAValidProfilePtr(
    VpVcpDeviceObjectType *pDevObj,
    VpProfileType type,
    VpProfilePtrType pProfile)
{

    if(pProfile == VP_PTABLE_NULL) {
        return TRUE;
    }else if((pProfile >= VP_PTABLE_INDEX1) && (pProfile <= VP_PTABLE_INDEX15)){
        /* A valid profile index */
        return IsThereAValidProfileTableEntry(pDevObj, type, pProfile);
    } else {
        /* Passed pointer points to a profile */
        return VpVerifyProfileType(type, pProfile);
    }
} /* IsItAValidProfilePtr() */

/**
 * IsThereAValidProfileTableEntry()
 * This function is used to enquire the validity of a profile table entry. That
 * is this function tells whether a profile table entry contains a valid profile
 * or not.
 *
 * Preconditions:
 * None.
 *
 * Postconditions:
 * Returns TRUE if a valid profile is present in a given profile table.
 * Else it returns FALSE.
 */
static bool
IsThereAValidProfileTableEntry(
    VpVcpDeviceObjectType *pDevObj,
    VpProfileType type,
    VpProfilePtrType pProfileIndex)
{
    uint16 *pData = VP_NULL;
    int tempIndex;
    uint16 mask = 0;

    /* Obatin profile table status variable pointer */
    pData = GetProfileTableStatusPtr(pDevObj, type);
    if(pData == VP_NULL) {
        /* Profile type not understood */
        return FALSE;
    }

    /* Obtain Profile index and validate the index */
    tempIndex = VpGetProfileIndex(pProfileIndex);
    if(tempIndex < 0) {
        /* A profile index is expected here */
        return FALSE;
    }
    /* Verify the index is a permissible index */
    if(tempIndex >= GetProfileTableMaxCount(type)) {
        return FALSE;
    }

    mask = 1<<tempIndex;
    if((*pData & mask) == mask){
        return TRUE;
    } else {
        return FALSE;
    }
} /* IsThereAValidProfileTableEntry() */

/**
 * GetProfileTableStatusPtr()
 * This function returns the pointer that holds the status corresponding to
 * a certain profile table.
 *
 * Preconditions:
 * None
 *
 * Postconditions:
 * Returns the pointer to the location where profile table status is stored
 * if successfull else returns VP_NULL.
 */
static uint16 *
GetProfileTableStatusPtr(
    VpVcpDeviceObjectType *pDevObj,
    VpProfileType type)
{
    uint16 *pData = VP_NULL;

    switch (type) {
        case VP_PROFILE_DEVICE: {
            pData = &(pDevObj->devicePrfTblSts);
            break;
        }
        case VP_PROFILE_AC: {
            pData = &(pDevObj->acPrfTblSts);
            break;
        }
        case VP_PROFILE_DC: {
            pData = &(pDevObj->dcPrfTblSts);
            break;
        }
        case VP_PROFILE_RING: {
            pData = &(pDevObj->ringPrfTblSts);
            break;
        }
        case VP_PROFILE_RINGCAD: {
            pData = &(pDevObj->ringCadPrfTblSts);
            break;
        }
        case VP_PROFILE_TONE: {
            pData = &(pDevObj->tonePrfTblSts);
            break;
        }
        case VP_PROFILE_METER: {
            pData = &(pDevObj->meterPrfTblSts);
            break;
        }
        case VP_PROFILE_CID: {
            pData = &(pDevObj->cidPrfTblSts);
            break;
        }
        case VP_PROFILE_TONECAD: {
            pData = &(pDevObj->toneCadPrfTblSts);
            break;
        }
        default: {
            /* This case is not expected */
        }
    }

    return pData;
} /* GetProfileTableStatusPtr() */

#endif /* VP_CC_VCP_SERIES */







